--------------------------------------------------------------------------------
--
-- Copyright (C) 2017
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Author:
-- . Eric Kooistra, 1 May 2018
-- Purpose:
-- . Reshape the block size by redefining the sop and eop for all streams.
-- Description:
--   See dp_block_reshape.vhd for functional description.
--
--   This dp_block_reshape_arr instantiate dp_block_reshape g_nof_streams times.
--   When g_shared_control = TRUE then stream 0 is used to control all streams.
--   When g_shared_control = FALSE then each stream controls itself, independent
--   of the other streams.
--

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY dp_block_reshape_arr IS
  GENERIC (
    g_nof_streams              : NATURAL := 1;       -- nof streams
    g_shared_control           : BOOLEAN := FALSE;   -- when TRUE then use snk_in_arr(0) for control, else use per snk_in_arr stream
    g_input_nof_data_per_sync  : NATURAL := 2;       -- nof data per input sync interval, used only for sop_index
    g_reshape_nof_data_per_blk : NATURAL := 2;       -- nof data per output block
    g_pipeline_src_out         : NATURAL := 1;       -- Pipeline source outputs (all sosi)
    g_pipeline_src_in          : NATURAL := 0        -- Pipeline source inputs (all siso). This will also pipeline src_out.
  );
  PORT (
    clk              : IN  STD_LOGIC;
    rst              : IN  STD_LOGIC;

    snk_in_arr       : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
    snk_out_arr      : OUT t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);

    src_out_arr      : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
    src_in_arr       : IN  t_dp_siso_arr(g_nof_streams-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);
    src_index_2arr_2 : OUT t_natural_2arr_2(g_nof_streams-1 DOWNTO 0)   -- [1] sop index, [0] valid index per stream
  );
END dp_block_reshape_arr;


ARCHITECTURE str OF dp_block_reshape_arr IS

  SIGNAL in_sosi_arr  : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL in_siso_arr  : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL out_sosi_arr : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL out_siso_arr : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL in_sosi      : t_dp_sosi;
  SIGNAL in_siso      : t_dp_siso;
  SIGNAL out_sosi     : t_dp_sosi;
  SIGNAL out_siso     : t_dp_siso;

  SIGNAL out_index_2arr_2 : t_natural_2arr_2(g_nof_streams-1 DOWNTO 0);   -- [1] sop index, [0] valid index per stream
  
BEGIN

  -- Dependent on g_shared_control use stream(0) to control all streams or use separate control per stream
  in_sosi  <= snk_in_arr(0);
  in_siso  <= in_siso_arr(0);
  out_sosi <= out_sosi_arr(0);
  out_siso <= src_in_arr(0);
  
  p_in : PROCESS(snk_in_arr, in_sosi, in_siso, in_siso_arr)
  BEGIN
    -- in_sosi_arr
    IF g_shared_control=FALSE THEN
      in_sosi_arr <= snk_in_arr;
    ELSE
      in_sosi_arr <= (OTHERS=>in_sosi);
      FOR I IN 0 TO g_nof_streams-1 LOOP
        in_sosi_arr(I).data <= snk_in_arr(I).data;
        in_sosi_arr(I).re   <= snk_in_arr(I).re;
        in_sosi_arr(I).im   <= snk_in_arr(I).im;
      END LOOP;
    END IF;
    
    -- snk_out_arr
    IF g_shared_control=FALSE THEN
      snk_out_arr <= in_siso_arr;
    ELSE
      snk_out_arr <= (OTHERS=>in_siso);
    END IF;
  END PROCESS;
  
  p_out : PROCESS(out_sosi_arr, out_sosi, out_siso, src_in_arr)
  BEGIN
    -- src_out_arr
    IF g_shared_control=FALSE THEN
      src_out_arr <= out_sosi_arr;
    ELSE
      src_out_arr <= (OTHERS=>out_sosi);
      FOR I IN 0 TO g_nof_streams-1 LOOP
        src_out_arr(I).data <= out_sosi_arr(I).data;
        src_out_arr(I).re   <= out_sosi_arr(I).re;
        src_out_arr(I).im   <= out_sosi_arr(I).im;
      END LOOP;
    END IF;
    
    -- out_siso_arr
    IF g_shared_control=FALSE THEN
      out_siso_arr <= src_in_arr;
    ELSE
      out_siso_arr <= (OTHERS=>out_siso);
    END IF;
  END PROCESS;

  -- Instantiate dp_block_reshape per stream
  gen_streams : FOR I IN 0 TO g_nof_streams-1 GENERATE
    u_block_reshape : ENTITY work.dp_block_reshape
    GENERIC MAP (
      g_input_nof_data_per_sync  => g_input_nof_data_per_sync,
      g_reshape_nof_data_per_blk => g_reshape_nof_data_per_blk,
      g_pipeline_src_out         => g_pipeline_src_out,
      g_pipeline_src_in          => g_pipeline_src_in
    )
    PORT MAP (
      clk           => clk,
      rst           => rst,
    
      snk_in        => in_sosi_arr(I),
      snk_out       => in_siso_arr(I),
    
      src_out       => out_sosi_arr(I),
      src_in        => out_siso_arr(I),
      src_index_arr => out_index_2arr_2(I)
    );
  END GENERATE;
  
  -- Wire index arr
  src_index_2arr_2 <= out_index_2arr_2 WHEN g_shared_control=FALSE ELSE (OTHERS=>out_index_2arr_2(0));

END str;
