-------------------------------------------------------------------------------
--
-- Copyright (C) 2018
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.all;
USE work.dp_stream_pkg.ALL;

-- Author: Eric Kooistra, 16 Mar 2018
-- Purpose:
--   Resize a block by moving the sop and or the eop of the block
-- Description:
--   The output sop and eop can be put at any index of the input block in range(g_input_block_size).
--   Any valid input data before the new sop index and after the new eop index is lost. This
--   component is useful to e.g. remove an unused head or tail of a block.
--
--   * The info at sop (sync, bsn, channel) is preserved by sop_sosi when g_out_sop_index >= 0
--   * The info at eop (empty, err) gets lost if g_out_eop_index < g_input_block_size-1, this is
--     often no issue. If it is an issue then use dp_split instead.
-- Remarks:
-- - This component supports flow control and was designed by keeping the functional
--   state registers and the pipeline registers seperate. Therefore the function is 
--   implemented using combinatorial logic and local state registers to keep its
--   state. The combinatorial function output preserves the snk_in ready latency and
--   is pipelined using dp_pipeline to ease timing closure on the output.
--   A similar approach is also used with dp_counter.vhd.

ENTITY dp_block_resize IS
  GENERIC (
    g_input_block_size : NATURAL := 256;
    g_out_sop_index    : NATURAL := 0;   -- >= 0
    g_out_eop_index    : NATURAL := 255  -- >= g_out_sop_index and < g_input_block_size
  );
  PORT (
    rst          : IN  STD_LOGIC;
    clk          : IN  STD_LOGIC;
    -- ST sink
    snk_out      : OUT t_dp_siso;
    snk_in       : IN  t_dp_sosi;
    -- ST source
    src_in       : IN  t_dp_siso := c_dp_siso_rdy;
    src_out      : OUT t_dp_sosi
  );
END dp_block_resize;


LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.all;
USE work.dp_stream_pkg.ALL;

ARCHITECTURE rtl OF dp_block_resize IS

  SIGNAL cnt_reg      : NATURAL RANGE 0 TO g_input_block_size-1;
  SIGNAL cnt          : NATURAL;
  SIGNAL sop_sosi_reg : t_dp_sosi;
  SIGNAL sop_sosi     : t_dp_sosi;
  SIGNAL block_sosi   : t_dp_sosi;
  
BEGIN

  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      cnt_reg      <= 0;
      sop_sosi_reg <= c_dp_sosi_rst;
    ELSIF rising_edge(clk) THEN
      cnt_reg      <= cnt;
      sop_sosi_reg <= sop_sosi;
    END IF;
  END PROCESS;

  -- Preserve snk_in @ sop  
  sop_sosi <= snk_in WHEN snk_in.sop='1' ELSE sop_sosi_reg;
  
  -- Count valid per block
  p_cnt : PROCESS(snk_in, cnt_reg)
  BEGIN
    cnt <= cnt_reg;
    IF snk_in.sop='1' THEN
      cnt <= 0;
    ELSIF snk_in.valid='1' THEN
      cnt <= cnt_reg + 1;
    END IF;
  END PROCESS;

  -- Resize snk_in combinatorially into block_sosi, so no impact on RL
  p_block_sosi : PROCESS(snk_in, cnt, sop_sosi)
  BEGIN
    -- Default keep snk_in info and data fields
    block_sosi       <= snk_in;
    block_sosi.sync  <= '0';
    block_sosi.sop   <= '0';
    block_sosi.eop   <= '0';
    block_sosi.valid <= '0';
    IF snk_in.valid='1' THEN
      -- Set output sop and preserve info @ sop
      IF cnt=g_out_sop_index THEN
        block_sosi.sop     <= '1';
        block_sosi.sync    <= sop_sosi.sync;
        block_sosi.bsn     <= sop_sosi.bsn;
        block_sosi.channel <= sop_sosi.channel;
      END IF;
      -- Set output eop, info @ eop gets lost if g_out_eop_index < g_input_block_size-1
      IF cnt=g_out_eop_index THEN
        block_sosi.eop   <= '1';
      END IF;
      IF cnt>=g_out_sop_index AND cnt<=g_out_eop_index THEN
        block_sosi.valid <= '1';
      END IF;
    END IF;
  END PROCESS;

  -- Register block_sosi to easy timing closure
  u_pipeline : ENTITY work.dp_pipeline
  GENERIC MAP (
    g_pipeline   => 1  -- 0 for wires, > 0 for registers, 
  )
  PORT MAP (
    rst          => rst,
    clk          => clk,
    -- ST sink
    snk_out      => snk_out,
    snk_in       => block_sosi,
    -- ST source
    src_in       => src_in,
    src_out      => src_out
  );
  
END rtl;
