--------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Future note:
--   The dp_bsn_align aligns at BSN level. In retrospect it is probably 
--   sufficient and easier to align at sync level. This can be done in a new
--   component dp_sync_align. The dp_sync_align then waits for the sync in
--   all streams and then declares alignment. The sync interval is longer than
--   a BSN interval so sync alignment will take longer to achieve. However this
--   is no disadvantage, because typically the system remains aligned after the
--   initial alignment.
--   Per stream it or for all streams together it may be necessary to have
--   monitors to detect misalignment. These can be used to detect:
--   . corrupted blocks due to input FIFO overflow by means of valid cnt from
--     sop to eop = g_block_size and valid cnt from eop to sop = 0.
--   . missing blocks or out of order blocks by means of the incrementing BSN.
--   . sync timeout (similar to sop timeout in dp_bsn_align).
--
-- Purpose:
--   Align frames from multiple input streams
--
-- Description:
--
--   BSN alignment:
--   The frame alignment is based on the BSN of the enabled inputs. Initially
--   and when the BSN differ too much then the inputs get flushed via s_xoff.
--   The inputs are then flushed for g_xoff_timeout cycles. Using a preset
--   timeout is easier and more robust dynamically determining the flush time
--   based on monitoring the inputs states.
--   When the BSN differ less than g_bsn_latency, then the maximum of all BSN
--   is determined and the inputs that have not yet reached this BSN are read 
--   until they have. When all inputs have the same BSN pending, then they are
--   output per frame. After every frame the BSN alignment is checked and the
--   frame output continues if the alignment is still OK, else the function
--   tries to recover the BSN aligmnent.
--   The assumption is that the data blocks arrive almost at the same time at
--   the inputs. The g_bsn_latency defines the maximum number of blocks that
--   their arrival may differ. Therefor each input stream needs an external 
--   FIFO that can at least buffer g_bsn_latency number of blocks, so
--   c_input_fifo_size >= g_bsn_latency * g_block_size. Typically if all inputs
--   are transmitted at the same instant then they will arrive with at within
--   +-1 BSN difference, but it can be more dependent on the buffering and
--   multiplexing that occurs for each input from tx to rx. Default using +-3
--   seems a save marging so g_bsn_latency = 3. Choosing g_bsn_latency larger
--   than needed costs extra logic for the BSN comparison and extra RAM for
--   the input FIFO buffering.
--   The data blocks arrive with gaps in between them, the block size + average
--   gap size is equal to the g_block_period with which the data blocks were
--   generated. This g_block_period is e.g. 256 cycles and is used as a unit
--   for the g_xoff_timeout and the g_sop_timeout.
--
--   Input FIFO size and sop and xoff timeouts:
--
--   For example for:
--     g_bsn_latency      = 3
--     g_block_size       = 32
--     g_block_period     = 256
--   use:
--     c_input_fifo_size >= (g_bsn_latency + 2) * g_block_size  -- of external FIFO
--     g_sop_timeout      = (g_bsn_latency + 1) * g_block_period
--     g_xoff_timeout     =  g_bsn_latency * 2  * g_block_period
--
--   The g_sop_timeout makes sure that an attempt to align the input stream is
--   ended if it does not succeed within g_sop_timeout clock cycles. The
--   worst case time to align is about g_block_period * g_bsn_latency, because
--   more time difference between the input streams can not occur during 
--   normal operation. Choose g_sop_timeout +1 period larger.
--   The g_sop_timeout also avoids that the input FIFOs will get full.
--   Choosing g_sop_timeout too large may cause some input FIFOs to get full,
--   but that is probably no problem because the BSN aligner will then still
--   recover anyway. Therefore choose the input FIFO size a little larger
--   to fit all blocks that can arrive during a sop timeout, hence choose
--   c_input_fifo_size +2 blocks larger.
--   To recover from a failed alignment the xoff is used. The xoff makes sure
--   that the rx of new input frames is stopped and that the input FIFOs get
--   flushed. The g_xoff_timeout must be long enough to ensure that all input
--   FIFOs are empty. Hence typically g_xoff_timeout must be equal to the input
--   FIFO size, so about g_bsn_latency*g_block_size. However it is bad to flush
--   to short and fine to flush longer than needed. Therefor add a factor 2
--   extra to g_xoff_timeout and use g_block_period instead of g_block_size.
--
--   External input fill FIFO:
--   The BSN aligner can only cope with gaps in the snk_in_arr valid if they
--   occur the same for all inputs. Otherwise the inputs need to be put through
--   a dp_fifo_fill that fills >= 1 * g_block_period to ensure that after the
--   alignment at the sop the inputs deliver valid data for the whole frame,
--   so for every src_in ready there then is valid output.
--
-- Remark:
-- . Using g_nof_input=1 is transparent.
-- . The xoff is used for coarse flow control over blocks of data to recover
--   from start up mis alignment or an rx input that briefly misses a burst of
--   data. The ready is used for the fine flow per clock cycle to align on the
--   inputs at their sop's.
-- . The assumption is that data from the upstream inputs can be stopped via
--   xoff or have a dp_flush that can be enabled via xoff to empty all input
--   streams.
-- . The number of outputs is also g_nof_input. All outputs, including those
--   for the inputs that are not enabled, have the same sosi control signalling
--   (sync, valid, sop and eop). Therefore a downstream sink only needs to use
--   the control from one output, typically output 0. For the inputs that are
--   not enabled the sosi data fields are forced to 0.
-- . The input block sizes must match the g_block_size. Using g_block_size an
--   internally generated reference block is used to set the sosi control 
--   during the frame output. The advantage of using this internal reference
--   block generator is that it does not depend on the in_en_arr settings.
-- . It was also considered to have a state machine per input, but later on it
--   was though easier to use only a single state machine and vectored input 
--   control. 
--

ENTITY dp_bsn_align IS
  GENERIC (
    g_block_size           : NATURAL := 32;  -- > 1, g_block_size=1 is not supported, g_block_size <= g_block_period
    g_block_period         : NATURAL := 256; -- number of clock cycles per block period or block interval (not used in the architecture, used in the comment only)
    g_nof_input            : NATURAL := 2;   -- >= 1
    g_xoff_timeout         : NATURAL := 1536;  -- e.g.  g_bsn_latency * 2  * g_block_period
    g_sop_timeout          : NATURAL := 1024;  -- e.g. (g_bsn_latency + 1) * g_block_period
    g_bsn_latency          : NATURAL := 3;
    g_bsn_request_pipeline : NATURAL := 2    -- = 1, 2, ..., c_bsn_nof_stages, requested total pipelining of the BSN max and BSN min operation
  );
  PORT (
    rst         : IN  STD_LOGIC;
    clk         : IN  STD_LOGIC;
    -- ST sinks
    snk_out_arr : OUT t_dp_siso_arr(g_nof_input-1 DOWNTO 0);
    snk_in_arr  : IN  t_dp_sosi_arr(g_nof_input-1 DOWNTO 0);
    -- ST source
    src_in_arr  : IN  t_dp_siso_arr(g_nof_input-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);  -- only src_in_arr(0) is used for src_out flow control because all src_out need to stay aligned
    src_out_arr : OUT t_dp_sosi_arr(g_nof_input-1 DOWNTO 0);
    -- MM
    in_en_evt   : IN  STD_LOGIC := '0';                                          -- pulse '1' indicates that the in_en_arr user input enables have been updated
    in_en_arr   : IN  STD_LOGIC_VECTOR(g_nof_input-1 DOWNTO 0) := (OTHERS=>'1')  -- default all user inputs are enabled
  );
END dp_bsn_align;


ARCHITECTURE rtl OF dp_bsn_align IS

  CONSTANT c_timeout_max            : NATURAL := largest(g_xoff_timeout, g_sop_timeout);
  CONSTANT c_timeout_w              : NATURAL := ceil_log2(c_timeout_max+1);
  
  -- Use as much BSN range as necessary. At the end of the range the limited range BSN will wrap. This will cause g_bsn_latency out of 2**c_bsn_align_w
  -- possible limited BSN values to fail alignment initially, but for these instants the alignment will be possible some BSN later. Using +2 provides
  -- sufficient opportunity for BSN alignment at the first pend_sop_all attempt and certainly at the next.
  CONSTANT c_bsn_align_w            : NATURAL := ceil_log2(g_bsn_latency) + 2;
  
  -- Control pipelining of pend_bsn_max, pend_bsn_max and pend_sop_all to ease achieving timing closure for comparing the BSN of g_nof_input.
  -- . Use g_bsn_request_pipeline=1 for fastest alignment detection with only 1 cycle lost
  -- . Use g_bsn_request_pipeline>1 to more easily achieve timing closure at expense of more cycles lost
  -- . The maximum g_bsn_request_pipeline = c_bsn_nof_stages provided that there is sufficient data not valid gap between the input packets to allow for these lost cycles
  CONSTANT c_bsn_nof_stages         : NATURAL := ceil_log2(g_nof_input);  -- = nof operation stages for determining max BSN and min BSN
  CONSTANT c_bsn_stage_pipeline     : NATURAL := 1;                       -- = fixed 1, always use at least one register
  CONSTANT c_bsn_request_pipeline   : NATURAL := smallest(c_bsn_nof_stages, g_bsn_request_pipeline);
  CONSTANT c_bsn_stage_pipeline_mod : NATURAL := c_bsn_nof_stages / c_bsn_request_pipeline;
  CONSTANT c_bsn_actual_pipeline    : NATURAL := c_bsn_nof_stages / c_bsn_stage_pipeline_mod;  -- = 1, 2, ..., c_bsn_nof_stages, actual nof operation stages that will get a register
  
  TYPE t_state IS (s_xoff, s_align, s_data);
  
  -- Define the local registers in t_reg record to more clearly show the single registered process, conform the two process method of Gaisler. However
  -- still do use multiple combinatorial processes, instead of the only one combinatiorial process of the Gaisler method that uses variables.
  TYPE t_reg IS RECORD  -- local registers
    in_en_arr          : STD_LOGIC_VECTOR(g_nof_input-1 DOWNTO 0);
    timeout_cnt        : STD_LOGIC_VECTOR(c_timeout_w-1 DOWNTO 0);
    pend_sop_all       : STD_LOGIC_VECTOR(c_bsn_actual_pipeline-1 DOWNTO 0);
    state              : t_state;
    blk_buf            : t_dp_sosi;
    src_buf_arr        : t_dp_sosi_arr(g_nof_input-1 DOWNTO 0);
    src_out_arr        : t_dp_sosi_arr(g_nof_input-1 DOWNTO 0);
  END RECORD;
  
  CONSTANT c_reg_rst        : t_reg := ((OTHERS=>'0'), (OTHERS=>'0'), (OTHERS=>'0'), s_xoff, c_dp_sosi_rst, (OTHERS=>c_dp_sosi_rst), (OTHERS=>c_dp_sosi_rst));
  
  SIGNAL r                  : t_reg;
  SIGNAL nxt_r              : t_reg;
  
  SIGNAL in_en_ack          : STD_LOGIC;
  SIGNAL in_en_new          : STD_LOGIC;
  
  SIGNAL blk_en             : STD_LOGIC;
  SIGNAL blk_siso           : t_dp_siso;
  SIGNAL blk_sosi           : t_dp_sosi;
  SIGNAL next_blk_buf       : t_dp_sosi;
  SIGNAL hold_blk_in        : t_dp_siso;
  
  SIGNAL pend_bsn_arr       : t_natural_arr(g_nof_input-1 DOWNTO 0);
  SIGNAL pend_bsn_vec       : STD_LOGIC_VECTOR(g_nof_input*c_bsn_align_w-1 DOWNTO 0);
  SIGNAL pend_bsn_max       : STD_LOGIC_VECTOR(c_bsn_align_w-1 DOWNTO 0);
  SIGNAL pend_bsn_min       : STD_LOGIC_VECTOR(c_bsn_align_w-1 DOWNTO 0);
  
  SIGNAL i_snk_out_arr      : t_dp_siso_arr(g_nof_input-1 DOWNTO 0);
  SIGNAL next_src_buf_arr   : t_dp_sosi_arr(g_nof_input-1 DOWNTO 0);
  SIGNAL pend_src_buf_arr   : t_dp_sosi_arr(g_nof_input-1 DOWNTO 0);
  SIGNAL hold_src_in_arr    : t_dp_siso_arr(g_nof_input-1 DOWNTO 0);
  
  SIGNAL src_in             : t_dp_siso;
  
BEGIN

  snk_out_arr <= i_snk_out_arr;
  src_out_arr <= r.src_out_arr;

  src_in <= src_in_arr(0);  -- all src_in_arr are identical, so use element (0)
  
  p_clk: PROCESS(clk, rst)
  BEGIN
    IF rst='1' THEN
      r <= c_reg_rst;
    ELSIF rising_edge(clk) THEN
      r <= nxt_r;
    END IF;
  END PROCESS;
  
  -- control user input enable updates to not occur during a block
  u_in_en_new : ENTITY common_lib.common_switch
  GENERIC MAP (
    g_rst_level    => '0',
    g_priority_lo  => FALSE,
    g_or_high      => FALSE,
    g_and_low      => FALSE
  )
  PORT MAP (
    rst         => rst,
    clk         => clk,
    switch_high => in_en_evt,
    switch_low  => in_en_ack,
    out_level   => in_en_new
  );
  
  -- block reference
  u_block_gen : ENTITY work.dp_block_gen
  GENERIC MAP (
    g_nof_data => g_block_size,
    g_empty    => 0,
    g_channel  => 0,
    g_error    => 0
  )
  PORT MAP (
    rst        => rst,
    clk        => clk,
    -- Streaming source
    src_in     => blk_siso,
    src_out    => blk_sosi,
    -- MM control
    en         => blk_en
  );

  -- Hold the sink input to be able to register the source output
  u_block_hold : ENTITY work.dp_hold_input
  PORT MAP (
    rst          => rst,
    clk          => clk,
    -- ST sink
    snk_out      => blk_siso,  -- wired blk_siso = hold_blk_in
    snk_in       => blk_sosi,
    -- ST source
    src_in       => hold_blk_in,
    next_src_out => next_blk_buf,
    pend_src_out => OPEN,
    src_out_reg  => r.blk_buf
  );
    
  gen_inputs : FOR I IN g_nof_input-1 DOWNTO 0 GENERATE
    u_hold : ENTITY work.dp_hold_input
    PORT MAP (
      rst          => rst,
      clk          => clk,
      -- ST sink
      snk_out      => OPEN,
      snk_in       => snk_in_arr(I),
      -- ST source
      src_in       => hold_src_in_arr(I),
      next_src_out => next_src_buf_arr(I),
      pend_src_out => pend_src_buf_arr(I),
      src_out_reg  => r.src_buf_arr(I)
    );
  END GENERATE;
  
  -- This is not Erlang or Haskell, but that does not mean that we can not do some functional programming in VHDL as well.  
  -- By using functions that operate on the t_dp_siso_arr and t_dp_sosi_arr we avoid FOR-LOOPs here. However using
  -- functions can obscure the understanding of the code. Therefore it is important to define the functions such that they
  -- are generic so that they can be reused.
  
  -- User input enable control
  nxt_r.in_en_arr <= in_en_arr WHEN in_en_ack='1' ELSE r.in_en_arr;

  -- Detect whether the sop is pending at the enabled user inputs, yields '1' when all enabled user inputs show a sop pending
  gen_reg : IF c_bsn_actual_pipeline=1 GENERATE
    nxt_r.pend_sop_all(0) <= func_dp_stream_arr_and(pend_src_buf_arr, r.in_en_arr, "SOP");
  END GENERATE;
  gen_pipe : IF c_bsn_actual_pipeline>1 GENERATE
    nxt_r.pend_sop_all    <= func_dp_stream_arr_and(pend_src_buf_arr, r.in_en_arr, "SOP") & r.pend_sop_all(c_bsn_actual_pipeline-1 DOWNTO 1);
  END GENERATE;
    
  -- Determine pending bsn range for the enabled user inputs, it is valid at the r.pend_sop_all(0)  
  p_pend_bsn_arr : PROCESS(pend_src_buf_arr)
  BEGIN
    FOR I IN g_nof_input-1 DOWNTO 0 LOOP
      pend_bsn_vec((I+1)*c_bsn_align_w-1 DOWNTO I*c_bsn_align_w) <=         pend_src_buf_arr(I).bsn(c_bsn_align_w-1 DOWNTO 0);
      pend_bsn_arr(I)                                            <= TO_UINT(pend_src_buf_arr(I).bsn(c_bsn_align_w-1 DOWNTO 0));  -- for debugging purposes in wave window
    END LOOP;
  END PROCESS;
    
  -- Use tree instead of folding to ease timing closure:
  --   pend_bsn_max <= func_dp_stream_arr_bsn_max(pend_src_buf_arr, r.in_en_arr, c_bsn_align_w);
  u_pend_bsn_max : ENTITY common_lib.common_operation_tree
  GENERIC MAP (
    g_operation      => "MAX",
    g_representation => "UNSIGNED",
    g_pipeline       => c_bsn_stage_pipeline,          -- amount of output pipelining per stage
    g_pipeline_mod   => c_bsn_stage_pipeline_mod,      -- only pipeline the stage output by g_pipeline when the stage number MOD g_pipeline_mod = 0
    g_nof_inputs     => g_nof_input,
    g_dat_w          => c_bsn_align_w
  )
  PORT MAP (
    clk         => clk,
    in_data_vec => pend_bsn_vec,
    in_en_vec   => r.in_en_arr,
    result      => pend_bsn_max
  );
  
  -- Use tree instead of folding to ease timing closure:
  --   pend_bsn_min <= func_dp_stream_arr_bsn_min(pend_src_buf_arr, r.in_en_arr, c_bsn_align_w);
  u_pend_bsn_min : ENTITY common_lib.common_operation_tree
  GENERIC MAP (
    g_operation      => "MIN",
    g_representation => "UNSIGNED",
    g_pipeline       => c_bsn_stage_pipeline,          -- amount of output pipelining per stage
    g_pipeline_mod   => c_bsn_stage_pipeline_mod,      -- only pipeline the stage output by g_pipeline when the stage number MOD g_pipeline_mod = 0
    g_nof_inputs     => g_nof_input,
    g_dat_w          => c_bsn_align_w
  )
  PORT MAP (
    clk         => clk,
    in_data_vec => pend_bsn_vec,
    in_en_vec   => r.in_en_arr,
    result      => pend_bsn_min
  );
  
  -- Hold input registers
  nxt_r.blk_buf     <= next_blk_buf;
  nxt_r.src_buf_arr <= next_src_buf_arr;
  
  p_state : PROCESS(r, src_in, pend_src_buf_arr, pend_bsn_max, pend_bsn_min, next_src_buf_arr, next_blk_buf, in_en_new)
    VARIABLE v_siso_arr_hold  : t_dp_siso_arr(g_nof_input-1 DOWNTO 0);
    VARIABLE v_siso_arr_rdy   : t_dp_siso_arr(g_nof_input-1 DOWNTO 0);
    VARIABLE v_siso_arr_flush : t_dp_siso_arr(g_nof_input-1 DOWNTO 0);
    VARIABLE v_siso_arr_src   : t_dp_siso_arr(g_nof_input-1 DOWNTO 0);
    VARIABLE v_sosi_arr       : t_dp_sosi_arr(g_nof_input-1 DOWNTO 0);  -- auxiliary sosi variable
    VARIABLE v_blk_buf        : t_dp_sosi;                              -- auxiliary sosi variable
    CONSTANT c_init_all_zeros : STD_LOGIC_VECTOR(c_dp_stream_dsp_data_w-1 DOWNTO 0) := array_init('0', c_dp_stream_dsp_data_w);    
  BEGIN
    -- default semi-constant v_siso_arr controls taking account of the dynamic r.in_en_arr
    v_siso_arr_hold  := func_dp_stream_arr_select(r.in_en_arr, c_dp_siso_hold,  c_dp_siso_flush);
    v_siso_arr_rdy   := func_dp_stream_arr_select(r.in_en_arr, c_dp_siso_rdy,   c_dp_siso_flush);
    v_siso_arr_flush := func_dp_stream_arr_select(r.in_en_arr, c_dp_siso_flush, c_dp_siso_flush);  -- = (OTHERS=>c_dp_siso_flush)
    v_siso_arr_src   := func_dp_stream_arr_select(r.in_en_arr, src_in,          c_dp_siso_flush);
    
    -- assume if any input has a sync then all enabled inputs will have the sync
    v_blk_buf      := next_blk_buf;
    v_blk_buf.sync := vector_or(func_dp_stream_arr_get(next_src_buf_arr, "SYNC"));
    
    -- control user input enable
    in_en_ack <= '0';
    
    -- control reference block input
    blk_en <= '0';
    
    -- default reset timeout
    nxt_r.timeout_cnt <= (OTHERS=>'0');
  
    -- default flow control flush reference block input and all user inputs
    hold_blk_in     <= c_dp_siso_flush;
    i_snk_out_arr   <= v_siso_arr_flush;
    hold_src_in_arr <= v_siso_arr_flush;
    
    nxt_r.state       <= r.state;
    nxt_r.src_out_arr <= func_dp_stream_arr_reset_control(r.src_out_arr);  -- default hold sosi data and reset sosi control
    
    CASE r.state IS
      WHEN s_xoff =>
        in_en_ack <= '1';  -- accept any in_en_arr updates
        -- remain in s_xoff for g_xoff_timeout cycles 
        IF UNSIGNED(r.timeout_cnt)=g_xoff_timeout THEN
          hold_blk_in     <= c_dp_siso_rdy;
          i_snk_out_arr   <= v_siso_arr_rdy;
          hold_src_in_arr <= v_siso_arr_rdy;
          blk_en          <= '1';                -- enable reference block input
          nxt_r.state     <= s_align;
        ELSE
          nxt_r.timeout_cnt <= INCR_UVEC(r.timeout_cnt, 1);
        END IF;
        
      WHEN s_align =>
        blk_en <= '1';  -- keep reference block input enabled
        
        hold_blk_in     <= c_dp_siso_hold;       -- the reference block sop is already there, because blk_siso.ready was '1' in previous state and dp_block_gen can always deliver valid data
        i_snk_out_arr   <= v_siso_arr_rdy;       -- wait for sop at the user inputs and then hold the sop of the enabled user inputs
        hold_src_in_arr <= v_siso_arr_hold;
        FOR I IN g_nof_input-1 DOWNTO 0 LOOP
          IF pend_src_buf_arr(I).sop='1' THEN
            i_snk_out_arr(I) <= c_dp_siso_hold;  -- flush user input until sop, then stop further input and hold this sosi
          END IF;
        END LOOP;
        
        -- remain in s_align until all enabled user inputs have a sop pending  
        nxt_r.timeout_cnt <= INCR_UVEC(r.timeout_cnt, 1);
        IF r.pend_sop_all(0)='1' THEN
          -- now i_snk_out_arr = v_siso_arr_hold
          -- use registered r.pend_sop_all to be able to use registered pend_bsn_max and pend_bsn_min to ease timing closure at the expense of an idle clock cycle
          IF UNSIGNED(pend_bsn_max)=UNSIGNED(pend_bsn_min) THEN
            -- bsn are aligned, so start outputing the frames
            
            -- >>> Begin of same code as in s_data
            -- let downstream flow control the enabled inputs
            hold_blk_in     <= src_in;
            i_snk_out_arr   <= v_siso_arr_src;
            hold_src_in_arr <= v_siso_arr_src;
            
            -- output the aligned user input frames
            v_sosi_arr := (OTHERS=>c_dp_sosi_rst);
            v_sosi_arr := func_dp_stream_arr_select(r.in_en_arr, next_src_buf_arr, v_sosi_arr);
            v_sosi_arr := func_dp_stream_arr_set_control(v_sosi_arr, v_blk_buf);
            v_sosi_arr := func_dp_stream_arr_copy_valid_bsn(v_sosi_arr, r.in_en_arr);           -- Copy the BSN of one of the valid streams to all other streams
            v_sosi_arr := func_dp_stream_set_data(v_sosi_arr, c_init_all_zeros, "ALL", r.in_en_arr); -- Fill the RE and IM fields of the disabled channels with ones.
            nxt_r.src_out_arr <= v_sosi_arr;
            -- <<< End of same code as in s_data
            
            nxt_r.state     <= s_data;
          ELSIF UNSIGNED(pend_bsn_max)<=UNSIGNED(pend_bsn_min)+g_bsn_latency THEN
            -- bsn differ but are within alignment range
            FOR I IN g_nof_input-1 DOWNTO 0 LOOP
              IF UNSIGNED(pend_src_buf_arr(I).bsn(c_bsn_align_w-1 DOWNTO 0)) < UNSIGNED(pend_bsn_max) THEN
                i_snk_out_arr(I)   <= c_dp_siso_rdy;  -- flush the pending sop and remain in s_align to wait for the next sop, then stop further input and hold this sosi
                hold_src_in_arr(I) <= c_dp_siso_rdy;
              END IF;
            END LOOP;
          ELSE
            -- bsn differ too much, so flush all inputs and try to align again
            nxt_r.timeout_cnt <= (OTHERS=>'0');
            nxt_r.state <= s_xoff;
          END IF;
        -- check user input enable control (must use 'ELSIF' and not 'END IF; IF', to properly handle case where both conditions are TRUE at same clock cycle)
        ELSIF in_en_new='1' THEN
          -- need to use updated user input enable settings, so flush all inputs and try to align again
          nxt_r.timeout_cnt <= (OTHERS=>'0');
          nxt_r.state <= s_xoff;
        ELSIF UNSIGNED(r.timeout_cnt)=g_sop_timeout THEN
          -- input sop timeout, so flush all inputs and try to align again
          nxt_r.timeout_cnt <= (OTHERS=>'0');
          nxt_r.state <= s_xoff;
        END IF;

      WHEN OTHERS =>  -- s_data
        blk_en <= '1';  -- keep reference block input enabled
        
        -- let downstream flow control the enabled inputs
        hold_blk_in     <= src_in;
        i_snk_out_arr   <= v_siso_arr_src;
        hold_src_in_arr <= v_siso_arr_src;
        
        -- output the aligned user input frames
        v_sosi_arr := (OTHERS=>c_dp_sosi_rst);                                               -- default force the sosi data of unused user inputs to 0
        v_sosi_arr := func_dp_stream_arr_select(r.in_en_arr, next_src_buf_arr, v_sosi_arr);  -- output the pending input frame of all enabled user inputs
        v_sosi_arr := func_dp_stream_arr_set_control(v_sosi_arr, v_blk_buf);                 -- use the same reference block sosi control for all outputs
        v_sosi_arr := func_dp_stream_arr_copy_valid_bsn(v_sosi_arr, r.in_en_arr);            -- Copy the BSN of one of the valid streams to all other streams
        v_sosi_arr := func_dp_stream_set_data(v_sosi_arr, c_init_all_zeros, "ALL", r.in_en_arr); -- Fill the RE and IM fields of the disabled channels with ones.
        nxt_r.src_out_arr <= v_sosi_arr;
        
        -- at end of every block
        IF v_blk_buf.eop='1' THEN
          -- check bsn alignment
          nxt_r.state <= s_align;
        
          -- check user input enable control
          IF in_en_new='1' THEN
            -- need to use updated user input enable settings, so flush all inputs and try to align again
            nxt_r.state <= s_xoff;
          END IF;
        END IF;
    END CASE;     
  END PROCESS;
    
END rtl;
