-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Provide MM slave register for dp_bsn_align
-- Description:
--
--  Read/Write registers for all streams. The number of registers dpeneds on the
--  g_nof_inputs that are used. Each input has it's own enable register.
--
--   31             24 23             16 15              8 7               0  wi
--  |-----------------|-----------------|-----------------|-----------------|
--  |                                                   enable_input0 = [0] |  0
--  |-----------------------------------------------------------------------|
--  |                                                   enable_input1 = [0] |  1
--  |-----------------------------------------------------------------------|
--  |                                                   enable_input2 = [0] |  2
--  |-----------------------------------------------------------------------|
--                                      |   
--                                      |
--  |-----------------------------------------------------------------------|
--  |                                                  enable_input.. = [0] |  g_nof_inputs-1
--  |-----------------------------------------------------------------------|


LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;

ENTITY dp_bsn_align_reg IS
  GENERIC (
    g_nof_input          : POSITIVE := 3; 
    g_cross_clock_domain : BOOLEAN  := TRUE  -- use FALSE when mm_clk and st_clk are the same, else use TRUE to cross the clock domain
  );
  PORT (
    -- Clocks and reset
    mm_rst     : IN  STD_LOGIC;   -- reset synchronous with mm_clk
    mm_clk     : IN  STD_LOGIC;   -- memory-mapped bus clock
    st_rst     : IN  STD_LOGIC;   -- reset synchronous with st_clk
    st_clk     : IN  STD_LOGIC;   -- other clock domain clock
    
    -- Memory Mapped Slave in mm_clk domain
    sla_in     : IN  t_mem_mosi;  -- actual ranges defined by c_mm_reg
    sla_out    : OUT t_mem_miso;  -- actual ranges defined by c_mm_reg
    
    -- MM registers in st_clk domain
    out_en_evt : OUT  STD_LOGIC;  -- pulses when new enable data is available
    out_en_arr : OUT  STD_LOGIC_VECTOR
  );
END dp_bsn_align_reg;


ARCHITECTURE str OF dp_bsn_align_reg IS

  -- Define the actual size of the MM slave register
  CONSTANT c_mm_reg : t_c_mem := (latency  => 1,
                                  adr_w    => ceil_log2(g_nof_input),
                                  dat_w    => c_word_w,       -- Use MM bus data width = c_word_w = 32 for all MM registers
                                  nof_dat  => g_nof_input,
                                  init_sl  => '0');

  --  FUNCTION array_init(init : NATURAL;   nof, width : NATURAL) RETURN STD_LOGIC_VECTOR;  -- useful to init an unconstrained std_logic_vector with repetitive content
  CONSTANT c_reg_init    : STD_LOGIC_VECTOR(g_nof_input*c_word_w-1 DOWNTO 0) := array_init(1, g_nof_input, c_word_w);  

  -- Registers in st_clk domain
  SIGNAL out_en_arr_reg  : STD_LOGIC_VECTOR(g_nof_input*c_word_w-1 DOWNTO 0); -- := (OTHERS => '1');
  SIGNAL reg_wr_arr      : STD_LOGIC_VECTOR(g_nof_input-1 DOWNTO 0) := (OTHERS => '0');  

BEGIN

  u_reg : ENTITY common_lib.common_reg_r_w_dc
  GENERIC MAP (
    g_cross_clock_domain => g_cross_clock_domain,
    g_in_new_latency     => 1,   
    g_readback           => TRUE,
    g_reg                => c_mm_reg,
    g_init_reg           => RESIZE_UVEC(c_reg_init, c_mem_reg_init_w) --RESIZE_UVEC(TO_UVEC(1, c_word_w), c_mem_reg_init_w)
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,
    st_rst      => st_rst,
    st_clk      => st_clk,
    
    -- Memory Mapped Slave in mm_clk domain
    sla_in      => sla_in,
    sla_out     => sla_out,
    
    -- MM registers in st_clk domain
    reg_wr_arr  => reg_wr_arr,
    reg_rd_arr  => OPEN,
    in_new      => OPEN,
    in_reg      => out_en_arr_reg, -- read 
    out_reg     => out_en_arr_reg, -- write
    out_new     => out_en_evt
  );
  
  gen_out_arr : FOR I IN 0 TO g_nof_input-1 GENERATE
    out_en_arr(I) <= out_en_arr_reg(I*c_word_w);
  END GENERATE;
  
END str;

