-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose : Delay the input sync and BSN so they remain aligned to the data
--           path processing.
-- Description:
--   The out_release of the data path processing forms the release trigger for
--   the sync and the BSN. The delay is static (fixed) regarding the number of
--   blocks and can be 0 or more blocks and is set by g_nof_block_latency. The
--   delay is dynamic regarding the back pressure ready signal and can handle
--   any data processing logic latency >= 1 clock cycles.
-- Issues:
-- . The in_sync can occur during data invalid, so when in_eop = '0'. The 
--   BSN timing must be robust for this.
-- . Connecting out_release to an early out_sop can cause that the out_sync
--   occurs during data invalid.
-- . Connecting out_release to out_eop ensures that the out_sync occurs during
--   valid but can cause the output sync and BSN to be 1 block early in case
--   the out_eop occurs before the in_eop (e.g. if the DP block outputs smaller
--   blocks than that it gets at its input).
-- ==> The dp_latency_adapter can now pass on a sync during data not valid. So
--     it is now possible to connect out_release to an early out_sop.
-- Remarks:
-- . The sync and BSN can not be delayed by a fixed latency, because the data
--   path processing may use dynamic back pressure using the ready signal.
--   Therefore this dp_bsn_delay uses out_release to release the sync and BSN.
-- . About dynamic block delay support. 
--   - A data path function that can store and forward a dynamic number of
--     blocks can not use this dp_bsn_delay, because g_nof_block_latency is a
--     generic. Perhaps the block latency can be made dynamic by monitoring the
--     number of out_release that occured relative to the number of in_eop. The
--     g_nof_block_latency then sets the maximum that is supported.
--   - Alternatively, and probably better, the sync and BSN can be put through
--     a separate FIFO that is written by the in_eop and read by the
--     out_release. The FIFO can then have RL=0 so that the read sync becomes
--     out_sync directly and the read bsn then becomes out_bsn after a register
--     stage.
--     Putting the sync and BSN through the same FIFO as the data and sop, eop
--     consumes quite some RAM since the BSN is typically 48 bit wide and it
--     then needs to be pass through through the FIFO for every valid clock
--     cycle.

ENTITY dp_bsn_delay IS
  GENERIC (
    g_nof_block_latency  : NATURAL := 0;  -- >= 0
    g_bsn_w              : NATURAL := 48  -- use g_bsn_w to allow leaving in_bsn NC
  );
  PORT (
    rst         : IN  STD_LOGIC;
    clk         : IN  STD_LOGIC;
    in_sync     : IN  STD_LOGIC;
    in_bsn      : IN  STD_LOGIC_VECTOR(g_bsn_w-1 DOWNTO 0) := (OTHERS=>'0');
    in_eop      : IN  STD_LOGIC;
    out_release : IN  STD_LOGIC;   -- input from the data path processing, used as output release pulse
    out_sync    : OUT STD_LOGIC;
    out_bsn     : OUT STD_LOGIC_VECTOR(g_bsn_w-1 DOWNTO 0)
  );
END dp_bsn_delay;


ARCHITECTURE rtl OF dp_bsn_delay IS

  TYPE t_bsn_arr  IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(g_bsn_w-1 DOWNTO 0);
  
  SIGNAL hold_sync        : STD_LOGIC;
  SIGNAL sync_dly         : STD_LOGIC_VECTOR(0 TO g_nof_block_latency);  -- [0] is hold_sync, the combinatorial in_sync is not delayed, because the DP latency must be > 0
  SIGNAL nxt_sync_dly     : STD_LOGIC_VECTOR(1 TO g_nof_block_latency);
  
  SIGNAL bsn_dly          : t_bsn_arr(0 TO g_nof_block_latency);         -- [0] is combinatorial in_bsn
  SIGNAL nxt_bsn_dly      : t_bsn_arr(1 TO g_nof_block_latency);
  
  SIGNAL bsn_reg          : STD_LOGIC_VECTOR(g_bsn_w-1 DOWNTO 0);
  SIGNAL nxt_bsn_reg      : STD_LOGIC_VECTOR(g_bsn_w-1 DOWNTO 0);
  
BEGIN

  p_clk : PROCESS(clk, rst)
  BEGIN
    IF rst='1' THEN
      sync_dly(1 TO g_nof_block_latency)  <= (OTHERS=>'0');
      bsn_dly( 1 TO g_nof_block_latency)  <= (OTHERS=>(OTHERS=>'0'));
      bsn_reg                             <= (OTHERS=>'0');
    ELSIF rising_edge(clk) THEN
      sync_dly(1 TO g_nof_block_latency)  <= nxt_sync_dly;
      bsn_dly( 1 TO g_nof_block_latency)  <= nxt_bsn_dly;
      bsn_reg                             <= nxt_bsn_reg;
    END IF;
  END PROCESS;
  
  -- Hold the in_sync during the block until the in_eop
  u_hold_sync : ENTITY common_lib.common_switch
  GENERIC MAP (
    g_priority_lo  => FALSE,  -- in_sync has priority over in_eop, because they may occur simultaneously
    g_or_high      => FALSE,  -- hold_sync goes high after in_sync
    g_and_low      => FALSE   -- hold_sync goes low  after in_eop
  )
  PORT MAP (
    rst         => rst,
    clk         => clk,
    switch_high => in_sync,
    switch_low  => in_eop,
    out_level   => hold_sync
  );
  
  -- Delay line for in_sync and in_bsn at block level
  sync_dly(0)  <= hold_sync;
  nxt_sync_dly <= sync_dly(0 TO g_nof_block_latency-1) WHEN in_eop='1' ELSE sync_dly(1 TO g_nof_block_latency);
  
  bsn_dly(0)   <= in_bsn;
  nxt_bsn_dly  <= bsn_dly( 0 TO g_nof_block_latency-1) WHEN in_eop='1' ELSE bsn_dly( 1 TO g_nof_block_latency);
  
  -- Output block alignment using the out_release
  out_sync    <= sync_dly(g_nof_block_latency) WHEN out_release='1' ELSE '0';      -- combinatorial pulse
  nxt_bsn_reg <= bsn_dly(g_nof_block_latency)  WHEN out_release='1' ELSE bsn_reg;  -- register the BSN to hold it during the output block
  out_bsn     <= bsn_reg;
  
END rtl;

