-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Monitor the sosi and siso control of a stream per sync interval
-- Description:
--   The mon_evt pulses when new monitor output is available regarding the
--   previous in_sosi.sync interval:
--   . mon_sync_timeout        = '1' when the in_sosi.sync did not occur
--   . mon_ready_stable        = '1' when ready was always '1' during last sync interval
--   . mon_xon_stable          = '1' when xon   was always '1' during last sync interval
--   . mon_bsn_at_sync         = BSN at sync
--   . mon_nof_sop             = number of sop during last sync interval
--   . mon_nof_err             = number of err at eop during last sync interval
--   . mon_nof_valid           = number of valid during last sync interval;
--   . mon_bsn_first           = BSN at the first SOP that came in (When g_log_first_bsn = TRUE)
--   . mon_bsn_first_cycle_cnt = number of cycles after rst at which mon_bsn_first was logged (When g_log_first_bsn = TRUE)
--
--   . mon_bsn_first           = BSN at the moment sync_in is pulsed (When g_log_first_bsn = FALSE)
--   . mon_bsn_first_cycle_cnt = number of cycles after the last sop befor sync_in is pulsed (When g_log_first_bsn = FALSE)
--
-- Remarks:
-- . Assumes RL > 0 so each active valid indicates a new data (with RL = 0 the
--   valid remains active until an acknowledge by ready)

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.all;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;


ENTITY dp_bsn_monitor IS
  GENERIC (
    g_sync_timeout  : NATURAL := 200*10**6;  -- choose >= nof clk cycles per sync period
    g_error_bi      : NATURAL := 0;
    g_log_first_bsn : BOOLEAN := TRUE -- Logs first BSN + cycle count. Useful when using BSN aligner.
  );
  PORT (
    rst                     : IN  STD_LOGIC := '0';
    clk                     : IN  STD_LOGIC;

    -- ST interface
    in_siso                 : IN  t_dp_siso;
    in_sosi                 : IN  t_dp_sosi;
    sync_in                 : IN  STD_LOGIC := '0';  -- Sync input that samples the current bsn number and the number of clockcycles that have expired since the last sop    
    -- MM interface
    -- . control   
    mon_evt                 : OUT STD_LOGIC;  -- pulses when new monitor output is available regarding the previous sync interval
    mon_sync                : OUT STD_LOGIC;  -- pulses every in_sosi.sync interval
    mon_sync_timeout        : OUT STD_LOGIC;
    -- . siso
    mon_ready_stable        : OUT STD_LOGIC;
    mon_xon_stable          : OUT STD_LOGIC;
    -- . sosi
    mon_bsn_at_sync         : OUT STD_LOGIC_VECTOR;
    mon_nof_sop             : OUT STD_LOGIC_VECTOR;
    mon_nof_err             : OUT STD_LOGIC_VECTOR;
    mon_nof_valid           : OUT STD_LOGIC_VECTOR;
    mon_bsn_first           : OUT STD_LOGIC_VECTOR;
    mon_bsn_first_cycle_cnt : OUT STD_LOGIC_VECTOR
  );
END dp_bsn_monitor;


ARCHITECTURE rtl OF dp_bsn_monitor IS

  CONSTANT c_sync_timeout_w : NATURAL := ceil_log2(g_sync_timeout);
  CONSTANT c_bsn_w          : NATURAL := mon_bsn_at_sync'LENGTH;
  CONSTANT c_cnt_sop_w      : NATURAL := mon_nof_sop'LENGTH;
  CONSTANT c_cnt_valid_w    : NATURAL := mon_nof_valid'LENGTH;

  SIGNAL ready                       : STD_LOGIC;
  SIGNAL nxt_ready                   : STD_LOGIC;
  SIGNAL ready_stable                : STD_LOGIC;
  SIGNAL xon                         : STD_LOGIC;
  SIGNAL nxt_xon                     : STD_LOGIC;
  SIGNAL xon_stable                  : STD_LOGIC;
  
  SIGNAL err                         : STD_LOGIC;
  SIGNAL nxt_err                     : STD_LOGIC;
  SIGNAL valid                       : STD_LOGIC;
  SIGNAL nxt_valid                   : STD_LOGIC;
  SIGNAL sop                         : STD_LOGIC;
  SIGNAL nxt_sop                     : STD_LOGIC;
  SIGNAL first_sop                   : STD_LOGIC;
  SIGNAL nxt_first_sop               : STD_LOGIC;
  SIGNAL sync                        : STD_LOGIC;
  SIGNAL nxt_sync                    : STD_LOGIC;
  SIGNAL bsn                         : STD_LOGIC_VECTOR(c_bsn_w-1 DOWNTO 0);
  SIGNAL nxt_bsn                     : STD_LOGIC_VECTOR(c_bsn_w-1 DOWNTO 0);
  SIGNAL cnt_sop                     : STD_LOGIC_VECTOR(c_cnt_sop_w-1 DOWNTO 0);
  SIGNAL nof_sop                     : STD_LOGIC_VECTOR(c_cnt_sop_w-1 DOWNTO 0);
  SIGNAL cnt_err                     : STD_LOGIC_VECTOR(c_cnt_sop_w-1 DOWNTO 0);   -- use c_cnt_sop_w, because at maximium all frames have an error
  SIGNAL nof_err                     : STD_LOGIC_VECTOR(c_cnt_sop_w-1 DOWNTO 0);
  SIGNAL cnt_valid                   : STD_LOGIC_VECTOR(c_cnt_valid_w-1 DOWNTO 0);
  SIGNAL nof_valid                   : STD_LOGIC_VECTOR(c_cnt_valid_w-1 DOWNTO 0);

  SIGNAL i_mon_ready_stable          : STD_LOGIC;
  SIGNAL i_mon_xon_stable            : STD_LOGIC;
  SIGNAL i_mon_bsn_at_sync           : STD_LOGIC_VECTOR(c_bsn_w-1 DOWNTO 0);
  SIGNAL i_mon_nof_sop               : STD_LOGIC_VECTOR(c_cnt_sop_w-1 DOWNTO 0);
  SIGNAL i_mon_nof_err               : STD_LOGIC_VECTOR(c_cnt_sop_w-1 DOWNTO 0);
  SIGNAL i_mon_nof_valid             : STD_LOGIC_VECTOR(c_cnt_valid_w-1 DOWNTO 0);
  SIGNAL i_mon_bsn_first             : STD_LOGIC_VECTOR(c_bsn_w-1 DOWNTO 0);
  SIGNAL i_mon_bsn_first_cycle_cnt   : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
  SIGNAL i_current_bsn               : STD_LOGIC_VECTOR(c_bsn_w-1 DOWNTO 0);
  SIGNAL cnt_cycle                   : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);

  SIGNAL sync_timeout_cnt            : STD_LOGIC_VECTOR(c_sync_timeout_w-1 DOWNTO 0);
  SIGNAL sync_timeout                : STD_LOGIC;
  SIGNAL sync_timeout_n              : STD_LOGIC;
  SIGNAL nxt_sync_timeout            : STD_LOGIC;
  SIGNAL sync_timeout_revt           : STD_LOGIC;
  
  SIGNAL nxt_mon_evt                 : STD_LOGIC;
  SIGNAL nxt_mon_ready_stable        : STD_LOGIC;
  SIGNAL nxt_mon_xon_stable          : STD_LOGIC;
  SIGNAL nxt_mon_bsn_at_sync         : STD_LOGIC_VECTOR(c_bsn_w-1 DOWNTO 0);
  SIGNAL nxt_mon_nof_sop             : STD_LOGIC_VECTOR(c_cnt_sop_w-1 DOWNTO 0);
  SIGNAL nxt_mon_nof_err             : STD_LOGIC_VECTOR(c_cnt_sop_w-1 DOWNTO 0);
  SIGNAL nxt_mon_nof_valid           : STD_LOGIC_VECTOR(c_cnt_valid_w-1 DOWNTO 0);
  SIGNAL nxt_mon_bsn_first           : STD_LOGIC_VECTOR(c_bsn_w-1 DOWNTO 0);
  SIGNAL nxt_mon_bsn_first_cycle_cnt : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
  SIGNAL nxt_current_bsn             : STD_LOGIC_VECTOR(c_bsn_w-1 DOWNTO 0);
  
BEGIN

  mon_sync                <= sync;
  mon_sync_timeout        <= sync_timeout;
  mon_ready_stable        <= i_mon_ready_stable;
  mon_xon_stable          <= i_mon_xon_stable;
  mon_bsn_at_sync         <= i_mon_bsn_at_sync;
  mon_nof_sop             <= i_mon_nof_sop;
  mon_nof_err             <= i_mon_nof_err;
  mon_nof_valid           <= i_mon_nof_valid;
  mon_bsn_first           <= i_mon_bsn_first;
  mon_bsn_first_cycle_cnt <= i_mon_bsn_first_cycle_cnt;
  
  nxt_mon_evt          <= sync OR sync_timeout_revt;
  nxt_mon_ready_stable <= ready_stable WHEN sync='1' ELSE i_mon_ready_stable;
  nxt_mon_xon_stable   <= xon_stable   WHEN sync='1' ELSE i_mon_xon_stable;
  nxt_mon_bsn_at_sync  <= bsn          WHEN sync='1' ELSE i_mon_bsn_at_sync;
  nxt_mon_nof_sop      <= nof_sop      WHEN sync='1' ELSE i_mon_nof_sop;
  nxt_mon_nof_err      <= nof_err      WHEN sync='1' ELSE i_mon_nof_err;
  nxt_mon_nof_valid    <= nof_valid    WHEN sync='1' ELSE i_mon_nof_valid;

  nof_sop   <= INCR_UVEC(cnt_sop, 1);    -- +1 because the sop at the sync also counts
  nof_err   <= cnt_err;
  nof_valid <= INCR_UVEC(cnt_valid, 1);  -- +1 because the valid at the sync also counts
  
  u_sync_timeout_cnt : ENTITY common_lib.common_counter
  GENERIC MAP (
    g_width => c_sync_timeout_w
  )
  PORT MAP (
    rst     => rst,
    clk     => clk,
    cnt_clr => sync,
    cnt_en  => sync_timeout_n,
    count   => sync_timeout_cnt
  );
  
  sync_timeout_n <= NOT nxt_sync_timeout;
  
  nxt_sync_timeout <= '1' WHEN UNSIGNED(sync_timeout_cnt)>=g_sync_timeout ELSE '0';
  
  u_sync_timeout_revt : ENTITY common_lib.common_evt
  GENERIC MAP (
    g_evt_type   => "RISING",
    g_out_invert => FALSE,
    g_out_reg    => FALSE
  )
  PORT MAP (
    rst      => rst,
    clk      => clk,
    in_sig   => sync_timeout,
    out_evt  => sync_timeout_revt
  );
    
  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst = '1' THEN
      -- internal
      ready                     <= '0';
      xon                       <= '0';
      valid                     <= '0';
      sop                       <= '0';
      err                       <= '0';
      sync                      <= '0';
      bsn                       <= (OTHERS=>'0');
      -- output
      mon_evt                   <= '0';
      sync_timeout              <= '0';
      i_mon_ready_stable        <= '0';
      i_mon_xon_stable          <= '0';
      i_mon_bsn_at_sync         <= (OTHERS=>'0');
      i_mon_nof_sop             <= (OTHERS=>'0');
      i_mon_nof_err             <= (OTHERS=>'0');
      i_mon_nof_valid           <= (OTHERS=>'0');
      i_mon_bsn_first           <= (OTHERS=>'0');
      i_mon_bsn_first_cycle_cnt <= (OTHERS=>'0');
      i_current_bsn             <= (OTHERS=>'0');
    ELSIF rising_edge(clk) THEN
      -- internal
      ready                     <= nxt_ready;
      xon                       <= nxt_xon;
      valid                     <= nxt_valid;
      sop                       <= nxt_sop;
      err                       <= nxt_err;
      sync                      <= nxt_sync;
      bsn                       <= nxt_bsn;  
      first_sop                 <= nxt_first_sop;
      -- output
      mon_evt                   <= nxt_mon_evt;
      sync_timeout              <= nxt_sync_timeout;
      i_mon_ready_stable        <= nxt_mon_ready_stable;
      i_mon_xon_stable          <= nxt_mon_xon_stable;
      i_mon_bsn_at_sync         <= nxt_mon_bsn_at_sync;
      i_mon_nof_sop             <= nxt_mon_nof_sop;
      i_mon_nof_err             <= nxt_mon_nof_err;
      i_mon_nof_valid           <= nxt_mon_nof_valid;
      i_mon_bsn_first           <= nxt_mon_bsn_first;        
      i_mon_bsn_first_cycle_cnt <= nxt_mon_bsn_first_cycle_cnt;
      i_current_bsn             <= nxt_current_bsn;
    END IF;
  END PROCESS;
  
  -- siso
  nxt_ready <= in_siso.ready;
  nxt_xon   <= in_siso.xon;
  
  u_ready_stable : ENTITY common_lib.common_stable_monitor
  PORT MAP (
    rst          => rst,
    clk          => clk,
    -- MM
    r_in         => ready,
    r_stable     => ready_stable,
    r_stable_ack => sync
  );
  
  u_xon_stable : ENTITY common_lib.common_stable_monitor
  PORT MAP (
    rst          => rst,
    clk          => clk,
    -- MM
    r_in         => xon,
    r_stable     => xon_stable,
    r_stable_ack => sync
  );
  
  -- Sample the BSN, because BSN is only valid during sop. 
  nxt_current_bsn <= in_sosi.bsn(c_bsn_w-1 DOWNTO 0) WHEN in_sosi.sop='1' ELSE i_current_bsn;
  
  -- sosi
  -- . no need to AND sop, eop with valid, because can only be active when valid = '1'
  -- . no need to AND sync     with sop,   because can only be active when sop   = '1'
  nxt_valid                   <= in_sosi.valid;
  nxt_sop                     <= in_sosi.sop;   
  nxt_sync                    <= in_sosi.sync;
  nxt_err                     <= in_sosi.err(g_error_bi)         WHEN in_sosi.eop='1' ELSE '0';  -- assume sosi.err(g_error_bi) = '1' at eop indicates an error
  nxt_bsn                     <= in_sosi.bsn(c_bsn_w-1 DOWNTO 0) WHEN in_sosi.sop='1' ELSE bsn;  -- keep bsn as defined at sop
  
  gen_log_first_bsn : IF g_log_first_bsn = TRUE GENERATE
    nxt_mon_bsn_first           <= in_sosi.bsn(c_bsn_w-1 DOWNTO 0) WHEN first_sop='0' AND in_sosi.sop='1' ELSE i_mon_bsn_first;
    nxt_mon_bsn_first_cycle_cnt <= cnt_cycle                       WHEN first_sop='0' AND in_sosi.sop='1' ELSE i_mon_bsn_first_cycle_cnt;
  END GENERATE;

  nxt_first_sop <= '1' WHEN first_sop='0' AND in_sosi.sop='1' ELSE first_sop;  
  
  gen_sync_in_enable : IF g_log_first_bsn = FALSE GENERATE
    nxt_mon_bsn_first           <= i_current_bsn WHEN sync_in = '1' ELSE i_mon_bsn_first; 
    nxt_mon_bsn_first_cycle_cnt <= cnt_cycle     WHEN sync_in = '1' ELSE i_mon_bsn_first_cycle_cnt;
  END GENERATE;
  
  u_cnt_sop : ENTITY common_lib.common_counter
  GENERIC MAP (
    g_width => c_cnt_sop_w
  )
  PORT MAP (
    rst     => rst,
    clk     => clk,
    cnt_clr => sync,
    cnt_en  => sop,
    count   => cnt_sop
  );
  
  u_nof_err : ENTITY common_lib.common_counter
  GENERIC MAP (
    g_width => c_cnt_sop_w
  )
  PORT MAP (
    rst     => rst,
    clk     => clk,
    cnt_clr => sync,
    cnt_en  => err,
    count   => cnt_err
  );
  
  u_cnt_valid : ENTITY common_lib.common_counter
  GENERIC MAP (
    g_width => c_cnt_valid_w
  )
  PORT MAP (
    rst     => rst,
    clk     => clk,
    cnt_clr => sync,
    cnt_en  => valid,
    count   => cnt_valid
  );

  u_cnt_cycle : ENTITY common_lib.common_counter
  GENERIC MAP (
    g_width => c_word_w
  )
  PORT MAP (
    rst     => rst,
    clk     => clk,
    cnt_clr => in_sosi.sop,
    cnt_en  => '1',
    count   => cnt_cycle
  );
  
END rtl;



