-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Provide MM slave register for dp_bsn_monitor
-- Description:
--
--  Read only monitor register for streams with sync.
--
--   31             24 23             16 15              8 7               0  wi
--  |-----------------|-----------------|-----------------|-----------------|
--  |              sync timeout = [2], ready_stable = [1], xon_stable = [0] |  0
--  |-----------------------------------------------------------------------|
--  |                            bsn_at_sync[31: 0]                         |  1
--  |-----------------------------------------------------------------------|
--  |                            bsn_at_sync[63:32]                         |  2
--  |-----------------------------------------------------------------------|
--  |                                nof_sop[31: 0]                         |  3
--  |-----------------------------------------------------------------------|
--  |                              nof_valid[31: 0]                         |  4
--  |-----------------------------------------------------------------------|
--  |                                nof_err[31: 0]                         |  5
--  |-----------------------------------------------------------------------|
--  |                              bsn_first[31: 0]                         |  6
--  |-----------------------------------------------------------------------|
--  |                              bsn_first[63:32]                         |  7
--  |-----------------------------------------------------------------------|
--  |                    bsn_first_cycle_cnt[31: 0]                         |  8
--  |-----------------------------------------------------------------------|


LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;

ENTITY dp_bsn_monitor_reg IS
  GENERIC (
    g_cross_clock_domain : BOOLEAN := TRUE  -- use FALSE when mm_clk and st_clk are the same, else use TRUE to cross the clock domain
  );
  PORT (
    -- Clocks and reset
    mm_rst                  : IN  STD_LOGIC;   -- reset synchronous with mm_clk
    mm_clk                  : IN  STD_LOGIC;   -- memory-mapped bus clock
    st_rst                  : IN  STD_LOGIC;   -- reset synchronous with st_clk
    st_clk                  : IN  STD_LOGIC;   -- other clock domain clock
    
    -- Memory Mapped Slave in mm_clk domain
    sla_in                  : IN  t_mem_mosi;  -- actual ranges defined by c_mm_reg
    sla_out                 : OUT t_mem_miso;  -- actual ranges defined by c_mm_reg
    
    -- MM registers in st_clk domain
    mon_evt                 : IN  STD_LOGIC;  -- pulses when new monitor data is available regarding the previous sync interval
    mon_sync_timeout        : IN  STD_LOGIC;
    -- . siso
    mon_ready_stable        : IN  STD_LOGIC;
    mon_xon_stable          : IN  STD_LOGIC;
    -- . sosi
    mon_bsn_at_sync         : IN  STD_LOGIC_VECTOR;
    mon_nof_sop             : IN  STD_LOGIC_VECTOR;
    mon_nof_err             : IN  STD_LOGIC_VECTOR;
    mon_nof_valid           : IN  STD_LOGIC_VECTOR;
    mon_bsn_first           : IN  STD_LOGIC_VECTOR;
    mon_bsn_first_cycle_cnt : IN  STD_LOGIC_VECTOR
  );
END dp_bsn_monitor_reg;


ARCHITECTURE str OF dp_bsn_monitor_reg IS

  -- Define the actual size of the MM slave register
  CONSTANT c_mm_reg : t_c_mem := (latency  => 1,
                                  adr_w    => 4,
                                  dat_w    => c_word_w,       -- Use MM bus data width = c_word_w = 32 for all MM registers
                                  nof_dat  => 9,
                                  init_sl  => '0');

  -- Registers in st_clk domain
  SIGNAL mon_reg      : STD_LOGIC_VECTOR(c_mm_reg.nof_dat*c_mm_reg.dat_w-1 DOWNTO 0) := (OTHERS=>'0');

BEGIN

  -- Register mapping
  mon_reg(         3-1 DOWNTO          0) <= mon_sync_timeout & mon_ready_stable & mon_xon_stable;
  mon_reg(1*c_word_w-1 DOWNTO          3) <= (OTHERS=>'0');
  mon_reg(3*c_word_w-1 DOWNTO 1*c_word_w) <= RESIZE_UVEC(mon_bsn_at_sync, c_longword_w);
  mon_reg(4*c_word_w-1 DOWNTO 3*c_word_w) <= RESIZE_UVEC(mon_nof_sop,     c_word_w);
  mon_reg(5*c_word_w-1 DOWNTO 4*c_word_w) <= RESIZE_UVEC(mon_nof_valid,   c_word_w);
  mon_reg(6*c_word_w-1 DOWNTO 5*c_word_w) <= RESIZE_UVEC(mon_nof_err,     c_word_w);
  mon_reg(8*c_word_w-1 DOWNTO 6*c_word_w) <= RESIZE_UVEC(mon_bsn_first,   c_longword_w);
  mon_reg(9*c_word_w-1 DOWNTO 8*c_word_w) <= RESIZE_UVEC(mon_bsn_first_cycle_cnt, c_word_w);

  u_reg : ENTITY common_lib.common_reg_r_w_dc
  GENERIC MAP (
    g_cross_clock_domain => g_cross_clock_domain,
    g_in_new_latency     => 1,    -- mon_evt to mon_reg has latency 1 in dp_bsn_monitor
    g_readback           => FALSE,
    g_reg                => c_mm_reg
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,
    st_rst      => st_rst,
    st_clk      => st_clk,
    
    -- Memory Mapped Slave in mm_clk domain
    sla_in      => sla_in,
    sla_out     => sla_out,
    
    -- MM registers in st_clk domain
    reg_wr_arr  => OPEN,
    reg_rd_arr  => OPEN,
    in_new      => mon_evt,
    in_reg      => mon_reg,   -- read only
    out_reg     => OPEN       -- no write
  );
  
END str;
