-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose : Schedule an output trigger based on the Data Path BSN[]
-- Description:
--   When the snk_in.bsn matches the scheduled_bsn then trigger_out will pulse
--   at the block sync of that block.
-- Remark:
-- . The block sync is defined by the snk_in.eop. The trigger_out is in phase
--   with this block sync.
-- . If the scheduled_bsn is in the past then no trigger_out pulse will occur.
-- . A scheduled_bsn that has not yet occured can be canceled simply by applying
--   a new scheduled_bsn value.
-- . The bsn scheduler can only handle one scheduled_bsn at a time.


ENTITY dp_bsn_scheduler IS
  GENERIC (
    g_bsn_w       : NATURAL := 48
  );
  PORT (
    rst           : IN  STD_LOGIC;
    clk           : IN  STD_LOGIC;
    -- MM control
    scheduled_bsn : IN  STD_LOGIC_VECTOR(g_bsn_w-1 DOWNTO 0) := (OTHERS=>'0');
    -- Streaming
    snk_in        : IN  t_dp_sosi;  -- only uses eop (= block sync), bsn[]
    trigger_out   : OUT STD_LOGIC
  );
END dp_bsn_scheduler;


ARCHITECTURE rtl OF dp_bsn_scheduler IS

  SIGNAL trigger_en     : STD_LOGIC;
  SIGNAL nxt_trigger_en : STD_LOGIC;
  
BEGIN

  -- Use the block sync combinatorially, to avoid one cycle latency and have block synchronisation for the trigger_out
  trigger_out <= snk_in.eop AND trigger_en;
  
  nxt_trigger_en <= '1' WHEN UNSIGNED(scheduled_bsn) = UNSIGNED(snk_in.bsn(scheduled_bsn'RANGE)) + 1 ELSE '0';
  
  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      trigger_en <= '0';
    ELSIF rising_edge(clk) THEN
      trigger_en <= nxt_trigger_en;
    END IF;
  END PROCESS;
  
END rtl;

