-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose : 
--   Start a periodic block sync interval and maintain a block sequence
--   number
-- Description:
--   When dp_on is low then all outputs are low. When dp_on is high, the 
--   output sync starts pulsing with a period of g_block_size number of clk 
--   cycles and the output valid, sop and eop will be active.
--   Alternatively, one can assert dp_on while dp_on_pps is high to 
--   start the data path on the next PPS.
-- Remarks:
--   Starting the data path is only possible from the dp_off state, so one
--   has to disable (dp_on='0') the data path before restarting it.

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY dp_bsn_source IS
  GENERIC (
    g_block_size         : NATURAL := 256;  
    g_nof_block_per_sync : NATURAL := 8;   
    g_bsn_w              : NATURAL := 48
  );
  PORT (
    rst                : IN  STD_LOGIC;
    clk                : IN  STD_LOGIC;
    pps                : IN  STD_LOGIC := '1';

    dp_on              : IN  STD_LOGIC;
    dp_on_pps          : IN  STD_LOGIC;

    dp_on_status       : OUT STD_LOGIC;        
 
    init_bsn           : IN  STD_LOGIC_VECTOR(g_bsn_w-1 DOWNTO 0) := (OTHERS=>'0');
    nof_block_per_sync : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0) := TO_UVEC(g_nof_block_per_sync, c_word_w);

    src_out            : OUT t_dp_sosi  -- only uses sync, bsn[], valid, sop and eop
  );
END dp_bsn_source;


ARCHITECTURE rtl OF dp_bsn_source IS

  CONSTANT c_block_size_cnt_w : NATURAL := ceil_log2(g_block_size);
  CONSTANT c_block_cnt_zero   : STD_LOGIC_VECTOR(g_bsn_w-1 DOWNTO 0) := (OTHERS => '0');

  TYPE t_state_enum IS (s_init, s_dp_off, s_dp_on_sop, s_dp_on, s_dp_on_eop);

  SIGNAL state       : t_state_enum;
  SIGNAL nxt_state   : t_state_enum; 
  SIGNAL prev_state  : t_state_enum; 

  SIGNAL block_size_cnt     : STD_LOGIC_VECTOR(c_block_size_cnt_w-1 DOWNTO 0);
  SIGNAL nxt_block_size_cnt : STD_LOGIC_VECTOR(c_block_size_cnt_w-1 DOWNTO 0);
  
  SIGNAL block_cnt          : STD_LOGIC_VECTOR(g_bsn_w-1 DOWNTO 0);
  SIGNAL nxt_block_cnt      : STD_LOGIC_VECTOR(g_bsn_w-1 DOWNTO 0);

  SIGNAL i_src_out          : t_dp_sosi := c_dp_sosi_init;
  SIGNAL nxt_src_out        : t_dp_sosi;
 
  SIGNAL nxt_dp_on_status   : STD_LOGIC;
  SIGNAL i_dp_on_status     : STD_LOGIC;
 
BEGIN

  src_out <= i_src_out;
  dp_on_status <= i_dp_on_status;

  p_state : PROCESS(state, prev_state, dp_on, dp_on_pps, pps, block_size_cnt, block_cnt, nof_block_per_sync, init_bsn, i_src_out, i_dp_on_status)
  BEGIN  
    nxt_state           <= state;   
    nxt_src_out         <= i_src_out;
    nxt_src_out.sync    <= '0';
    nxt_src_out.valid   <= '0';
    nxt_src_out.sop     <= '0';
    nxt_src_out.eop     <= '0';
    nxt_block_size_cnt  <= block_size_cnt;
    nxt_block_cnt       <= block_cnt;
    nxt_dp_on_status    <= i_dp_on_status;

    CASE state IS

      WHEN s_dp_on_eop =>
        nxt_src_out.eop    <= '1';
        nxt_src_out.valid  <= '1'; 
        nxt_state          <= s_dp_on_sop;
        nxt_block_size_cnt <= INCR_UVEC(block_size_cnt, 1);
        IF UNSIGNED(block_cnt) = UNSIGNED(nof_block_per_sync) -1 THEN
          nxt_block_cnt    <=  (OTHERS=>'0');
        ELSE 
          nxt_block_cnt    <= INCR_UVEC(block_cnt, 1);        
        END IF;
        IF dp_on = '0' THEN
          nxt_state <= s_dp_off;
        END IF;

      WHEN s_dp_on =>
        nxt_src_out.valid  <= '1';
        nxt_block_size_cnt <= INCR_UVEC(block_size_cnt, 1);
        IF UNSIGNED(block_size_cnt) = g_block_size -3 THEN
          nxt_state        <= s_dp_on_eop;
        END IF;

      WHEN s_dp_on_sop =>
        nxt_dp_on_status   <= '1';
        nxt_src_out.sop    <= '1';
        nxt_src_out.valid  <= '1';
        nxt_state          <= s_dp_on;
        nxt_block_size_cnt <= (OTHERS=>'0');
        IF prev_state = s_dp_on_eop THEN
          nxt_src_out.bsn(g_bsn_w-1 DOWNTO 0) <= INCR_UVEC(i_src_out.bsn(g_bsn_w-1 DOWNTO 0), 1);
        END IF;
        IF UNSIGNED(block_cnt) = UNSIGNED(c_block_cnt_zero) THEN 
          nxt_src_out.sync <= '1';
        END IF;

      WHEN s_dp_off =>
        nxt_dp_on_status <= '0';
        nxt_block_cnt <=  (OTHERS=>'0');
        nxt_src_out.bsn <= RESIZE_DP_BSN(init_bsn);
        IF dp_on = '1' THEN 
          IF dp_on_pps = '1' THEN
            IF pps = '1' THEN
              nxt_state        <= s_dp_on_sop;
            END IF; 
          ELSE 
            nxt_state        <= s_dp_on_sop;
          END IF;
        END IF;
    
      WHEN OTHERS => -- s_init
        nxt_state <= s_dp_off;
      
    END CASE;
  END PROCESS;


  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      prev_state      <= s_init;
      state           <= s_init;
      i_src_out       <= c_dp_sosi_rst;
      block_cnt       <= (OTHERS=>'0');
      block_size_cnt  <= (OTHERS=>'0');
      i_dp_on_status  <= '0';
    ELSIF rising_edge(clk) THEN
      prev_state      <= state;  
      state           <= nxt_state;
      i_src_out       <= nxt_src_out;
      block_cnt       <= nxt_block_cnt;
      block_size_cnt  <= nxt_block_size_cnt;
      i_dp_on_status  <= nxt_dp_on_status;
    END IF;
  END PROCESS;

END rtl;


