--------------------------------------------------------------------------------
--
-- Copyright (C) 2017
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Author:
-- . Daniel van der Schuur
-- Purpose:
-- . Pipeline wrapper around dp_counter_func
-- Description:
-- . See dp_counter_func.
-- . dp_counter_func contains only functional logic (no pipelining).
-- . This wrapper adds pipelining for the source side outputs (g_pipeline_src_out)
--   and source side inputs (g_pipeline_src_in).
-- Usage:
-- . When g_nof_counters = 1 then dp_counter can reshape the block sizes in a
--   sync intervals, as is done in dp_block_reshape.vhd
-- . When g_nof_counters = 2 then dp_counter can reshape the block sizes and the
--   sync interval, as is done in dp_block_reshape_sync.vhd
-- . When g_nof_counters > 1 then the dp_counter count_src_out_arr provides the 
--   block indices for an array of data. The g_nof_counters then defines the 
--   dimension of the array within a sync interval. For example:
--
--     a[N][M][P] with N=3, M=5, p=10 will have N*M*P = 150 valid per sync interval. 
--
--   The dp_counter can then provide for each valid in the sync interval the
--   corresponding array index (n, m, p), using:
--
--     g_nof_counters = 3
--     g_range_start  = (0, 0,  0)
--     g_range_stop   = (3, 5, 10)
--     g_range_step   = (1, 1,  1)
--
-- Remarks:
-- . Unfortunately it is not possible to use a generic in the definition of another generic. Therefore the
--   g_range_* generics are defined as unconstraint.
-- . When dp_counter is instanciated the length of the g_range_* generics must at least fit g_nof_counters.
--   In theory count_src_out_arr could be declared as unconstraint and then g_nof_counters could be derived
--   from the length of count_src_out_arr, but it is more clear to explitely to declare g_nof_counters.
--
--   31-05-2018 J Hargreaves added count_offset_in_arr input
--     This signal can be used to apply an offset to the start count
--   USAGE
--     Do not connect the signal if it is not needed: It will consume extra resources
--     It only works when g_range_step(i) is 1
--     It only works when g_range_start(i) + count_offset_in_arr(i) < g_range_stop(i)
--     Any other useage will break counters >= stage i

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY dp_counter IS
  GENERIC (
    g_nof_counters     : NATURAL := 1;
    g_range_start      : t_nat_natural_arr;  -- range must fit (g_nof_counters-1 DOWNTO 0)
    g_range_stop       : t_nat_natural_arr;  -- range must fit (g_nof_counters-1 DOWNTO 0)
    g_range_step       : t_nat_natural_arr;  -- range must fit (g_nof_counters-1 DOWNTO 0)
    g_pipeline_src_out : NATURAL := 1;       -- Pipeline source outputs (data,valid,sop,eop etc)
    g_pipeline_src_in  : NATURAL := 0        -- Pipeline source inputs (ready,xon). This will also pipeline src_out.
  );
  PORT (                                                    
    clk         : IN  STD_LOGIC;
    rst         : IN  STD_LOGIC;

    snk_in      : IN  t_dp_sosi;
    snk_out     : OUT t_dp_siso;

    src_out     : OUT t_dp_sosi;
    src_in      : IN  t_dp_siso := c_dp_siso_rdy;

    count_offset_in_arr : IN  t_nat_natural_arr(g_nof_counters-1 DOWNTO 0) := (OTHERS=>0);  
    count_src_out_arr : OUT t_dp_sosi_arr(g_nof_counters-1 DOWNTO 0)
  );
END dp_counter;


ARCHITECTURE wrap OF dp_counter IS

  -- force downto range for unconstraint g_range generics
  CONSTANT c_range_len             : NATURAL := g_range_start'LENGTH;  -- g_nof_counters must be <= c_range_len
  CONSTANT c_range_start           : t_nat_natural_arr(c_range_len-1 DOWNTO 0) := g_range_start;
  CONSTANT c_range_stop            : t_nat_natural_arr(c_range_len-1 DOWNTO 0) := g_range_stop;
  CONSTANT c_range_step            : t_nat_natural_arr(c_range_len-1 DOWNTO 0) := g_range_step;

  CONSTANT c_use_dp_pipeline       : BOOLEAN := (g_pipeline_src_out>0 AND g_pipeline_src_in=0);
  CONSTANT c_use_dp_pipeline_ready : BOOLEAN := (g_pipeline_src_in>0);

  SIGNAL dp_counter_func_src_out_arr : t_dp_sosi_arr(g_nof_counters-1 DOWNTO 0);

BEGIN

  ------------------------------------------------------------------------------
  -- dp_counter_func
  ------------------------------------------------------------------------------ 
  u_dp_counter_func : ENTITY work.dp_counter_func
  GENERIC MAP (
    g_nof_counters => g_nof_counters,
    g_range_start  => c_range_start,
    g_range_stop   => c_range_stop,
    g_range_step   => c_range_step
  )
  PORT MAP (
    rst       => rst,
    clk       => clk,

    count_en  => snk_in.valid,

    count_offset_in_arr => count_offset_in_arr,
    count_src_out_arr   => dp_counter_func_src_out_arr
  );
  
  ------------------------------------------------------------------------------
  -- dp_pipeline
  ------------------------------------------------------------------------------
  gen_dp_pipeline : IF c_use_dp_pipeline = TRUE GENERATE
    u_dp_pipeline_snk_in : ENTITY work.dp_pipeline
    GENERIC MAP (
      g_pipeline => g_pipeline_src_out
    )
    PORT MAP (
      clk         => clk,
      rst         => rst,
  
      snk_in      => snk_in,
      snk_out     => snk_out,
  
      src_out     => src_out,
      src_in      => src_in
    );

    gen_dp_pipeline_count_src_out_arr : FOR i IN 0 TO g_nof_counters-1 GENERATE
      u_dp_pipeline_count_src_out_arr : ENTITY work.dp_pipeline
      GENERIC MAP (
        g_pipeline => g_pipeline_src_out
      )
      PORT MAP (
        clk     => clk,
        rst     => rst,
    
        snk_in  => dp_counter_func_src_out_arr(i),
    
        src_out => count_src_out_arr(i),
        src_in  => src_in
      );
    END GENERATE;
  END GENERATE;

  ------------------------------------------------------------------------------
  -- dp_pipeline_ready
  ------------------------------------------------------------------------------
  gen_dp_pipeline_ready : IF c_use_dp_pipeline_ready = TRUE GENERATE
    u_dp_pipeline_ready : ENTITY work.dp_pipeline_ready
    GENERIC MAP (
      g_in_latency   => 1
    )
    PORT MAP (
      clk          => clk,
      rst          => rst,
  
      snk_in       => snk_in,
      snk_out      => snk_out,
  
      src_out      => src_out,
      src_in       => src_in
    );

    gen_dp_pipeline_ready_count_src_out_arr : FOR i IN 0 TO g_nof_counters-1 GENERATE
      u_dp_pipeline_ready_count_src_out_arr : ENTITY work.dp_pipeline_ready
      GENERIC MAP (
        g_in_latency => 1
      )
      PORT MAP (
        clk     => clk,
        rst     => rst,
    
        snk_in  => dp_counter_func_src_out_arr(i),
    
        src_out => count_src_out_arr(i),
        src_in  => src_in
      );
    END GENERATE;
  END GENERATE;  

  ------------------------------------------------------------------------------
  -- No pipelining
  ------------------------------------------------------------------------------
  no_dp_pipeline : IF c_use_dp_pipeline=FALSE AND c_use_dp_pipeline_ready=FALSE GENERATE
    src_out <= snk_in;
    snk_out <= src_in;
    
    count_src_out_arr <= dp_counter_func_src_out_arr;
  END GENERATE;

END wrap;
