--------------------------------------------------------------------------------
--
-- Copyright (C) 2017
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Author:
-- . Daniel van der Schuur
-- Purpose:
-- . Simple counter with start, stop, step; used in dp_counter_func.
-- Description:
-- . range(0,2,1) = [0, 1] is the smallest count range allowed
-- Usage:
-- . Not for standalone use; part of dp_counter_func.
 
LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY dp_counter_func_single IS --FIXME move this to common
  GENERIC (
    g_range_start : NATURAL; -- (start,stop,step) like python range(start, stop, step)
    g_range_stop  : NATURAL;
    g_range_step  : NATURAL
  );
  PORT (                                                    
    clk        : IN  STD_LOGIC;
    rst        : IN  STD_LOGIC;

    count_en   : IN STD_LOGIC;
    check_max  : IN STD_LOGIC := '1';
    count_offset : NATURAL := 0;
   
    count      : OUT STD_LOGIC_VECTOR(31 DOWNTO 0);
    count_init : OUT STD_LOGIC; -- Pulses at first init
    count_min  : OUT STD_LOGIC; --Pulses when count=start
    count_max  : OUT STD_LOGIC  --Pulses when count=max
  );
END dp_counter_func_single;


ARCHITECTURE rtl OF dp_counter_func_single IS

 -- The user defines the counters like a Python range(start,stop,step) in which the stop value
  -- is never actually reached. Calculate the actual maximum values here.
  -- . Example:
  --   . range(0,4,2) = [0, 2]
  --   . range(0,5,2) = [0, 2, 4]
  --   . range(0,6,2) = [0, 2, 4]
  --   . range(0,7,2) = [0, 2, 4, 6]
  --   . range(1,7,2) = [1, 3, 5]
  -- . The maximum value is: start+((stop-1-start)/step)*step
  CONSTANT c_nof_count : NATURAL := (g_range_stop-1-g_range_start)/g_range_step + 1;
  CONSTANT c_count_max : NATURAL := g_range_start+(c_nof_count-1)*g_range_step;
  CONSTANT c_count_w   : NATURAL := ceil_log2(c_count_max+1);

  TYPE t_reg IS RECORD
    count_en  : STD_LOGIC;
    count     : STD_LOGIC_VECTOR(c_count_w-1 DOWNTO 0);
    count_min : STD_LOGIC;
    count_max : STD_LOGIC;
    count_init : STD_LOGIC;
  END RECORD;

  SIGNAL r, nxt_r : t_reg;

BEGIN
  
  --------------------------------------------------------------------------------
  -- Combinational logic
  --------------------------------------------------------------------------------
  p_comb : PROCESS(rst, r, count_en, check_max, count_offset) 
    VARIABLE v : t_reg;
  BEGIN
    v           := r;
    v.count_min := '0';
    v.count_max := '0';
    v.count_init := '0';

    IF count_en='1' THEN

      -- Start counting / init
      IF r.count_en='0' THEN
        v.count_en := '1';
        v.count := TO_UVEC(g_range_start+count_offset, c_count_w);
        v.count_min := '1';
        v.count_init := '1';     
      -- keep counting
      ELSE
        v.count := INCR_UVEC(r.count, g_range_step);
        IF c_count_max>0 AND check_max='1' AND r.count = TO_UVEC(c_count_max-g_range_step, c_count_w) THEN -- count max almost reached
          v.count_max := '1';
        ELSIF r.count = TO_UVEC(c_count_max, c_count_w) THEN -- count max reached
          -- Reset count to start value
          v.count := TO_UVEC(g_range_start, c_count_w);
          v.count_min := '1';
        END IF;
      END IF;

      -- If the maximum count is 0, count_max is always high.
      IF c_count_max=0 THEN
        v.count_max := '1';       
      END IF;   

    ELSIF check_max='1' AND r.count = TO_UVEC(c_count_max, c_count_w) THEN -- count max reached
      v.count_max := '1';
    END IF;
   
    IF rst = '1' THEN
      v.count_en  := '0';
      v.count     := (OTHERS=>'0');
      v.count_min := '0';
      v.count_max := '0';
      v.count_init := '0';
    END IF;
 
    nxt_r <= v;
  END PROCESS;

  --------------------------------------------------------------------------------
  -- Register stage
  --------------------------------------------------------------------------------
  r <= nxt_r WHEN rising_edge(clk);

  --------------------------------------------------------------------------------
  -- Outputs
  --------------------------------------------------------------------------------
  count(31 DOWNTO c_count_w)  <= (OTHERS=>'0');
  count(c_count_w-1 DOWNTO 0) <= nxt_r.count;
  count_init <= nxt_r.count_init;
  count_min  <= nxt_r.count_min;
  count_max  <= nxt_r.count_max;


END rtl;
