--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Author:
-- . Daniel van der Schuur
-- Purpose: 
-- . Deinterleave the input into g_nof_out outputs, outputting g_block_size_int words before switching to next output.
-- Description: 
-- . DP wrapper for common_deinterleave
--
-- Status (erko):
-- . Obselete, do not use in new designs. Instead use:
--   - dp_deinterleave_one_to_n.vhd for 1 to N, preserves all sosi fields
--   - dp_unfolder.vhd for 1 to power of 2, but only preserves the valid
-- . The tb_tb_dp_deinterleave_interleave_to_one.vhd shows that dp_deinterleave.vhd fails if the
--   snk_in.valid is random.

ENTITY dp_deinterleave IS
  GENERIC (
    g_dat_w             : NATURAL;
    g_nof_out           : NATURAL;
    g_block_size_int    : NATURAL;         -- Deinterleaver block size; outputs g_block_size_int words before switching to next output
    g_block_size_output : NATURAL;         -- Output block size: The number of samles in the blocks at the output
    g_use_ctrl          : BOOLEAN := TRUE; -- Requires: [input block size (sop:eop)] / [g_nof_out]/ [g_block_size_output] = integer number!
    g_use_sync_bsn      : BOOLEAN := TRUE; -- forwards (stored) input Sync+BSN to all output streams
    g_use_complex       : BOOLEAN;
    g_align_out         : BOOLEAN := FALSE -- Aligns the output streams
  );
  PORT (
    rst         : IN  STD_LOGIC;
    clk         : IN  STD_LOGIC;

    snk_in      : IN  t_dp_sosi;
    src_out_arr : OUT t_dp_sosi_arr(g_nof_out-1 DOWNTO 0)
  );
END dp_deinterleave;


ARCHITECTURE wrap OF dp_deinterleave IS

  CONSTANT c_complex_w  : NATURAL := g_dat_w/2;

  SIGNAL common_deinterleave_in_dat  : STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
  SIGNAL common_deinterleave_in_val  : STD_LOGIC;
  SIGNAL common_deinterleave_out_dat : STD_LOGIC_VECTOR(g_nof_out*g_dat_w-1 DOWNTO 0);
  SIGNAL common_deinterleave_out_val : STD_LOGIC_VECTOR(g_nof_out-1 DOWNTO 0);

  SIGNAL common_deinterleave_src_out_arr : t_dp_sosi_arr(g_nof_out-1 DOWNTO 0);
  SIGNAL dp_block_gen_src_out_arr        : t_dp_sosi_arr(g_nof_out-1 DOWNTO 0);

BEGIN

  -----------------------------------------------------------------------------
  -- DP wrapper around common_deinterleave
  -----------------------------------------------------------------------------
  gen_sosi_dat_in: IF g_use_complex = FALSE GENERATE
    common_deinterleave_in_dat <= snk_in.data(g_dat_w-1 DOWNTO 0);
  END GENERATE;

  gen_sosi_compl_in: IF g_use_complex = TRUE GENERATE
    common_deinterleave_in_dat <= snk_in.im(c_complex_w-1 DOWNTO 0) & snk_in.re(c_complex_w-1 DOWNTO 0);
  END GENERATE;

  common_deinterleave_in_val <= snk_in.valid;

  u_deinterleave : ENTITY common_lib.common_deinterleave
  GENERIC MAP (
    g_nof_out    => g_nof_out,
    g_block_size => g_block_size_int,
    g_dat_w      => g_dat_w,
    g_align_out  => g_align_out
  )
  PORT MAP (
    rst        => rst,
    clk        => clk,
    
    in_dat     => common_deinterleave_in_dat,
    in_val     => common_deinterleave_in_val, 
    
    out_dat    => common_deinterleave_out_dat,
    out_val    => common_deinterleave_out_val
  );

  -----------------------------------------------------------------------------
  -- Use complex fields if required
  -----------------------------------------------------------------------------
  gen_wires_out : FOR i IN 0 TO g_nof_out-1 GENERATE
    gen_sosi_dat_out: IF g_use_complex = FALSE GENERATE
      common_deinterleave_src_out_arr(i).data(g_dat_w-1 DOWNTO 0) <= common_deinterleave_out_dat( i*g_dat_w+g_dat_w -1 DOWNTO i*g_dat_w);
    END GENERATE;
    
    gen_sosi_compl_out: IF g_use_complex = TRUE GENERATE
      common_deinterleave_src_out_arr(i).im <= RESIZE_DP_DSP_DATA(common_deinterleave_out_dat( i*g_dat_w + g_dat_w     -1 DOWNTO i*g_dat_w + c_complex_w));
      common_deinterleave_src_out_arr(i).re <= RESIZE_DP_DSP_DATA(common_deinterleave_out_dat( i*g_dat_w + c_complex_w -1 DOWNTO i*g_dat_w));
    END GENERATE;

    common_deinterleave_src_out_arr(i).valid <= common_deinterleave_out_val(i);
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Add SOP and EOP to the outputs
  -----------------------------------------------------------------------------
  gen_ctrl : IF g_use_ctrl=TRUE GENERATE
    gen_dp_block_gen : FOR i IN 0 TO g_nof_out-1 GENERATE
      u_dp_block_gen : ENTITY work.dp_block_gen
      GENERIC MAP (
        g_use_src_in       => FALSE,
        g_nof_data         => g_block_size_output,
        g_preserve_sync    => TRUE,
        g_preserve_bsn     => TRUE
      )
      PORT MAP(
        rst        => rst,
        clk        => clk,
       
        snk_in     => common_deinterleave_src_out_arr(i),
        src_out    => dp_block_gen_src_out_arr(i)
      );    
    END GENERATE;
  END GENERATE;

  no_ctrl : IF g_use_ctrl=FALSE GENERATE
    dp_block_gen_src_out_arr <= common_deinterleave_src_out_arr;
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Re-add input sync + BSN to all output streams
  -----------------------------------------------------------------------------
  align_out : IF g_use_sync_bsn = TRUE GENERATE
    gen_dp_fifo_info: FOR i IN 0 TO g_nof_out-1 GENERATE
      u_dp_fifo_info : ENTITY work.dp_fifo_info
      GENERIC MAP (
        g_use_sync => TRUE,
        g_use_bsn  => TRUE
      )
      PORT MAP (
        rst          => rst,
        clk          => clk,
    
        data_snk_in  => dp_block_gen_src_out_arr(i),  -- delayed snk_in data
        info_snk_in  => snk_in,                       -- original snk_in info
    
        src_in       => c_dp_siso_rdy,
        src_out      => src_out_arr(i)
      );
    END GENERATE;
  END GENERATE;

  no_align_out : IF g_use_sync_bsn = FALSE GENERATE
    src_out_arr <= dp_block_gen_src_out_arr;
  END GENERATE;

END wrap;
