--------------------------------------------------------------------------------
--
-- Copyright (C) 2019
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Author:
-- . Eric Kooistra
-- Purpose:
-- . Deinterleave the input into g_nof_outputs parallel outputs, with one input
--   sample per output and all outputs valid at the same cycle.
-- Description:
-- . The snk_in blocks size should typically be an integer multiple of
--   g_nof_outputs. If the snk_in block size is an integer multiple of
--   g_nof_outputs, then the snk_in.eop occurs when cnt = g_nof_outputs-1, else
--   the snk_in.eop will break the deinterleaving and output immediately to
--   maintain output blocks with sop and eop.
-- . The snk_in.sop info (bsn, channel) and snk_in.eop info (empty, err) are
--   preserved and passed on to the src_out_arr outputs.
-- . The implementation supports siso flow control, thanks to that the
--   function is implemented combinatorially (so RL=0) using internal state
--   registers. The function only outputs when thereis valid input, therefore
--   it does not need to use src_in_arr().ready. The combinatorial output is
--   pipelined to ease timing closure.
-- Remark:
-- . This function is similar to dp_deinterleave by only supports m=1 to
--   g_nof_outputs and therefore the implementation is simpler.

ENTITY dp_deinterleave_one_to_n IS
  GENERIC (
    g_pipeline      : NATURAL := 1;  -- 0 for combinatorial, > 0 for registers
    g_nof_outputs   : NATURAL
  );
  PORT (
    rst         : IN  STD_LOGIC;
    clk         : IN  STD_LOGIC;

    snk_out     : OUT t_dp_siso;
    snk_in      : IN  t_dp_sosi;
    src_in_arr  : IN  t_dp_siso_arr(g_nof_outputs-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);
    src_out_arr : OUT t_dp_sosi_arr(g_nof_outputs-1 DOWNTO 0)
  );
END dp_deinterleave_one_to_n;


ARCHITECTURE rtl OF dp_deinterleave_one_to_n IS

  -- local state registers
  TYPE t_reg IS RECORD
    cnt         : NATURAL RANGE 0 TO g_nof_outputs-1;
    input_arr   : t_dp_sosi_arr(g_nof_outputs-1 DOWNTO 0);
  END RECORD;

  SIGNAL s            : t_reg;   -- combinatoral state
  SIGNAL r            : t_reg;   -- register state

  SIGNAL out_siso_arr : t_dp_siso_arr(g_nof_outputs-1 DOWNTO 0);
  SIGNAL out_sosi_arr : t_dp_sosi_arr(g_nof_outputs-1 DOWNTO 0);

BEGIN

  p_reg : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      r.cnt       <= 0;
      r.input_arr <= (OTHERS=>c_dp_sosi_rst);
    ELSIF rising_edge(clk) THEN
      r <= s;
    END IF;
  END PROCESS;

  -- Count every g_nof_outputs valid input samples in snk_in, or restart at sop
  p_cnt : PROCESS(snk_in, r)
  BEGIN
    s.cnt <= r.cnt;
    IF snk_in.sop='1' THEN
      s.cnt <= 0;   -- wrap or force restart at sop
    ELSIF snk_in.valid='1' THEN
      IF r.cnt < g_nof_outputs-1 THEN
        s.cnt <= r.cnt + 1;
      ELSE
        s.cnt <= 0;  -- wrap after every g_nof_outputs valid input samples
      END IF;
    END IF;
  END PROCESS;

  -- Load in series of g_nof_outputs valid snk_in
  -- Alternative solution is to shift in snk_in to avoid selection logic based
  -- on cnt. However then possibly selection logic based on cnt is needed to
  -- know where the sop info is in case the deinterleaving needs to break on
  -- eop.
  p_input_arr : PROCESS(snk_in, s, r)
  BEGIN
    s.input_arr <= r.input_arr;
    IF snk_in.valid='1'THEN
      s.input_arr(s.cnt) <= snk_in;
    END IF;
  END PROCESS;

  -- Output s.input_arr to parallel out_sosi_arr including all sosi data fields
  p_out_sosi_arr : PROCESS(snk_in, s)
  BEGIN
    -- Default pass on data
    out_sosi_arr <= s.input_arr;
    FOR I IN 0 TO g_nof_outputs-1 LOOP
      out_sosi_arr(I).valid <= '0';
      out_sosi_arr(I).sync  <= '0';
      out_sosi_arr(I).sop   <= '0';
      out_sosi_arr(I).eop   <= '0';
    END LOOP;

    -- Output at valid and end of cnt or break at eop
    IF snk_in.valid='1' THEN
      IF s.cnt = g_nof_outputs-1 OR snk_in.eop = '1'THEN
        FOR I IN 0 TO g_nof_outputs-1 LOOP
          out_sosi_arr(I).valid <= '1';
          -- Use sop info from cnt=0
          out_sosi_arr(I).sync    <= s.input_arr(0).sync;
          out_sosi_arr(I).sop     <= s.input_arr(0).sop;
          out_sosi_arr(I).bsn     <= s.input_arr(0).bsn;
          out_sosi_arr(I).channel <= s.input_arr(0).channel;
          -- Use eop info directly from snk_in
          out_sosi_arr(I).eop     <= snk_in.eop;
          out_sosi_arr(I).empty   <= snk_in.empty;
          out_sosi_arr(I).err     <= snk_in.err;
        END LOOP;
      END IF;
    END IF;
  END PROCESS;

  -- Register and pass on flow control
  snk_out <= out_siso_arr(0);  -- all out_siso_arr have the same siso, so wire output 0

  u_pipeline_outputs : ENTITY work.dp_pipeline_arr
  GENERIC MAP (
    g_nof_streams => g_nof_outputs,
    g_pipeline    => g_pipeline  -- 0 for wires, > 0 for registers
  )
  PORT MAP (
    rst          => rst,
    clk          => clk,
    -- ST sink
    snk_out_arr  => out_siso_arr,
    snk_in_arr   => out_sosi_arr,
    -- ST source
    src_in_arr   => src_in_arr,
    src_out_arr  => src_out_arr
  );

END rtl;
