--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;


-- Purpose:
--   De-multiplex frames from one input stream with one or more channels onto
--   separate output streams.
-- Description:
--   If g_combined=TRUE then all outputs must be ready for the dp_demux to be
--   ready. This is because due to RL=1 the snk_in.channel and
--   src_in_arr().ready do not apply to the same clock cycle. The adavantage
--   of using g_combined=TRUE is that it takes minimal logic. The fact that it
--   works combinatorially may cause problems with achieving timing closure.
--   If g_combined=FALSE then only the output for the current channel needs to
--   be ready. This is possible because internally the dp_demux then operates
--   at RL=0 so it has a show-ahead on the valid snk_in.channel. The output is 
--   again at RL=1 which has the additional benefit that the src_in_arr().ready
--   and the snk_out.ready SISO ready signals are connected via a register
--   stage. The RL adapter to get to RL=0 also registers the SOSI data signals.
--   The output selection scheme depends on g_mode:
--   0: Use current channel to select the output port.
--      The snk_in.channel[log2(g_nof_output)-1:0] bits are used to select the
--      output port. The remaining snk_in.channel[high:log2(g_nof_output)]
--      bits are passed on to src_out_arr().channel (assuming typically 
--      g_remove_channel_lo = TRUE).
--   1: Use eop to select next output.
--   2: Use external MM control input to select the output
-- Remark:
-- . For g_nof_output=1 the dp_demux collapses to wires, because ceil_log2(1)=1.
-- . The dp_mux adds true_log2(nof ports) low bits to src_out_arr.channel and
--   this dp_demux removes true_log2(nof ports) low bits from snk_in().channel.
-- . For demultiplexing time series frames or sample it can be applicable to
--   use g_remove_channel_lo=FALSE in combination with g_mode=1.
-- . Mode 0 and 2 two work for any sosi stream, mode 1 requires framed data i.e.
--   with eop.

ENTITY dp_demux IS
  GENERIC (
    g_mode              : NATURAL := 0;
    g_nof_output        : NATURAL := 2;  -- >= 1
    g_remove_channel_lo : BOOLEAN := TRUE;
    g_combined          : BOOLEAN := TRUE;
    g_sel_ctrl_invert   : BOOLEAN := FALSE;  -- Use default FALSE when stream array IO are indexed (0 TO g_nof_input-1), else use TRUE when indexed (g_nof_input-1 DOWNTO 0)
    g_sel_ctrl_pkt      : BOOLEAN := FALSE   -- When TRUE and g_mode=2 and g_combined=FALSE, this selects a new output on packet boundaries only.
  );
  PORT (
    rst         : IN  STD_LOGIC;
    clk         : IN  STD_LOGIC;
    -- Control & Status
    sel_ctrl    : IN  NATURAL RANGE 0 TO g_nof_output-1 := 0;  -- used by g_mode = 2
    sel_stat    : OUT NATURAL RANGE 0 TO g_nof_output-1;       -- used by g_sel_ctrl_pkt = TRUE
    -- ST sinks
    snk_out     : OUT t_dp_siso;
    snk_in      : IN  t_dp_sosi;
    -- ST source
    src_in_arr  : IN  t_dp_siso_arr(0 TO g_nof_output-1) := (OTHERS=>c_dp_siso_rdy);
    src_out_arr : OUT t_dp_sosi_arr(0 TO g_nof_output-1)
  );
END dp_demux;


ARCHITECTURE rtl OF dp_demux IS

  -- The low part of snk_in.channel is used to select the output port and the high part of snk_in.channel used for src_out_arr[].channel
  CONSTANT c_sel_w  : NATURAL := true_log2(g_nof_output);
  
  -- Output selection scheme
  SIGNAL sel_channel        : STD_LOGIC_VECTOR(c_dp_stream_channel_w-1 DOWNTO 0);  -- used by g_mode = 0
  SIGNAL sel_eop            : STD_LOGIC;                                           -- used by g_mode = 1
  SIGNAL output_select      : NATURAL RANGE 0 TO g_nof_output-1;
  SIGNAL prev_output_select : NATURAL RANGE 0 TO g_nof_output-1;
  SIGNAL pkt_det            : STD_LOGIC;

  -- g_combined=FALSE
  SIGNAL adapt_siso       : t_dp_siso;
  SIGNAL adapt_sosi       : t_dp_sosi;
  
  SIGNAL prev_src_in_arr  : t_dp_siso_arr(0 TO g_nof_output-1);
  SIGNAL pend_src_out_arr : t_dp_sosi_arr(0 TO g_nof_output-1);
  
BEGIN

  -- Collapse to wires
  gen_one : IF g_nof_output=1 GENERATE
    snk_out        <= src_in_arr(0);
    src_out_arr(0) <= snk_in;  -- so c_sel_w = 0
  END GENERATE;
  
  ------------------------------------------------------------------------------
  -- Output selection scheme
  ------------------------------------------------------------------------------
  
  -- Use current channel to select the output port.
  gen_mode_0 : IF g_mode=0 GENERATE
    output_select <= TO_UINT(sel_channel(c_sel_w-1 DOWNTO 0));
  END GENERATE;
  
  -- Use eop to select next output.
  gen_mode_1 : IF g_mode=1 GENERATE
    p_clk : PROCESS(rst, clk)
    BEGIN
      IF rst='1' THEN
        output_select <= 0;
      ELSIF rising_edge(clk) THEN
        IF sel_eop='1' THEN
          IF output_select=g_nof_output-1 THEN
            output_select <= 0;
          ELSE
            output_select <= output_select + 1;
          END IF;
        END IF;
      END IF;
    END PROCESS;
  END GENERATE;
  
  -- Use external MM control input to select the output
  gen_mode_2 : IF g_mode=2 GENERATE

    no_pkt_ctrl: IF g_sel_ctrl_pkt=FALSE GENERATE
      -- Select new output immediately (registered: one cycle after sel_ctrl change)
      p_clk : PROCESS(rst, clk)
      BEGIN
        IF rst='1' THEN
          output_select <= 0;
        ELSIF rising_edge(clk) THEN
          IF g_sel_ctrl_invert=FALSE THEN
            output_select <= sel_ctrl;
          ELSE
            output_select <= g_nof_output-1-sel_ctrl;
          END IF;
        END IF;
      END PROCESS;
    END GENERATE;

    gen_pkt_ctrl: IF g_sel_ctrl_pkt=TRUE AND g_combined=FALSE GENERATE
      -- Select new output while respecting packet boundaries. We need RL=0 (g_combined=FALSE) to do
      -- this in a reliable way.
      -- Changing the output is only allowed:
      --   -> When there is no packet present on the bus, or
      --   -> When there is a SOP present on the bus (this allows us to redirect back-to-back packets, thanks to RL=0)
      p_select : PROCESS(prev_output_select, sel_ctrl, pkt_det, adapt_sosi)
      BEGIN
        output_select <= prev_output_select;
        IF pkt_det='0' OR adapt_sosi.sop='1' THEN
          IF g_sel_ctrl_invert=FALSE THEN
            output_select <= sel_ctrl;
          ELSE
            output_select <= g_nof_output-1-sel_ctrl;
          END IF;
        -- User might need this status port to indicate if/when the output has actually been switched
        sel_stat <= output_select; 
        END IF;
      END PROCESS;   

      -- Keep the current selection until we're allowed to switch (on a packet bounday)
      p_clk : PROCESS(rst, clk)
      BEGIN
        IF rst='1' THEN
          prev_output_select <= 0;
        ELSIF rising_edge(clk) THEN
          prev_output_select <= output_select;
        END IF;
      END PROCESS;   

      -- The dp_packet_detect component simply asserts its output from SOP to EOP
      u_dp_packet_detect : ENTITY work.dp_packet_detect
      GENERIC MAP (
        g_latency => 0
      )
      PORT MAP (
        rst       => rst,
        clk       => clk,  
        sosi      => adapt_sosi,
        siso      => adapt_siso,  -- We're using the adapted sink_in with RL=0
        pkt_det   => pkt_det
      );    
  
    END GENERATE;

  END GENERATE;
    

  ------------------------------------------------------------------------------
  -- g_combined=TRUE : All outputs must be ready
  ------------------------------------------------------------------------------
    
  gen_combined : IF g_combined=TRUE AND g_nof_output>1 GENERATE
    snk_out.xon   <= func_dp_stream_arr_and(src_in_arr, "XON");
    snk_out.ready <= func_dp_stream_arr_and(src_in_arr, "READY");
  
    sel_channel <= snk_in.channel;
    sel_eop     <= snk_in.eop;
    
    p_output : PROCESS(snk_in, output_select)
    BEGIN
      FOR I IN 0 TO g_nof_output-1 LOOP
        src_out_arr(I)         <= snk_in;
        src_out_arr(I).sync    <= '0';
        src_out_arr(I).valid   <= '0';
        src_out_arr(I).sop     <= '0';
        src_out_arr(I).eop     <= '0';
        IF g_remove_channel_lo=TRUE THEN
          src_out_arr(I).channel <= SHIFT_UVEC(snk_in.channel, c_sel_w);  -- strip the low part that is used for the demux from the channel field
        END IF;
        IF output_select=I THEN
          src_out_arr(I).sync  <= snk_in.sync;
          src_out_arr(I).valid <= snk_in.valid;
          src_out_arr(I).sop   <= snk_in.sop;
          src_out_arr(I).eop   <= snk_in.eop;
        END IF;
      END LOOP;
    END PROCESS;
  END GENERATE;

  
  ------------------------------------------------------------------------------
  -- g_combined=FALSE : Only the output for the current channel needs to be ready
  ------------------------------------------------------------------------------

  gen_individual : IF g_combined=FALSE AND g_nof_output>1 GENERATE
    -- Adapt input to RL = 0 to have show-ahead
    u_rl_0 : ENTITY work.dp_latency_adapter
    GENERIC MAP (
      g_in_latency  => 1,
      g_out_latency => 0
    )
    PORT MAP (
      rst       => rst,
      clk       => clk,
      -- ST sink
      snk_out   => snk_out,
      snk_in    => snk_in,
      -- ST source
      src_in    => adapt_siso,
      src_out   => adapt_sosi
    );
    
    sel_channel <= adapt_sosi.channel;
    sel_eop     <= adapt_sosi.eop AND adapt_siso.ready;  -- RL = 0, so eop is only valid when the ready acknowledges it
    
    -- Demux at RL = 0
    p_output : PROCESS(adapt_sosi, prev_src_in_arr, output_select)
    BEGIN
      adapt_siso <= c_dp_siso_rst;
      FOR I IN 0 TO g_nof_output-1 LOOP
        pend_src_out_arr(I)         <= adapt_sosi;
        pend_src_out_arr(I).sync    <= '0';
        pend_src_out_arr(I).valid   <= '0';
        pend_src_out_arr(I).sop     <= '0';
        pend_src_out_arr(I).eop     <= '0';
        IF g_remove_channel_lo=TRUE THEN
          pend_src_out_arr(I).channel <= SHIFT_UVEC(adapt_sosi.channel, c_sel_w);  -- strip the low part that is used for the demux from the channel field
        END IF;
        IF output_select=I THEN
          adapt_siso <= prev_src_in_arr(I);
          pend_src_out_arr(I).sync  <= adapt_sosi.sync;
          pend_src_out_arr(I).valid <= adapt_sosi.valid;
          pend_src_out_arr(I).sop   <= adapt_sosi.sop;
          pend_src_out_arr(I).eop   <= adapt_sosi.eop;
        END IF;
      END LOOP;
      
      -- Pass on frame level flow control
      adapt_siso.xon <= func_dp_stream_arr_and(prev_src_in_arr, "XON");
    END PROCESS;
    
    -- Back to output RL = 1
    gen_rl_1 : FOR I IN 0 TO g_nof_output-1 GENERATE
      u_incr : ENTITY work.dp_latency_increase
      GENERIC MAP (
        g_in_latency   => 0,
        g_incr_latency => 1
      )
      PORT MAP (
        rst         => rst,
        clk         => clk,
        -- ST sink
        snk_out     => prev_src_in_arr(I),
        snk_in      => pend_src_out_arr(I),
        -- ST source
        src_in      => src_in_arr(I),
        src_out     => src_out_arr(I)
      );
    END GENERATE;
  END GENERATE;
  
END rtl;
