--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose: Distribute the frames from n input streams on to m output streams.
-- Description:
-- * First the n input streams are demultiplexed in n*m internal streams and
--   then these are multiplexed on to the m output streams.
-- * The input streams can have an input FIFO dependent on g_use_fifo.
-- * The dp_distribute has two operation modes transmitter and receiver that
--   are set by g_tx.
--   When g_tx=TRUE then transmit:
--   . demux g_mode=1 selects next output when eop is there to distribute the
--     incomming frames evenly over all outputs.
--   . demux uses c_demux_remove_channel_lo=FALSE because the input frames all
--     have the same channel, they are demuxed in order.
--   . default non-blocking mux g_tx_mux_mode = 0 selects the next input when
--     the input has had a fair chance. Alternatively mux g_tx_mux_mode = 1
--     selects the next input when the eop is there. This causes that all
--     inputs will get to send equal number of frames, so this input traffic
--     must fit that and no input frames must get lost.
--   . mux uses c_mux_append_channel_lo=TRUE because the frames that will be
--     output to the serial link come from different the input sources. The
--     channel_lo field allows the other side of the link to reliably demux
--     these frames even when a frame gets lost on the link and disturbs the
--     order.
--   When g_tx=FALSE then receive:
--   . demux g_mode=0 uses the channel_lo field to select the output.
--   . demux uses c_demux_remove_channel_lo=TRUE because it uses the channel_lo
--     field for the output selection.
--   . default non-blocking mux g_rx_mux_mode = 0 selects the next input when
--     the input has had a fair chance.
--   . mux uses c_mux_append_channel_lo=FALSE because the frames that will be
--     output are in time order from the same source.
-- * The input streams and output streams can have DP packet coding to ease
--   transporting the other sosi fields over the link. If the snk_in stream
--   uses DP packets then g_code_channel_lo=TRUE may be used to also transport
--   the dp_distribute mux channel_lo bits via the CHAN field of the DP packet.
--   For g_tx=TRUE the mux channel_lo bits get inserted at the output and for
--   g_tx=FALSE the mux channel_lo bits get removed at the input. The removal
--   is done after the FIFO to avoid having to transport the channel_lo field
--   in parallel through the FIFO.
-- * When g_transpose = TRUE then the dp_distribute still does the demux-mux
--   transpose of frames in time over multiple serial links when g_nof_input
--   = g_nof_output. When g_transpose = FALSE then the inputs are wired on to
--   the corresponding serial link if g_nof_input = g_nof_output.
-- Data flow:
--      snk--> FIFO      -- when g_use_fifo
--       in--> DECODE    -- when NOT g_tx AND g_code_channel_lo
--       rx--> DEMUX
--   demux2--> rewire
--     mux2--> MUX
--       tx--> ENCODE    -- when     g_tx AND g_code_channel_lo
--         -->src
-- Remark:
-- . The assumption is that all links are equaly fast so the time order of the
--   frames at the input is preserved at the output.
-- . Thanks to the non-blocking dp_mux the distribution continuous when a frame
--   gets lost or an input is not used.

LIBRARY IEEE,common_lib, technology_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY dp_distribute IS
  GENERIC (
    g_technology      : NATURAL := c_tech_select_default;
    -- Distribution IO
    g_tx              : BOOLEAN;
    g_nof_input       : NATURAL := 4;  -- >= 1
    g_nof_output      : NATURAL := 3;  -- >= 1
    g_transpose       : BOOLEAN := FALSE;
    g_code_channel_lo : BOOLEAN := FALSE;  -- for DP packet optionaly transport the mux channel_lo bits via the CHAN field
    g_data_w          : NATURAL := 16;     -- actual sosi data width, used for FIFO and for dp_packet_en_channel_lo
    -- Scheduling
    g_tx_mux_mode     : NATURAL := 0;  -- default 0 for non-blocking mux, but 1 also works in simulation and on hardware provided that the inputs are garantueed to arrive
    g_rx_mux_mode     : NATURAL := 0;  -- default 0 for non-blocking mux, but 1 also works in simulation and on hardware provided that the inputs are garantueed to arrive
    -- Input FIFO
    g_use_fifo        : BOOLEAN := FALSE;
    g_bsn_w           : NATURAL := 1;
    g_empty_w         : NATURAL := 1;
    g_channel_w       : NATURAL := 1;
    g_error_w         : NATURAL := 1;
    g_use_bsn         : BOOLEAN := FALSE;
    g_use_empty       : BOOLEAN := FALSE;
    g_use_channel     : BOOLEAN := FALSE;
    g_use_error       : BOOLEAN := FALSE;
    g_use_sync        : BOOLEAN := FALSE;
    g_fifo_af_margin  : NATURAL := 4;    -- Nof words below max (full) at which fifo is considered almost full
    g_fifo_fill       : NATURAL := 0;
    g_fifo_size       : NATURAL := 256  -- (32+2) * 256 = 1 M9K, g_data_w+2 for sop and eop
  );
  PORT (
    rst         : IN  STD_LOGIC;
    clk         : IN  STD_LOGIC;
    -- ST sinks
    snk_out_arr : OUT t_dp_siso_arr(0 TO g_nof_input-1);
    snk_in_arr  : IN  t_dp_sosi_arr(0 TO g_nof_input-1);
    -- ST source
    src_in_arr  : IN  t_dp_siso_arr(0 TO g_nof_output-1) := (OTHERS=>c_dp_siso_rdy);
    src_out_arr : OUT t_dp_sosi_arr(0 TO g_nof_output-1)
  );
END dp_distribute;


ARCHITECTURE str OF dp_distribute IS

  CONSTANT c_demux_mode              : NATURAL := sel_a_b(g_tx, 1, 0);
  CONSTANT c_demux_remove_channel_lo : BOOLEAN := NOT g_tx;
  
  CONSTANT c_mux_mode                : NATURAL := sel_a_b(g_tx, g_tx_mux_mode, g_rx_mux_mode);
  CONSTANT c_mux_append_channel_lo   : BOOLEAN := g_tx;
  
  CONSTANT g_encode_channel_lo       : BOOLEAN :=     g_tx AND g_code_channel_lo;  -- true only for transmit to link
  CONSTANT g_decode_channel_lo       : BOOLEAN := NOT g_tx AND g_code_channel_lo;  -- true only for receive from link
  CONSTANT c_link_channel_lo         : NATURAL := sel_a_b(g_tx, ceil_log2(g_nof_input), ceil_log2(g_nof_output));

  TYPE t_demux_siso_2arr IS ARRAY (INTEGER RANGE <>) OF t_dp_siso_arr(0 TO g_nof_output-1);
  TYPE t_demux_sosi_2arr IS ARRAY (INTEGER RANGE <>) OF t_dp_sosi_arr(0 TO g_nof_output-1);
  TYPE t_mux_siso_2arr   IS ARRAY (INTEGER RANGE <>) OF t_dp_siso_arr(0 TO g_nof_input-1);
  TYPE t_mux_sosi_2arr   IS ARRAY (INTEGER RANGE <>) OF t_dp_sosi_arr(0 TO g_nof_input-1);

  SIGNAL in_siso_arr     : t_dp_siso_arr(0 TO g_nof_input-1);      -- from snk_* for ST after optional input FIFO
  SIGNAL in_sosi_arr     : t_dp_sosi_arr(0 TO g_nof_input-1);
  SIGNAL rx_siso_arr     : t_dp_siso_arr(0 TO g_nof_input-1);      -- ST after optional decoding
  SIGNAL rx_sosi_arr     : t_dp_sosi_arr(0 TO g_nof_input-1);
  SIGNAL demux_siso_2arr : t_demux_siso_2arr(0 TO g_nof_input-1);  -- ST at output of demux
  SIGNAL demux_sosi_2arr : t_demux_sosi_2arr(0 TO g_nof_input-1);
  SIGNAL mux_siso_2arr   : t_mux_siso_2arr(0 TO g_nof_output-1);   -- ST rewired for input to mux
  SIGNAL mux_sosi_2arr   : t_mux_sosi_2arr(0 TO g_nof_output-1);
  SIGNAL tx_siso_arr     : t_dp_siso_arr(0 TO g_nof_output-1);     -- ST at output of mux, to src_* for ST output after optional encoding
  SIGNAL tx_sosi_arr     : t_dp_sosi_arr(0 TO g_nof_output-1);
  
BEGIN

  no_fifo : IF g_use_fifo=FALSE GENERATE
    snk_out_arr <= in_siso_arr;
    in_sosi_arr <= snk_in_arr;
  END GENERATE;
    
  gen_fifo : IF g_use_fifo=TRUE GENERATE
    gen_input : FOR I IN 0 TO g_nof_input-1 GENERATE
      u_fifo : ENTITY work.dp_fifo_fill
      GENERIC MAP (
        g_technology     => g_technology,
        g_bsn_w          => g_bsn_w,
        g_data_w         => g_data_w,
        g_empty_w        => g_empty_w,
        g_channel_w      => g_channel_w,
        g_error_w        => g_error_w,
        g_use_bsn        => g_use_bsn,
        g_use_empty      => g_use_empty,
        g_use_channel    => g_use_channel,
        g_use_error      => g_use_error,
        g_use_sync       => g_use_sync,
        g_fifo_fill      => g_fifo_fill,
        g_fifo_size      => g_fifo_size,
        g_fifo_af_margin => g_fifo_af_margin,
        g_fifo_rl        => 1
      )
      PORT MAP (
        rst       => rst,
        clk       => clk,
        -- ST sink
        snk_out   => snk_out_arr(I),
        snk_in    => snk_in_arr(I),
        -- ST source
        src_in    => in_siso_arr(I),
        src_out   => in_sosi_arr(I)
      );
    END GENERATE;
  END GENERATE;

  no_dec : IF g_decode_channel_lo=FALSE GENERATE
    in_siso_arr <= rx_siso_arr;
    rx_sosi_arr <= in_sosi_arr;
  END GENERATE;
  
  gen_dec : IF g_decode_channel_lo=TRUE GENERATE
    gen_i : FOR I IN 0 TO g_nof_input-1 GENERATE
      u_dec : ENTITY work.dp_packet_dec_channel_lo
      GENERIC MAP (
        g_data_w      => g_data_w,
        g_channel_lo  => c_link_channel_lo
      )
      PORT MAP (
        rst       => rst,
        clk       => clk,
        -- ST sinks
        snk_out   => in_siso_arr(I),
        snk_in    => in_sosi_arr(I),
        -- ST source
        src_in    => rx_siso_arr(I),
        src_out   => rx_sosi_arr(I)
      );
    END GENERATE;
  END GENERATE;
  
  no_transpose : IF g_nof_input=g_nof_output AND g_transpose=FALSE GENERATE
    rx_siso_arr <= tx_siso_arr;
    tx_sosi_arr <= rx_sosi_arr;
  END GENERATE;
  
  gen_transpose : IF g_nof_input/=g_nof_output OR g_transpose=TRUE GENERATE
    gen_demux : FOR I IN 0 TO g_nof_input-1 GENERATE
      u_demux : ENTITY work.dp_demux
      GENERIC MAP (
        g_mode              => c_demux_mode,
        g_nof_output        => g_nof_output,
        g_remove_channel_lo => c_demux_remove_channel_lo,
        g_combined          => FALSE
      )
      PORT MAP (
        rst         => rst,
        clk         => clk,
        -- ST sinks
        snk_out     => rx_siso_arr(I),
        snk_in      => rx_sosi_arr(I),
        -- ST source
        src_in_arr  => demux_siso_2arr(I),
        src_out_arr => demux_sosi_2arr(I)
      );
    END GENERATE;
  
    -- Rewire to distribute
    gen_in : FOR I IN 0 TO g_nof_input-1 GENERATE
      gen_out : FOR J IN 0 TO g_nof_output-1 GENERATE
        demux_siso_2arr(I)(J) <= mux_siso_2arr(J)(I);
        mux_sosi_2arr(J)(I) <= demux_sosi_2arr(I)(J);
      END GENERATE;
    END GENERATE;
  
    gen_mux : FOR J IN 0 TO g_nof_output-1 GENERATE
      u_mux : ENTITY work.dp_mux
      GENERIC MAP (
        -- MUX
        g_mode              => c_mux_mode,
        g_nof_input         => g_nof_input,
        g_append_channel_lo => c_mux_append_channel_lo,
        -- Input FIFO
        g_use_fifo          => FALSE,
        g_fifo_size         => array_init(1024, g_nof_input),  -- must match g_nof_input, even when g_use_fifo=FALSE
        g_fifo_fill         => array_init(   0, g_nof_input)   -- must match g_nof_input, even when g_use_fifo=FALSE
      )
      PORT MAP (
        rst         => rst,
        clk         => clk,
        -- ST sinks
        snk_out_arr => mux_siso_2arr(J),
        snk_in_arr  => mux_sosi_2arr(J),
        -- ST source
        src_in      => tx_siso_arr(J),
        src_out     => tx_sosi_arr(J)
      );
    END GENERATE;
  END GENERATE;
  
  no_enc : IF g_encode_channel_lo=FALSE GENERATE
    tx_siso_arr <= src_in_arr;
    src_out_arr <= tx_sosi_arr;
  END GENERATE;
  
  gen_enc : IF g_encode_channel_lo=TRUE GENERATE
    gen_j : FOR J IN 0 TO g_nof_output-1 GENERATE
      u_enc : ENTITY work.dp_packet_enc_channel_lo
      GENERIC MAP (
        g_data_w      => g_data_w,
        g_channel_lo  => c_link_channel_lo
      )
      PORT MAP (
        -- ST sinks
        snk_out   => tx_siso_arr(J),
        snk_in    => tx_sosi_arr(J),
        -- ST source
        src_in    => src_in_arr(J),
        src_out   => src_out_arr(J)
      );
    END GENERATE;
  END GENERATE;

END str;
