-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose:
--   Provide MM access to the ST input of a FIFO.
-- Description:
--   An MM master can directly write the FIFO, provided that the MM data width
--   (g_mm_word_w) fits within the ST data width.
--   The MM master can read the mm_usedw[] value to know how many valid data
--   words are in the FIFO.
--   The MM master can read the mm_availw[] value to know how many valid data
--   it can write to the FIFO in a loop, until it needs to check mm_availw[]
--   again. The mm_availw[] is defined as g_fifo_size - g_fifo_af_margin -
--   mm_usedw[].
--   When mm_availw[] > 0 then the MM master can be sure that the mm_wrdata
--   will fit in the FIFO. Therefore it is not needed to use src_out.ready and
--   in this way the MM master does not need to check whether the FIFO has free
--   space left for every write FIFO data access.
-- Remark:
-- . This dp_fifo_from_mm connects to the MM clock domain side of the FIFO,
--   hence is operates entirely in the MM clock domain.
-- . The ST sop, eop, empty, channel and error fields are not supported.


ENTITY dp_fifo_from_mm IS
  GENERIC (
    g_fifo_size      : NATURAL := 512;
    g_fifo_af_margin : NATURAL := 1;  -- >=1, so > 0 to avoid FIFO full (which causes usedw=0, because value g_fifo_size then wraps to 0)
    g_mm_word_w      : NATURAL := 32
  );
  PORT (
    rst           : IN  STD_LOGIC;    -- MM clock domain
    clk           : IN  STD_LOGIC;
    -- ST soure connected to FIFO input
    src_out       : OUT t_dp_sosi;
    usedw         : IN  STD_LOGIC_VECTOR(ceil_log2(g_fifo_size)-1 DOWNTO 0);
    -- Control for FIFO read access
    mm_wr         : IN  STD_LOGIC;    -- MM write pulse to write the mm_wrdata to src_out.data
    mm_wrdata     : IN  STD_LOGIC_VECTOR(g_mm_word_w-1 DOWNTO 0);
    mm_usedw      : OUT STD_LOGIC_VECTOR(g_mm_word_w-1 DOWNTO 0);
    mm_availw     : OUT STD_LOGIC_VECTOR(g_mm_word_w-1 DOWNTO 0)
  );
END dp_fifo_from_mm;


ARCHITECTURE str OF dp_fifo_from_mm IS

  CONSTANT c_fifo_almost_full : NATURAL := g_fifo_size-g_fifo_af_margin;  -- FIFO almost full level, usedw will always be <= c_fifo_almost_full
  CONSTANT c_usedw_w          : NATURAL := ceil_log2(g_fifo_size);
  
  SIGNAL availw        : STD_LOGIC_VECTOR(c_usedw_w-1 DOWNTO 0);  -- no need to use ceil_log2(g_fifo_size+1), because availw = g_fifo_size - g_fifo_af_margin < g_fifo_size when the FIFO is empty
  
BEGIN

  ASSERT g_fifo_af_margin/=0 REPORT "dp_fifo_from_mm.vhd : the g_fifo_af_margin must be > 0" SEVERITY FAILURE;

  -- Combinatorial logic
  availw <= STD_LOGIC_VECTOR(TO_UNSIGNED(c_fifo_almost_full, c_usedw_w) - UNSIGNED(usedw));  -- No danger for underflow (< 0) because usedw will always be <= c_fifo_almost_full
  
  -- Wires
  mm_usedw  <= RESIZE_UVEC(usedw,  g_mm_word_w);
  mm_availw <= RESIZE_UVEC(availw, g_mm_word_w);  

  src_out.valid <= mm_wr;
  src_out.data  <= RESIZE_DP_DATA(mm_wrdata);  -- Unsigned resize to ST data width
  
END str;
