-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose:
--   Realign the snk_in info with the snk_in data that got delayed
-- Description:
--   The snk_in info consists of the BSN and channel fields at the sop and the
--   empty and err fields at the eop. Dependent on the number of blocks that
--   the data gets delayed these info fields need to be buffered to be able to
--   realign them again with the delayed data.
--   The data gets delayed e.g. due to processing latency some DSP function
--   or because it is passed through a FIFO.
--   The info typically only needs to be passed along with the same delay as
--   the data. To save logic and RAM and to not burden the data function with
--   the handling of the info this dp_fifo_info can realign the original
--   snk_in info with the delayed snk_in data.
--   The dp_fifo_info uses the info_snk_in.sop and info_snk_in.eop to write
--   the info_snk_in info to the FIFO and it uses the data_snk_in.sop and
--   data_snk_in.eop to read the info from the FIFO. Therefore the DSP
--   component must pass on the snk_in.sop and snk_in.eop, to be able to
--   attach the info_snk_in sop info and eop info again at the output. The
--   g_fifo_size must be sufficient to fit the maximum latency in number of
--   blocks (sop's) that the DSP component can introduce.
--
-- Usage:
--                           data_snk_out
--                           data_snk_in
--                               .
--                   |-----|     .         |------|
--   snk_out <------ |     | <------------ |      | <---- src_in
--   snk_in  ---*--> | DP/ | ------------> | dp   | ----> src_out
--              |    | DSP |               | fifo |
--              |    |-----|               | info |
--              |            info_snk_in   |      |
--              \------------------------> |------|
--
--                   
-- Remark:
-- . See also dp/doc/sketch_dp_fifo_info.pdf
-- . The data is passed on unchanged by dp_fifo_info.
-- . The dp_fifo_info preserves the RL=1 but is does delay the data by one
--   cycle.
-- . If the DSP component does not pass on the sop and eop, then it can be
--   possible to recreate these at the output based on the valid using e.g.
--   dp_block_gen. This assumes that the DSP does pass on the valid, that the
--   block size is known and that the first valid at the output corresponds
--   to a sop.

LIBRARY IEEE, common_lib, technology_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY dp_fifo_info IS
  GENERIC (
    g_technology     : NATURAL := c_tech_select_default;
    g_use_sync       : BOOLEAN := FALSE;
    g_use_bsn        : BOOLEAN := FALSE;
    g_use_channel    : BOOLEAN := FALSE;
    g_use_empty      : BOOLEAN := FALSE;
    g_use_error      : BOOLEAN := FALSE;
    g_bsn_w          : NATURAL := c_dp_stream_bsn_w;
    g_empty_w        : NATURAL := c_dp_stream_empty_w;
    g_channel_w      : NATURAL := c_dp_stream_channel_w;
    g_error_w        : NATURAL := c_dp_stream_error_w;
    g_fifo_size      : NATURAL := 4
  );
  PORT (
    rst          : IN  STD_LOGIC;
    clk          : IN  STD_LOGIC;
    -- Monitor info FIFO filling
    fifo_wr_ful  : OUT STD_LOGIC;  -- corresponds to the carry bit of usedw when FIFO is full
    fifo_usedw   : OUT STD_LOGIC_VECTOR(ceil_log2(g_fifo_size)-1 DOWNTO 0);
    fifo_rd_emp  : OUT STD_LOGIC;
    -- ST sink
    data_snk_out : OUT t_dp_siso;
    data_snk_in  : IN  t_dp_sosi;  -- delayed snk_in data
    info_snk_in  : IN  t_dp_sosi;  -- original snk_in info
    -- ST source
    src_in       : IN  t_dp_siso;
    src_out      : OUT t_dp_sosi
  );
END dp_fifo_info;

ARCHITECTURE str OF dp_fifo_info IS

  CONSTANT c_fifo_sop_dat_w   : NATURAL := func_slv_concat_w(g_use_sync, g_use_bsn, g_use_channel, 1, g_bsn_w, g_channel_w);  -- concat sop info via FIFO
  CONSTANT c_fifo_eop_dat_w   : NATURAL := func_slv_concat_w(g_use_empty, g_use_error, g_empty_w, g_error_w);                 -- concat eop info via FIFO
  CONSTANT c_fifo_info_dat_w  : NATURAL := c_fifo_sop_dat_w + c_fifo_eop_dat_w;
  
  SIGNAL dp_pipeline_data_src_in    : t_dp_siso;
  SIGNAL dp_pipeline_data_src_out   : t_dp_sosi;
  
  SIGNAL info_src_out               : t_dp_sosi;
  
  SIGNAL wr_sync                    : STD_LOGIC_VECTOR(0 DOWNTO 0);
  SIGNAL rd_sync                    : STD_LOGIC_VECTOR(0 DOWNTO 0);
  
  SIGNAL dp_latency_adapter_snk_out : t_dp_siso;
  SIGNAL dp_latency_adapter_snk_in  : t_dp_sosi;
  
  SIGNAL fifo_sop_wr_dat            : STD_LOGIC_VECTOR(c_fifo_sop_dat_w-1 DOWNTO 0);
  SIGNAL fifo_sop_wr_req            : STD_LOGIC;
  SIGNAL fifo_sop_wr_ful            : STD_LOGIC;
  SIGNAL fifo_sop_usedw             : STD_LOGIC_VECTOR(fifo_usedw'RANGE);
  SIGNAL fifo_sop_rd_dat            : STD_LOGIC_VECTOR(c_fifo_sop_dat_w-1 DOWNTO 0) := (OTHERS=>'0');
  SIGNAL fifo_sop_rd_val            : STD_LOGIC;
  SIGNAL fifo_sop_rd_req            : STD_LOGIC;
  SIGNAL fifo_sop_rd_emp            : STD_LOGIC;
  
  SIGNAL fifo_eop_wr_dat            : STD_LOGIC_VECTOR(c_fifo_eop_dat_w-1 DOWNTO 0);
  SIGNAL fifo_eop_wr_req            : STD_LOGIC;
  SIGNAL fifo_eop_wr_ful            : STD_LOGIC;
  SIGNAL fifo_eop_usedw             : STD_LOGIC_VECTOR(fifo_usedw'RANGE);
  SIGNAL fifo_eop_rd_dat            : STD_LOGIC_VECTOR(c_fifo_eop_dat_w-1 DOWNTO 0) := (OTHERS=>'0');
  SIGNAL fifo_eop_rd_val            : STD_LOGIC;
  SIGNAL fifo_eop_rd_req            : STD_LOGIC;
  SIGNAL fifo_eop_rd_emp            : STD_LOGIC;
  
BEGIN

  no_info : IF c_fifo_info_dat_w=0 GENERATE
    data_snk_out <= src_in;
    src_out      <= data_snk_in;
  END GENERATE;
  
  gen_info : IF c_fifo_info_dat_w>0 GENERATE
    -- Output monitor FIFO filling
    fifo_wr_ful <= fifo_sop_wr_ful WHEN c_fifo_sop_dat_w>0 ELSE fifo_eop_wr_ful;
    fifo_usedw  <= fifo_sop_usedw  WHEN c_fifo_sop_dat_w>0 ELSE fifo_eop_usedw;
    fifo_rd_emp <= fifo_sop_rd_emp WHEN c_fifo_sop_dat_w>0 ELSE fifo_eop_rd_emp;
    
    -- Data pipeline register to compensate for the fifo rd_req to rd_val latency of 1
    u_dp_pipeline : ENTITY work.dp_pipeline
    GENERIC MAP (
      g_pipeline => 1
    )
    PORT MAP (
      rst          => rst,
      clk          => clk,
      -- ST sink
      snk_out      => data_snk_out,
      snk_in       => data_snk_in,
      -- ST source
      src_in       => dp_pipeline_data_src_in,
      src_out      => dp_pipeline_data_src_out
    );
  
    -- Buffer sop info
    gen_info_sop : IF c_fifo_sop_dat_w>0 GENERATE
      fifo_sop_wr_req <= info_snk_in.sop;
      fifo_sop_rd_req <= data_snk_in.sop;
      
      -- concatenate the sop data to use a single FIFO
      wr_sync(0) <= info_snk_in.sync;
      fifo_sop_wr_dat <= func_slv_concat(g_use_sync, g_use_bsn, g_use_channel,
                                         wr_sync,
                                         info_snk_in.bsn(g_bsn_w-1 DOWNTO 0),
                                         info_snk_in.channel(g_channel_w-1 DOWNTO 0));
                                         
      u_common_fifo_sc : ENTITY common_lib.common_fifo_sc
      GENERIC MAP (
        g_technology  => g_technology,
        g_note_is_ful => FALSE,  -- when TRUE report NOTE when FIFO goes full, fifo overflow is always reported as FAILURE
        g_use_lut     => TRUE,   -- when TRUE then force using LUTs via Altera eab="OFF",
        g_dat_w       => c_fifo_sop_dat_w,
        g_nof_words   => g_fifo_size
      )
      PORT MAP (
        rst      => rst,
        clk      => clk,
        wr_dat   => fifo_sop_wr_dat,
        wr_req   => fifo_sop_wr_req,
        wr_ful   => fifo_sop_wr_ful,
        rd_dat   => fifo_sop_rd_dat,
        rd_req   => fifo_sop_rd_req,
        rd_emp   => fifo_sop_rd_emp,
        rd_val   => fifo_sop_rd_val,
        usedw    => fifo_sop_usedw
      );
  
      -- Extract the sop data from the FIFO output SLV
      info_src_out.sync <= rd_sync(0);
      rd_sync                                      <= func_slv_extract(g_use_sync, g_use_bsn, g_use_channel, 1, g_bsn_w, g_channel_w, fifo_sop_rd_dat, 0);
      info_src_out.bsn(        g_bsn_w-1 DOWNTO 0) <= func_slv_extract(g_use_sync, g_use_bsn, g_use_channel, 1, g_bsn_w, g_channel_w, fifo_sop_rd_dat, 1);
      info_src_out.channel(g_channel_w-1 DOWNTO 0) <= func_slv_extract(g_use_sync, g_use_bsn, g_use_channel, 1, g_bsn_w, g_channel_w, fifo_sop_rd_dat, 2);
    END GENERATE;
  
    -- Buffer eop info
    gen_info_eop : IF c_fifo_eop_dat_w>0 GENERATE
      fifo_eop_wr_req <= info_snk_in.eop;
      fifo_eop_rd_req <= data_snk_in.eop;
      
      -- concatenate the eop data to use a single FIFO
      fifo_eop_wr_dat <= func_slv_concat(g_use_empty, g_use_error,
                                         info_snk_in.empty(g_empty_w-1 DOWNTO 0),
                                         info_snk_in.err(g_error_w-1 DOWNTO 0));
                                         
      u_common_fifo_sc : ENTITY common_lib.common_fifo_sc
      GENERIC MAP (
        g_technology  => g_technology,
        g_note_is_ful => FALSE,  -- when TRUE report NOTE when FIFO goes full, fifo overflow is always reported as FAILURE
        g_use_lut     => TRUE,   -- when TRUE then force using LUTs via Altera eab="OFF",
        g_dat_w       => c_fifo_eop_dat_w,
        g_nof_words   => g_fifo_size
      )
      PORT MAP (
        rst      => rst,
        clk      => clk,
        wr_dat   => fifo_eop_wr_dat,
        wr_req   => fifo_eop_wr_req,
        wr_ful   => fifo_eop_wr_ful,
        rd_dat   => fifo_eop_rd_dat,
        rd_req   => fifo_eop_rd_req,
        rd_emp   => fifo_eop_rd_emp,
        rd_val   => fifo_eop_rd_val,
        usedw    => fifo_eop_usedw
      );
  
      -- Extract the eop data from the FIFO output SLV
      info_src_out.empty(g_empty_w-1 DOWNTO 0) <= func_slv_extract(g_use_empty, g_use_error, g_empty_w, g_error_w, fifo_eop_rd_dat, 0);
      info_src_out.err(g_error_w-1 DOWNTO 0)   <= func_slv_extract(g_use_empty, g_use_error, g_empty_w, g_error_w, fifo_eop_rd_dat, 1);
    END GENERATE;
    
    -- Realign the info with the data
    dp_pipeline_data_src_in <= src_in;
    src_out                 <= func_dp_stream_combine_info_and_data(info_src_out, dp_pipeline_data_src_out);
  END GENERATE;

END str;
