--------------------------------------------------------------------------------
--
-- Copyright (C) 2016
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Status: 
-- . Active (recommended for new designs)
-- Author:
-- . Daniel van der Schuur
-- Purpose:
-- . Fold n input streams into n/2, n/2/2, n/2/2/2, .. output streams
-- Description:
-- . This component assumes the user has scheduled the input streams properly.
--   . Input data that is to be folded onto one stream must not be valid at the same clock cycle
-- . nof_outputs = ceil_div(g_nof_inputs, 2^(g_nof_folds)) for g_nof_folds>=0
--   . Examples:
--     . g_nof_inputs=10, g_nof_folds=0 -> nof_outputs=10
--     . g_nof_inputs=10, g_nof_folds=1 -> nof_outputs= 5
--     . g_nof_inputs=10, g_nof_folds=2 -> nof_outputs= 3
--     . g_nof_inputs=10, g_nof_folds=3 -> nof_outputs= 2
--     . g_nof_inputs=10, g_nof_folds=4 -> nof_outputs= 1
--     . g_nof_inputs=10, g_nof_folds<0 -> nof_outputs= 1
-- . This entity recursively instantiates (registered) stages of itself when folding streams multiple times.
--
-- Remark:
-- . Use dp_deinterleave_n_to_one.vhd when:
--   - for 1 to N, so any N >= 1 (so not only powers of 2)
--   - when all sosi fields need to be preserved (so not only the valid)
--   - if flow control is needed

ENTITY dp_folder IS
  GENERIC (
    g_nof_inputs        : NATURAL;         -- Number of inputs to fold >0
    g_nof_folds         : INTEGER := -1;   -- >0: Number of folds; 0: Wire out to in; <0: Fold until one output remains
    g_output_block_size : NATURAL := 0;    -- >0: Create SOP/EOP tagged output blocks of this size. 0: Forward incoming SOP,EOP.
    g_fwd_sync_bsn      : BOOLEAN := FALSE;-- TRUE: forwards (stored) input Sync+BSN (from snk_in_arr(0)) to all output streams
    g_use_channel       : BOOLEAN := FALSE -- TRUE: Also fold the channel field
  );
  PORT (
    clk            : IN  STD_LOGIC;
    rst            : IN  STD_LOGIC;

    snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
    src_out_arr    : OUT t_dp_sosi_arr(sel_a_b(g_nof_folds>=0, ceil_div(g_nof_inputs, ceil_pow2(g_nof_folds)), 1)-1 DOWNTO 0)
  );
END dp_folder;

ARCHITECTURE str OF dp_folder IS

  COMPONENT dp_folder IS
    GENERIC (
      g_nof_inputs        : NATURAL;
      g_nof_folds         : INTEGER := -1;
      g_output_block_size : NATURAL := 0;   
      g_fwd_sync_bsn      : BOOLEAN := FALSE;
      g_use_channel       : BOOLEAN := FALSE
    ); 
    PORT (
      rst            : IN  STD_LOGIC;
      clk            : IN  STD_LOGIC;
  
      snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  
      src_out_arr    : OUT t_dp_sosi_arr(sel_a_b(g_nof_folds>=0, ceil_div(g_nof_inputs, ceil_pow2(g_nof_folds)), 1)-1 DOWNTO 0)
    );
  END COMPONENT;

  CONSTANT c_nof_muxes : NATURAL := ceil_div(g_nof_inputs, 2);  -- Using ceil_div always yields an even number of mux inputs

  SIGNAL mux_snk_in_arr      : t_dp_sosi_arr(2*c_nof_muxes-1 DOWNTO 0);
  SIGNAL mux_snk_in_2arr_2   : t_dp_sosi_2arr_2(c_nof_muxes-1 DOWNTO 0);

  SIGNAL mux_src_out_arr     : t_dp_sosi_arr(c_nof_muxes-1 DOWNTO 0);
  SIGNAL nxt_mux_src_out_arr : t_dp_sosi_arr(c_nof_muxes-1 DOWNTO 0);

  SIGNAL dp_folder_src_out_arr    : t_dp_sosi_arr(sel_a_b(g_nof_folds>=0, ceil_div(g_nof_inputs, ceil_pow2(g_nof_folds)), 1)-1 DOWNTO 0);

  SIGNAL dp_block_gen_snk_in_arr  : t_dp_sosi_arr(sel_a_b(g_nof_folds>=0, ceil_div(g_nof_inputs, ceil_pow2(g_nof_folds)), 1)-1 DOWNTO 0);
  SIGNAL dp_block_gen_src_out_arr : t_dp_sosi_arr(sel_a_b(g_nof_folds>=0, ceil_div(g_nof_inputs, ceil_pow2(g_nof_folds)), 1)-1 DOWNTO 0);

BEGIN

  gen_arch: IF g_nof_folds/=0 GENERATE
    -----------------------------------------------------------------------------
    -- Wire input array to mux_snk_in_arr to make sure we have an even number
    -- of buses to work with in case of an odd number of inputs
    -- . We want an even number of buses because we will wire up 2-input muxes
    -----------------------------------------------------------------------------
    gen_even_nof_buses: FOR i IN 0 TO g_nof_inputs-1 GENERATE   
      mux_snk_in_arr(i) <= snk_in_arr(i);
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- Wire inputs to the 2-input muxes
    -----------------------------------------------------------------------------
    gen_mux_inputs_0: FOR i IN 0 TO c_nof_muxes-1 GENERATE
      mux_snk_in_2arr_2(i)(0) <= mux_snk_in_arr(2*i);
      mux_snk_in_2arr_2(i)(1) <= mux_snk_in_arr(2*i+1);
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- Simple 2-input mux logic
    -----------------------------------------------------------------------------
    p_mux : PROCESS(mux_snk_in_2arr_2)
    BEGIN
      FOR i IN 0 TO c_nof_muxes-1 LOOP
        nxt_mux_src_out_arr(i) <= c_dp_sosi_rst;
        IF mux_snk_in_2arr_2(i)(0).valid='1' THEN
          nxt_mux_src_out_arr(i).data  <= mux_snk_in_2arr_2(i)(0).data;
          nxt_mux_src_out_arr(i).re    <= mux_snk_in_2arr_2(i)(0).re;
          nxt_mux_src_out_arr(i).im    <= mux_snk_in_2arr_2(i)(0).im;
          nxt_mux_src_out_arr(i).valid <= '1';
          IF g_use_channel=TRUE THEN
            nxt_mux_src_out_arr(i).channel <= mux_snk_in_2arr_2(i)(0).channel;
          END IF;
        ELSIF mux_snk_in_2arr_2(i)(1).valid='1' THEN
          nxt_mux_src_out_arr(i).data  <= mux_snk_in_2arr_2(i)(1).data;
          nxt_mux_src_out_arr(i).re    <= mux_snk_in_2arr_2(i)(1).re;
          nxt_mux_src_out_arr(i).im    <= mux_snk_in_2arr_2(i)(1).im;
          nxt_mux_src_out_arr(i).valid <= '1';
          IF g_use_channel=TRUE THEN
            nxt_mux_src_out_arr(i).channel <= mux_snk_in_2arr_2(i)(1).channel;
          END IF;
        END IF;
      END LOOP;
    END PROCESS;
 
    -- Registers
    p_clk: PROCESS(clk, rst)
    BEGIN
      IF rst='1' THEN
        mux_src_out_arr <= (OTHERS=>c_dp_sosi_rst);
      ELSIF rising_edge(clk) THEN
        mux_src_out_arr <= nxt_mux_src_out_arr;
      END IF;
    END PROCESS;

    -----------------------------------------------------------------------------
    -- Not done folding; add a stage.
    -- . g_nof_folds <0 : user wants to fold all the way to one output
    -- . g_nof_folds >0 : user wants to fold n times
    -----------------------------------------------------------------------------
    gen_dp_folder: IF (g_nof_folds<0 AND c_nof_muxes>1) OR g_nof_folds>1 GENERATE
      u_dp_folder : dp_folder
      GENERIC MAP (
        g_nof_inputs        => c_nof_muxes,
        g_nof_folds         => g_nof_folds-1,
        g_output_block_size => g_output_block_size,
        g_fwd_sync_bsn      => g_fwd_sync_bsn,
        g_use_channel       => g_use_channel
      )
      PORT MAP (
        rst         => rst,
        clk         => clk,
                    
        snk_in_arr  => mux_src_out_arr,                   
        src_out_arr => dp_folder_src_out_arr
      );

      src_out_arr <= dp_folder_src_out_arr;

    END GENERATE;

    -----------------------------------------------------------------------------
    -- c_nof_muxes=1 or g_nof_folds=1, so this is the last stage.
    -----------------------------------------------------------------------------
    gen_src_out_arr: IF (g_nof_folds<0 AND c_nof_muxes=1) OR g_nof_folds=1 GENERATE
      dp_block_gen_snk_in_arr <= mux_src_out_arr;

      -----------------------------------------------------------------------------
      -- Add SOP and EOP to the outputs
      -----------------------------------------------------------------------------
      gen_ctrl : IF g_output_block_size>0 GENERATE
        gen_dp_block_gen : FOR i IN 0 TO c_nof_muxes-1 GENERATE
          u_dp_block_gen : ENTITY work.dp_block_gen
          GENERIC MAP (
            g_use_src_in       => FALSE,
            g_nof_data         => g_output_block_size,
            g_preserve_sync    => TRUE,
            g_preserve_bsn     => TRUE,
            g_preserve_channel => g_use_channel
          )
          PORT MAP(
            rst        => rst,
            clk        => clk,
           
            snk_in     => dp_block_gen_snk_in_arr(i),
            src_out    => dp_block_gen_src_out_arr(i)
          );    
        END GENERATE;
      END GENERATE;
    
      no_ctrl : IF g_output_block_size=0 GENERATE
        dp_block_gen_src_out_arr <= dp_block_gen_snk_in_arr;
      END GENERATE;
    
      -----------------------------------------------------------------------------
      -- Re-add input sync + BSN to all output streams
      -----------------------------------------------------------------------------
      gen_sync_bsn : IF g_fwd_sync_bsn = TRUE GENERATE
        gen_dp_fifo_info: FOR i IN 0 TO c_nof_muxes-1 GENERATE
          u_dp_fifo_info : ENTITY work.dp_fifo_info
          GENERIC MAP (
            g_use_sync => TRUE,
            g_use_bsn  => TRUE
          )
          PORT MAP (
            rst          => rst,
            clk          => clk,
        
            data_snk_in  => dp_block_gen_src_out_arr(i),  -- delayed snk_in data
            info_snk_in  => snk_in_arr(0),                -- original snk_in info
        
            src_in       => c_dp_siso_rdy,
            src_out      => src_out_arr(i)
          );
        END GENERATE;
      END GENERATE;
    
      no_sync_bsn : IF g_fwd_sync_bsn = FALSE GENERATE
        src_out_arr <= dp_block_gen_src_out_arr;
      END GENERATE;
    END GENERATE;  
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Wire output to input if g_nof_folds=0
  -----------------------------------------------------------------------------
  gen_wire_out_to_in: IF g_nof_folds=0 GENERATE
    dp_block_gen_snk_in_arr <= snk_in_arr;
  END GENERATE;

END str;

