-------------------------------------------------------------------------------
--
-- Copyright (C) 2017
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------
--
-- Author: E. Kooistra, 7 apr 2017
-- Purpose:
--   Force the sosi data,re,im field to constant or incrementing value in a
--   block or pass on.
--
--   See also dp_force_data_serial.vhd to force data at a specific index in
--   a block.
-- Description:
--   The force_en input dynamically enables or disables the entire force
--   function.
--
--   Default the force value is constant, but using the generics it can be 
--   made to automatically increment for every valid and restart at sync or
--   sop.
--   . set force_data, re, im via MM interface
--   . set increment step via generic g_increment_data, re, im. Use
--     g_increment_data, re, im = 0 for constant value in a block.
--   . Dependent on g_increment_data_on_sop, re, im the increment is applied
--     for every valid or only at the sop.
--   . restart the data increment on sync or on sop via generic
--     g_restart_data_on_sync or g_restart_data_on_sop. Similar for
--     g_restart_re_on_sync, sop and g_restart_im_on_sync, sop.
--     If g_increment_data, re, im = 0 then the g_restart_data_on_sop, sync
--     setting is void.
-- Usage:
-- . DSP tests: The amplitude of constant data is useful to verify scaling and
--   requantization.
-- . Data move test: Forcing constant (zero) data is useful for to trace data
--   when the data gets reordered. A combination of constant re and
--   incrementing im can also be useful to trace data.
-- Remarks:
-- . Ready latency = 1
-- . When force_en='0' then this dp_force_data_parallel is equivalent
--   to dp_pipeline.
-- . The sosi.data and sosi.re, sosi.im fields can be forced independently.
--   However in the stream typically only the sosi.data is used or only the
--   complex sosi.re, sosi.im. Hence, when the sosi.data is used then the
--   force_re and force_im are void and when the sosi.re, sosi.im are used
--   then the force_data is void.

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.all;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY dp_force_data_parallel IS
  GENERIC (
    g_dat_w                 : NATURAL := 32;    -- must be <= 32 to fit INTEGER range
    g_increment_data        : INTEGER := 0;
    g_increment_re          : INTEGER := 0;
    g_increment_im          : INTEGER := 0;
    g_increment_data_on_sop : BOOLEAN := FALSE;
    g_increment_re_on_sop   : BOOLEAN := FALSE;
    g_increment_im_on_sop   : BOOLEAN := FALSE;
    g_restart_data_on_sync  : BOOLEAN := FALSE;
    g_restart_re_on_sync    : BOOLEAN := FALSE;
    g_restart_im_on_sync    : BOOLEAN := FALSE;
    g_restart_data_on_sop   : BOOLEAN := FALSE;
    g_restart_re_on_sop     : BOOLEAN := FALSE;
    g_restart_im_on_sop     : BOOLEAN := FALSE
  );
  PORT (
    rst           : IN  STD_LOGIC;
    clk           : IN  STD_LOGIC;
    -- MM control
    force_en      : IN  STD_LOGIC := '0';
    force_data    : IN  INTEGER := 0;       -- used for sosi.data
    force_re      : IN  INTEGER := 0;       -- used for sosi.re
    force_im      : IN  INTEGER := 0;       -- used for sosi.im
    -- ST sink
    snk_out       : OUT t_dp_siso;
    snk_in        : IN  t_dp_sosi;
    -- ST source
    src_in        : IN  t_dp_siso := c_dp_siso_rdy;
    src_out       : OUT t_dp_sosi
  );
END dp_force_data_parallel;


LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.all;
USE work.dp_stream_pkg.ALL;

ARCHITECTURE str OF dp_force_data_parallel IS

  SIGNAL data_in          : t_dp_sosi;
  SIGNAL data_out         : t_dp_sosi;
  
BEGIN

  p_comb : PROCESS(snk_in, force_en, force_data, force_re, force_im, data_out)
  BEGIN
    data_in <= snk_in;
    IF force_en='1' THEN
      -- default force value
      data_in.data <= TO_DP_SDATA(force_data);
      data_in.re   <= TO_DP_DSP_DATA(force_re);
      data_in.im   <= TO_DP_DSP_DATA(force_im);
      -- optional increment on valid (no need to check snk_in.valid)
      IF g_increment_data/=0 THEN
        data_in.data <= RESIZE_DP_SDATA(STD_LOGIC_VECTOR(SIGNED(data_out.data(g_dat_w-1 DOWNTO 0)) + g_increment_data));
      END IF;
      IF g_increment_re/=0 THEN
        data_in.re   <= RESIZE_DP_DSP_DATA(STD_LOGIC_VECTOR(SIGNED(data_out.re(g_dat_w-1 DOWNTO 0)) + g_increment_re));
      END IF;
      IF g_increment_im/=0 THEN
        data_in.im   <= RESIZE_DP_DSP_DATA(STD_LOGIC_VECTOR(SIGNED(data_out.im(g_dat_w-1 DOWNTO 0)) + g_increment_im));
      END IF;
      -- optional increment only on sop
      IF snk_in.sop='0' THEN
        IF g_increment_data_on_sop=TRUE THEN data_in.data <= data_out.data; END IF;
        IF g_increment_re_on_sop=TRUE   THEN data_in.re   <= data_out.re;   END IF;
        IF g_increment_im_on_sop=TRUE   THEN data_in.im   <= data_out.im;   END IF;
      END IF;
      -- optional restart increment
      IF snk_in.sync='1' THEN
        IF g_restart_data_on_sync=TRUE THEN data_in.data <= TO_DP_SDATA(force_data);  END IF;
        IF g_restart_re_on_sync=TRUE   THEN data_in.re   <= TO_DP_DSP_DATA(force_re); END IF;
        IF g_restart_im_on_sync=TRUE   THEN data_in.im   <= TO_DP_DSP_DATA(force_im); END IF;
      END IF;
      IF snk_in.sop='1' THEN
        IF g_restart_data_on_sop=TRUE THEN data_in.data <= TO_DP_SDATA(force_data);  END IF;
        IF g_restart_re_on_sop=TRUE   THEN data_in.re   <= TO_DP_DSP_DATA(force_re); END IF;
        IF g_restart_im_on_sop=TRUE   THEN data_in.im   <= TO_DP_DSP_DATA(force_im); END IF;
      END IF;
    END IF;
  END PROCESS;
  
  u_dp_pipeline : ENTITY work.dp_pipeline
  GENERIC MAP (
    g_pipeline   => 1
  )
  PORT MAP (
    rst          => rst,
    clk          => clk,
    -- ST sink
    snk_out      => snk_out,
    snk_in       => data_in,
    -- ST source
    src_in       => src_in,
    src_out      => data_out
  );
    
  src_out <= data_out;
END str;
