--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_packetizing_pkg.ALL;


-- Reuse from LOFAR rad_frame.vhd and rad_frame(rtl).vhd

-- Purpose:
--   Puts the input data path signals:
--     . val
--     . dat[]
--     . sof
--     . eof
--     . fsn[]
--   into the internal data path (dp) frame data format:
--
--     -----------------------
--     | fsn | dat ... | brc |
--     -----------------------
--       sof             eof
--       val   val ...   val
--
--   To make an external frame with sfd and crc instead of brc use dp_frame_tx.
--
-- Description:
--   The input sof and eof mark the first respectively the last valid word of a
--   block of input data. The output sof indicates the fsn word. After that
--   follows the block of input data. At the end a brc is appended indicated by
--   the output eof. If necessary invalid data is inserted between frames.
--   The brc is a boolean redundancy check and will be 0 indicating OK when
--   in_err='0', else 1. The in_err is valid during the active in_eof. The brc
--   can act as a place holder for a true crc to be inserted when the frame
--   is sent outside the FPGA using dp_frame_tx and dp_frame_rx. Alternatively
--   the brc can be kept in the dp frame in addition to a PHY level CRC.
--   The width of the fsn and brc is set by the dat width. The fsn is sign
--   extended to g_dat_w, in this way all bits g_dat_w:g_fsn_w-1 indicate
--   fsync, so a signed resize suits for every g_dat_w >= g_fsn_w.

ENTITY dp_frame IS
  GENERIC (
    g_fsn_w          : POSITIVE := c_dp_max_w;
    g_dat_w          : POSITIVE := c_dp_max_w
  );
  PORT (
    rst              : IN  STD_LOGIC;
    clk              : IN  STD_LOGIC;
    
    in_fsn           : IN  STD_LOGIC_VECTOR(g_fsn_w-1 DOWNTO 0);
    in_dat           : IN  STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
    in_val           : IN  STD_LOGIC;
    in_sof           : IN  STD_LOGIC;
    in_eof           : IN  STD_LOGIC;
    in_err           : IN  STD_LOGIC;
    
    out_dat          : OUT STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
    out_val          : OUT STD_LOGIC;
    out_sof          : OUT STD_LOGIC;
    out_eof          : OUT STD_LOGIC
  );
END dp_frame;


ARCHITECTURE rtl OF dp_frame IS

  -- Use unregistered input to save logic, use registered input to ease achieving timing constrains.
  CONSTANT c_input_reg   : BOOLEAN := FALSE;
  
  CONSTANT c_brc_ok      : STD_LOGIC_VECTOR(in_dat'RANGE) := STD_LOGIC_VECTOR(TO_UNSIGNED(c_dp_brc_ok ,in_dat'LENGTH));
  CONSTANT c_brc_err     : STD_LOGIC_VECTOR(in_dat'RANGE) := STD_LOGIC_VECTOR(TO_UNSIGNED(c_dp_brc_err,in_dat'LENGTH));
  
  -- Registered inputs (see c_input_reg)
  SIGNAL in_fsn_reg      : STD_LOGIC_VECTOR(in_fsn'RANGE);
  SIGNAL in_dat_reg      : STD_LOGIC_VECTOR(in_dat'RANGE);
  SIGNAL in_val_reg      : STD_LOGIC;
  SIGNAL in_sof_reg      : STD_LOGIC;
  SIGNAL in_eof_reg      : STD_LOGIC;
  SIGNAL in_err_reg      : STD_LOGIC;
  
  -- Delayed inputs to create space to insert fsn word
  SIGNAL in_dat_dly      : STD_LOGIC_VECTOR(in_dat'RANGE);
  SIGNAL in_val_dly      : STD_LOGIC;
  SIGNAL in_eof_dly      : STD_LOGIC;
  SIGNAL in_err_dly      : STD_LOGIC;
  
  -- Delay input eof and err one more to append brc word
  SIGNAL in_eof_dly2     : STD_LOGIC;
  SIGNAL in_err_dly2     : STD_LOGIC;
  
  SIGNAL i_out_dat       : STD_LOGIC_VECTOR(out_dat'RANGE);
  SIGNAL nxt_out_dat     : STD_LOGIC_VECTOR(out_dat'RANGE);
  SIGNAL nxt_out_val     : STD_LOGIC;
  SIGNAL nxt_out_sof     : STD_LOGIC;
  SIGNAL nxt_out_eof     : STD_LOGIC;
  
BEGIN

  out_dat <= i_out_dat;
  
  gen_input_reg : IF c_input_reg=TRUE GENERATE
    p_reg : PROCESS (clk, rst)
    BEGIN
      IF rst='1' THEN
        in_fsn_reg <= (OTHERS => '0');
        in_dat_reg <= (OTHERS => '0');
        in_val_reg <= '0';
        in_sof_reg <= '0';
        in_eof_reg <= '0';
        in_err_reg <= '0';
      ELSIF rising_edge(clk) THEN
        in_fsn_reg <= in_fsn;
        in_dat_reg <= in_dat;
        in_val_reg <= in_val;
        in_sof_reg <= in_sof;
        in_eof_reg <= in_eof;
        in_err_reg <= in_err;
      END IF;
    END PROCESS;
  END GENERATE;
  no_input_reg : IF c_input_reg=FALSE GENERATE
    in_fsn_reg <= in_fsn;
    in_dat_reg <= in_dat;
    in_val_reg <= in_val;
    in_sof_reg <= in_sof;
    in_eof_reg <= in_eof;
    in_err_reg <= in_err;
  END GENERATE;
  
  
  p_clk: PROCESS(clk, rst)
  BEGIN
    IF rst='1' THEN
      in_dat_dly  <= (OTHERS => '0');
      in_val_dly  <= '0';
      in_eof_dly  <= '0';
      in_err_dly  <= '0';
      in_eof_dly2 <= '0';
      in_err_dly2 <= '0';
      i_out_dat   <= (OTHERS => '0');
      out_val     <= '0';
      out_sof     <= '0';
      out_eof     <= '0';
    ELSIF rising_edge(clk) THEN
      in_dat_dly  <= in_dat_reg;
      in_val_dly  <= in_val_reg;
      in_eof_dly  <= in_eof_reg;
      in_err_dly  <= in_err_reg;
      in_eof_dly2 <= in_eof_dly;
      in_err_dly2 <= in_err_dly;
      i_out_dat   <= nxt_out_dat;
      out_val     <= nxt_out_val;
      out_sof     <= nxt_out_sof;
      out_eof     <= nxt_out_eof;
    END IF;
  END PROCESS;

  p_output: PROCESS(i_out_dat, in_fsn_reg, in_dat_dly, in_val_dly, in_sof_reg, in_eof_dly2, in_err_dly2)
  BEGIN
    -- set default nxt_out_dat
    nxt_out_dat <= i_out_dat;          -- hold output dat when input is inactive
    nxt_out_val <= '0';
    nxt_out_sof <= '0';
    nxt_out_eof <= '0';
    
    -- overrule default nxt_out_dat
    IF in_sof_reg='1' THEN             -- insert fsn, no need to delay input fsn, because it holds
      nxt_out_dat <= RESIZE_SVEC(in_fsn_reg,out_dat'LENGTH); -- sign extend is fsync, 
      nxt_out_val <= '1';
      nxt_out_sof <= '1';
    ELSIF in_eof_dly2='1' THEN
      nxt_out_dat <= c_brc_ok;         -- fill in brc, account for input err
      IF in_err_dly2='1' THEN
        nxt_out_dat <= c_brc_err;
      END IF;
      nxt_out_val <= '1';
      nxt_out_eof <= '1';
    ELSIF in_val_dly='1' THEN
      nxt_out_dat <= in_dat_dly;       -- pass on input dat
      nxt_out_val <= '1';
    END IF;
  END PROCESS;
  
END rtl;
