--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.all;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;

-- Reuse from LOFAR rad_frame_hdr.vhd and rad_frame_hdr(rtl).vhd

-- Purpose:
--   Determine start and end of data blocks and introduce the data path (dp)
--   frame sequence number (fsn) with sync and sequence number information.
-- Description:
--   The input data are the data path signals:
--     . sync
--     . val
--     . dat[]
--   The output data are the data paths signals:
--     . val
--     . dat[]
--     . sof
--     . eof
--     . fsn[]
--   The blocks have size g_block_size. There must be gaps between data blocks
--   to allow space for dp_frame and dp_frame_tx which insert >= 1 idle, sfd,
--   fsn and brc/crc into the stream. Therefore in_val must go low for at
--   least 4 cycles between data blocks. The sof marks the start of block and
--   the eof marks the end of the data block.
--   For the first block after a sync the msbit of the fsn is '1' and the frame
--   sequence number is forced to restart at 0. The fsn increments for the
--   subsequent data blocks. The fsn may wrap several times during a sync
--   interval. The fsn word is valid at the sof and during the frame. The new
--   fsn becomes valid directly after the eof or after an in_sync, that
--   restarts the fsn. The in_sync may be high for an entire data block. The
--   falling edge of in_sync is used internally and therefor the first cycle
--   after the last active in_sync must have in_val low (i.e. fall in the gap
--   between data blocks).
--   If g_use_sync is FALSE then the periodic in_sync is not used and the
--   function starts packetizing the data at the first valid input and will
--   remain in state s_data forever.
--   The function verifies that subsequent in_sync occur on a block boundary,
--   if not then the output flushes the rest of the block and then disables and
--   tries to recover at the next in_sync. In practise such an block error can
--   not occur, because the input signals come from internal circuitry in the
--   FPGA so if that goes wrong, then the entire FPGA logic state is undefined. 
--   If the input data can be reset independent of this function then a
--   fractional data block can occur, so then the flush capability becomes
--   useful. Furthermore flush capability ensures that only full blocks are
--   output or nothing, similar as for dp_frame_rx. The error is output via
--   out_err which is valid in the cycle directly after out_eof. The g_fsn_w
--   must be <= g_dat_w.

ENTITY dp_frame_fsn IS
  GENERIC (
    g_fsn_w        : POSITIVE := 16;    
    g_dat_w        : POSITIVE := 16;    
    g_use_sync     : BOOLEAN  := TRUE;
    g_block_size   : POSITIVE := 1000
  );
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;    
    in_sync        : IN  STD_LOGIC := '0';
    in_val         : IN  STD_LOGIC;
    in_dat         : IN  STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);    
    out_fsn        : OUT STD_LOGIC_VECTOR(g_fsn_w-1 DOWNTO 0);
    out_dat        : OUT STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);    
    out_val        : OUT STD_LOGIC;
    out_sof        : OUT STD_LOGIC;
    out_eof        : OUT STD_LOGIC;
    out_err        : OUT STD_LOGIC
  );
END dp_frame_fsn;


ARCHITECTURE rtl OF dp_frame_fsn IS
  
  TYPE state_type IS (s_idle, s_data, s_flush);
  
  -- Use unregistered input to save logic, use registered input to ease achieving timing constrains.
  CONSTANT c_input_reg     : BOOLEAN := FALSE;
  
  -- Registered inputs (see c_input_reg)
  SIGNAL in_sync_reg       : STD_LOGIC;
  SIGNAL in_val_reg        : STD_LOGIC;
  SIGNAL in_dat_reg        : STD_LOGIC_VECTOR(in_dat'RANGE);
  
  SIGNAL nxt_out_dat       : STD_LOGIC_VECTOR(out_dat'RANGE);
  SIGNAL nxt_out_val       : STD_LOGIC;
  
  SIGNAL state             : state_type;
  SIGNAL nxt_state         : state_type;
  SIGNAL rst_state         : state_type;
  
  -- sync
  SIGNAL prev_sync         : STD_LOGIC;
  SIGNAL sync_fevt         : STD_LOGIC;
  
  -- frame sync and frame sequence number    
  SIGNAL fsync             : STD_LOGIC;
  SIGNAL nxt_fsync         : STD_LOGIC;
  SIGNAL fsn               : STD_LOGIC_VECTOR(g_fsn_w-2 DOWNTO 0);  -- MSbit of out_fsn will hold fsync
  SIGNAL nxt_fsn           : STD_LOGIC_VECTOR(fsn'RANGE);  
  
  -- sample counter 
  SIGNAL cnt               : NATURAL RANGE 0 TO g_block_size-1;
  SIGNAL nxt_cnt           : NATURAL;
  SIGNAL cnt_is_zero       : STD_LOGIC;
  SIGNAL cnt_is_block_size : STD_LOGIC;
  
  -- start, end of frame  
  SIGNAL nxt_out_sof       : STD_LOGIC;
  SIGNAL i_out_eof         : STD_LOGIC;
  SIGNAL nxt_out_eof       : STD_LOGIC;
  SIGNAL nxt_out_err       : STD_LOGIC;
  
BEGIN

  out_eof <= i_out_eof;
  
  -- Frame sequence number is composed of fsync and fsn
  out_fsn <= fsync & fsn;

  gen_input_reg : IF c_input_reg=TRUE GENERATE
    p_reg : PROCESS (clk, rst)
    BEGIN
      IF rst='1' THEN
        in_sync_reg <= '0';
        in_val_reg  <= '0';
        in_dat_reg  <= (OTHERS => '0');
      ELSIF rising_edge(clk) THEN
        in_sync_reg <= in_sync;
        in_val_reg  <= in_val;
        in_dat_reg  <= in_dat;
      END IF;
    END PROCESS;
  END GENERATE;
  no_input_reg : IF c_input_reg=FALSE GENERATE
    in_sync_reg <= in_sync;
    in_val_reg  <= in_val;
    in_dat_reg  <= in_dat;
  END GENERATE;
    
  p_clk: PROCESS(clk, rst)
  BEGIN
    IF rst='1' THEN
      fsync     <= '0';
      fsn       <= (OTHERS => '0');
      prev_sync <= '0';
      cnt       <= 0;
      out_dat   <= (OTHERS => '0');
      out_val   <= '0';
      out_sof   <= '0';
      i_out_eof <= '0';
      out_err   <= '0';
      state     <= rst_state;
    ELSIF rising_edge(clk) THEN
      fsync     <= nxt_fsync;
      fsn       <= nxt_fsn;
      prev_sync <= in_sync_reg;
      cnt       <= nxt_cnt;
      out_dat   <= nxt_out_dat;
      out_val   <= nxt_out_val;
      out_sof   <= nxt_out_sof;
      i_out_eof <= nxt_out_eof;
      out_err   <= nxt_out_err;
      state     <= nxt_state;
    END IF;
  END PROCESS;
  
  rst_state <= s_idle WHEN g_use_sync=TRUE ELSE s_data;
  
  sync_fevt <= prev_sync AND NOT in_sync_reg;
  
  cnt_is_zero       <= '1' WHEN cnt=0              ELSE '0';
  cnt_is_block_size <= '1' WHEN cnt=g_block_size-1 ELSE '0';

  p_fsn: PROCESS(sync_fevt, fsync, fsn, i_out_eof)
  BEGIN
    nxt_fsync <= fsync;
    nxt_fsn   <= fsn;
    IF sync_fevt='1' THEN
      nxt_fsync <= '1';                 -- store sync
      nxt_fsn   <= (OTHERS => '0');     -- restart fsn
    ELSIF i_out_eof='1' THEN
      nxt_fsync <= '0';
      nxt_fsn   <= STD_LOGIC_VECTOR(UNSIGNED(fsn)+1);
    END IF;
  END PROCESS;
  
  p_state: PROCESS(state, cnt, cnt_is_zero, cnt_is_block_size, sync_fevt, in_dat_reg, in_val_reg)
  BEGIN
    nxt_out_dat <= in_dat_reg;
    nxt_out_val <= in_val_reg;
    nxt_out_sof <= '0';
    nxt_out_eof <= '0';
    nxt_out_err <= '0';
    nxt_state   <= state;
    nxt_cnt     <= cnt;
    IF in_val_reg='1' THEN
      nxt_cnt <= cnt+1;
      IF cnt_is_block_size='1' THEN
        nxt_cnt <= 0;
      END IF;
    END IF;
    
    CASE state IS
      WHEN s_idle =>                    -- no output until in_sync_reg goes inactive
        nxt_out_val <= '0';
        nxt_cnt <= 0;
        IF sync_fevt='1' THEN
          nxt_state <= s_data;
        END IF;
      WHEN s_data =>
        IF in_val_reg='1' THEN          -- block dat sof and eof
          IF cnt_is_zero='1' THEN
            nxt_out_sof <='1';
          ELSIF cnt_is_block_size='1' THEN
            nxt_out_eof <='1';
          END IF;
        END IF;
        IF sync_fevt='1' AND cnt_is_zero/='1' THEN
          nxt_state <= s_flush;         -- fractional nof data blocks in sync interval
        END IF;
      WHEN OTHERS => -- s_flush
        nxt_out_val <= '1';
        nxt_cnt <= cnt+1;               -- flush full block with undefined out_dat[]
        IF cnt_is_block_size='1' THEN
          nxt_out_err <= '1';           -- signal error, will propagate via brc, crc
          nxt_out_eof <='1';
          nxt_cnt <= 0;
          nxt_state <= s_idle;          -- try to recover at next sync_fevt
        END IF;
    END CASE;
  END PROCESS;
  
END rtl;
