--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_packetizing_pkg.ALL;

-- Reuse from LOFAR rad_frame_rx.vhd and rad_frame_rx(rtl).vhd

-- Purpose:
--   From the input dp frame PHY stream:
--
--          -----------------------------
--     idle | sfd | fsn | dat ... | crc | idle ...
--          -----------------------------
--
--   detect and output the dp frames for g_sfd:
--
--                -----------------------
--                | fsn | dat ... | brc |
--                -----------------------
--                  sof             eof
--                  val   val ...   val
--
--   Mark the sof and eof of each frame and check the received CRC
-- Description:
--   A dp PHY frame consists of g_sfd, fsn, data and crc. The dp PHY frames
--   are seperated by at least one c_dp_idle word. The received crc is passed
--   on as boolean crc called bcr, where '0' indicates OK and '1' indicates
--   that an crc error.
--   The sof is detected based on a c_dp_idle to g_sfd word transition and the
--   eof is then determined by counting g_block_size data words. The g_sfd word
--   is not passed on, because we have the sof signal. Therefore the internal
--   frame consists of fsn, dat, bcr fields and seperate sof and eof signals.
--   The sof signal is active during the fsn and the eof is active during the
--   brc.
--   While counting valid data words the module is insensitive to c_dp_idle/
--   g_sfd. Hence the output is always a full frame, if for some reason the
--   input framing got corrupted, then this will reflect in the crc so the brc
--   will note error. Perhaps even a frame may get lost, but the frame
--   reception will recover on a next sof detection. Idle words between
--   internal frames are simply passed on, which is fine since out_val is
--   inactive.
-- Remark:
-- . Frame output is flow controlled on frame level only, by the out_xon input.
-- . Typically choose the rx timeout g_timeout_w=0 if a link is fixed after
--   power up (e.g. for RSP RI). If the link can stop and start asynchronously
--   (e.g. changing cables, transmitter restart) then choose g_timeout_w
--   sufficient to fit the frame period (e.g. for RSP Serdes).
-- . In p_state the rx timeout restart after every active in_val_reg, this in
--   principle allows using a smaller timeout that is sufficient to span the
--   inactive gaps of in_val_reg, so then g_timeout_w <= c_rsp_slice_size is
--   suficient. An alternative implementation for the timeout would be to let
--   the timeout run from the sop, the timeout must then at least span the
--   frame period. The frame period can vary a somewhat due to mutliplexing
--   so with c_rsp_slice_size the g_timeout_w would then have to be set to
--   c_rsp_slice_size+1

ENTITY dp_frame_rx IS
  GENERIC (
    g_sfd          : STD_LOGIC_VECTOR := c_dp_sfd;
    g_lofar        : BOOLEAN := FALSE;  -- when TRUE use same CRC as for LOFAR
    g_dat_w        : NATURAL := c_dp_max_w;
    g_block_size   : NATURAL := 1000;
    g_timeout_w    : NATURAL := 0    -- when 0 then no timeout else when > 0 then flush pending rx payload after 2**g_timeout_w clk cylces of inactive in_val
  );
  PORT (
    rst      : IN  STD_LOGIC;
    clk      : IN  STD_LOGIC;
    in_dat   : IN  STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
    in_val   : IN  STD_LOGIC := '1';
    out_dat  : OUT STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
    out_val  : OUT STD_LOGIC;
    out_sof  : OUT STD_LOGIC;
    out_eof  : OUT STD_LOGIC;
    out_xon  : IN  STD_LOGIC := '1'
  );
END dp_frame_rx;


ARCHITECTURE rtl OF dp_frame_rx IS

  -- Use unregistered input to save logic, use registered input to ease achieving timing constrains.
  CONSTANT c_input_reg   : BOOLEAN := FALSE;
  
  CONSTANT c_idle        : STD_LOGIC_VECTOR := c_dp_idle(in_dat'RANGE);
  CONSTANT c_sfd         : STD_LOGIC_VECTOR := g_sfd(in_dat'RANGE);
  CONSTANT c_brc_ok      : STD_LOGIC_VECTOR := STD_LOGIC_VECTOR(TO_UNSIGNED(c_dp_brc_ok ,g_dat_w));
  CONSTANT c_brc_err     : STD_LOGIC_VECTOR := STD_LOGIC_VECTOR(TO_UNSIGNED(c_dp_brc_err,g_dat_w));
  
  CONSTANT c_timeout_cnt_w : NATURAL := g_timeout_w+1;
  
  TYPE t_state IS (s_sof, s_fsn, s_eof, s_flush);
  
  SIGNAL crc             : STD_LOGIC_VECTOR(c_dp_crc_w-1 DOWNTO 0);
  SIGNAL nxt_crc         : STD_LOGIC_VECTOR(crc'RANGE);
  SIGNAL cnt             : NATURAL RANGE 0 TO g_block_size;
  SIGNAL nxt_cnt         : NATURAL;

  -- Registered inputs (see c_input_reg)
  SIGNAL in_val_reg      : STD_LOGIC;
  SIGNAL in_dat_reg      : STD_LOGIC_VECTOR(in_dat'RANGE);
  
  SIGNAL prev_in_dat_reg : STD_LOGIC_VECTOR(in_dat'RANGE);
  SIGNAL nxt_out_dat     : STD_LOGIC_VECTOR(out_dat'RANGE);
  SIGNAL nxt_out_val     : STD_LOGIC;
  SIGNAL nxt_out_sof     : STD_LOGIC;
  SIGNAL nxt_out_eof     : STD_LOGIC;
  
  SIGNAL xon_reg         : STD_LOGIC;
  SIGNAL nxt_xon_reg     : STD_LOGIC;
  
  SIGNAL state           : t_state;
  SIGNAL nxt_state       : t_state;
  
  -- handle rx timeout
  SIGNAL timeout_cnt          : STD_LOGIC_VECTOR(c_timeout_cnt_w-1 DOWNTO 0);
  SIGNAL timeout_cnt_clr      : STD_LOGIC;
  SIGNAL nxt_timeout_cnt_clr  : STD_LOGIC;
  SIGNAL timeout_evt          : STD_LOGIC;
  SIGNAL flush_en             : STD_LOGIC;
  SIGNAL nxt_flush_en         : STD_LOGIC;
  SIGNAL flush_val            : STD_LOGIC;
  
  PROCEDURE proc_handle_rx_timeout(SIGNAL   valid       : IN    STD_LOGIC;
                                   SIGNAL   timeout_evt : IN    STD_LOGIC;
                                   SIGNAL   clr         : OUT   STD_LOGIC;
                                   VARIABLE v_state     : INOUT t_state) IS  -- use variable v_state instead of signal to avoid getting latches
  BEGIN
    IF valid='1' THEN
      clr <= '1';              -- restart timeout_cnt during frame rx and remain in current state
    ELSE
      clr <= '0';              -- let valid inactive timeout count increment
      IF timeout_evt='1' THEN  -- can only occur when g_timeout_w>0
        v_state := s_flush;    -- exit to flush state to finish the current rx frame with void data
      END IF;
    END IF;
  END proc_handle_rx_timeout;
  
BEGIN

  gen_input_reg : IF c_input_reg=TRUE GENERATE
    p_reg : PROCESS (clk, rst)
    BEGIN
      IF rst='1' THEN
        in_val_reg <= '0';
        in_dat_reg <= (OTHERS => '0');
      ELSIF rising_edge(clk) THEN
        in_val_reg <= in_val;
        in_dat_reg <= in_dat;
      END IF;
    END PROCESS;
  END GENERATE;
  no_input_reg : IF c_input_reg=FALSE GENERATE
    in_val_reg <= in_val;
    in_dat_reg <= in_dat;
  END GENERATE;
  
  p_clk : PROCESS (clk, rst)
  BEGIN
    IF rst='1' THEN
      timeout_cnt_clr <= '1';
      flush_en        <= '0';
      crc             <= (OTHERS => '1');
      cnt             <= 0;
      prev_in_dat_reg <= (OTHERS => '0');
      out_dat         <= (OTHERS => '0');
      out_val         <= '0';
      out_sof         <= '0';
      out_eof         <= '0';
      xon_reg         <= '0';
      state           <= s_sof;
    ELSIF rising_edge(clk) THEN
      timeout_cnt_clr <= nxt_timeout_cnt_clr;
      flush_en        <= nxt_flush_en;
      crc             <= nxt_crc;
      cnt             <= nxt_cnt;
      prev_in_dat_reg <= in_dat_reg;
      out_dat         <= nxt_out_dat;
      out_val         <= nxt_out_val;
      out_sof         <= nxt_out_sof;
      out_eof         <= nxt_out_eof;
      xon_reg         <= nxt_xon_reg;
      state           <= nxt_state;
    END IF;
  END PROCESS;
  
  ------------------------------------------------------------------------------
  -- Rx timeout control
  ------------------------------------------------------------------------------
  no_timeout : IF g_timeout_w=0 GENERATE
    timeout_evt <= '0';
    flush_val   <= '0';
  END GENERATE;
  
  gen_timeout : IF g_timeout_w>0 GENERATE
    u_timeout_cnt : ENTITY common_lib.common_counter
    GENERIC MAP (
      g_width     => c_timeout_cnt_w
    )
    PORT MAP (
      rst     => rst,
      clk     => clk,
      cnt_clr => timeout_cnt_clr,
      count   => timeout_cnt
    );
    
    timeout_evt <= timeout_cnt(g_timeout_w);  -- check MSbit for timeout of 2**g_timeout_w clk cycles
    
    ASSERT NOT(rising_edge(clk) AND timeout_evt='1') REPORT "rad_frame_rx timeout occurred!" SEVERITY WARNING;
    
    -- define flush_val to prepare for throttle support of out_val during flush, default no throttle support yet
    flush_val <= flush_en;
  END GENERATE;
  
  ------------------------------------------------------------------------------
  -- XON/XOFF rx frame flow control
  ------------------------------------------------------------------------------
  nxt_xon_reg <= out_xon;  -- register out_xon to ease timing closure, because out_xon may come from far on the chip and some more cycles latency for xon is acceptable
  
  ------------------------------------------------------------------------------
  -- Rx state machine
  ------------------------------------------------------------------------------
  p_state: PROCESS(state, crc, cnt, in_val_reg, in_dat_reg, prev_in_dat_reg, xon_reg, timeout_evt, flush_val)
    VARIABLE v_nxt_state : t_state;  -- use variable instead of signal to avoid getting latches with proc_handle_rx_timeout()
  BEGIN
    nxt_timeout_cnt_clr  <= '1';                                 -- default no timeout_cnt
    nxt_flush_en         <= '0';
    
    nxt_crc     <= crc;
    nxt_cnt     <= cnt;
    nxt_out_dat <= in_dat_reg;                                   -- default pass on input dat
    nxt_out_val <= '0';
    nxt_out_sof <= '0';
    nxt_out_eof <= '0';
    v_nxt_state := state;
    
    IF in_val_reg='1' THEN
      nxt_crc <= func_dp_next_crc(g_lofar, in_dat_reg, crc);
    END IF;
      
    CASE state IS
      WHEN s_sof =>
        nxt_crc <= (OTHERS => '1');
        nxt_cnt <= 0;
        IF xon_reg = '1' THEN                                    -- flush any input rx frame if xon_reg is XOFF
          IF in_val_reg='1' THEN
            IF in_dat_reg=c_sfd AND prev_in_dat_reg=c_idle THEN  -- ignore in_val during c_idle
              v_nxt_state := s_fsn;
            END IF;
          END IF;
        END IF;
      WHEN s_fsn =>
        IF in_val_reg='1' THEN
          nxt_out_val <= '1';
          nxt_out_sof <= '1';
          v_nxt_state := s_eof;
        END IF;
        -- Exit to s_flush in case of rx timeout
        proc_handle_rx_timeout(in_val_reg, timeout_evt, nxt_timeout_cnt_clr, v_nxt_state);
      WHEN s_eof =>
        IF in_val_reg='1' THEN
          nxt_out_val <= '1';
          IF cnt=g_block_size THEN                             -- fill in brc based on rx crc result
            nxt_out_eof <= '1';
            nxt_out_dat <= c_brc_ok;
            IF c_dp_crc_w>g_dat_w THEN
              IF crc(in_dat'RANGE) /= in_dat_reg THEN
                nxt_out_dat <= c_brc_err;
              END IF;
            ELSE
              IF crc /= in_dat_reg(crc'RANGE) THEN
                nxt_out_dat <= c_brc_err;
              END IF;
            END IF;
            v_nxt_state := s_sof;
          ELSE
            nxt_cnt <= cnt+1;                                  -- count payload data
          END IF;
        END IF;
        -- Exit to s_flush in case of rx timeout
        proc_handle_rx_timeout(in_val_reg, timeout_evt, nxt_timeout_cnt_clr, v_nxt_state);
      WHEN OTHERS => -- s_flush                                -- this state can only be reached if g_timeout_w>0
        nxt_flush_en <= '1';                                   -- enable flush frame due to in_val_reg inactive timeout
        IF flush_val='1' THEN
          nxt_out_val <= '1';                                  -- flush rx frame with void data
          IF cnt>=g_block_size THEN
            nxt_out_eof <= '1';
            nxt_out_dat <= c_brc_err;                          -- fill in c_brc_err for flushed rx frame
            v_nxt_state := s_sof;
          ELSE
            nxt_cnt <= cnt+1;                                  -- count payload data
          END IF;
        END IF;
    END CASE;
    
    nxt_state <= v_nxt_state;
  END PROCESS;
  
END rtl;
