--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Reuse from LOFAR rad_frame_scheduler.vhd and rad_frame_scheduler(rtl).vhd

-- Purpose:
--   Schedule one or more input frames streams into one output stream.
-- Description:
--   The input frame that arrives first is transmitted. Then if the next input
--   has a frame pending then that is transmitted.
--   The scheduler starts transmitting data from a FIFO if it has been filled
--   with more than g_fifo_fill words. Given a fixed frame length, this is
--   useful when the in_val is throttled while the out_val should not be
--   inactive valid between out_sof to out_eof.
--   The required g_fifo_size per input depends on the relative frame rates,
--   typically it must be sufficient to fit at least two frames.
-- Remark:
-- . Using g_nof_input=1 is useful to have an input FIFO that can be filled
--   sufficiently to level set by g_fifo_fill to ensure that the output frame
--   has data valid without gaps.
-- . The inputs can dynamically be disabled or enabled via in_dis. The
--   dp_xonoff ensures that in_dis is applied during entire frames and before
--   they are written into the input FIFO.

LIBRARY IEEE,common_lib, technology_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY dp_frame_scheduler IS
  GENERIC (
    g_technology : NATURAL := c_tech_select_default;
    g_dat_w      : NATURAL := 16;
    g_nof_input  : NATURAL := 2;                   -- >= 1
    g_fifo_rl    : NATURAL := 1;                   -- for all input use 0 for look ahead FIFO, 1 for normal FIFO
    g_fifo_size  : t_natural_arr := (1024, 1024);  -- must match g_nof_input-1 DOWNTO 0
    g_fifo_fill  : t_natural_arr := (0, 0)         -- must match g_nof_input-1 DOWNTO 0
  );
  PORT (
    rst      : IN  STD_LOGIC;
    clk      : IN  STD_LOGIC;
    gp_out   : OUT STD_LOGIC_VECTOR(g_nof_input        -1 DOWNTO 0);
    in_dat   : IN  STD_LOGIC_VECTOR(g_nof_input*g_dat_w-1 DOWNTO 0);
    in_val   : IN  STD_LOGIC_VECTOR(g_nof_input        -1 DOWNTO 0);
    in_sof   : IN  STD_LOGIC_VECTOR(g_nof_input        -1 DOWNTO 0);
    in_eof   : IN  STD_LOGIC_VECTOR(g_nof_input        -1 DOWNTO 0);
    in_dis   : IN  STD_LOGIC_VECTOR(g_nof_input        -1 DOWNTO 0) := (OTHERS=>'0');  -- optional input disable control at frame level
    out_dat  : OUT STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
    out_val  : OUT STD_LOGIC;
    out_sof  : OUT STD_LOGIC;
    out_eof  : OUT STD_LOGIC
  );
END dp_frame_scheduler;


ARCHITECTURE rtl OF dp_frame_scheduler IS

  -- Use unregistered input to save logic, use registered input to ease achieving timing constrains.
  CONSTANT c_input_reg       : BOOLEAN := FALSE;
  
  -- Convert unconstrained range (that starts at INTEGER'LEFT) to g_nof_input-1 DOWNTO 0 range
  CONSTANT c_fifo_size       : t_natural_arr(g_nof_input-1 DOWNTO 0) := g_fifo_size;
  CONSTANT c_fifo_fill       : t_natural_arr(g_nof_input-1 DOWNTO 0) := g_fifo_fill;
  
  TYPE t_dat_arr        IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);

  SIGNAL dbg_c_fifo_size : t_natural_arr(c_fifo_size'RANGE) := c_fifo_size;
  SIGNAL dbg_c_fifo_fill : t_natural_arr(c_fifo_fill'RANGE) := c_fifo_fill;
  
  -- Registered inputs (see c_input_reg)
  SIGNAL in_dat_reg      : t_dat_arr(0 TO g_nof_input-1);
  SIGNAL in_val_reg      : t_sl_arr( 0 TO g_nof_input-1);
  SIGNAL in_sof_reg      : t_sl_arr( 0 TO g_nof_input-1);
  SIGNAL in_eof_reg      : t_sl_arr( 0 TO g_nof_input-1);
  SIGNAL snk_in          : t_dp_sosi_arr(0 TO g_nof_input-1);
  
  SIGNAL wr_siso         : t_dp_siso_arr(0 TO g_nof_input-1);
  SIGNAL wr_sosi         : t_dp_sosi_arr(0 TO g_nof_input-1);

  SIGNAL rd_req          : t_sl_arr( 0 TO g_nof_input-1);
  SIGNAL rd_dat          : t_dat_arr(0 TO g_nof_input-1);
  SIGNAL rd_val          : t_sl_arr( 0 TO g_nof_input-1);
  SIGNAL rd_sof          : t_sl_arr( 0 TO g_nof_input-1);
  SIGNAL rd_eof          : t_sl_arr( 0 TO g_nof_input-1);
  SIGNAL rd_siso         : t_dp_siso_arr(0 TO g_nof_input-1);
  SIGNAL rd_sosi         : t_dp_sosi_arr(0 TO g_nof_input-1);
  
  TYPE state_type IS (s_idle, s_output);
  
  SIGNAL state           : state_type;
  SIGNAL nxt_state       : state_type;
  
  SIGNAL in_sel          : NATURAL RANGE 0 TO g_nof_input-1;
  SIGNAL nxt_in_sel      : NATURAL;
  
  SIGNAL nxt_out_dat     : STD_LOGIC_VECTOR(out_dat'RANGE);
  SIGNAL nxt_out_val     : STD_LOGIC;
  SIGNAL nxt_out_sof     : STD_LOGIC;
  SIGNAL nxt_out_eof     : STD_LOGIC;
  
BEGIN

  gen_input_reg : IF c_input_reg=TRUE GENERATE
    p_reg : PROCESS (clk, rst)
    BEGIN
      IF rst='1' THEN
        in_dat_reg <= (OTHERS =>(OTHERS => '0'));
        in_val_reg <= (OTHERS => '0');
        in_sof_reg <= (OTHERS => '0');
        in_eof_reg <= (OTHERS => '0');
      ELSIF rising_edge(clk) THEN
        FOR I IN 0 TO g_nof_input-1 LOOP
          in_dat_reg(I) <= in_dat((I+1)*g_dat_w-1 DOWNTO I*g_dat_w);
          in_val_reg(I) <= in_val(                       I);
          in_sof_reg(I) <= in_sof(                       I);
          in_eof_reg(I) <= in_eof(                       I);
        END LOOP;
      END IF;
    END PROCESS;
  END GENERATE;
  no_input_reg : IF c_input_reg=FALSE GENERATE
    p_in_reg : PROCESS(in_dat, in_val, in_sof, in_eof)
    BEGIN
      FOR I IN 0 TO g_nof_input-1 LOOP
        in_dat_reg(I) <= in_dat((I+1)*g_dat_w-1 DOWNTO I*g_dat_w);
        in_val_reg(I) <= in_val(                       I);
        in_sof_reg(I) <= in_sof(                       I);
        in_eof_reg(I) <= in_eof(                       I);
      END LOOP;
    END PROCESS;
  END GENERATE;

  gen_fifo : FOR I IN 0 TO g_nof_input-1 GENERATE
    snk_in(I).data(g_dat_w-1 DOWNTO 0) <= in_dat_reg(I);
    snk_in(I).valid                    <= in_val_reg(I);
    snk_in(I).sop                      <= in_sof_reg(I);
    snk_in(I).eop                      <= in_eof_reg(I);
    
    -- Input disable control at frame level
    wr_siso(I).ready <= '1';
    wr_siso(I).xon   <= NOT in_dis(I);
    
    u_xonoff : ENTITY work.dp_xonoff
    PORT MAP (
      rst           => rst,
      clk           => clk,
      -- Frame in
      in_siso       => OPEN,
      in_sosi       => snk_in(I),
      -- Frame out
      out_siso      => wr_siso(I),  -- flush control via xon, ready is not used and only passed on
      out_sosi      => wr_sosi(I)
    );
      
    -- Input FIFO
    rd_siso(I).ready <= rd_req(I);
    
    rd_dat(I) <= rd_sosi(I).data(g_dat_w-1 DOWNTO 0);
    rd_val(I) <= rd_sosi(I).valid;
    rd_sof(I) <= rd_sosi(I).sop;
    rd_eof(I) <= rd_sosi(I).eop;
    
    u_fill : ENTITY work.dp_fifo_fill
    GENERIC MAP (
      g_technology  => g_technology,
      g_data_w      => g_dat_w,
      g_empty_w     => 1,
      g_channel_w   => 1,
      g_error_w     => 1,
      g_use_empty   => FALSE,
      g_use_channel => FALSE,
      g_use_error   => FALSE,
      g_fifo_fill   => c_fifo_fill(I),
      g_fifo_size   => c_fifo_size(I),
      g_fifo_rl     => g_fifo_rl
    )
    PORT MAP (
      rst         => rst,
      clk         => clk,
      -- ST sink
      snk_out     => OPEN,        -- OUT = request to upstream ST source
      snk_in      => wr_sosi(I),
      -- ST source
      src_in      => rd_siso(I),  -- IN  = request from downstream ST sink
      src_out     => rd_sosi(I),

      wr_ful      => gp_out(I)
    );
  END GENERATE;
  
  -- Output select
  nxt_out_dat <= rd_dat(in_sel);
  nxt_out_val <= rd_val(in_sel);
  nxt_out_sof <= rd_sof(in_sel);
  nxt_out_eof <= rd_eof(in_sel);
    
  p_clk: PROCESS(clk, rst)
  BEGIN
    IF rst='1' THEN
      in_sel     <= 0;
      state      <= s_idle;
      out_dat    <= (OTHERS=>'0');
      out_val    <= '0';
      out_sof    <= '0';
      out_eof    <= '0';
    ELSIF rising_edge(clk) THEN
      in_sel     <= nxt_in_sel;
      state      <= nxt_state;
      out_dat    <= nxt_out_dat;
      out_val    <= nxt_out_val;
      out_sof    <= nxt_out_sof;
      out_eof    <= nxt_out_eof;
    END IF;
  END PROCESS;
    
  p_state : PROCESS(state, in_sel, rd_sof, rd_eof)
  BEGIN
    rd_req      <= (OTHERS=>'0');
    nxt_in_sel  <= in_sel;
    nxt_state   <= state;
    
    CASE state IS
      WHEN s_idle =>
        IF rd_sof(in_sel)='1' THEN
          -- if the sof is there, then start outputting the frame
          rd_req(in_sel) <= '1';
          nxt_state <= s_output;
        ELSE
          -- else select next input
          IF in_sel=g_nof_input-1 THEN
            nxt_in_sel       <= 0;
            rd_req(0)        <= '1';
          ELSE
            nxt_in_sel       <= in_sel+1;
            rd_req(in_sel+1) <= '1';
          END IF;
        END IF;
      WHEN OTHERS => -- s_output
        -- output the frame
        rd_req(in_sel) <= '1';
        IF rd_eof(in_sel)='1' THEN
          rd_req(in_sel) <= '0'; -- eof so stop reading the FIFO, this also prevents reading an immediate next rd_sof for this in_sel to allow selecting next input in s_idle
          nxt_state <= s_idle;
        END IF;
    END CASE;     
  END PROCESS;
    
END rtl;
