-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Reuse from LOFAR rad_frame_status.vhd and rad_frame_status(rtl).vhd

-- Purpose:
--   This function the status of the frame processing in RAD. It counts the number of received frames and
--   reports whether there occured brc errors and whether frames were discarded due to misalignment. The
--   status is reset for every sync interval.
-- Description:
--   The status output is aggregate input discarded & frame sync & aggregate brc & count nof rx frames. The
--   status applies to the interval before the previous sync interval, due to the data processing latency.


LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;

ENTITY dp_frame_status IS
  GENERIC (
    g_dat_w    : POSITIVE := 32;
    g_hdr_w    : POSITIVE := 16;
    g_cnt_w    : POSITIVE := 18;
    g_status_w : POSITIVE := 32
  );
  PORT (
    rst        : IN  STD_LOGIC;
    clk        : IN  STD_LOGIC;
    sync       : IN  STD_LOGIC;
    
    in_dat     : IN  STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
    in_val     : IN  STD_LOGIC;
    in_sof     : IN  STD_LOGIC;
    in_eof     : IN  STD_LOGIC;
    in_dis     : IN  STD_LOGIC;
    
    status     : OUT STD_LOGIC_VECTOR(g_status_w-1 DOWNTO 0)
  );
END dp_frame_status;

ARCHITECTURE rtl OF dp_frame_status IS

  SIGNAL i_status      : STD_LOGIC_VECTOR(status'RANGE); 
  SIGNAL debug_status  : STD_LOGIC_VECTOR(status'RANGE); 
  SIGNAL debug_dis     : STD_LOGIC;
  SIGNAL debug_fsync   : STD_LOGIC;
  SIGNAL debug_brc     : STD_LOGIC;
  SIGNAL debug_cnt     : STD_LOGIC_VECTOR(g_cnt_w-1 DOWNTO 0);
  
  SIGNAL sync_dly      : STD_LOGIC;                             -- fdelay sync one cycle to compensate for payevent delay in ei_evt.
  
  SIGNAL fsync         : STD_LOGIC;                             -- frame sync from in_dat header at in_sof
  SIGNAL fsync_det     : STD_LOGIC;
  SIGNAL fsync_hld     : STD_LOGIC;
  SIGNAL nxt_fsync_hld : STD_LOGIC;
  
  SIGNAL brc           : STD_LOGIC;                             -- frame brc from in_dat brc at in_eof
  SIGNAL brc_det       : STD_LOGIC;                             -- aggregate brc
  SIGNAL brc_hld       : STD_LOGIC;
  SIGNAL nxt_brc_hld   : STD_LOGIC;
  
  SIGNAL dis_det       : STD_LOGIC;                             -- frame got discarded due to mis alignment
  SIGNAL dis_hld       : STD_LOGIC;                             -- aggregate discarded
  SIGNAL nxt_dis_hld   : STD_LOGIC;
  
  SIGNAL cnt           : STD_LOGIC_VECTOR(g_cnt_w-1 DOWNTO 0);  -- frame count
  SIGNAL nxt_cnt       : STD_LOGIC_VECTOR(cnt'RANGE);
  SIGNAL cnt_hld       : STD_LOGIC_VECTOR(cnt'RANGE);
  SIGNAL nxt_cnt_hld   : STD_LOGIC_VECTOR(cnt'RANGE);
  
BEGIN

  status   <= i_status;
  i_status <= hton(RESIZE_UVEC(dis_hld & fsync_hld & brc_hld & cnt_hld, g_status_w), c_word_sz);
  -- synthesis translate_off
  -- Signals for simulation wave window
  debug_status <= hton(i_status, c_word_sz);
  debug_dis    <= debug_status(g_cnt_w+2);
  debug_fsync  <= debug_status(g_cnt_w+1);
  debug_brc    <= debug_status(g_cnt_w);
  debug_cnt    <= debug_status(g_cnt_w-1 DOWNTO 0);
  -- synthesis translate_on

  p_clk: PROCESS(clk, rst)
  BEGIN
    IF rst='1' THEN
      sync_dly  <= '0';
      fsync_hld <= '0';
      brc_hld   <= '0';
      dis_hld   <= '0';
      cnt       <= (OTHERS => '0');
      cnt_hld   <= (OTHERS => '0');
    ELSIF rising_edge(clk) THEN
      sync_dly  <= sync;
      fsync_hld <= nxt_fsync_hld;
      brc_hld   <= nxt_brc_hld;
      dis_hld   <= nxt_dis_hld;
      cnt       <= nxt_cnt;
      cnt_hld   <= nxt_cnt_hld;
    END IF;
  END PROCESS;
  
  fsync <= '1' WHEN in_dat(g_hdr_w-1)='1' AND in_sof='1' ELSE '0';
  brc   <= '1' WHEN in_dat(0)        ='1' AND in_eof='1' ELSE '0';
  
  -- use sync_dly also for fsync_hld to let all status fields apply to the same sync interval
  nxt_fsync_hld <= fsync_det                                     WHEN sync_dly='1' ELSE fsync_hld;
  nxt_brc_hld   <= brc_det   WHEN fsync='1' ELSE            '0'  WHEN sync_dly='1' ELSE brc_hld;
  nxt_dis_hld   <= dis_det   WHEN fsync='1' ELSE            '0'  WHEN sync_dly='1' ELSE dis_hld;
  nxt_cnt_hld   <= cnt       WHEN fsync='1' ELSE (OTHERS => '0') WHEN sync_dly='1' ELSE cnt_hld;
  
  -- frame sync detection
  u_fsync_det : ENTITY common_lib.common_switch
  PORT MAP (
    clk         => clk,
    rst         => rst,
    switch_high => fsync,     -- frame sync
    switch_low  => sync_dly,  -- system sync
    out_level   => fsync_det
  );
  
  -- any frame brc over fsync interval
  u_brc_det : ENTITY common_lib.common_switch
  PORT MAP (
    clk         => clk,
    rst         => rst,
    switch_high => brc,       -- frame brc
    switch_low  => fsync,     -- frame sync
    out_level   => brc_det
  );
  
  -- any frame discarded over fsync interval
  u_dis_det : ENTITY common_lib.common_switch
  PORT MAP (
    clk         => clk,
    rst         => rst,
    switch_high => in_dis,    -- frame discarded
    switch_low  => fsync,     -- frame sync
    out_level   => dis_det
  );

  -- frame count over fsync interval
  p_frame_cnt : PROCESS(cnt, fsync, in_eof)
  BEGIN
    nxt_cnt <= cnt;
    IF fsync='1' THEN
      nxt_cnt <= (OTHERS => '0');
    ELSIF in_eof='1' THEN
      nxt_cnt <= STD_LOGIC_VECTOR(UNSIGNED(cnt) + 1);
    END IF;
  END PROCESS;

END ARCHITECTURE;
