--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_packetizing_pkg.ALL;

-- Reuse from LOFAR rad_frame_tx.vhd and rad_frame_tx(rtl).vhd

-- Purpose:
--   Pass on the input data path frame:
--
--                -----------------------
--                | fsn | dat ... | brc |
--                -----------------------
--                  sof             eof
--                  val   val ...   val
--
--   into an output data path PHY frame:
--  
--          ----------------------------------
--     idle |idle| sfd | fsn | dat ... | crc | idle ...
--          ----------------------------------
--            sof                        eof
--            val  val   val   val ...   val
--
-- Description:
--   An input internal frame consists of fsn, data and brc. The in_sof marks
--   fsn and the in_eof marks the brc. When in_val is '0' then c_dp_idle data
--   is output.
--   The dp PHY output frame suits point it point links only. A valid CRC is
--   inserted when the brc is OK else insert NOT crc to propagate the brc error
--   at dp PHY level. For dat'LENGTH < c_dp_crc_w the transmitted crc word gets
--   extended. For dat'LENGTH > c_dp_crc_w the transmitted crc word gets
--   truncated.
--   The dp PHY frame is prepended with at least 1 idle word and a start of
--   frame delimiter (sfd) word. Different sfd words can be used to have
--   multiple independent dp frame streams via one dp PHY link. The dat block
--   size is fixed per dp frame stream.
--   More idle words are sent between frames if necessary. The out_val, out_sof
--   and out_eof signals are available to ease subsequent tx frame scheduling,
--   but they are not used on the actual dp PHY link. The idle word just before
--   the sfd is also marked valid and indicated by the out_sof to ensure that
--   there is at least one idle word between frames.

ENTITY dp_frame_tx IS
  GENERIC (
    g_sfd        : STD_LOGIC_VECTOR := c_dp_sfd;
    g_lofar      : BOOLEAN := FALSE;  -- when TRUE use same CRC as for LOFAR
    g_dat_w      : NATURAL := c_dp_max_w
  );
  PORT (
    rst      : IN  STD_LOGIC;
    clk      : IN  STD_LOGIC;
    in_dat   : IN  STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
    in_val   : IN  STD_LOGIC;
    in_sof   : IN  STD_LOGIC;
    in_eof   : IN  STD_LOGIC;
    out_dat  : OUT STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
    out_val  : OUT STD_LOGIC;
    out_sof  : OUT STD_LOGIC;
    out_eof  : OUT STD_LOGIC
  );
END dp_frame_tx;


ARCHITECTURE rtl OF dp_frame_tx IS

  -- Use unregistered input to save logic, use registered input to ease achieving timing constrains.
  CONSTANT c_input_reg   : BOOLEAN := FALSE;
  
  CONSTANT c_idle        : STD_LOGIC_VECTOR := c_dp_idle(in_dat'RANGE);
  CONSTANT c_sfd         : STD_LOGIC_VECTOR := g_sfd(in_dat'RANGE);
  CONSTANT c_brc_ok      : STD_LOGIC_VECTOR := STD_LOGIC_VECTOR(TO_UNSIGNED(c_dp_brc_ok ,g_dat_w));
  
  SIGNAL crc             : STD_LOGIC_VECTOR(c_dp_crc_w-1 DOWNTO 0);
  SIGNAL nxt_crc         : STD_LOGIC_VECTOR(crc'RANGE);

  -- Registered inputs (see c_input_reg)
  SIGNAL in_dat_reg      : STD_LOGIC_VECTOR(in_dat'RANGE);
  SIGNAL in_val_reg      : STD_LOGIC;
  SIGNAL in_sof_reg      : STD_LOGIC;
  SIGNAL in_eof_reg      : STD_LOGIC;
  
  -- Delayed inputs to create space to insert idle and sfd word
  SIGNAL in_dat_dly      : STD_LOGIC_VECTOR(in_dat'RANGE);
  SIGNAL in_val_dly      : STD_LOGIC;
  SIGNAL in_sof_dly      : STD_LOGIC;
  SIGNAL in_eof_dly      : STD_LOGIC;
  SIGNAL in_dat_dly2     : STD_LOGIC_VECTOR(in_dat'RANGE);
  SIGNAL in_val_dly2     : STD_LOGIC;
  SIGNAL in_eof_dly2     : STD_LOGIC;
  
  SIGNAL frame_busy      : STD_LOGIC;
  SIGNAL nxt_frame_busy  : STD_LOGIC;
  
  SIGNAL nxt_out_dat     : STD_LOGIC_VECTOR(out_dat'RANGE);
  SIGNAL nxt_out_val     : STD_LOGIC;
  SIGNAL nxt_out_sof     : STD_LOGIC;
  SIGNAL nxt_out_eof     : STD_LOGIC;
  
BEGIN

  gen_input_reg : IF c_input_reg=TRUE GENERATE
    p_reg : PROCESS (clk, rst)
    BEGIN
      IF rst='1' THEN
        in_dat_reg <= (OTHERS => '0');
        in_val_reg <= '0';
        in_sof_reg <= '0';
        in_eof_reg <= '0';
      ELSIF rising_edge(clk) THEN
        in_dat_reg <= in_dat;
        in_val_reg <= in_val;
        in_sof_reg <= in_sof;
        in_eof_reg <= in_eof;
      END IF;
    END PROCESS;
  END GENERATE;
  no_input_reg : IF c_input_reg=FALSE GENERATE
    in_dat_reg <= in_dat;
    in_val_reg <= in_val;
    in_sof_reg <= in_sof;
    in_eof_reg <= in_eof;
  END GENERATE;
  
  p_clk : PROCESS (clk, rst)
  BEGIN
    IF rst='1' THEN
      in_dat_dly  <= (OTHERS => '0');
      in_val_dly  <= '0';
      in_sof_dly  <= '0';
      in_eof_dly  <= '0';
      in_dat_dly2 <= (OTHERS => '0');
      in_val_dly2 <= '0';
      in_eof_dly2 <= '0';
      crc         <= (OTHERS => '1');
      out_dat     <= (OTHERS => '0');
      out_val     <= '0';
      out_sof     <= '0';
      out_eof     <= '0';
      frame_busy  <= '0';
    ELSIF rising_edge(clk) THEN
      in_dat_dly  <= in_dat_reg;
      in_val_dly  <= in_val_reg;
      in_sof_dly  <= in_sof_reg;
      in_eof_dly  <= in_eof_reg;
      in_dat_dly2 <= in_dat_dly;
      in_val_dly2 <= in_val_dly;
      in_eof_dly2 <= in_eof_dly;
      crc         <= nxt_crc;
      out_dat     <= nxt_out_dat;
      out_val     <= nxt_out_val;
      out_sof     <= nxt_out_sof;
      out_eof     <= nxt_out_eof;
      frame_busy  <= nxt_frame_busy;
    END IF;
  END PROCESS;
  
  p_output: PROCESS(crc, in_dat_dly2, in_val_dly2, in_sof_reg, in_sof_dly, in_eof_dly2, frame_busy)
  BEGIN
    -- set default nxt_crc
    nxt_crc <= crc;
    IF in_val_dly2='1' THEN
      nxt_crc <= func_dp_next_crc(g_lofar, in_dat_dly2, crc);
    END IF;
    
    -- set default nxt_out_dat
    nxt_out_dat <= in_dat_dly2;        -- pass on input dat
    nxt_out_val <= in_val_dly2;
    nxt_out_sof <= '0';
    nxt_out_eof <= '0';
    
    nxt_frame_busy <= frame_busy;      -- hold to know when a frame is being transmitted or not
    
    -- overrule default nxt_crc, nxt_out_dat
    IF in_sof_reg='1' THEN
      nxt_crc     <= (OTHERS => '1');  -- init crc
      nxt_out_dat <= c_idle;           -- prepend idle word
      nxt_out_val <= '1';
      nxt_out_sof <= '1';
      nxt_frame_busy <= '1';
    ELSIF in_sof_dly='1' THEN
      nxt_crc     <= (OTHERS => '1');  -- init crc
      nxt_out_dat <= c_sfd;            -- prepend sfd word
      nxt_out_val <= '1';
    ELSIF in_eof_dly2='1' THEN
      nxt_frame_busy <= '0';
      nxt_out_dat <= RESIZE_UVEC(crc, g_dat_w);  -- fill in calculated tx crc and account for brc
      IF in_dat_dly2/=c_brc_ok THEN
        nxt_out_dat <= NOT RESIZE_UVEC(crc, g_dat_w);
      END IF;
      nxt_out_eof <= '1';
    ELSIF in_val_dly2='0' THEN
      IF frame_busy='0' THEN
        nxt_out_dat <= c_idle;         -- insert idle word when out_val is inactive between frames
      --ELSE
      -- do not transmit idle when out_val is inactive during a frame to reduce potential false sync at receiver
      END IF;
    END IF;
  END PROCESS;
    
END rtl;
