--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Adds a header of g_hdr_nof_words long and g_data_w wide to a data stream with
-- the same width g_data_w.
-- g_data_w must be a multiple of both c_word_w (write data width used by the RAM) 
-- and g_symbol_w.

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY dp_hdr_insert IS
  GENERIC (
    g_data_w          : NATURAL;
    g_symbol_w        : NATURAL; 
    g_hdr_nof_words   : NATURAL;
    g_internal_bypass : BOOLEAN := FALSE; -- can be used to eliminate external GENERATE-statements/wiring
    g_init_hdr        : STRING := "UNUSED";
    g_dp_on_at_init   : STD_LOGIC := '0' -- Enable data path at default
  );
  PORT (
    mm_rst    : IN  STD_LOGIC;
    mm_clk    : IN  STD_LOGIC;

    st_rst    : IN  STD_LOGIC;
    st_clk    : IN  STD_LOGIC;

    ram_mosi : IN  t_mem_mosi;  --Hdr write port
    reg_mosi : IN  t_mem_mosi;  --Hdr release

    snk_out   : OUT t_dp_siso;
    snk_in    : IN  t_dp_sosi;

    src_in    : IN  t_dp_siso;
    src_out   : OUT t_dp_sosi
  );
END dp_hdr_insert;


ARCHITECTURE str OF dp_hdr_insert IS

  SIGNAL hdr_siso        : t_dp_siso;
  SIGNAL hdr_sosi        : t_dp_sosi;

  -- element (0) becomes the head, element (1) becomes the tail
  SIGNAL concat_siso_arr : t_dp_siso_arr(0 TO 1);
  SIGNAL concat_sosi_arr : t_dp_sosi_arr(0 TO 1);
  
BEGIN

  no_bypass: IF g_internal_bypass = FALSE GENERATE
    u_dp_ram_from_mm : ENTITY work.mms_dp_ram_from_mm
    GENERIC MAP (
      g_ram_wr_nof_words => g_hdr_nof_words * (g_data_w/c_word_w),
      g_ram_rd_dat_w     => g_data_w,
      g_init_file        => g_init_hdr,
      g_dp_on_at_init    => g_dp_on_at_init
    )
    PORT MAP (
      mm_rst      => mm_rst,
      mm_clk      => mm_clk,
  
      st_rst      => st_rst,
      st_clk      => st_clk,
  
      reg_mosi    => reg_mosi,  
      ram_mosi    => ram_mosi,
  
      src_in      => hdr_siso,
      src_out     => hdr_sosi
    );
  
    hdr_siso           <= concat_siso_arr(0);
    snk_out            <= concat_siso_arr(1);  
    concat_sosi_arr(0) <= hdr_sosi;
    concat_sosi_arr(1) <= snk_in;

    u_dp_concat : ENTITY work.dp_concat  -- RL = 1
    GENERIC MAP (
      g_data_w    => g_data_w,
      g_symbol_w  => g_symbol_w 
    )
    PORT MAP (
      rst         => st_rst,
      clk         => st_clk,
      -- ST sinks
      snk_out_arr => concat_siso_arr,
      snk_in_arr  => concat_sosi_arr,
      -- ST source
      src_in      => src_in,
      src_out     => src_out
    );
  END GENERATE; 

  gen_bypass : IF g_internal_bypass = TRUE GENERATE
    src_out <= snk_in;
    snk_out <= src_in;
  END GENERATE;  
  
END str;
