-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE common_lib.common_field_pkg.ALL;

ENTITY dp_offload_rx IS
  GENERIC (
    g_nof_streams   : NATURAL;
    g_data_w        : NATURAL;
    g_hdr_field_arr : t_common_field_arr;
    g_remove_crc    : BOOLEAN;
    g_crc_nof_words : NATURAL
  ); 
  PORT (
    mm_rst             : IN  STD_LOGIC;
    mm_clk             : IN  STD_LOGIC; 
    
    dp_rst             : IN  STD_LOGIC;
    dp_clk             : IN  STD_LOGIC; 

    reg_hdr_dat_mosi   : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_hdr_dat_miso   : OUT t_mem_miso;

    snk_in_arr         : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0); 
    snk_out_arr        : OUT t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);

    src_out_arr        : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0); 
    src_in_arr         : IN  t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);

    hdr_fields_out_arr : OUT t_slv_1024_arr(g_nof_streams-1 DOWNTO 0); -- Valid at src_out_arr(i).sop, use for sosi.sync
    hdr_fields_raw_arr : OUT t_slv_1024_arr(g_nof_streams-1 DOWNTO 0)  -- Valid at src_out_arr(i).sop and beyond, use for sosi.bsn
  );
END dp_offload_rx;


ARCHITECTURE str OF dp_offload_rx IS 

  CONSTANT c_nof_header_words         : NATURAL := field_slv_len(g_hdr_field_arr) / g_data_w;

  CONSTANT c_field_sel                 : STD_LOGIC_VECTOR(g_hdr_field_arr'RANGE) := (OTHERS=>'0');-- Not used in sink mode but requires set range

  SIGNAL dp_split_src_out_2arr        : t_dp_sosi_2arr_2(g_nof_streams-1 DOWNTO 0);
  SIGNAL dp_split_src_in_2arr         : t_dp_siso_2arr_2(g_nof_streams-1 DOWNTO 0);

  SIGNAL reg_hdr_dat_mosi_arr         : t_mem_mosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL reg_hdr_dat_miso_arr         : t_mem_miso_arr(g_nof_streams-1 DOWNTO 0); 

  SIGNAL dp_field_blk_slv_out         : t_slv_1024_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL dp_field_blk_slv_out_val     : STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0);

  SIGNAL dp_field_blk_src_out_arr     : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0); 

  SIGNAL reg_dp_field_blk_slv_out     : t_slv_1024_arr(g_nof_streams-1 DOWNTO 0); 
  SIGNAL nxt_reg_dp_field_blk_slv_out : t_slv_1024_arr(g_nof_streams-1 DOWNTO 0); 

  SIGNAL dp_tail_remove_src_out_arr   : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0); 
  SIGNAL dp_tail_remove_src_in_arr    : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);

  CONSTANT c_dp_field_blk_snk_data_w : NATURAL := g_data_w;
  CONSTANT c_dp_field_blk_src_data_w : NATURAL := field_slv_in_len(field_arr_set_mode(g_hdr_field_arr , "RO"));

BEGIN

  ---------------------------------------------------------------------------------------
  -- Remove header 
  ---------------------------------------------------------------------------------------    
  gen_nof_streams: FOR i IN 0 TO g_nof_streams-1 GENERATE       
    u_dp_split : ENTITY work.dp_split 
    GENERIC MAP (
      g_data_w        => g_data_w,
      g_symbol_w      => g_data_w,
      g_nof_symbols   => c_nof_header_words
    )
    PORT MAP (
      rst         => dp_rst,
      clk         => dp_clk,

      snk_out     => snk_out_arr(i),
      snk_in      => snk_in_arr(i),

      src_in_arr  => dp_split_src_in_2arr(i),
      src_out_arr => dp_split_src_out_2arr(i)
    );

    dp_split_src_in_2arr(i)(1) <= c_dp_siso_rdy;

  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- Feed header block to dp_field_blk that enables MM reading the fields and
  -- field forwarding to data path
  ---------------------------------------------------------------------------------------
  gen_dp_field_blk : FOR i IN 0 TO g_nof_streams-1 GENERATE
    u_dp_field_blk : ENTITY work.dp_field_blk
    GENERIC MAP (
      g_field_arr   => field_arr_set_mode(g_hdr_field_arr , "RO"),
      g_field_sel   => c_field_sel,
      g_snk_data_w  => c_dp_field_blk_snk_data_w, --g_data_w,
      g_src_data_w  => c_dp_field_blk_src_data_w  --field_slv_in_len(field_arr_set_mode(g_hdr_field_arr , "RO"))
    )
    PORT MAP (
      dp_rst       => dp_rst,
      dp_clk       => dp_clk,

      mm_rst       => mm_rst,
      mm_clk       => mm_clk,
  
      snk_in       => dp_split_src_out_2arr(i)(1),

      src_out      => dp_field_blk_src_out_arr(i),

--      slv_out      => dp_field_blk_slv_out(i)(field_slv_len(g_hdr_field_arr)-1 DOWNTO 0),
--      slv_out_val  => dp_field_blk_slv_out_val(i),

      reg_slv_mosi => reg_hdr_dat_mosi_arr(i),
      reg_slv_miso => reg_hdr_dat_miso_arr(i)
    );

    dp_field_blk_slv_out(i)(c_dp_field_blk_src_data_w-1 DOWNTO 0) <= dp_field_blk_src_out_arr(i).data(c_dp_field_blk_src_data_w-1 DOWNTO 0);
    dp_field_blk_slv_out_val(i) <= dp_field_blk_src_out_arr(i).valid;

  END GENERATE;

  ---------------------------------------------------------------------------------------
  -- Remove CRC if phy link is used
  ---------------------------------------------------------------------------------------
  gen_dp_tail_remove : FOR i IN 0 TO g_nof_streams-1 GENERATE
    u_dp_tail_remove : ENTITY work.dp_tail_remove
    GENERIC MAP (
      g_data_w      => g_data_w,
      g_symbol_w    => g_data_w,
      g_nof_symbols => sel_a_b(g_remove_crc, g_crc_nof_words, 0) 
    )
    PORT MAP (
      st_rst  => dp_rst,
      st_clk  => dp_clk,
   
      snk_out => dp_split_src_in_2arr(i)(0), 
      snk_in  => dp_split_src_out_2arr(i)(0),
  
      src_in  => dp_tail_remove_src_in_arr(i),   
      src_out => dp_tail_remove_src_out_arr(i)
    );
  END GENERATE;

  src_out_arr <= dp_tail_remove_src_out_arr;
  dp_tail_remove_src_in_arr <= src_in_arr;

  ---------------------------------------------------------------------------------------
  -- Sync header fields to dp_tail_remove_src_out_arr(i).sop 
  ---------------------------------------------------------------------------------------
 
  -- Store the header fields  
  p_clk : PROCESS(dp_clk, dp_rst)
  BEGIN
    IF dp_rst='1' THEN
      reg_dp_field_blk_slv_out <= (OTHERS=>(OTHERS=>'0'));
    ELSIF rising_edge(dp_clk) THEN
      reg_dp_field_blk_slv_out <= nxt_reg_dp_field_blk_slv_out;
    END IF;
  END PROCESS;

  gen_reg_dp_fields_blk_slv : FOR i IN 0 TO g_nof_streams-1 GENERATE
    nxt_reg_dp_field_blk_slv_out(i) <= dp_field_blk_slv_out(i) WHEN dp_field_blk_slv_out_val(i) = '1' ELSE reg_dp_field_blk_slv_out(i);
  END GENERATE;

  -- Put the fields that can only be active at sop on the entity output at src_out.sop (e.g. sosi sync field)
  gen_hdr_fields_out_arr : FOR i IN 0 TO g_nof_streams-1 GENERATE
    hdr_fields_out_arr(i) <= reg_dp_field_blk_slv_out(i) WHEN dp_tail_remove_src_out_arr(i).sop='1' ELSE (OTHERS=>'0');
  END GENERATE;
  -- Put the fields that can be valid during entire packet on the entity combinatorially (e.g. all sosi slv fields)
  hdr_fields_raw_arr <= reg_dp_field_blk_slv_out;
  
  ---------------------------------------------------------------------------------------
  -- MM control & monitoring
  ---------------------------------------------------------------------------------------
  u_common_mem_mux_hdr_dat : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_streams,
    g_mult_addr_w => ceil_log2(field_nof_words(g_hdr_field_arr, c_word_w))
  )
  PORT MAP (
    mosi     => reg_hdr_dat_mosi,
    miso     => reg_hdr_dat_miso,
    mosi_arr => reg_hdr_dat_mosi_arr,
    miso_arr => reg_hdr_dat_miso_arr
  );

END str;
