-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, mm_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE common_lib.common_field_pkg.ALL;

ENTITY dp_offload_rx_filter_mm IS
  GENERIC (   
    g_bypass              : BOOLEAN  := FALSE;
    g_nof_streams         : POSITIVE := 1;
    g_data_w              : NATURAL; 
    g_hdr_field_arr       : t_common_field_arr    
  ); 
  PORT (
    
    dp_rst            : IN  STD_LOGIC;
    dp_clk            : IN  STD_LOGIC;     
    
    mm_rst            : IN  STD_LOGIC;
    mm_clk            : IN  STD_LOGIC;
    
    reg_dp_offload_rx_filter_hdr_fields_mosi  : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_dp_offload_rx_filter_hdr_fields_miso  : OUT t_mem_miso;
                                                             
    snk_in_arr        : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0); 
    snk_out_arr       : OUT t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
                      
    src_out_arr       : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0); 
    src_in_arr        : IN  t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
    
    hdr_fields_to_check_arr: IN  t_slv_1024_arr(g_nof_streams-1 DOWNTO 0);
    
    hdr_fields_val    : IN STD_LOGIC
  );
END dp_offload_rx_filter_mm;


ARCHITECTURE str OF dp_offload_rx_filter_mm  IS 
                                         
  CONSTANT c_header_w : NATURAL := field_slv_in_len(field_arr_set_mode(g_hdr_field_arr , "RO"));
  CONSTANT c_nof_words: NATURAL := c_header_w/g_data_w;
  
  CONSTANT c_head : NATURAL := 1;
  CONSTANT c_tail : NATURAL := 0;
  
  CONSTANT c_field_slv_out_w : NATURAL := field_slv_out_len(field_arr_set_mode(g_hdr_field_arr , "RW"));
  
  CONSTANT c_nof_ena				 : NATURAL := 4;
  CONSTANT c_reg_w					 : NATURAL := c_nof_ena*c_word_w;
  CONSTANT c_adr_w					 : NATURAL := ceil_log2(c_nof_ena);
  CONSTANT c_ena_reg         : t_c_mem := (1, c_adr_w, 32, c_nof_ena, '0');
  
  SUBTYPE eth_dst_mac_range     IS NATURAL RANGE field_hi(g_hdr_field_arr, "eth_dst_mac"    ) DOWNTO field_lo(g_hdr_field_arr, "eth_dst_mac");
  SUBTYPE ip_total_length_range IS NATURAL RANGE field_hi(g_hdr_field_arr, "ip_total_length") DOWNTO field_lo(g_hdr_field_arr, "ip_total_length");    
  SUBTYPE ip_dst_addr_range     IS NATURAL RANGE field_hi(g_hdr_field_arr, "ip_dst_addr"    ) DOWNTO field_lo(g_hdr_field_arr, "ip_dst_addr");
  SUBTYPE udp_dst_port_range    IS NATURAL RANGE field_hi(g_hdr_field_arr, "udp_dst_port"   ) DOWNTO field_lo(g_hdr_field_arr, "udp_dst_port");
  
  TYPE    t_field_slv_in_arr    IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(field_slv_in_len(g_hdr_field_arr)-1 DOWNTO 0);  
  TYPE    t_field_slv_out_arr   IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(c_field_slv_out_w-1 DOWNTO 0);  
   
  SIGNAL  mm_fields_slv_in_arr    : t_field_slv_in_arr(g_nof_streams-1 DOWNTO 0); 
  SIGNAL  mm_fields_slv_out_arr   : t_field_slv_out_arr(g_nof_streams-1 DOWNTO 0);
  
  SIGNAL  mult_streams_mosi_arr   : t_mem_mosi_arr(g_nof_streams-1 DOWNTO 0);   
  SIGNAL  mult_streams_miso_arr   : t_mem_miso_arr(g_nof_streams-1 DOWNTO 0);
  
  SIGNAL  common_mosi_arr 				: t_mem_mosi_arr(1 DOWNTO 0);   
  SIGNAL  common_miso_arr 				: t_mem_miso_arr(1 DOWNTO 0);
  
  TYPE t_bool_arr    IS ARRAY (INTEGER RANGE <>) OF BOOLEAN;
  TYPE t_reg_sig_arr IS ARRAY (INTEGER RANGE <>) OF STD_LOGIC_VECTOR(c_reg_w-1 DOWNTO 0);
  
  SIGNAL  reg_ena_sig    	       	: t_reg_sig_arr(g_nof_streams-1 DOWNTO 0);
  
  SIGNAL  eth_dst_mac_ena         : t_bool_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL  ip_dst_addr_ena         : t_bool_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL  ip_total_length_ena     : t_bool_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL  udp_dst_port_ena        : t_bool_arr(g_nof_streams-1 DOWNTO 0); 
  
  TYPE reg_type IS RECORD
    valid       : STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0);
    src_out_arr : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  END RECORD;

  SIGNAL r, rin : reg_type;
  
BEGIN
  
  snk_out_arr <= src_in_arr;
  
  gen_bypass : IF g_bypass=TRUE GENERATE
    src_out_arr <= snk_in_arr;
  END GENERATE; 
  
  no_bypass : IF g_bypass=FALSE GENERATE 
       
    p_comb : PROCESS(hdr_fields_to_check_arr, snk_in_arr, r, dp_rst)
      VARIABLE v : reg_type;
    BEGIN   
      v := r; 
          
      v.src_out_arr := snk_in_arr; 
      check : FOR i in 0 TO g_nof_streams-1 LOOP
        IF snk_in_arr(i).sop = '1' THEN   
          v.valid(i) := '1';
          
          IF eth_dst_mac_ena(i)     = TRUE AND NOT(hdr_fields_to_check_arr(i)(eth_dst_mac_range)     = mm_fields_slv_out_arr(i)(eth_dst_mac_range)     ) THEN
            v.valid(i) := '0';
          END IF;
          
          IF ip_total_length_ena(i) = TRUE AND NOT(hdr_fields_to_check_arr(i)(ip_total_length_range) = mm_fields_slv_out_arr(i)(ip_total_length_range) ) THEN
            v.valid(i) := '0';
          END IF;            
                             
          IF ip_dst_addr_ena(i)     = TRUE AND NOT(hdr_fields_to_check_arr(i)(ip_dst_addr_range)     = mm_fields_slv_out_arr(i)(ip_dst_addr_range)     ) THEN
            v.valid(i) := '0';   
          END IF;
           
          IF udp_dst_port_ena(i)    = TRUE AND NOT(hdr_fields_to_check_arr(i)(udp_dst_port_range)    = mm_fields_slv_out_arr(i)(udp_dst_port_range)    ) THEN
            v.valid(i) := '0';   
          END IF;      
         
        END IF; 
      END LOOP check;        
    
      IF(dp_rst = '1') THEN 
        v.src_out_arr := (OTHERS => c_dp_sosi_rst);
        v.valid       := (OTHERS => '0');
      END IF;
      
      rin <= v;
      
    END PROCESS;
    
    
    PROCESS(r)
    BEGIN 
      src_out_arr   <= r.src_out_arr;
      set_val : FOR i in 0 TO g_nof_streams-1 LOOP  
        src_out_arr(i).valid <= r.valid(i) AND r.src_out_arr(i).valid; 
        src_out_arr(i).sop   <= r.valid(i) AND r.src_out_arr(i).sop;
        src_out_arr(i).eop   <= r.valid(i) AND r.src_out_arr(i).eop;
      END LOOP set_val;
    END PROCESS; 
    
    p_regs : PROCESS(dp_clk)
    BEGIN 
      IF RISING_EDGE(dp_clk) THEN 
        r <= rin; 
      END IF; 
    END PROCESS;
      
    
		------------------------------------------- 
		-- mm_fields for MM access to each field -- 
		-------------------------------------------  
    u_mult_mem_mux : ENTITY common_lib.common_mem_mux
    GENERIC MAP (    
      g_nof_mosi    => g_nof_streams,
      g_mult_addr_w => ceil_log2(field_nof_words(g_hdr_field_arr, c_word_w))+1
    )
    PORT MAP (
      mosi     => reg_dp_offload_rx_filter_hdr_fields_mosi,
      miso     => reg_dp_offload_rx_filter_hdr_fields_miso,
      mosi_arr => mult_streams_mosi_arr,
      miso_arr => mult_streams_miso_arr 
    ); 
    
    gen_mm_fields : FOR i IN 0 TO g_nof_streams-1 GENERATE        

      u_common_mem_mux : ENTITY common_lib.common_mem_mux
      GENERIC MAP (    
        g_nof_mosi    => 2,
        g_mult_addr_w => ceil_log2(field_nof_words(g_hdr_field_arr, c_word_w))
      )
      PORT MAP (
        mosi     => mult_streams_mosi_arr(i),
        miso     => mult_streams_miso_arr(i),
        mosi_arr => common_mosi_arr,
        miso_arr => common_miso_arr 
      ); 
      
      
      eth_dst_mac_ena(i)     <= is_true(reg_ena_sig(i)(0));
    	ip_dst_addr_ena(i)     <= is_true(reg_ena_sig(i)(32));
    	ip_total_length_ena(i) <= is_true(reg_ena_sig(i)(64));
    	udp_dst_port_ena(i)    <= is_true(reg_ena_sig(i)(96));
      
      cr : ENTITY common_lib.common_reg_r_w_dc
  	  GENERIC MAP(
  	    g_reg       => c_ena_reg
  	  )
  	    
  	  PORT MAP(
  	    -- Clocks and reset
  	    mm_rst      => mm_rst,   -- reset synchronous with mm_clk
  	    mm_clk      => mm_clk,   -- memory-mapped bus clock
  	    st_rst      => dp_rst,   -- reset synchronous with st_clk
  	    st_clk      => dp_clk,   -- other clock domain clock
  	    
  	    -- Memory Mapped Slave in mm_clk domain
  	    sla_in      => common_mosi_arr(1),  -- IN  t_mem_mosi;  -- actual ranges defined by g_reg
  	    sla_out     => common_miso_arr(1),  -- OUT t_mem_miso;  -- actual ranges defined by g_reg
  	    
  	    -- MM registers in st_clk domain
  	    reg_wr_arr  => OPEN,
  	    reg_rd_arr  => OPEN,
  	    in_new      => OPEN,
  	    in_reg      => reg_ena_sig(i),
  	    out_reg     => reg_ena_sig(i),
  	    out_new     => OPEN
  	  );
    
      u_mm_fields_slv: ENTITY mm_lib.mm_fields
      GENERIC MAP(
        g_field_arr => field_arr_set_mode(g_hdr_field_arr , "RW")
      )
      PORT MAP (
        mm_clk     => mm_clk,
        mm_rst     => mm_rst,
      
        mm_mosi    => common_mosi_arr(0),
        mm_miso    => common_miso_arr(0),
        
        slv_clk    => dp_clk,
        slv_rst    => dp_rst, 
      
        slv_in     => mm_fields_slv_in_arr(i),
        slv_in_val => hdr_fields_val,
        slv_out    => mm_fields_slv_out_arr(i)
      );     
    END GENERATE;     
    
  END GENERATE;        
END str;
