-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, common_mult_lib, technology_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_field_pkg.ALL;

-- Purpose:
-- . Calculate UDP total length and IP total length fields
-- Description:
-- . The length fields are calculated based on the status outputs
--   of dp_split and dp_packet_merge:
--   . udp_total_length = c_udp_header_len + nof_data_bytes + c_user_hdr_len
--     . c_udp_header_len = 8 (fixed) bytes
--     . nof_data_bytes   = nof_words_per_block * nof_blocks_per_packet * 4 bytes
--     . c_user_hdr_len   = nof bytes in remaining header
--   . ip_total_length = udp_total_length - c_ip_header_len

ENTITY dp_offload_tx_len_calc IS
  GENERIC (
    g_technology      : NATURAL := c_tech_select_default;
    g_hdr_field_arr   : t_common_field_arr;
    g_data_w          : NATURAL; -- Must be a power of 2
    g_word_align_len  : NATURAL := 2 
  ); 
  PORT (
    rst     : IN  STD_LOGIC;
    clk     : IN  STD_LOGIC;  

    nof_words_per_block   : IN STD_LOGIC_VECTOR;
    nof_blocks_per_packet : IN STD_LOGIC_VECTOR;

    udp_total_length : OUT STD_LOGIC_VECTOR(15 DOWNTO 0);
    ip_total_length  : OUT STD_LOGIC_VECTOR(15 DOWNTO 0)
  );
END dp_offload_tx_len_calc;


ARCHITECTURE str OF dp_offload_tx_len_calc IS 

  CONSTANT c_eth_header_len    : NATURAL := 14;
  CONSTANT c_ip_header_len     : NATURAL := 20;
  CONSTANT c_udp_header_len    : NATURAL := 8;

  CONSTANT c_total_hdr_len     : NATURAL := field_slv_len(g_hdr_field_arr) / c_byte_w;
  CONSTANT c_user_hdr_len      : NATURAL := c_total_hdr_len - g_word_align_len - c_eth_header_len - c_ip_header_len - c_udp_header_len;

  CONSTANT c_product_w         : NATURAL := nof_words_per_block'LENGTH + nof_blocks_per_packet'LENGTH;
  CONSTANT c_nof_bits_to_shift : NATURAL := true_log2(g_data_w/c_byte_w); --number of bits to shift left to yield nof_data_bytes)
  CONSTANT c_shift_left_int    : INTEGER := -1 * c_nof_bits_to_shift;
  CONSTANT c_adder_in_w        : NATURAL := 16;

  SIGNAL nof_data_words        : STD_LOGIC_VECTOR(c_product_w-1 DOWNTO 0);
  SIGNAL nof_data_bytes        : STD_LOGIC_VECTOR(c_product_w+c_nof_bits_to_shift-1 DOWNTO 0);
 
  SIGNAL udp_adder_in          : STD_LOGIC_VECTOR(3*c_adder_in_w-1 DOWNTO 0);
  SIGNAL udp_adder_out         : STD_LOGIC_VECTOR(c_adder_in_w+2-1 DOWNTO 0);

  SIGNAL ip_adder_in           : STD_LOGIC_VECTOR(2*c_adder_in_w-1 DOWNTO 0);
  SIGNAL ip_adder_out          : STD_LOGIC_VECTOR(c_adder_in_w+1-1 DOWNTO 0);
 
BEGIN

  ---------------------------------------------------------------------------------------
  -- Calculate number of payload words
  --------------------------------------------------------------------------------------- 
  u_common_mult: ENTITY common_mult_lib.common_mult
  GENERIC MAP (
    g_technology       => g_technology,
    g_variant          => "RTL",
    g_in_a_w           => nof_words_per_block'LENGTH,
    g_in_b_w           => nof_blocks_per_packet'LENGTH,
    g_out_p_w          => c_product_w,
    g_pipeline_input   => 0,
    g_pipeline_product => 0,
    g_pipeline_output  => 0,
    g_representation   => "UNSIGNED"
  )
  PORT MAP (
    rst     => rst,
    clk     => clk,
    clken   => '1',
    in_a    => nof_words_per_block,
    in_b    => nof_blocks_per_packet,
    out_p   => nof_data_words
  );  

  ---------------------------------------------------------------------------------------
  -- Calculate number of payload bytes by bit shifting to the left
  --------------------------------------------------------------------------------------- 
  nof_data_bytes <= SHIFT_UVEC( RESIZE_UVEC(nof_data_words, c_product_w+c_nof_bits_to_shift), c_shift_left_int);

  ---------------------------------------------------------------------------------------
  -- Calculate UDP total length
  --------------------------------------------------------------------------------------- 
  udp_adder_in <= RESIZE_UVEC(nof_data_bytes, c_adder_in_w) & TO_UVEC(c_udp_header_len, c_adder_in_w) & TO_UVEC(c_user_hdr_len, c_adder_in_w);
 
  u_common_adder_tree_udp : ENTITY common_lib.common_adder_tree(str)
  GENERIC MAP (
    g_representation => "UNSIGNED",
    g_pipeline       => 0,
    g_nof_inputs     => 3,
    g_dat_w          => c_adder_in_w,
    g_sum_w          => c_adder_in_w+2
  )
  PORT MAP (
    clk    => clk,
    in_dat => udp_adder_in,
    sum    => udp_adder_out
  );

  udp_total_length <= udp_adder_out(c_adder_in_w-1 DOWNTO 0);

  ---------------------------------------------------------------------------------------
  -- Calculate IP total length
  --------------------------------------------------------------------------------------- 
  ip_adder_in <= RESIZE_UVEC(udp_adder_out, c_adder_in_w) & TO_UVEC(c_ip_header_len, c_adder_in_w);

  u_common_adder_tree_ip : ENTITY common_lib.common_adder_tree(str)
  GENERIC MAP (
    g_representation => "UNSIGNED",
    g_pipeline       => 0,
    g_nof_inputs     => 2,
    g_dat_w          => c_adder_in_w,
    g_sum_w          => c_adder_in_w+1
  )
  PORT MAP (
    clk    => clk,
    in_dat => ip_adder_in,
    sum    => ip_adder_out
  );

  ip_total_length <= ip_adder_out(c_adder_in_w-1 DOWNTO 0);
 
END str;
