--------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE work.dp_packet_pkg.ALL;

-- Purpose: Decode DP packet to DP sosi.
-- Description:
-- 
--   From an input DP packet:
--  
--    ----------------------------------------------
--    |chan| bsn        | data ...             |err|
--    ----------------------------------------------
--      sop                                     eop
--      val  val   val    val ...               val
--                                              err, when g_use_phy_err = TRUE
--
--   extract the block of output SOSI data:
--
--                      -----------------------
--                      | data ...            |
--                      -----------------------
--                        sop             eop
--                        val   val  ...  val
--                        sync
--                        bsn
--                        channel
--                                        empty
--                                        err
--
--   The data length (eop) and empty do not show as fields in the packet
--   because they are implitely known per channel.
--
--   The dp_packet_enc does the reverse.
--
--   The g_channel_lo parameter allows using the LSbit channel fields to be 
--   used for defining different groups of DP packet streams. This is e.g. used
--   to map these groups to corresponding 'tlen' types in case of transport
--   went by using UTH packets over a transceiver link.
--
-- Remark:
-- . This dp_packet_dec has a similar structure as uth_rx(rtl_hold) for the 
--   flow control and as dp_packet_enc for the states. However dp_packet_dec
--   uses the sop and eop so it does not (need to) know the DP packet nof 
--   data. Therefor it uses dp_shiftreg to adjust the output eop to the last
--   payload data instead of at the last err-field data.
-- . When the DP packet comes from a PHY link then it may have gotten corrupted
--   contents so then it is useful to set g_use_phy_err to TRUE. The PHY 
--   error status is based on the received CRC and fits in 1 bit. The
--   g_phy_err_bi defines where the PHY error bit is put in src_out.err in
--   case g_use_phy_err = TRUE.
-- . The output SOSI sync is defined such that it can only be active at sop, so
--   not during entire frame or even until the next sop.


ENTITY dp_packet_dec IS
  GENERIC (
    g_data_w        : NATURAL := 32;
    g_channel_lo    : NATURAL := 0;
    g_phy_err_bi    : NATURAL := 0;      -- bit index in src_out.err for where to put the PHY snk_in.err bit when g_use_phy_err=TRUE, else not used
    g_use_phy_err   : BOOLEAN := FALSE;  -- when FALSE pass on the DP packet err field, else merge the PHY snk_in.err field with the DP packet err field to get the output src_out.err.
    g_use_this_siso : BOOLEAN := TRUE    -- default use TRUE for best throughput performance
  );
  PORT (
    rst       : IN  STD_LOGIC;
    clk       : IN  STD_LOGIC;
    -- ST sinks
    snk_out   : OUT t_dp_siso;
    snk_in    : IN  t_dp_sosi;
    -- ST source
    src_in    : IN  t_dp_siso;
    src_out   : OUT t_dp_sosi
  );
END dp_packet_dec;


ARCHITECTURE rtl OF dp_packet_dec IS

  CONSTANT c_output_reg     : BOOLEAN := TRUE;  -- register dp_shiftreg output to ease timing closure
  
  CONSTANT c_channel_len    : NATURAL := ceil_div(c_dp_packet_channel_w, g_data_w);
  CONSTANT c_bsn_len        : NATURAL := ceil_div(c_dp_packet_bsn_w,     g_data_w);
  CONSTANT c_error_len      : NATURAL := ceil_div(c_dp_packet_error_w,   g_data_w);
  CONSTANT c_len_max        : NATURAL := c_bsn_len;  -- the BSN is the widest field
    
  CONSTANT c_channel_vec_w  : NATURAL := c_channel_len * g_data_w;
  CONSTANT c_bsn_vec_w      : NATURAL := c_bsn_len     * g_data_w;
  CONSTANT c_error_vec_w    : NATURAL := c_error_len   * g_data_w;
  
  CONSTANT c_shiftreg_len   : NATURAL := c_error_len + 1;
  
  TYPE t_state IS (s_sop, s_channel, s_bsn, s_data_sop, s_data_eop);
  
  -- declare data field slv signals to ease debugging in wave window
  SIGNAL in_data          : STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0);
  SIGNAL buf_data         : STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0);
  SIGNAL blk_data         : STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0);
  SIGNAL out_data         : STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0);
  
  SIGNAL state            : t_state;
  SIGNAL nxt_state        : t_state;
  
  SIGNAL cnt              : NATURAL RANGE 0 TO c_len_max-1;
  SIGNAL nxt_cnt          : NATURAL;

  SIGNAL channel          : STD_LOGIC_VECTOR(c_channel_vec_w-1 DOWNTO 0);
  SIGNAL nxt_channel      : STD_LOGIC_VECTOR(c_channel_vec_w-1 DOWNTO 0);
  SIGNAL bsn              : STD_LOGIC_VECTOR(c_bsn_vec_w-1 DOWNTO 0);
  SIGNAL nxt_bsn          : STD_LOGIC_VECTOR(c_bsn_vec_w-1 DOWNTO 0);    
  SIGNAL src_err          : STD_LOGIC_VECTOR(c_error_vec_w-1 DOWNTO 0);

  SIGNAL this_siso        : t_dp_siso;
  SIGNAL hold_src_in      : t_dp_siso;
  SIGNAL src_buf          : t_dp_sosi;
  SIGNAL nxt_src_buf      : t_dp_sosi;
  SIGNAL next_src_buf     : t_dp_sosi;
  
  SIGNAL blk_sosi         : t_dp_sosi;
  SIGNAL nxt_blk_sosi     : t_dp_sosi;
  SIGNAL cur_src_out_inputs : t_dp_sosi_arr(0 TO c_shiftreg_len-1);
  SIGNAL new_src_out_inputs : t_dp_sosi_arr(0 TO c_shiftreg_len-1);
  
  SIGNAL i_src_out        : t_dp_sosi;
  
BEGIN

  src_out <= i_src_out;
  
  -- Map to slv to ease monitoring in wave window
  in_data     <= snk_in.data(g_data_w-1 DOWNTO 0);
  buf_data    <= src_buf.data(g_data_w-1 DOWNTO 0);
  blk_data    <= blk_sosi.data(g_data_w-1 DOWNTO 0);
  out_data    <= i_src_out.data(g_data_w-1 DOWNTO 0);
  
  -- Input flow control
  snk_out <= this_siso WHEN g_use_this_siso=TRUE ELSE src_in;
  
  p_clk : PROCESS (clk, rst)
  BEGIN
    IF rst='1' THEN
      channel        <= (OTHERS => '0');
      bsn            <= (OTHERS => '0');
      src_buf        <= c_dp_sosi_rst;
      blk_sosi       <= c_dp_sosi_rst;
      state          <= s_channel;
      cnt            <= c_channel_len-1;
    ELSIF rising_edge(clk) THEN
      channel        <= nxt_channel;
      bsn            <= nxt_bsn;
      src_buf        <= nxt_src_buf;
      blk_sosi       <= nxt_blk_sosi;
      state          <= nxt_state;
      cnt            <= nxt_cnt;
    END IF;
  END PROCESS;
  
  -- Hold input register
  nxt_src_buf <= next_src_buf;
  
  u_hold : ENTITY work.dp_hold_input
  PORT MAP (
    rst          => rst,
    clk          => clk,
    -- ST sink
    snk_out      => OPEN,
    snk_in       => snk_in,
    -- ST source
    src_in       => hold_src_in,
    next_src_out => next_src_buf,
    pend_src_out => OPEN,
    src_out_reg  => src_buf
  );
  
  -- State machine
  p_state : PROCESS(state, src_in, channel, bsn, cnt, blk_sosi, next_src_buf)
    VARIABLE v_in_channel   : STD_LOGIC_VECTOR(g_channel_lo-1 DOWNTO 0);
    VARIABLE v_channel_data : STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0);
    VARIABLE v_bsn_data     : STD_LOGIC_VECTOR(g_data_w-1 DOWNTO 0);
  BEGIN
    v_in_channel   := next_src_buf.channel(g_channel_lo-1 DOWNTO 0);
    v_channel_data := RESIZE_UVEC(next_src_buf.data, g_data_w);
    v_bsn_data     := RESIZE_UVEC(next_src_buf.data, g_data_w);

    nxt_state          <= state;
    this_siso          <= src_in;
    hold_src_in        <= src_in;
    
    nxt_channel        <= channel;
    nxt_bsn            <= bsn;
    nxt_cnt            <= cnt;
    
    nxt_blk_sosi       <= blk_sosi;  -- blk_sosi outputs payload data and err field, the err field gets stripped later on
    nxt_blk_sosi.sync  <= '0';
    nxt_blk_sosi.valid <= '0';
    nxt_blk_sosi.sop   <= '0';
    nxt_blk_sosi.eop   <= '0';
    
    CASE state IS
      WHEN s_sop =>
        this_siso   <= c_dp_siso_rdy;
        hold_src_in <= c_dp_siso_rdy;
        IF next_src_buf.sop='1' THEN  
          nxt_channel((cnt+1)*g_data_w-1 DOWNTO cnt*g_data_w) <= v_channel_data;
          IF cnt=0 THEN
            nxt_cnt <= c_bsn_len-1;
            nxt_state <= s_bsn;
          ELSE
            nxt_cnt <= cnt - 1;
            nxt_state <= s_channel;
          END IF;
        END IF;
      WHEN s_channel =>
        this_siso   <= c_dp_siso_rdy;
        hold_src_in <= c_dp_siso_rdy;
        IF next_src_buf.valid='1' THEN  
          nxt_channel((cnt+1)*g_data_w-1 DOWNTO cnt*g_data_w) <= v_channel_data;
          IF cnt=0 THEN
            nxt_cnt <= c_bsn_len-1;
            nxt_state <= s_bsn;
          ELSE
            nxt_cnt <= cnt - 1;
          END IF;
        END IF;
      WHEN s_bsn =>
        this_siso   <= c_dp_siso_rdy;
        hold_src_in <= c_dp_siso_rdy;
        IF next_src_buf.valid='1' THEN  
          nxt_bsn((cnt+1)*g_data_w-1 DOWNTO cnt*g_data_w) <= v_bsn_data;
          IF cnt=0 THEN
            this_siso <= src_in;                   -- maintain active hold_src_in and now use the src_in for this_siso
            nxt_cnt <= c_channel_len-1;            -- prepare cnt for next DP packet, the error field c_error_len for this DP packet gets handled by dp_shiftreg
            nxt_state <= s_data_sop;
          ELSE
            nxt_cnt <= cnt - 1;
          END IF;
        END IF;
      WHEN s_data_sop =>
        IF next_src_buf.valid='1' THEN             -- first payload data
          nxt_blk_sosi.data    <= next_src_buf.data;
          nxt_blk_sosi.valid   <= '1';
          nxt_blk_sosi.sop     <= '1';
          nxt_blk_sosi.channel <= RESIZE_DP_CHANNEL(channel & v_in_channel);
          nxt_blk_sosi.sync    <= bsn(bsn'HIGH);                            -- support SOSI sync via MSbit of BSN field (sync active only at sop),
          nxt_blk_sosi.bsn     <= RESIZE_DP_BSN(bsn(bsn'HIGH-1 DOWNTO 0));  -- the rest of the BSN field transports the SOSI BSN
          nxt_blk_sosi.err     <= TO_DP_ERROR(0);                           -- initialy assume block has no error
          nxt_state <= s_data_eop;                 -- no need to check for eop, because DP packet has at least 1 data word and an error field, so eop can not occur yet
        END IF;
      WHEN OTHERS => -- s_data_eop
        IF next_src_buf.valid='1' THEN             -- rest of payload data and the error field
          nxt_blk_sosi.data  <= next_src_buf.data;
          nxt_blk_sosi.valid <= '1';
          IF next_src_buf.eop='1' THEN
            nxt_blk_sosi.eop <= '1';
            nxt_blk_sosi.err <= next_src_buf.err;  -- the sosi error field contains the one bit PHY error status
            nxt_state <= s_sop;
          END IF;
        END IF;
    END CASE;
    
    -- Pass on frame level flow control
    this_siso.xon   <= src_in.xon;
    hold_src_in.xon <= src_in.xon;
  END PROCESS;
  
  
  -- Handle output error field and eop
  u_src_shift : ENTITY work.dp_shiftreg
  GENERIC MAP (
    g_output_reg     => c_output_reg,
    g_flush_eop      => TRUE,
    g_modify_support => TRUE,
    g_nof_words      => c_shiftreg_len
  )
  PORT MAP (
    rst                 => rst,
    clk                 => clk,
    -- ST sink
    snk_out             => OPEN,
    snk_in              => blk_sosi,
    -- Control shift register contents
    cur_shiftreg_inputs => cur_src_out_inputs,
    new_shiftreg_inputs => new_src_out_inputs,
    -- ST source
    src_in              => src_in,
    src_out             => i_src_out
  );
  
  p_src_err : PROCESS(cur_src_out_inputs)
  BEGIN
    FOR I IN 0 TO c_error_len-1 LOOP
      src_err((I+1)*g_data_w-1 DOWNTO I*g_data_w) <= RESIZE_UVEC(cur_src_out_inputs(I).data, g_data_w);
    END LOOP;
  END PROCESS;
   
  p_src_out : PROCESS(cur_src_out_inputs, src_err)
  BEGIN
    -- default
    new_src_out_inputs <= cur_src_out_inputs;
    
    -- handle the SOSI err field
    IF cur_src_out_inputs(0).eop='1' THEN
      -- strip the error data words by clearing their control
      FOR I IN 0 TO c_error_len-1 LOOP
        new_src_out_inputs(I).valid <= '0';
        new_src_out_inputs(I).sync <= '0';
      END LOOP;
      -- clear the eop at the last error data word
      new_src_out_inputs(0).eop <= '0';
      -- move the eop and move the error data word to the SOSI error field at last payload data
      new_src_out_inputs(c_shiftreg_len-1).eop <= '1';
      new_src_out_inputs(c_shiftreg_len-1).err <= RESIZE_DP_ERROR(src_err);                      -- output transported DP packet err field
      IF g_use_phy_err=TRUE THEN
        new_src_out_inputs(c_shiftreg_len-1).err(g_phy_err_bi) <= cur_src_out_inputs(0).err(0);  -- merge one bit PHY error status at bit index g_phy_err_bi
      END IF;
    END IF;
  END PROCESS;
  
END rtl;

