--------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose: Decode sosi.channel low bits from the high part of the CHAN field
--          of a DP packet.
-- Description:
--   The snk_in data must be DP packet data. The [high-1:high-g_channel_lo]
--   bits of the CHAN field get set to 0 and placed into the [g_channel_lo-1:0]
--   bits of the sosi.channel field. The [high] bit of the CHAN field is not
--   changed because it is reserved for future use.  
-- Remark:
-- . See also dp_packet_enc_channel_lo.
-- . The dp_packet_dec_channel_lo.vhd is verified in tb_dp_distribute.vhd.
-- . The sosi.channel field must be valid during the entire frame, so not only 
--   at the sop. Therefore it is necessary to use channel_lo_hold.

ENTITY dp_packet_dec_channel_lo IS
  GENERIC (
    g_data_w      : NATURAL := 32;
    g_channel_lo  : NATURAL := 0
  );
  PORT (
    rst       : IN STD_LOGIC;
    clk       : IN STD_LOGIC;
    -- ST sinks
    snk_out   : OUT t_dp_siso;
    snk_in    : IN  t_dp_sosi;
    -- ST source
    src_in    : IN  t_dp_siso;
    src_out   : OUT t_dp_sosi
  );
END dp_packet_dec_channel_lo;


ARCHITECTURE rtl OF dp_packet_dec_channel_lo IS

  SIGNAL channel_lo_hold      : STD_LOGIC_VECTOR(g_channel_lo-1 DOWNTO 0);
  SIGNAL nxt_channel_lo_hold  : STD_LOGIC_VECTOR(g_channel_lo-1 DOWNTO 0);
  
BEGIN

  ASSERT g_channel_lo+1 <= g_data_w REPORT "dp_packet_dec_channel_lo : g_channel_lo does not fit in g_data_w" SEVERITY FAILURE;
  
  -- Pass on the flow control
  snk_out <= src_in;
  
  -- Pass on the DP packet and extract the g_channel_lo bits from CHAN field
  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      channel_lo_hold <= (OTHERS=>'0');
    ELSIF rising_edge(clk) THEN
      channel_lo_hold <= nxt_channel_lo_hold;
    END IF;
  END PROCESS;
  
  p_channel_lo : PROCESS (snk_in, channel_lo_hold)
    VARIABLE v_channel_lo : STD_LOGIC_VECTOR(g_channel_lo-1 DOWNTO 0);
  BEGIN
    src_out <= snk_in;
    nxt_channel_lo_hold <= channel_lo_hold;
    
    IF g_channel_lo>0 THEN
      src_out.channel(g_channel_lo-1 DOWNTO 0) <= channel_lo_hold;                 -- default combinatorially assign the held channel_lo bits to the sosi.channel field after the sop
      IF snk_in.sop='1' THEN
        -- clear the channel_lo bits in the MSWord of the CHAN field
        src_out.data(g_data_w-2 DOWNTO g_data_w-1-g_channel_lo) <= (OTHERS=>'0');
        
        -- extract the channel_lo bits from the MSWord of the CHAN field
        v_channel_lo := snk_in.data(g_data_w-2 DOWNTO g_data_w-1-g_channel_lo);
        src_out.channel(g_channel_lo-1 DOWNTO 0) <= v_channel_lo;                  -- combinatorially assign the channel_lo bits to the sosi.channel field at the sop
        nxt_channel_lo_hold                      <= v_channel_lo;                  -- register the channel_lo bits so they can be assigned to the rest of the frame as well
      END IF;
    END IF;
  END PROCESS;
      
END rtl;