--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose:
-- . Merge g_nof_pkt input packets into one output packet
-- Description:
-- . The output is the same as the registered input, except for the SOP and EOP 
--   signal; these are overridden to create output packets as long as multiple (g_nof_pkt)
--   input packets.
--
-- Remarks:
-- . g_nof_pkt statically sets the number of packets to merge into one in absence of 
--   dynamic nof_pkt control. When the nof_pkt control input is used, g_nof_pkt sets
--   the maximum number of packets that can be merged.
-- . the nof_pkt control input is read on completion of the current block. The current 
--   working value is kept in r (t_reg). nof_pkt = 0 is ignored.
-- . with g_align_at_sync=true the merge can be forced to restart at a snk_in.sync.
--
-- Issues:
-- . merged err should preferrably be the bit wise or of the merged input packet err field instead
--   of only the err field of the last.
--
-- Design steps:
-- A) No flow control
-- . Start with the Gaisler p_comb, p_reg template:
--
--                  /----------------------------\
--                  |            nxt_r       r   |
--                  |   /------\   .         .   |
--                  \-->|      |   .         .   |
--     snk_in  -------->|      | ---> p_reg -----*--> src_out
--     snk_out <--------|p_comb| <------------------- src_in
--                      \------/                         .
--                                                       .
--                                                  snk --> src : latency + 1
--
--   The following signals are registered in r: outputs, internal states. Typically do not yet
--   add pipelining registers, because these are not functional and can be added later using
--   common_pipeline or dp_pipeline components. For the design start with a timing diagram of the
--   input and output for merging e.g. 3 packets. Draw the sop and eop and also draw the case that
--   there is a data invalid gap between packets. There needs to be a packet counter so, also
--   draw this signal.
-- . Using the timing diagram now assing each of the v fields. Try to keep the implementation
--   for each field seperate, i.e. do not combine multiple fields into one if then else statement.
--   The reason is that keeping each field implementation seperate makes the code more clear and
--   also eases making modifications later on. Mark the implementation section by a comment line.
-- . Typically use r in the if condition and at the right of the := assignment. Depending only
--   on r for implementing v has the advantage that the order of the code section is not important,
--   i.e. similar as with seperate combinatorial processes. However for some cases it can be
--   beneficial to use v in the condition or in the assignment, e.g. as with v_busy.
-- . If the component will not support flow control, then the snk - src latency may be made
--   > 1. However to prepare for adding flow control later on it is prefered to keep the input -
--   output latency at 1. Keeping the snk - src latency at 1 typically also makes the
--   implementation more clear, because if more latency is needed then it may be better to 
--   partition the function over two or more components. The snk - src latency of 1 is not 
--   functional, because the nxt_r with latency 0 also could be used as output. However the
--   registering of outputs can be functional. From digital implementation point of view it is
--   appropriate though to continue with the r outputs and assign these to the entity outputs.
-- . When the architecture HDL implementation looks clean coded then compile it. When it compiles
--   OK commit it.
-- . Next step is to add a test bench. For dp_packet_merge it is beneficial to first also implement
--   the reverse function dp_packet_unmerge, because that allows the test bench to have the same
--   stimuli input counter data also appear as counter data at the output for verification. 
--   Furthermore even though the reverse function may not be needed yet in an application it 
--   happens quite often that it will become useful in the future.
-- . For the stimuli and verification use the counter data procedures from tb_dp_pkg.vhd. The
--   test bench consists of stimuli --> DUT --> reverse DUT --> verification. First keep data
--   valid active and use nof_pkt > 1. When that works OK, commit the DUT and the test bench.
--   After that try the extreme values for control, i.e. nof_pkt = 0 should yield no output and
--   nof_pkt = 1 should yield output = input. Then add dynamic in_en to get data not valid gaps
--   during and between the packets. Also try dynamic changing of nof_pkt. When that all
--   works commit the VHDL files.
--   
-- B) Add flow control using dp_latency_adapter:
--
--                  /----------------------------\
--                  |            nxt_r       r   |           
--                  |   /------\   .         .   |     /-------\         
--                  \-->|      |   .         .   |     |dp     |         
--     snk_in  -------->|      | ---> p_reg -----*---->|latency|---> src_out
--     snk_out <--------|p_comb| <---------------------|adapter|<--- src_in
--        .             \------/             .         \-------/  .
--        .                                  .                    .
--        .                                  .                    .
--        .                                  .                    .
--      RL=1                               RL=2                 RL=1
--
-- C) Add flow control using dp_hold_input
-- . Adding flow control implies adding dp_hold_input to connect to the snk_in/skn_out and using
--   pend_src_out/hold_src_in instead in p_comb:
--
--                 |-----| <---------------------------------------------\
--                 |     |                               nxt_r       r   |
--                 |     |                      /------\   .         .   |
--                 | dp  |                      |      |   .         .   |
--     snk_in  --> |hold | --> pend_src_out --> |      | ---> p_reg -----*--> src_out
--     snk_out <-- |input| <-- hold_src_in  <-- |p_comb| <------------------- src_in
--        .        |-----|         .            \------/                         .
--        .                        .                                             .
--      RL=1                     RL=0                                          RL=1
--
--   The ready latency (RL) at the input is RL=1. The RL at the output must also be RL=1. The
--   dp_hold_input effectively achieves that the pend_src_out/hold_src_in have RL=0.
--   The snk - src latency of the function is 1, so at the output the RL then becomes RL=1 again.
--   Inside p_comb the v output valid now only get assigned with the pend_src_out valid when
--   src_in.ready='1'. Internally in dp_hold_input the next_src_out.valid = pend_src_out.valid
--   AND src_in.ready, and similar for the sop, eop amd sync. Therefore the p_comb can also use
--   next_src_out instead of pend_src_out in combination with src_in.ready. The advantage of
--   using next_src_out is that it makes the code more compact. The advantange of using
--   pend_src_out and src_in.ready explicitly is that it may make it more clear how src_in.ready
--   is used.
-- . Typically the p_comb can simply pass on src_in to hold_src_in, however it may also control
--   hold_src_in by itself if this function also needs some extra flow control, e.g. to halt or
--   flush the input.
-- . Most components only need to pass on the src_in.xon to hold_src_in.xon. Hence this can be
--   the last statement in p_comb.
-- . In the test bench now also apply the src_in.ready to verify that the component behaves 
--   correctly for allways active ready and random ready flow control.
-- . Make a multi-test bench tb_tb_dp_packet_merge to verify different tb settings for the DUT
--   stimuli and add it to the overall DP module tb_tb_tb_dp_backpressure regression test bench.
--
-- D) Potential alternative to add flow control using dp_pipeline
--
-- . This scheme has not been verified yet, but it could work:
--
--                  /-----------------------------\
--                  |            nxt_r        r   |     dp_pipeline_stream
--                  |   /------\   .          .   |  <--select none or one-->
--                  \-->|      |---o-> p_reg -----/      ____    /-----\
--                      |      |   |                    |dp  |   |dp   | 
--     snk_in  -------->|      |   \------------------->|pipe|-->|pipe |--> src_out
--     snk_out <--------|p_comb|<-----------------------|line|<--|line |--- src_in
--        .             \------/   .                    |____|   |ready|      .
--        .               RL=0     .                             \-----/      .
--      RL=1                     RL=1                                        RL=1
--
-- . dp_pipeline_stream = wires or dp_pipeline or dp_pipeline_ready
--   - dp_pipeline pipelines the sosi
--   - dp_pipeline_ready pipelines the siso and also the sosi
-- . without flow control there is no siso, so then dp_pipeline_stream degenerates
--   to common_pipeline.
-- . The nxt_r contains combinatorially (= functionally) already what src_out will become in
--   the next cycle and because it is combinatorially related to the input it preserves the
--   RL=1 (whereas p_reg has RL=2). Leading nxt_r through dp_pipeline includes the flow control
--   to it.
-- . Compared to using dp_hold_input using dp_pipeline costs twice more registers, because
--   dp_pipeline internally has a register stage and p_reg is still needed. Still synthesis
--   may be able to optimize away some redundant registers.
-- . The advantage of this scheme it that it allows designing the function without flow
--   control according to scheme A). Then make nxt_r available instead of r and lead nxt_r
--   through dp_pipeline to register it and to add the flow control.
--

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY dp_packet_merge IS
  GENERIC (
    g_nof_pkt       : NATURAL;
    g_align_at_sync : BOOLEAN := FALSE
  );
  PORT (
    rst         : IN  STD_LOGIC;
    clk         : IN  STD_LOGIC;

    nof_pkt     : IN  STD_LOGIC_VECTOR(ceil_log2(g_nof_pkt+1)-1 DOWNTO 0) := TO_UVEC(g_nof_pkt, ceil_log2(g_nof_pkt+1));
    nof_pkt_out : OUT STD_LOGIC_VECTOR(ceil_log2(g_nof_pkt+1)-1 DOWNTO 0); -- Valid at src_out.sop

    snk_out     : OUT t_dp_siso;
    snk_in      : IN  t_dp_sosi;

    src_in      : IN  t_dp_siso;
    src_out     : OUT t_dp_sosi
  );
END dp_packet_merge;


ARCHITECTURE rtl OF dp_packet_merge IS

  TYPE t_reg IS RECORD
    nof_pkt     : NATURAL RANGE 0 TO g_nof_pkt+1;
    pkt_cnt     : NATURAL RANGE 0 TO g_nof_pkt+1;
    align_cnt   : NATURAL RANGE 0 TO g_nof_pkt+1;
    busy        : STD_LOGIC;
    src_out     : t_dp_sosi;
  END RECORD;

  CONSTANT c_use_dp_latency_adapter : BOOLEAN := TRUE;  -- when TRUE adjust RL from 2 to 1, else use dp_hold_input to keep RL at 1.
  
  SIGNAL r, nxt_r : t_reg;
  
  SIGNAL dp_latency_adapter_snk_out : t_dp_siso;
  SIGNAL dp_latency_adapter_snk_in  : t_dp_sosi;
  SIGNAL dp_latency_adapter_src_in  : t_dp_siso;
  SIGNAL dp_latency_adapter_src_out : t_dp_sosi;

BEGIN
  
  -- Map t_reg outputs to entity outputs
  nof_pkt_out <= TO_UVEC(r.nof_pkt, ceil_log2(g_nof_pkt+1));

  no_dp_latency_adapter : IF c_use_dp_latency_adapter=FALSE GENERATE
    snk_out <= src_in;
    src_out <= r.src_out;
    
    -- can put dp_hold_input here --
    
  END GENERATE;
  
  gen_dp_latency_adapter : IF c_use_dp_latency_adapter=TRUE GENERATE  
    snk_out <= dp_latency_adapter_snk_out;
    dp_latency_adapter_snk_in <= r.src_out;
    
    u_dp_latency_adapter : ENTITY work.dp_latency_adapter
    GENERIC MAP (
      g_in_latency  => 2,
      g_out_latency => 1
    )
    PORT MAP (
      rst          => rst,
      clk          => clk,
      -- ST sink
      snk_out      => dp_latency_adapter_snk_out,
      snk_in       => dp_latency_adapter_snk_in,
      -- ST source
      src_in       => dp_latency_adapter_src_in,
      src_out      => dp_latency_adapter_src_out
    );
    
    dp_latency_adapter_src_in <= src_in;
    src_out <= dp_latency_adapter_src_out;
  END GENERATE;
  

  -- p_reg
  r <= nxt_r WHEN rising_edge(clk);

  p_comb : PROCESS(rst, r, snk_in, nof_pkt)
    VARIABLE v : t_reg;
  BEGIN
    ----------------------------------------------------------------------------
    -- Default
    v := r;

    ----------------------------------------------------------------------------
    -- Function
    --                       _         _         _         _         _         _            _         _         _  
    --   snk_in.sop    _____|0|_______|1|_______|2|_______|0|_______|1|_______|2|__________|0|_______|1|_______|2|_______
    --                               _         _         _         _         _         _            _         _         _   
    --   snk_in.eop    _____________|0|_______|1|_______|2|_______|0|_______|1|_______|2|__________|0|_______|1|_______|2
    --                                                                                                                   
    --   r.nof_pkt                                       3                             3                                3
    --   v.pkt_cnt     0             1         2         0         1         2         0            1         2         0        
    --   r.pkt_cnt     x               1         2         0         1         2            0         1         2        
    --                       ___________________________   ___________________________      _____________________________
    --   v.busy        _____|                           |_|                           |____|
    --                       _                             _                                _                            
    --   v.src_out.sop _____|0|___________________________|1|______________________________|2|___________________________
    --                                                   _                             _                                _
    --   v.src_out.eop _________________________________|0|___________________________|1|______________________________|2
    --

    -- input packet counter
    IF snk_in.eop = '1' THEN
      IF r.pkt_cnt < r.nof_pkt-1 THEN 
        v.pkt_cnt := r.pkt_cnt + 1;
      ELSE
        v.pkt_cnt := 0;
      END IF;
    END IF;
    
    -- capture nof_pkt between output packets when there is no merge busy
    IF snk_in.sop = '1' THEN
      v.busy := '1';
    END IF;
    IF snk_in.eop = '1' THEN
      IF r.pkt_cnt = r.nof_pkt-1 THEN
        v.busy := '0';
      END IF;
    END IF;
    
    IF g_align_at_sync THEN
      IF snk_in.sync='1' THEN
        IF r.pkt_cnt > 0 THEN
          v.align_cnt := r.nof_pkt - r.pkt_cnt;  -- = 0 when sync occurs at integer number of merged blocks, > 0 when (re)align to sync is needed
        END IF;
      END IF;
    END IF;
    
    IF v.busy = '0' AND v.src_out.eop = '1' THEN
      v.nof_pkt := TO_UINT(nof_pkt) - r.align_cnt; -- use shorter next merge to (re)align to sync (when g_align_at_sync=false then r.align_cnt=0)
      v.align_cnt := 0;
    END IF;
    
    -- output packet bsn, sync, sop and eop
    v.src_out         := snk_in;
    v.src_out.bsn     := r.src_out.bsn;
    v.src_out.channel := r.src_out.channel;
    v.src_out.empty   := r.src_out.empty;
    v.src_out.err     := r.src_out.err;
    v.src_out.sync    := '0';
    v.src_out.sop     := '0';
    v.src_out.eop     := '0';
    
    IF snk_in.sop = '1' THEN
      IF r.pkt_cnt = 0 THEN
        v.src_out.sop     := '1';
        v.src_out.sync    := snk_in.sync;     -- only preserve sync when sync interval and nof_pkt interval are aligned, else sync gets lost but could still be recovered from BSN
        v.src_out.bsn     := snk_in.bsn;      -- use BSN     of first packet for merged packet
        v.src_out.channel := snk_in.channel;  -- use channel of first packet for merged packet
      END IF;
    END IF;
    
    IF snk_in.eop = '1' THEN
      IF r.pkt_cnt = r.nof_pkt-1 THEN
        v.src_out.eop   := '1';
        v.src_out.empty := snk_in.empty;
        v.src_out.err   := snk_in.err;        -- use err of last packet for merged packet
      END IF;
    END IF;
    
    -- force sosi control to inactive for no output when nof_pkt=0    
    IF r.nof_pkt=0 THEN
      v.src_out.valid := '0';
      v.src_out.sync  := '0';
      v.src_out.sop   := '0';
      v.src_out.eop   := '0';
      v.busy          := '0';
    END IF;
    
    ----------------------------------------------------------------------------
    -- Reset and nxt_r
    IF rst = '1' THEN
      v.nof_pkt     := TO_UINT(nof_pkt);
      v.pkt_cnt     := 0;
      v.align_cnt   := 0;
      v.busy        := '0';
      v.src_out     := c_dp_sosi_rst;
    END IF;
 
    nxt_r <= v;
  END PROCESS;
  
END rtl;
