--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib, easics_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;

USE easics_lib.RAD_CRC16_D16.ALL;
USE easics_lib.RAD_CRC18_D18.ALL;
USE easics_lib.RAD_CRC20_D20.ALL;

USE easics_lib.PCK_CRC16_D4.ALL;
USE easics_lib.PCK_CRC16_D8.ALL;
USE easics_lib.PCK_CRC16_D9.ALL;
USE easics_lib.PCK_CRC16_D10.ALL;
USE easics_lib.PCK_CRC16_D16.ALL;
USE easics_lib.PCK_CRC16_D18.ALL;
USE easics_lib.PCK_CRC16_D20.ALL;
USE easics_lib.PCK_CRC16_D32.ALL;
USE easics_lib.PCK_CRC16_D36.ALL;
USE easics_lib.PCK_CRC16_D48.ALL;
USE easics_lib.PCK_CRC16_D64.ALL;
USE easics_lib.PCK_CRC16_D72.ALL;

USE easics_lib.PCK_CRC32_D4.ALL;
USE easics_lib.PCK_CRC32_D8.ALL;
USE easics_lib.PCK_CRC32_D9.ALL;
USE easics_lib.PCK_CRC32_D10.ALL;
USE easics_lib.PCK_CRC32_D16.ALL;
USE easics_lib.PCK_CRC32_D18.ALL;
USE easics_lib.PCK_CRC32_D20.ALL;
USE easics_lib.PCK_CRC32_D32.ALL;
USE easics_lib.PCK_CRC32_D36.ALL;
USE easics_lib.PCK_CRC32_D48.ALL;
USE easics_lib.PCK_CRC32_D64.ALL;
USE easics_lib.PCK_CRC32_D72.ALL;

PACKAGE dp_packetizing_pkg Is

  --<types>--

  --<functions>--
  FUNCTION func_dp_next_crc(c_lofar : BOOLEAN; dat, crc : STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR;
  
  --<constants>--
  CONSTANT c_dp_max_w          : NATURAL := 64;
  CONSTANT c_dp_gap_min        : NATURAL := 4;  -- minimal gap between valid data blocks (to insert >= 1 idle, sfd, fsn and brc/crc into the stream)
  CONSTANT c_dp_idle           : STD_LOGIC_VECTOR(c_dp_max_w-1 DOWNTO 0) := "0101010101010101010101010101010101010101010101010101010101010101";  -- 0x..5555
  CONSTANT c_dp_sfd            : STD_LOGIC_VECTOR(c_dp_max_w-1 DOWNTO 0) := "0101010101010101010101010101010101010101010101010101010101010100";  -- 0x..5554;
  -- define more c_dp_sfd_* to have multiple independent dp frame streams per link, e.g. as was used for LOFAR:
  CONSTANT c_dp_sfd_lofar      : STD_LOGIC_VECTOR(c_dp_max_w-1 DOWNTO 0) := "0101010101010101010101010101010101010101010101010001000011111010";  -- 0x..10FA;
  
  CONSTANT c_dp_brc_ok         : NATURAL := 0;
  CONSTANT c_dp_brc_err        : NATURAL := 1;
  
  CONSTANT c_dp_crc_w          : NATURAL := 16;
  
  CONSTANT c_dp_mode_w         : NATURAL := 2;
  CONSTANT c_dp_mode_disable   : STD_LOGIC_VECTOR(c_dp_mode_w-1 DOWNTO 0) := "01";
  CONSTANT c_dp_mode_buf       : STD_LOGIC_VECTOR(c_dp_mode_w-1 DOWNTO 0) := "00";
  CONSTANT c_dp_mode_stream    : STD_LOGIC_VECTOR(c_dp_mode_w-1 DOWNTO 0) := "11";
  CONSTANT c_dp_mode_combine   : STD_LOGIC_VECTOR(c_dp_mode_w-1 DOWNTO 0) := "10";

END dp_packetizing_pkg;


PACKAGE BODY dp_packetizing_pkg IS

  --<functions>--
  
  ------------------------------------------------------------------------------
  -- FUNCTION: Calculate next CRC for this data
  ------------------------------------------------------------------------------
  
  FUNCTION func_dp_next_crc(c_lofar : BOOLEAN; dat, crc : STD_LOGIC_VECTOR) RETURN STD_LOGIC_VECTOR IS
    VARIABLE nxt_crc : STD_LOGIC_VECTOR(crc'RANGE);
  BEGIN
    IF c_lofar=TRUE THEN
      CASE dat'LENGTH IS
        WHEN 16 => nxt_crc := nextRAD_CRC16_D16(RESIZE_UVEC(dat, 16), crc);
        WHEN 18 => nxt_crc := nextRAD_CRC18_D18(RESIZE_UVEC(dat, 18), crc);
        WHEN 20 => nxt_crc := nextRAD_CRC20_D20(RESIZE_UVEC(dat, 20), crc);
        WHEN OTHERS   => REPORT "DP data width not supported in easics_lib RAD for LOFAR" SEVERITY FAILURE;
      END CASE;
    ELSE
      -- CRC-16
      IF c_dp_crc_w=16 THEN
        CASE dat'LENGTH IS
          WHEN 1  TO 4  => nxt_crc := nextCRC16_D4( RESIZE_UVEC(dat,  4), crc);
          WHEN 5  TO 8  => nxt_crc := nextCRC16_D8( RESIZE_UVEC(dat,  8), crc);
          WHEN 9        => nxt_crc := nextCRC16_D9(             dat     , crc);
          WHEN 10       => nxt_crc := nextCRC16_D10(            dat     , crc);
          WHEN 11 TO 16 => nxt_crc := nextCRC16_D16(RESIZE_UVEC(dat, 16), crc);
          WHEN 17 TO 18 => nxt_crc := nextCRC16_D18(RESIZE_UVEC(dat, 18), crc);
          WHEN 19 TO 20 => nxt_crc := nextCRC16_D20(RESIZE_UVEC(dat, 20), crc);
          WHEN 21 TO 32 => nxt_crc := nextCRC16_D32(RESIZE_UVEC(dat, 32), crc);
          WHEN 33 TO 36 => nxt_crc := nextCRC16_D36(RESIZE_UVEC(dat, 36), crc);
          WHEN 37 TO 48 => nxt_crc := nextCRC16_D48(RESIZE_UVEC(dat, 48), crc);
          WHEN 49 TO 64 => nxt_crc := nextCRC16_D64(RESIZE_UVEC(dat, 64), crc);
          WHEN 65 TO 72 => nxt_crc := nextCRC16_D72(RESIZE_UVEC(dat, 72), crc);
          WHEN OTHERS   => REPORT "DP data width not supported in easics_lib CRC-16" SEVERITY FAILURE;
        END CASE;
      END IF;
      -- CRC-32
      IF c_dp_crc_w=32 THEN
        CASE dat'LENGTH IS
          WHEN 1  TO 4  => nxt_crc := nextCRC32_D4( RESIZE_UVEC(dat,  4), crc);
          WHEN 5  TO 8  => nxt_crc := nextCRC32_D8( RESIZE_UVEC(dat,  8), crc);
          WHEN 9        => nxt_crc := nextCRC32_D9(             dat     , crc);
          WHEN 10       => nxt_crc := nextCRC32_D10(            dat     , crc);
          WHEN 11 TO 16 => nxt_crc := nextCRC32_D16(RESIZE_UVEC(dat, 16), crc);
          WHEN 17 TO 18 => nxt_crc := nextCRC32_D18(RESIZE_UVEC(dat, 18), crc);
          WHEN 19 TO 20 => nxt_crc := nextCRC32_D20(RESIZE_UVEC(dat, 20), crc);
          WHEN 21 TO 32 => nxt_crc := nextCRC32_D32(RESIZE_UVEC(dat, 32), crc);
          WHEN 33 TO 36 => nxt_crc := nextCRC32_D36(RESIZE_UVEC(dat, 36), crc);
          WHEN 37 TO 48 => nxt_crc := nextCRC32_D48(RESIZE_UVEC(dat, 48), crc);
          WHEN 49 TO 64 => nxt_crc := nextCRC32_D64(RESIZE_UVEC(dat, 64), crc);
          WHEN 65 TO 72 => nxt_crc := nextCRC32_D72(RESIZE_UVEC(dat, 72), crc);
          WHEN OTHERS   => REPORT "DP data width not supported in easics_lib CRC-32" SEVERITY FAILURE;
        END CASE;
      END IF;
    END IF;
    
    RETURN nxt_crc;
  END func_dp_next_crc;

END dp_packetizing_pkg;

