--------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose:
--   Remove padding octects from the head of a padded frame
-- Description:
--   This dp_pad_remove merely uses dp_split. It throws away the head and it
--   passes on the tail.
-- Remark:
-- . Ready latency (RL) = 1
-- . Reverse operation is dp_pad_insert


ENTITY dp_pad_remove IS
  GENERIC (
    g_data_w          : NATURAL := 64;
    g_symbol_w        : NATURAL := 8;    -- g_data_w/g_symbol_w must be integer
    g_nof_padding     : NATURAL := 6;    -- must be >= 0 (because dp_split can support 0, although using g_nof_padding = 0 is void)
    g_internal_bypass : BOOLEAN := FALSE -- can be used to eliminate external GENERATE-statements/wiring
  );
  PORT (
    rst     : IN  STD_LOGIC;
    clk     : IN  STD_LOGIC;
    -- ST sinks
    snk_out : OUT t_dp_siso;
    snk_in  : IN  t_dp_sosi;
    -- ST source
    src_in  : IN  t_dp_siso;
    src_out : OUT t_dp_sosi
  );
END dp_pad_remove;


ARCHITECTURE str OF dp_pad_remove IS

  -- element (0) gets the head, element (1) gets the tail
  SIGNAL split_siso_arr : t_dp_siso_arr(0 TO 1);
  SIGNAL split_sosi_arr : t_dp_sosi_arr(0 TO 1);
  
BEGIN

  no_bypass : IF g_internal_bypass = FALSE GENERATE
    -- Remove the padding octets from the head and output the tail
    split_siso_arr(0) <= c_dp_siso_rdy;
    split_siso_arr(1) <= src_in;
    
    src_out <= split_sosi_arr(1);
    
    u_split : ENTITY work.dp_split  -- RL = 1
    GENERIC MAP (
      g_data_w        => g_data_w,
      g_symbol_w      => g_symbol_w,
      g_nof_symbols   => g_nof_padding
    )
    PORT MAP (
      rst         => rst,
      clk         => clk,
      -- ST sinks
      snk_out     => snk_out,
      snk_in      => snk_in,
      -- ST source
      src_in_arr  => split_siso_arr,
      src_out_arr => split_sosi_arr
    );
  END GENERATE;

  gen_bypass : IF g_internal_bypass = TRUE GENERATE
    src_out <= snk_in;
    snk_out <= src_in;
  END GENERATE;  

END str;
