-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Dual page register to pass on the sosi sop and eop related fields
-- Description:
--   The sync, bsn and channel sosi fields are valid at the sop.
--   The empty and err sosi fields are valid at the eop.
--   The sop related input sosi fields are shifted from page to page dependent
--   on sop_wr_en[]. Idem for the eop related input sosi fields with
--   eop_wr_en[].
-- Remarks:
-- . E.g. use g_nof_pages = 2 and
--     sop_wr_en <= snk_in.sop & snk_in.sop;
--     eop_wr_en <= snk_in.eop & snk_in.eop;
--   to capture the input at the first wr_en and hold it for output at the
--   next wr_en.

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY dp_paged_sop_eop_reg IS
  GENERIC (
    g_nof_pages  : NATURAL := 2   -- >= 0
  );
  PORT (
    rst         : IN  STD_LOGIC;
    clk         : IN  STD_LOGIC;
    -- page write enable ctrl
    sop_wr_en   : IN  STD_LOGIC_VECTOR(g_nof_pages-1 DOWNTO 0);
    eop_wr_en   : IN  STD_LOGIC_VECTOR(g_nof_pages-1 DOWNTO 0);
    -- ST sink
    snk_in      : IN  t_dp_sosi;
    -- ST source
    src_out     : OUT t_dp_sosi
  );
END dp_paged_sop_eop_reg;


ARCHITECTURE str OF dp_paged_sop_eop_reg IS
  
  SIGNAL snk_in_sync  : STD_LOGIC_VECTOR(0 DOWNTO 0);
  SIGNAL src_out_sync : STD_LOGIC_VECTOR(0 DOWNTO 0);
  
BEGIN

  -- Sosi info at sop
  snk_in_sync(0) <= snk_in.sync;      -- convert sl to slv
  src_out.sync   <= src_out_sync(0);  -- convert slv to sl

  u_paged_sync : ENTITY common_lib.common_paged_reg
  GENERIC MAP (
    g_data_w    => 1,
    g_nof_pages => g_nof_pages
  )
  PORT MAP (
    rst          => rst,
    clk          => clk,
    wr_en        => sop_wr_en,
    wr_dat       => snk_in_sync,
    out_dat      => src_out_sync
  );
  
  u_paged_bsn : ENTITY common_lib.common_paged_reg
  GENERIC MAP (
    g_data_w    => c_dp_stream_bsn_w,
    g_nof_pages => g_nof_pages
  )
  PORT MAP (
    rst          => rst,
    clk          => clk,
    wr_en        => sop_wr_en,
    wr_dat       => snk_in.bsn,
    out_dat      => src_out.bsn
  );
  
  u_paged_channel : ENTITY common_lib.common_paged_reg
  GENERIC MAP (
    g_data_w    => c_dp_stream_channel_w,
    g_nof_pages => g_nof_pages
  )
  PORT MAP (
    rst          => rst,
    clk          => clk,
    wr_en        => sop_wr_en,
    wr_dat       => snk_in.channel,
    out_dat      => src_out.channel
  );
  
  -- Sosi info at eop
  u_paged_empty : ENTITY common_lib.common_paged_reg
  GENERIC MAP (
    g_data_w    => c_dp_stream_empty_w,
    g_nof_pages => g_nof_pages
  )
  PORT MAP (
    rst          => rst,
    clk          => clk,
    wr_en        => eop_wr_en,
    wr_dat       => snk_in.empty,
    out_dat      => src_out.empty
  );
  
  u_paged_err : ENTITY common_lib.common_paged_reg
  GENERIC MAP (
    g_data_w    => c_dp_stream_error_w,
    g_nof_pages => g_nof_pages
  )
  PORT MAP (
    rst          => rst,
    clk          => clk,
    wr_en        => eop_wr_en,
    wr_dat       => snk_in.err,
    out_dat      => src_out.err
  );
  
END str;
