--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, technology_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY dp_ram_from_mm IS
  GENERIC (
    g_technology       : NATURAL := c_tech_select_default;
    g_ram_wr_nof_words : NATURAL;
    g_ram_rd_dat_w     : NATURAL;
    g_init_file        : STRING := "UNUSED" 
  );
  PORT (
    mm_rst    : IN  STD_LOGIC;
    mm_clk    : IN  STD_LOGIC;

    st_rst    : IN  STD_LOGIC;
    st_clk    : IN  STD_LOGIC;

    mm_addr   : IN  STD_LOGIC_VECTOR(c_mem_address_w-1 DOWNTO 0);
    mm_wr     : IN  STD_LOGIC;
    mm_wrdata : IN  STD_LOGIC_VECTOR(c_mem_data_w-1 DOWNTO 0);

    dp_on     : IN  STD_LOGIC := '0';

    src_in    : IN  t_dp_siso;
    src_out   : OUT t_dp_sosi
  );
END dp_ram_from_mm;

ARCHITECTURE rtl OF dp_ram_from_mm IS

  CONSTANT c_mm_ram_wr  : t_c_mem := (latency  => 1,
                                      adr_w    => ceil_log2(g_ram_wr_nof_words),
                                      dat_w    => c_word_w,
                                      nof_dat  => g_ram_wr_nof_words,
                                      init_sl  => '0');

  CONSTANT c_ram_rd_nof_words : NATURAL := (c_word_w * g_ram_wr_nof_words) / g_ram_rd_dat_w;

  CONSTANT c_mm_ram_rd  : t_c_mem := (latency  => 1,
                                      adr_w    => ceil_log2(c_ram_rd_nof_words),
                                      dat_w    => g_ram_rd_dat_w,
                                      nof_dat  => c_ram_rd_nof_words,
                                      init_sl  => '0');

  TYPE t_state_enum IS (s_init, s_wait_for_rdy, s_read);

  SIGNAL state         : t_state_enum;
  SIGNAL nxt_state     : t_state_enum; 

  SIGNAL src_out_ctrl_val     : t_dp_sosi;
  SIGNAL nxt_src_out_ctrl_val : t_dp_sosi;

  SIGNAL rd_en       : STD_LOGIC;
  SIGNAL rd_val      : STD_LOGIC;
  SIGNAL rd_data     : STD_LOGIC_VECTOR(c_mm_ram_rd.dat_w-1 DOWNTO 0);
  SIGNAL rd_addr     : STD_LOGIC_VECTOR(c_mm_ram_rd.adr_w-1 DOWNTO 0);
  SIGNAL nxt_rd_addr : STD_LOGIC_VECTOR(c_mm_ram_rd.adr_w-1 DOWNTO 0);

BEGIN

  p_src_out : PROCESS(src_out_ctrl_val, rd_data)
  BEGIN
    src_out <= src_out_ctrl_val;
    src_out.data(c_mm_ram_rd.dat_w-1 DOWNTO 0) <= rd_data;
  END PROCESS;

  p_state : PROCESS(state, rd_addr, src_in, dp_on, rd_data)
  BEGIN
    nxt_state   <= state;  
    nxt_rd_addr <= rd_addr;
    nxt_src_out_ctrl_val <= c_dp_sosi_rst;
    rd_en <= '0';

    CASE state IS
      WHEN s_init => 
        nxt_state <= s_wait_for_rdy;

      WHEN s_wait_for_rdy => 
        -- Only start stream when dp_on = '1'
        IF dp_on='1' THEN
          IF src_in.ready = '1' THEN
            rd_en <= '1';
            nxt_state <= s_read;
            nxt_rd_addr <= INCR_UVEC(rd_addr, 1); 
            nxt_src_out_ctrl_val.valid <= '1';
            nxt_src_out_ctrl_val.sop <= '1';
            IF UNSIGNED(rd_addr)=c_ram_rd_nof_words-1 THEN
              nxt_src_out_ctrl_val.eop <= '1';
              nxt_state   <= s_wait_for_rdy;
              nxt_rd_addr <= (OTHERS=>'0');
            END IF;
          END IF;
        END IF;

      WHEN s_read =>
        IF src_in.ready = '1' THEN
          rd_en <= '1';
          nxt_rd_addr <= INCR_UVEC(rd_addr, 1); 
          nxt_src_out_ctrl_val.valid <= '1';
          IF UNSIGNED(rd_addr)=c_ram_rd_nof_words-1 THEN
            nxt_src_out_ctrl_val.eop <= '1';
            nxt_state   <= s_wait_for_rdy;
            nxt_rd_addr <= (OTHERS=>'0');
            -- Only stop stream after EOP
            IF dp_on='0' THEN
              nxt_state <= s_wait_for_rdy;
            END IF; 
          END IF;
        END IF;
        
      WHEN OTHERS =>
        nxt_state <= s_init;
    END CASE;
    
  END PROCESS;


  p_clk : PROCESS(st_rst, st_clk)
  BEGIN
    IF st_rst='1' THEN
      state   <= s_init;  
      rd_addr <= (OTHERS=>'0');
      src_out_ctrl_val <= c_dp_sosi_rst;
    ELSIF rising_edge(st_clk) THEN
      state   <= nxt_state;
      rd_addr <= nxt_rd_addr;
      src_out_ctrl_val <= nxt_src_out_ctrl_val;
    END IF;
  END PROCESS;

  u_ram : ENTITY common_lib.common_ram_cr_cw_ratio
  GENERIC MAP (
    g_technology => g_technology,
    g_ram_wr    => c_mm_ram_wr,
    g_ram_rd    => c_mm_ram_rd,
    g_init_file => g_init_file
  )
  PORT MAP (
    -- Write port clock domain
    wr_rst    => mm_rst,
    wr_clk    => mm_clk,
    wr_en     => mm_wr,
    wr_adr    => mm_addr(c_mm_ram_wr.adr_w-1 DOWNTO 0),
    wr_dat    => mm_wrdata(c_mm_ram_wr.dat_w-1 DOWNTO 0),
    -- Read port clock domain
    rd_rst    => st_rst,
    rd_clk    => st_clk,
    rd_en     => rd_en,
    rd_adr    => rd_addr,
    rd_dat    => rd_data,
    rd_val    => OPEN 
  );  

END rtl;

