--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, technology_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY dp_ram_to_mm IS
  GENERIC (
    g_technology       : NATURAL := c_tech_select_default;
    g_ram_rd_nof_words : NATURAL;
    g_ram_wr_dat_w     : NATURAL
  );
  PORT (
    mm_rst   : IN  STD_LOGIC;
    mm_clk   : IN  STD_LOGIC;

    st_rst   : IN  STD_LOGIC;
    st_clk   : IN  STD_LOGIC;

    sla_in   : IN  t_mem_mosi; 
    sla_out  : OUT t_mem_miso; 

    snk_in   : IN  t_dp_sosi;
    snk_out  : OUT t_dp_siso
  );
END dp_ram_to_mm;

ARCHITECTURE rtl OF dp_ram_to_mm IS

  CONSTANT c_mm_ram_rd  : t_c_mem := (latency  => 1,
                                      adr_w    => ceil_log2(g_ram_rd_nof_words),
                                      dat_w    => c_word_w,
                                      nof_dat  => g_ram_rd_nof_words,
                                      init_sl  => '0');

  CONSTANT c_ram_wr_nof_words : NATURAL := (c_word_w * g_ram_rd_nof_words) / g_ram_wr_dat_w;

  CONSTANT c_mm_ram_wr  : t_c_mem := (latency  => 1,
                                      adr_w    => ceil_log2(c_ram_wr_nof_words),
                                      dat_w    => g_ram_wr_dat_w,
                                      nof_dat  => c_ram_wr_nof_words,
                                      init_sl  => '0');

  TYPE t_state_enum IS (s_init, s_wait_for_sop, s_write);

  SIGNAL state         : t_state_enum;
  SIGNAL nxt_state     : t_state_enum; 

  SIGNAL nxt_snk_out  : t_dp_siso;

  SIGNAL wr_en       : STD_LOGIC;
  SIGNAL wr_data     : STD_LOGIC_VECTOR(c_mm_ram_wr.dat_w-1 DOWNTO 0);
  SIGNAL wr_addr     : STD_LOGIC_VECTOR(c_mm_ram_wr.adr_w-1 DOWNTO 0);
  SIGNAL nxt_wr_addr : STD_LOGIC_VECTOR(c_mm_ram_wr.adr_w-1 DOWNTO 0);

BEGIN

  p_state : PROCESS(state, wr_addr, snk_in, wr_data)
  BEGIN
    nxt_state   <= state;  
    nxt_wr_addr <= wr_addr;
    wr_en <= '0';
    nxt_snk_out <= c_dp_siso_rst;

    CASE state IS
      WHEN s_init => 
        nxt_state <= s_wait_for_sop;

      WHEN s_wait_for_sop => 
        nxt_snk_out <= c_dp_siso_rdy;
        IF snk_in.sop = '1' THEN
          wr_en <= '1';
          nxt_state <= s_write;
          nxt_wr_addr <= INCR_UVEC(wr_addr, 1); 
          IF UNSIGNED(wr_addr)=c_ram_wr_nof_words-1 THEN
            nxt_state   <= s_wait_for_sop;
            nxt_wr_addr <= (OTHERS=>'0');
          END IF;
        END IF;

      WHEN s_write =>
        nxt_snk_out <= c_dp_siso_rdy;
        IF snk_in.valid = '1' THEN
          wr_en <= '1';
          nxt_wr_addr <= INCR_UVEC(wr_addr, 1); 
          IF UNSIGNED(wr_addr)=c_ram_wr_nof_words-1 THEN
            nxt_state   <= s_wait_for_sop;
            nxt_wr_addr <= (OTHERS=>'0');
          END IF;
        END IF;
        
      WHEN OTHERS =>
        nxt_state <= s_init;
    END CASE;
    
  END PROCESS;


  p_clk : PROCESS(st_rst, st_clk)
  BEGIN
    IF st_rst='1' THEN
      state   <= s_init;  
      wr_addr <= (OTHERS=>'0');
      snk_out <= c_dp_siso_rst;
    ELSIF rising_edge(st_clk) THEN
      state   <= nxt_state;
      wr_addr <= nxt_wr_addr;
      snk_out <= nxt_snk_out;
    END IF;
  END PROCESS;

  u_ram : ENTITY common_lib.common_ram_cr_cw_ratio
  GENERIC MAP (
    g_technology => g_technology,
    g_ram_wr    => c_mm_ram_wr,
    g_ram_rd    => c_mm_ram_rd,
    g_init_file => "UNUSED" 
  )
  PORT MAP (
    -- Write port clock domain
    wr_rst    => st_rst,
    wr_clk    => st_clk,
    wr_en     => snk_in.valid, 
    wr_adr    => wr_addr,
    wr_dat    => snk_in.data(c_mm_ram_wr.dat_w-1 DOWNTO 0),
    -- Read port clock domain
    rd_rst    => mm_rst,
    rd_clk    => mm_clk,
    rd_en     => sla_in.rd,
    rd_adr    => sla_in.address(c_mm_ram_rd.adr_w-1 DOWNTO 0),
    rd_dat    => sla_out.rddata(c_mm_ram_rd.dat_w-1 DOWNTO 0),
    rd_val    => sla_out.rdval
  );  

END rtl;

