--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Author:
-- . Daniel van der Schuur
-- Purpose: 
-- . DP wrapper for common_reinterleave
-- Description:
-- . The reinterleaver is an array of deinterleavers that feed an array of interleavers.          
--                   _
--      _           | \
--     / |----------| |----
-- ----| |---\   /--|_/
--     \_|--\ \ /    _
--           \ X----| \
--            X     | |----
--      _    / X----|_/
--     / |--/ / \    _ 
-- ----| |---/   \--| \
--     \_|----------| |----
--                  |_/
-- Remarks:
-- . Requirements: 
--   * All input streams must be synchronous;
--   * g_inter_block_size >= g_deint_block_size;
--    . g_deint_block_size = g_block_size out
--    . Input block length (!= g_deint_block_size) / nof_out / g_block_size_x = integer. 
--      If it is not an integer number, the input blocks cannot be evenly distributed over
--      the outputs with the chosen block_size_x.

ENTITY dp_reinterleave IS
  GENERIC (
    g_dat_w             : NATURAL; -- Width of the data. Either the width of the datafield or the re and im field combined. 
    g_nof_in            : NATURAL; -- Number of input streams
    g_deint_block_size  : NATURAL; -- Inputs streams will be de-interleaved using this block size. Should match block_size_out when g_use_ctrl=TRUE
    g_nof_out           : NATURAL; -- The number of output streams
    g_inter_block_size  : NATURAL; -- Outputs streams will be interleaved using this block size. Should match block_size_in when g_use_ctrl=TRUE
    g_use_ctrl          : BOOLEAN := TRUE;
    g_use_sync_bsn      : BOOLEAN := TRUE; -- forwards (stored) input Sync+BSN to all output streams
    g_use_complex       : BOOLEAN;
    g_align_out         : BOOLEAN := FALSE;
    g_block_size_output : NATURAL := 1 -- Output block size: The number of samles in the blocks at the output
  );
  PORT (
    rst         : IN  STD_LOGIC;
    clk         : IN  STD_LOGIC;

    snk_in_arr  : IN  t_dp_sosi_arr(g_nof_in-1 DOWNTO 0);
    src_out_arr : OUT t_dp_sosi_arr(g_nof_out-1 DOWNTO 0)
  );
END dp_reinterleave;


ARCHITECTURE wrap OF dp_reinterleave IS

  CONSTANT c_complex_w : NATURAL := g_dat_w/2;

  SIGNAL common_reinterleave_in_dat  : STD_LOGIC_VECTOR(g_nof_in*g_dat_w-1 DOWNTO 0);
  SIGNAL common_reinterleave_in_val  : STD_LOGIC;
  SIGNAL common_reinterleave_out_dat : STD_LOGIC_VECTOR(g_nof_out*g_dat_w-1 DOWNTO 0);
  SIGNAL common_reinterleave_out_val : STD_LOGIC_VECTOR(g_nof_out-1 DOWNTO 0);

  SIGNAL common_reinterleave_src_out_arr : t_dp_sosi_arr(g_nof_out-1 DOWNTO 0);
  SIGNAL dp_block_gen_src_out_arr        : t_dp_sosi_arr(g_nof_out-1 DOWNTO 0);

BEGIN

  -----------------------------------------------------------------------------
  -- Map input sosi_arr to SLV
  -----------------------------------------------------------------------------
  gen_wires_in : FOR i IN 0 TO g_nof_in-1 GENERATE 
    gen_sosi_dat_in: IF g_use_complex = FALSE GENERATE
      -----------------------------------------------------------------------------
      -- Forward SOSI data field
      -----------------------------------------------------------------------------
      common_reinterleave_in_dat( i*g_dat_w+g_dat_w -1 DOWNTO i*g_dat_w) <= snk_in_arr(i).data(g_dat_w-1 DOWNTO 0);
    END GENERATE;

    gen_sosi_compl_in: IF g_use_complex = TRUE GENERATE
      -----------------------------------------------------------------------------
      -- Forward concatenated im & re fields
      -----------------------------------------------------------------------------
      common_reinterleave_in_dat( i*g_dat_w+g_dat_w -1 DOWNTO i*g_dat_w) <= snk_in_arr(i).im(c_complex_w-1 DOWNTO 0) & snk_in_arr(i).re(c_complex_w-1 DOWNTO 0);
    END GENERATE;
  END GENERATE;

  common_reinterleave_in_val <= snk_in_arr(0).valid;

  u_common_reinterleave : ENTITY common_lib.common_reinterleave
  GENERIC MAP (
    g_nof_in           => g_nof_in,
    g_deint_block_size => g_deint_block_size,
    g_nof_out          => g_nof_out,
    g_inter_block_size => g_inter_block_size,
    g_dat_w            => g_dat_w,
    g_align_out        => g_align_out
  )
  PORT MAP (
    rst        => rst,
    clk        => clk,
    
    in_dat     => common_reinterleave_in_dat,
    in_val     => common_reinterleave_in_val,
                                             
    out_dat    => common_reinterleave_out_dat,
    out_val    => common_reinterleave_out_val
  );

  -----------------------------------------------------------------------------
  -- Map output SLV to sosi_arr
  -----------------------------------------------------------------------------
  gen_wires_out : FOR i IN 0 TO g_nof_out-1 GENERATE
    gen_sosi_dat_out: IF g_use_complex = FALSE GENERATE
      -----------------------------------------------------------------------------
      -- Forward SOSI data field
      -----------------------------------------------------------------------------
      common_reinterleave_src_out_arr(i).data(g_dat_w-1 DOWNTO 0) <= common_reinterleave_out_dat( i*g_dat_w+g_dat_w -1 DOWNTO i*g_dat_w);
    END GENERATE;
    
    gen_sosi_compl_out: IF g_use_complex = TRUE GENERATE
      -----------------------------------------------------------------------------
      -- Forward concatenated im & re fields
      -----------------------------------------------------------------------------
      common_reinterleave_src_out_arr(i).im <= RESIZE_DP_DSP_DATA(common_reinterleave_out_dat( i*g_dat_w + g_dat_w     -1 DOWNTO i*g_dat_w + c_complex_w));
      common_reinterleave_src_out_arr(i).re <= RESIZE_DP_DSP_DATA(common_reinterleave_out_dat( i*g_dat_w + c_complex_w -1 DOWNTO i*g_dat_w));
    END GENERATE;

    common_reinterleave_src_out_arr(i).valid <= common_reinterleave_out_val(i);
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Add SOP and EOP to the outputs
  -----------------------------------------------------------------------------
  gen_ctrl : IF g_use_ctrl=TRUE GENERATE
    gen_dp_block_gen : FOR i IN 0 TO g_nof_out-1 GENERATE
      u_dp_block_gen : ENTITY work.dp_block_gen
      GENERIC MAP (
        g_use_src_in       => FALSE,
        g_nof_data         => g_block_size_output,
        g_preserve_sync    => TRUE,
        g_preserve_bsn     => TRUE
      )
      PORT MAP(
        rst        => rst,
        clk        => clk,
       
        snk_in     => common_reinterleave_src_out_arr(i),
        src_out    => dp_block_gen_src_out_arr(i)
      );    
    END GENERATE;
  END GENERATE;

  no_ctrl : IF g_use_ctrl=FALSE GENERATE
    dp_block_gen_src_out_arr <= common_reinterleave_src_out_arr;
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Re-add input sync + BSN to all output streams
  -----------------------------------------------------------------------------
  align_out : IF g_use_sync_bsn = TRUE GENERATE
    gen_dp_fifo_info: FOR i IN 0 TO g_nof_out-1 GENERATE
      u_dp_fifo_info : ENTITY work.dp_fifo_info
      GENERIC MAP (
        g_use_sync => TRUE,
        g_use_bsn  => TRUE
      )
      PORT MAP (
        rst          => rst,
        clk          => clk,
    
        data_snk_in  => dp_block_gen_src_out_arr(i),  -- delayed snk_in data
        info_snk_in  => snk_in_arr(0),                -- original snk_in info
    
        src_in       => c_dp_siso_rdy,
        src_out      => src_out_arr(i)
      );
    END GENERATE;
  END GENERATE;

  no_align_out : IF g_use_sync_bsn = FALSE GENERATE
    src_out_arr <= dp_block_gen_src_out_arr;
  END GENERATE;

END wrap;
