--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;

-- Reuse from LOFAR rad_repack.vhd and rad_repack(rtl).vhd

-- Purpose:
--   Repack g_in_nof_words of width g_in_dat_w into g_out_nof_words of width
--   g_out_dat_w.
-- Description:
--   The dp_repack works both as packer and as unpacker.
--   If g_in_nof_words=g_out_nof_words then in_dat is simply passed on directly
--   to out_dat via wires.
--   The nof input words indicated by in_sof and in_eof must be a multiple of
--   g_in_nof_words, i.e. padding is not supported.
--   The in_val active duty-cycle must be <= g_in_nof_words/g_out_nof_words.

ENTITY dp_repack IS
  GENERIC (
    g_in_dat_w        : NATURAL;
    g_in_nof_words    : NATURAL;
    g_out_dat_w       : NATURAL;
    g_out_nof_words   : NATURAL;
    g_ls_to_ms        : BOOLEAN := TRUE -- Read input SLV from LS to MS word by default
  );
  PORT (
    rst              : IN  STD_LOGIC;
    clk              : IN  STD_LOGIC;
    
    in_dat           : IN  STD_LOGIC_VECTOR(g_in_dat_w-1 DOWNTO 0);
    in_val           : IN  STD_LOGIC := '1';
    in_sof           : IN  STD_LOGIC := '0';
    in_eof           : IN  STD_LOGIC := '0';
    in_sync          : IN  STD_LOGIC := '0';  -- DP style sync at sof
    
    out_dat          : OUT STD_LOGIC_VECTOR(g_out_dat_w-1 DOWNTO 0);
    out_val          : OUT STD_LOGIC;
    out_sof          : OUT STD_LOGIC;
    out_eof          : OUT STD_LOGIC;
    sof_sync         : OUT STD_LOGIC   -- DP style sync at sof, passes on in_sync
  );
END dp_repack;


ARCHITECTURE rtl OF dp_repack IS

  CONSTANT c_buf_dat_w      : NATURAL := g_in_dat_w*g_in_nof_words;

  SIGNAL in_dat_vec         : STD_LOGIC_VECTOR(c_buf_dat_w-1 DOWNTO 0);
  SIGNAL nxt_in_dat_vec     : STD_LOGIC_VECTOR(in_dat_vec'RANGE);
  SIGNAL in_val_vec         : STD_LOGIC_VECTOR(g_in_nof_words-1 DOWNTO 0);
  SIGNAL nxt_in_val_vec     : STD_LOGIC_VECTOR(in_val_vec'RANGE);  
  
  SIGNAL buf_dat_vec        : STD_LOGIC_VECTOR(c_buf_dat_w-1 DOWNTO 0);
  SIGNAL nxt_buf_dat_vec    : STD_LOGIC_VECTOR(buf_dat_vec'RANGE);
  SIGNAL buf_val            : STD_LOGIC;
  SIGNAL nxt_buf_val        : STD_LOGIC;  
  
  SIGNAL out_dat_vec        : STD_LOGIC_VECTOR(c_buf_dat_w-1 DOWNTO 0);
  SIGNAL nxt_out_dat_vec    : STD_LOGIC_VECTOR(out_dat_vec'RANGE);
  SIGNAL out_val_vec        : STD_LOGIC_VECTOR(g_out_nof_words-1 DOWNTO 0);
  SIGNAL nxt_out_val_vec    : STD_LOGIC_VECTOR(out_val_vec'RANGE);  
  SIGNAL out_eof_vec        : STD_LOGIC_VECTOR(g_out_nof_words-1 DOWNTO 0);  
  SIGNAL nxt_out_eof_vec    : STD_LOGIC_VECTOR(out_eof_vec'RANGE);  
  
  SIGNAL buf_load           : STD_LOGIC;              -- input vec to buf
  SIGNAL buf_flush          : STD_LOGIC;              -- buf to output vec
  SIGNAL buf_sof            : STD_LOGIC;
  SIGNAL nxt_buf_sof        : STD_LOGIC;
  SIGNAL buf_sync           : STD_LOGIC;
  SIGNAL nxt_buf_sync       : STD_LOGIC;
  SIGNAL buf_eof            : STD_LOGIC;
  SIGNAL nxt_buf_eof        : STD_LOGIC;
  SIGNAL nxt_out_sof        : STD_LOGIC;
  SIGNAL nxt_sof_sync       : STD_LOGIC;
  
BEGIN

  no_pack : IF g_in_nof_words=g_out_nof_words GENERATE
    out_dat <= RESIZE_UVEC(in_dat, out_dat'LENGTH);  -- any extra bits will get stripped again by dp_repack at the other end,
                                                     -- typically g_out_dat_w=g_in_dat_w
    out_val <= in_val;
    out_sof <= in_sof;
    out_eof <= in_eof;
    sof_sync <= in_sync;
  END GENERATE;

  gen_pack : IF g_in_nof_words/=g_out_nof_words GENERATE
    p_clk : PROCESS(clk, rst)
    BEGIN
      IF rst='1' THEN      
        in_dat_vec  <= (OTHERS => '0');
        in_val_vec  <= (OTHERS => '0');
        buf_dat_vec <= (OTHERS => '0');
        buf_val     <= '0';
        buf_sof     <= '0';
        buf_eof     <= '0';
        buf_sync    <= '0';
        out_dat_vec <= (OTHERS => '0');
        out_val_vec <= (OTHERS => '0');
        out_eof_vec <= (OTHERS => '0');
        out_sof     <= '0';
        sof_sync    <= '0';
      ELSIF rising_edge(clk) THEN
        in_dat_vec  <= nxt_in_dat_vec;
        in_val_vec  <= nxt_in_val_vec;
        buf_dat_vec <= nxt_buf_dat_vec;
        buf_val     <= nxt_buf_val;
        buf_sof     <= nxt_buf_sof;
        buf_eof     <= nxt_buf_eof;
        buf_sync    <= nxt_buf_sync;
        out_dat_vec <= nxt_out_dat_vec;
        out_val_vec <= nxt_out_val_vec;
        out_eof_vec <= nxt_out_eof_vec;
        out_sof     <= nxt_out_sof;
        sof_sync    <= nxt_sof_sync;
      END IF;
    END PROCESS;
    
    gen_ls_to_ms : IF g_ls_to_ms = TRUE GENERATE
      -- read out_dat_vec: LS word
      out_dat <= out_dat_vec(out_dat'RANGE);
    END GENERATE;

    gen_ms_to_ls : IF g_ls_to_ms = FALSE GENERATE
      -- read out_dat_vec: MS word
      out_dat <= out_dat_vec(out_dat_vec'HIGH DOWNTO out_dat_vec'HIGH-out_dat'LENGTH+1);
    END GENERATE;

    out_val <= out_val_vec(0);
    out_eof <= out_eof_vec(0);
    
    buf_load  <= '1' WHEN SIGNED(in_val_vec)=-1 ELSE '0';
                          -- in_val_vec=-1: input set complete, ready to be repacked
    buf_flush <= '1' WHEN (UNSIGNED(out_val_vec)=1 OR UNSIGNED(out_val_vec)=0) AND buf_val='1' ELSE '0';
                          -- out_val_vec=0: ready to repack first input set
                          -- out_val_vec=1: ready to repack next input set
                          -- buf_val: new input set available
    
    p_in: PROCESS(in_sof, buf_load, in_val, in_val_vec, in_dat, in_dat_vec)
    BEGIN
      nxt_in_val_vec <= in_val_vec;
      IF in_sof='1' OR buf_load='1' THEN
        nxt_in_val_vec <= (OTHERS => '0');
        nxt_in_val_vec(in_val_vec'HIGH) <= in_val;
      ELSIF in_val='1' THEN
        nxt_in_val_vec <= in_val & in_val_vec(in_val_vec'HIGH DOWNTO 1);
      END IF;
      
      nxt_in_dat_vec <= in_dat_vec;
      IF in_val='1' THEN
        IF g_ls_to_ms = TRUE THEN
          -- Left shift in_dat into in_dat_vec at each in_dat cycle, results in LS to MS word concatenation
          nxt_in_dat_vec <= in_dat & in_dat_vec(in_dat_vec'HIGH DOWNTO in_dat'LENGTH);
        ELSE
          -- Right shift in_dat into in_dat_vec at each in_dat cycle, results in MS to LS word concatenation
          nxt_in_dat_vec <=in_dat_vec(in_dat_vec'HIGH-in_dat'LENGTH DOWNTO 0) & in_dat;
        END IF;

      END IF;
    END PROCESS;
    
    p_buf : PROCESS(buf_load, buf_flush, in_sync, in_sof, in_eof, buf_sync, buf_sof, buf_eof, in_dat_vec, in_val_vec, buf_dat_vec, buf_val)
    BEGIN
      nxt_buf_sof <= buf_sof;                    -- buf_sof
      nxt_buf_sync <= buf_sync;                  -- buf_sync
      IF in_sof='1' THEN
        nxt_buf_sof <= '1';
        nxt_buf_sync <= in_sync;
      ELSIF buf_flush='1' THEN
        nxt_buf_sof <= '0';
        nxt_buf_sync <= '0';
      END IF;
      nxt_buf_eof <= buf_eof;                    -- buf_eof
      IF in_eof='1' THEN
        nxt_buf_eof <= '1';
      ELSIF buf_flush='1' AND buf_load='0' THEN
        nxt_buf_eof <= '0';
      END IF;
      nxt_buf_dat_vec <= buf_dat_vec;            -- buf_dat_vec, buf_val
      nxt_buf_val     <= buf_val;
      IF buf_load='1' THEN
        nxt_buf_dat_vec <= in_dat_vec;
        nxt_buf_val     <= in_val_vec(in_val_vec'HIGH);
      ELSIF buf_flush='1' THEN
        nxt_buf_dat_vec <= (OTHERS => '0');
        nxt_buf_val     <= '0';
      END IF;
    END PROCESS;
    
    p_out: PROCESS(buf_flush, buf_load, buf_sync, buf_sof, buf_eof, buf_dat_vec, out_val_vec, out_dat_vec, out_eof_vec)
    BEGIN
      nxt_out_dat_vec <= out_dat_vec;            -- out_dat, out_val, out_sof
      IF buf_flush='1' THEN
        nxt_out_dat_vec <= buf_dat_vec;
        nxt_out_val_vec <= (OTHERS => '1');
        nxt_out_sof     <= buf_sof;
        nxt_sof_sync    <= buf_sync;
      ELSE
        IF g_ls_to_ms = TRUE THEN
          -- Push SLV to the right so new word appears at LS position
          nxt_out_dat_vec <= STD_LOGIC_VECTOR(TO_UNSIGNED(0,out_dat'LENGTH)) &
                                              out_dat_vec(out_dat_vec'HIGH DOWNTO out_dat'LENGTH); 
        ELSE
          -- Push SLV to the left so new word appears at MS position
          nxt_out_dat_vec <= out_dat_vec(out_dat_vec'HIGH-out_dat'LENGTH DOWNTO 0) & 
                                         STD_LOGIC_VECTOR(TO_UNSIGNED(0,out_dat'LENGTH));
        END IF;
       
        nxt_out_val_vec <= '0' & out_val_vec(out_val_vec'HIGH DOWNTO 1);
        nxt_out_sof <= '0';
        nxt_sof_sync <= '0';
      END IF;
      nxt_out_eof_vec <= out_eof_vec;            -- out_eof
      IF buf_flush='1' AND buf_load='0' THEN
        nxt_out_eof_vec <= (OTHERS => '0');
        nxt_out_eof_vec(out_eof_vec'HIGH) <= buf_eof;
      ELSE
        nxt_out_eof_vec <= '0' & out_eof_vec(out_eof_vec'HIGH DOWNTO 1);
      END IF;
    END PROCESS;
  END GENERATE;
  
END rtl;
