--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose:
-- Description:

LIBRARY IEEE, common_lib, technology_lib, mm_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_field_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY dp_shiftram IS
  GENERIC (
    g_technology  : NATURAL := c_tech_select_default;
    g_nof_streams : NATURAL;
    g_nof_words   : NATURAL;
    g_data_w      : NATURAL;
    g_use_sync_in : BOOLEAN := TRUE
  );
  PORT (
    dp_rst      : IN  STD_LOGIC;
    dp_clk      : IN  STD_LOGIC;

    mm_rst      : IN  STD_LOGIC;
    mm_clk      : IN  STD_LOGIC;
    
    sync_in     : IN  STD_LOGIC := '0';

    reg_mosi    : IN  t_mem_mosi;
    reg_miso    : OUT t_mem_miso;

    snk_in_arr  : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);

    src_out_arr : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0)
  );
END dp_shiftram;


ARCHITECTURE str OF dp_shiftram IS

  CONSTANT c_shift_w : NATURAL := ceil_log2(g_nof_words);

  TYPE t_common_shiftram_shift_arr IS ARRAY(g_nof_streams-1 DOWNTO 0) OF STD_LOGIC_VECTOR(c_shift_w-1 DOWNTO 0);
  SIGNAL common_shiftram_shift_in_arr  : t_common_shiftram_shift_arr;

  CONSTANT c_field_arr : t_common_field_arr(0 DOWNTO 0) := (0=> ( field_name_pad("shift"), "RW", c_shift_w, field_default(0) ));

  SIGNAL reg_mosi_arr      : t_mem_mosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL reg_miso_arr      : t_mem_miso_arr(g_nof_streams-1 DOWNTO 0); 

  TYPE t_mm_fields_out_arr IS ARRAY(g_nof_streams-1 DOWNTO 0) OF STD_LOGIC_VECTOR(field_slv_out_len(c_field_arr)-1 DOWNTO 0);
  SIGNAL mm_fields_out_arr : t_mm_fields_out_arr;

BEGIN

  -----------------------------------------------------------------------------
  -- common_shiftram
  ----------------------------------------------------------------------------- 
  gen_common_shiftram : FOR i IN 0 TO g_nof_streams-1 GENERATE
    u_common_shiftram : ENTITY common_lib.common_shiftram
    GENERIC MAP (
      g_technology => g_technology,
      g_data_w     => g_data_w,
      g_nof_words  => g_nof_words
    )
    PORT MAP (
      rst            => dp_rst,
      clk            => dp_clk,
      
      data_in        => snk_in_arr(i).data(g_data_w-1 DOWNTO 0),
      data_in_val    => snk_in_arr(i).valid,
      data_in_shift  => common_shiftram_shift_in_arr(i),
   
      data_out       => src_out_arr(i).data(g_data_w-1 DOWNTO 0),
      data_out_val   => src_out_arr(i).valid,
      data_out_shift => OPEN
    );
  END GENERATE;

  -----------------------------------------------------------------------------
  -- MM control
  -----------------------------------------------------------------------------
  u_common_mem_mux : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_streams,
    g_mult_addr_w => ceil_log2(field_nof_words(c_field_arr, c_word_w))
  )
  PORT MAP (
    mosi     => reg_mosi,
    miso     => reg_miso,
    mosi_arr => reg_mosi_arr,
    miso_arr => reg_miso_arr
  );

  gen_stream : FOR i IN 0 TO g_nof_streams-1 GENERATE
      
    u_mm_fields: ENTITY mm_lib.mm_fields
    GENERIC MAP(
      g_field_arr => c_field_arr
    )
    PORT MAP (
      mm_clk  => mm_clk,
      mm_rst  => mm_rst,
  
      mm_mosi => reg_mosi_arr(i),
      mm_miso => reg_miso_arr(i),
      
      slv_clk => dp_clk,
      slv_rst => dp_rst, 
  
      slv_out => mm_fields_out_arr(i)
    );
    gen_no_sync : IF g_use_sync_in = FALSE GENERATE
      common_shiftram_shift_in_arr(i) <= mm_fields_out_arr(i)(field_hi(c_field_arr, "shift") DOWNTO field_lo(c_field_arr, "shift"));
    END GENERATE;

    gen_sync : IF g_use_sync_in = TRUE GENERATE
      p_clk : PROCESS(dp_rst, dp_clk, sync_in, mm_fields_out_arr)
      BEGIN
        IF dp_rst='1' THEN
          common_shiftram_shift_in_arr(i) <= (OTHERS => '0');
        ELSIF rising_edge(dp_clk) THEN 
          IF(sync_in = '1') THEN 
            common_shiftram_shift_in_arr(i) <= mm_fields_out_arr(i)(field_hi(c_field_arr, "shift") DOWNTO field_lo(c_field_arr, "shift"));
          END IF; 
        END IF;
      END PROCESS; 
    END GENERATE;
    
  END GENERATE;
 
END str;
