-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.all;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose:
--   Move the valid input data through a shift register to have access to
--   g_nof_words, before outputing them.
-- Description:
--   This dp_shiftreg puts the valid input data through a shift register. When
--   the shift register is full then every new valid input also causes a valid
--   output.
--   When g_flush_eop is TRUE then shift register gets flushed if it contains
--   an eop, provided that the output is ready.
--   Note that for g_nof_words = 1 the dp_shiftreg still differs from
--   dp_pipeline, because dp_pipeline outputs valid data when the
--   src_in.ready='1', while dp_shiftreg outputs valid data when it gets
--   pushed out by valid input data or by flush eop.
--   Via cur_shiftreg_inputs for all words of the current shift register is
--   available externally for monitoring. When g_modify_support is TRUE then
--   the input for all the words in de shift register can be modified via
--   new_shiftreg_inputs register. Note that cur_shiftreg_inputs(0) = snk_in.
--   Use g_output_reg=TRUE to register the src_out SOSI control signals,
--   because they have a combinatorial AND-delay if g_output_reg=FALSE. 
-- Remarks:
-- . Ready latency = 1

ENTITY dp_shiftreg IS
  GENERIC (
    g_output_reg     : BOOLEAN := FALSE;
    g_flush_eop      : BOOLEAN := TRUE;
    g_modify_support : BOOLEAN := FALSE;
    g_nof_words      : NATURAL := 1  -- >= 1, shift register length
  );
  PORT (
    rst                 : IN  STD_LOGIC;
    clk                 : IN  STD_LOGIC;
    -- ST sink
    snk_out             : OUT t_dp_siso;
    snk_in              : IN  t_dp_sosi;
    -- Control shift register contents
    cur_shiftreg_inputs : OUT t_dp_sosi_arr(0 TO g_nof_words-1);
    new_shiftreg_inputs : IN  t_dp_sosi_arr(0 TO g_nof_words-1) := (OTHERS=>c_dp_sosi_rst);
    -- ST source
    src_in              : IN  t_dp_siso;
    src_out             : OUT t_dp_sosi
  );
END dp_shiftreg;


ARCHITECTURE rtl OF dp_shiftreg IS
  
  CONSTANT c_high  : NATURAL := g_nof_words-1;  -- shift stream data from 0 to g_nof_words-1
  
  SIGNAL prev_src_in           : t_dp_siso;
  SIGNAL flush                 : STD_LOGIC;
  SIGNAL shift                 : STD_LOGIC;
  
  SIGNAL i_cur_shiftreg_inputs : t_dp_sosi_arr(0 TO c_high);
  SIGNAL shiftreg_inputs       : t_dp_sosi_arr(0 TO c_high);
  
  SIGNAL shiftreg              : t_dp_sosi_arr(0 TO c_high);
  SIGNAL nxt_shiftreg          : t_dp_sosi_arr(0 TO c_high);
  
  SIGNAL shiftreg_out          : t_dp_sosi;
  
BEGIN

  cur_shiftreg_inputs <= i_cur_shiftreg_inputs;
  
  -- Pass on frame level flow control
  snk_out.xon <= src_in.xon;
  
  -- No change in ready latency
  snk_out.ready <= src_in.ready;  -- no need to allow snk_out.ready when shift register is not full yet while src_in.ready is low
  
  ------------------------------------------------------------------------------
  -- Shift register
  ------------------------------------------------------------------------------
  
  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      prev_src_in <= c_dp_siso_rst;
      shiftreg    <= (OTHERS=>c_dp_sosi_rst);
    ELSIF rising_edge(clk) THEN
      prev_src_in <= src_in;
      shiftreg    <= nxt_shiftreg;
    END IF;
  END PROCESS;

  no_flush_eop : IF g_flush_eop=FALSE GENERATE
    flush <= '0';
  END GENERATE;
  gen_flush_eop : IF g_flush_eop=TRUE GENERATE
    p_eop : PROCESS(prev_src_in, shiftreg)
      VARIABLE v_eop : STD_LOGIC := '0';
    BEGIN
      flush <= '0';
      IF prev_src_in.ready='1' THEN  -- like valid, flush must also fit to RL=1, therefore use prev_src_in.ready 
        flush <= func_dp_stream_arr_or(shiftreg, "EOP");
      END IF;
    END PROCESS;
  END GENERATE;
  
  shift <= snk_in.valid OR flush;
  
  p_shiftreg : PROCESS(snk_in, shift, shiftreg, shiftreg_inputs)
  BEGIN
    nxt_shiftreg <= shiftreg;    
    IF shift='1' THEN
      nxt_shiftreg <= shiftreg_inputs;
    END IF;
  END PROCESS;

  ------------------------------------------------------------------------------
  -- Shift register contents control
  ------------------------------------------------------------------------------
  
  -- Make shift register available for external monitoring
  p_monitor : PROCESS(snk_in, shiftreg)
  BEGIN
    i_cur_shiftreg_inputs(0)           <= snk_in;
    --i_cur_shiftreg_inputs(1 TO c_high) <= shiftreg(0 TO c_high-1);  -- causes "Fatal: (SIGSEGV) Bad handle or reference" in Modelsim, therefore use LOOP
    FOR I IN 1 TO c_high LOOP
      i_cur_shiftreg_inputs(I) <= shiftreg(I-1);
    END LOOP;
  END PROCESS;

  -- Support for shift register modification
  no_modification : IF g_modify_support=FALSE GENERATE
    shiftreg_inputs <= i_cur_shiftreg_inputs;
  END GENERATE;
  
  gen_modification : IF g_modify_support=TRUE GENERATE
    shiftreg_inputs <= new_shiftreg_inputs;
  END GENERATE;
  
  ------------------------------------------------------------------------------
  -- Output control
  ------------------------------------------------------------------------------
  
  p_shiftreg_out : PROCESS(shiftreg, shift)
  BEGIN
    shiftreg_out       <= shiftreg(c_high);
    -- The shiftreg contents is valid, but the shiftreg output is only valid when it is shifted out,
    -- therefore AND the SOSI control with shift
    shiftreg_out.sync  <= shiftreg(c_high).sync  AND shift;
    shiftreg_out.valid <= shiftreg(c_high).valid AND shift;
    shiftreg_out.sop   <= shiftreg(c_high).sop   AND shift;
    shiftreg_out.eop   <= shiftreg(c_high).eop   AND shift;
  END PROCESS;
  
  no_output_reg : IF g_output_reg=FALSE GENERATE
    src_out <= shiftreg_out;
  END GENERATE;
  
  gen_output_reg : IF g_output_reg=TRUE GENERATE
    u_output : ENTITY work.dp_pipeline
    PORT MAP (
      rst         => rst,
      clk         => clk,
      -- ST sink
      snk_out     => OPEN,
      snk_in      => shiftreg_out,
      -- ST source
      src_in      => src_in,
      src_out     => src_out
    );
  END GENERATE;
  
END rtl;
