--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose:
--   Split one frame into two frames.
-- Description:
--   Split an input frame into a head and a tail output frame, so the reverse
--   of dp_concat. The g_nof_symbols determines how many symbols are used for
--   the head output frame via source port 0. The rest of the symbols are
--   passed on to tail output frame via source port 1.
-- Remark:
-- . Both frames will have the same data width and symbol width.
-- . The split is done at symbol level and the output frame will have the
--   aggregated empty.
-- . The split head frame gets the current channel and error field value. The
--   error field may be undefined because the true error information is not
--   known yet as it is only available at the snk_in.eop. The split tail frame
--   gets channel and error field from the input frame.
-- . If the dynamic nof_symbols input is used, then g_nof_symbols must be set
--   to fit the largest head frame and input nof_symbols. Otherwise if
--   nof_symbols is fixed, then g_nof_symbols can be set to that value and the
--   dynamic input can be left not connected.
-- . If nof_symbols=0, then the input frame is passed on entirely to the tail
--   frame output. If nof_symbols > nof input symbols then the input frame
--   is passed on entirely to the head frame output.
-- . Allow an idle cycle between head and tail frame if the split is not at a 
--   word boundary. Otherwise the statemachine would become more complicated
--   and require the use of prev_src_in to check that the src_in_arr(c_tail) was
--   ready.
-- . Allow an idle cycle between tail and next head frame to ease the state
--   machine. Hence the input data must have at least one cycle gap between
--   input frames. The state machine is complicated due to that dp_split
--   support dynamic MM control of nof_symbols over a range of 0 to max.
-- . nof_symbols_out is valid at the SOP of the head frame and gives the nof
--   symbols in this head frame. Is based on nof_symbols. Therefore
--   nof_symbols_out is only correct if the input frame has >= nof_symbols.
--
-- Design steps that were taken for dp_split (took about 3 days):
--   1 aim is a component that does the opposite of dp_concat
--   2 same steps were taken as for dp_concat
--   3 verify with tb_dp_split:
--     . always ready and input available
--     . nof_symbols = 0, 1, 2, ..., 10
--     . nof_symbols_per_data = 1, 2, 3, 4
--     . random ready
--     . random input availble
--   4 combinatorial zero-delay loops and truncation warnings wer debugged by
--     restart -f, then run x ns to just before the when the error occurs and
--     then single step until the error occurs.


ENTITY dp_split IS
  GENERIC (
    g_data_w            : NATURAL := 16;   -- >= 1
    g_symbol_w          : NATURAL := 8;    -- >= 1, and g_data_w/g_symbol_w must be an integer
    g_nof_symbols       : NATURAL := 8;    -- nof symbols for the head frame, rest will go to the tail frame
    g_use_head_tail_src : BOOLEAN := FALSE -- FALSE: Use src_out_arr, src_in_arr. TRUE: use separate head_src_*, tail_src_*.
  );
  PORT (
    rst             : IN  STD_LOGIC;
    clk             : IN  STD_LOGIC;
    -- Static control input (connect via MM or leave open to use default)
    nof_symbols     : IN  STD_LOGIC_VECTOR(ceil_log2(g_nof_symbols+1)-1 DOWNTO 0) := TO_UVEC(g_nof_symbols, ceil_log2(g_nof_symbols+1));  -- +1 to also support power of 2 values for g_nof_symbols
    nof_symbols_out : OUT STD_LOGIC_VECTOR(ceil_log2(g_nof_symbols+1)-1 DOWNTO 0);

    snk_out         : OUT t_dp_siso;
    snk_in          : IN  t_dp_sosi;

    src_in_arr      : IN  t_dp_siso_arr(0 TO 1) := (OTHERS=>c_dp_siso_rdy);
    src_out_arr     : OUT t_dp_sosi_arr(0 TO 1);

    head_src_in     : IN  t_dp_siso := c_dp_siso_rdy; -- We don't need to create an arrayed TYPE for src_*_arr when using this output, making things easier.
    head_src_out    : OUT t_dp_sosi;

    tail_src_in     : IN  t_dp_siso := c_dp_siso_rdy;
    tail_src_out    : OUT t_dp_sosi
  );
END dp_split;


ARCHITECTURE rtl OF dp_split IS

  CONSTANT c_nof_symbols_per_data  : NATURAL := g_data_w / g_symbol_w;  -- nof symbols per data
  CONSTANT c_empty_w               : NATURAL := ceil_log2(c_nof_symbols_per_data);
  
  CONSTANT c_nof_symbols_w         : NATURAL := nof_symbols'LENGTH;
  CONSTANT c_symbol_cnt_w          : NATURAL := ceil_log2(g_nof_symbols+1+c_nof_symbols_per_data);  -- total symbols count with c_nof_symbols_per_data increments, +1 to also support power of 2 value
  
  CONSTANT c_head                  : NATURAL := 0;  -- head frame snk_in port number
  CONSTANT c_tail                  : NATURAL := 1;  -- tail frame snk_in port number
  
  TYPE t_state IS (s_head, s_tail, s_eop);
  
  SIGNAL state               : t_state;
  SIGNAL nxt_state           : t_state;
  
  SIGNAL i_snk_out           : t_dp_siso;
  SIGNAL hold_src_in         : t_dp_siso;
  SIGNAL next_src_buf        : t_dp_sosi;
  SIGNAL src_buf             : t_dp_sosi;
  SIGNAL nxt_src_buf         : t_dp_sosi;
  SIGNAL tail                : t_dp_sosi;
  SIGNAL nxt_tail            : t_dp_sosi;
  SIGNAL i_src_out_arr       : t_dp_sosi_arr(0 TO 1);
  SIGNAL nxt_src_out_arr     : t_dp_sosi_arr(0 TO 1);
  SIGNAL i_src_in_arr        : t_dp_siso_arr(0 TO 1); 

  SIGNAL busy                : STD_LOGIC;
  SIGNAL nxt_busy            : STD_LOGIC;
  SIGNAL sync                : STD_LOGIC;
  SIGNAL nxt_sync            : STD_LOGIC;
  SIGNAL nof_symbols_reg     : STD_LOGIC_VECTOR(c_nof_symbols_w-1 DOWNTO 0);
  SIGNAL nxt_nof_symbols_reg : STD_LOGIC_VECTOR(c_nof_symbols_w-1 DOWNTO 0);
  SIGNAL symbol_cnt          : STD_LOGIC_VECTOR(c_symbol_cnt_w-1  DOWNTO 0);
  SIGNAL nxt_symbol_cnt      : STD_LOGIC_VECTOR(c_symbol_cnt_w-1  DOWNTO 0);
  
  SIGNAL i_nof_symbols_out   : STD_LOGIC_VECTOR(c_nof_symbols_w-1 DOWNTO 0);
  SIGNAL nxt_nof_symbols_out : STD_LOGIC_VECTOR(c_nof_symbols_w-1 DOWNTO 0);

  SIGNAL head_empty_reg      : STD_LOGIC_VECTOR(c_empty_w-1 DOWNTO 0);
  SIGNAL nxt_head_empty_reg  : STD_LOGIC_VECTOR(c_empty_w-1 DOWNTO 0);
  SIGNAL input_empty_reg     : STD_LOGIC_VECTOR(c_empty_w-1 DOWNTO 0);
  SIGNAL nxt_input_empty_reg : STD_LOGIC_VECTOR(c_empty_w-1 DOWNTO 0);
  SIGNAL nof_remain_reg      : STD_LOGIC_VECTOR(c_empty_w-1 DOWNTO 0);
  SIGNAL nxt_nof_remain_reg  : STD_LOGIC_VECTOR(c_empty_w-1 DOWNTO 0);
  
BEGIN

  ASSERT (g_data_w MOD g_symbol_w) = 0 REPORT "dp_split : g_data_w/g_symbol_w must be an integer" SEVERITY FAILURE;
  
  snk_out <= i_snk_out;
  src_out_arr <= i_src_out_arr;

  head_src_out <= i_src_out_arr(c_head);
  tail_src_out <= i_src_out_arr(c_tail);
  i_src_in_arr <= src_in_arr WHEN g_use_head_tail_src=FALSE ELSE (head_src_in, tail_src_in);

  nof_symbols_out <= i_nof_symbols_out;
  
  p_clk: PROCESS(clk, rst)
  BEGIN
    IF rst='1' THEN
      busy              <= '0';
      sync              <= '0';
      nof_symbols_reg   <= (OTHERS=>'0');
      state             <= s_head;
      input_empty_reg   <= (OTHERS=>'0');
      head_empty_reg    <= (OTHERS=>'0');
      nof_remain_reg    <= (OTHERS=>'0');
      symbol_cnt        <= (OTHERS=>'0');
      tail              <= c_dp_sosi_rst;
      src_buf           <= c_dp_sosi_rst;
      i_src_out_arr     <= (OTHERS=>c_dp_sosi_rst);
      i_nof_symbols_out <= (OTHERS=>'0');
    ELSIF rising_edge(clk) THEN
      busy              <= nxt_busy;
      sync              <= nxt_sync;
      nof_symbols_reg   <= nxt_nof_symbols_reg;
      state             <= nxt_state;
      input_empty_reg   <= nxt_input_empty_reg;
      head_empty_reg    <= nxt_head_empty_reg;
      nof_remain_reg    <= nxt_nof_remain_reg;
      symbol_cnt        <= nxt_symbol_cnt;
      tail              <= nxt_tail;
      src_buf           <= nxt_src_buf;
      i_src_out_arr     <= nxt_src_out_arr;
      i_nof_symbols_out <= nxt_nof_symbols_out;
    END IF;
  END PROCESS;
  
  -- Hold the sink input to be able to register the source output
  u_hold : ENTITY work.dp_hold_input
  PORT MAP (
    rst          => rst,
    clk          => clk,
    -- ST sink
    snk_out      => OPEN,
    snk_in       => snk_in,
    -- ST source
    src_in       => hold_src_in,
    next_src_out => next_src_buf,
    pend_src_out => OPEN,
    src_out_reg  => src_buf
  );
  
  -- Hold input register
  nxt_src_buf <= next_src_buf;
  
  -- Only accept new nof_symbols setting when no frame is busy being output
  nxt_nof_symbols_reg <= nof_symbols_reg WHEN next_src_buf.valid='1' OR busy='1' ELSE nof_symbols;
  
  p_state : PROCESS(state, busy, sync, nof_symbols_reg, i_nof_symbols_out, symbol_cnt, i_src_out_arr, src_in_arr, next_src_buf, tail, input_empty_reg, head_empty_reg, nof_remain_reg)
    VARIABLE v_input_empty     : STD_LOGIC_VECTOR(c_empty_w-1 DOWNTO 0);
    VARIABLE v_head_empty      : STD_LOGIC_VECTOR(c_empty_w-1 DOWNTO 0);
    VARIABLE v_symbol_cnt      : STD_LOGIC_VECTOR(c_symbol_cnt_w-1 DOWNTO 0);  -- total symbols count with c_nof_symbols_per_data increments
    VARIABLE v_head_symbol_cnt : STD_LOGIC_VECTOR(c_symbol_cnt_w-1 DOWNTO 0);  -- head symbols count
    VARIABLE v_cnt             : STD_LOGIC_VECTOR(c_symbol_cnt_w-1 DOWNTO 0);
    VARIABLE v_nof_from_this   : NATURAL RANGE 0 TO c_nof_symbols_per_data;
  BEGIN

    -- SISO default  
    i_snk_out   <= c_dp_siso_rst;                    -- no input request
    hold_src_in <= c_dp_siso_rst;                    -- no input request
    
    -- SOSI defaults
    nxt_src_out_arr(c_head)       <= i_src_out_arr(c_head);  -- default no output
    nxt_src_out_arr(c_head).sync  <= '0';
    nxt_src_out_arr(c_head).valid <= '0';
    nxt_src_out_arr(c_head).sop   <= '0';
    nxt_src_out_arr(c_head).eop   <= '0';
    
    nxt_src_out_arr(c_tail)       <= i_src_out_arr(c_tail);  -- default no output
    nxt_src_out_arr(c_tail).sync  <= '0';
    nxt_src_out_arr(c_tail).valid <= '0';
    nxt_src_out_arr(c_tail).sop   <= '0';
    nxt_src_out_arr(c_tail).eop   <= '0';

    nxt_nof_symbols_out <= i_nof_symbols_out;
      
    -- detect frame busy
    nxt_busy <= busy;
    IF next_src_buf.eop='1' THEN
      nxt_busy <= '0';
    ELSIF next_src_buf.valid='1' THEN
      nxt_busy <= '1';
    END IF;
    
    -- preserve sync at sop also for tail, so head get sync directly from snk_in. tail gets the preserved sync
    -- no need to preserve other sosi control fields channel and bsn at sop, because they keep their value during snk_in frame
    nxt_sync <= sync;
    IF next_src_buf.sop='1' THEN
      nxt_sync <= next_src_buf.sync;
    END IF;
    
    -- update head symbols counter
    v_symbol_cnt := symbol_cnt;
    IF next_src_buf.sop='1' THEN
      v_symbol_cnt := TO_UVEC(c_nof_symbols_per_data, c_symbol_cnt_w);
      nxt_nof_symbols_out <= nof_symbols_reg;          -- Report the current nof symbols as status at sop of head frame
    ELSIF next_src_buf.valid='1' THEN
      v_symbol_cnt := INCR_UVEC(v_symbol_cnt, c_nof_symbols_per_data);
    END IF;
    
    v_input_empty := (OTHERS=>'0');
    v_head_symbol_cnt := v_symbol_cnt;
    IF next_src_buf.eop='1' THEN
      v_input_empty     := next_src_buf.empty(c_empty_w-1 DOWNTO 0);
      v_head_symbol_cnt := TO_UVEC(TO_UINT(v_symbol_cnt) - TO_UINT(v_input_empty), c_symbol_cnt_w);
    END IF;
    
    nxt_symbol_cnt <= (OTHERS=>'0');  -- default reset symbol count

    -- preserve empty information for next state
    nxt_input_empty_reg <= input_empty_reg;
    nxt_head_empty_reg  <= head_empty_reg;
    nxt_nof_remain_reg  <= nof_remain_reg;
    
    -- preserve tail data information for next state
    nxt_tail <= tail;  -- keep the tail data part in case the split is at symbol boundary and not at word boundary,
                       -- keep the sop of the tail output, the valid and eop of tail are not used.
    
    -- pass on output
    nxt_state <= state;
    
    CASE state IS
      WHEN s_head =>
        nxt_input_empty_reg <= (OTHERS=>'0');  -- default assume tail will           end at word boundary
        nxt_head_empty_reg  <= (OTHERS=>'0');  -- default assume tail will start         at word boundary
        nxt_nof_remain_reg  <= (OTHERS=>'0');  -- default assume tail will start and end at word boundary
        nxt_tail <= c_dp_sosi_rst;
        nxt_symbol_cnt <= symbol_cnt;          -- default keep symbol count
        
        -- Default ready for input to head or tail dependent on nof_symbols_reg
        IF TO_UINT(nof_symbols_reg)/=0 THEN
          i_snk_out   <= src_in_arr(c_head);  -- default input request for head output when head src_in_arr is ready 
          hold_src_in <= src_in_arr(c_head);  -- hold_src_in will flush the next_src_buf output when i_snk_out is changed to tail src_in_arr
        ELSE
          i_snk_out   <= src_in_arr(c_tail);  -- else in special of tail only split then input request for tail output when tail src_in_arr is ready 
          hold_src_in <= src_in_arr(c_tail);  -- hold_src_in will flush the next_src_buf output when i_snk_out is changed to tail src_in_arr
        END IF;
        
        -- Act when there is valid input ready for output
        IF next_src_buf.valid='1' THEN
          IF TO_UINT(nof_symbols_reg)=0 THEN                -- special case with no output head, so tail src_in_arr ready is already being used
            -- no head output, begin tail output
            nxt_symbol_cnt <= (OTHERS=>'0');                -- reset symbol count
            nxt_src_out_arr(c_tail) <= next_src_buf;        -- output tail start data immediately
            IF next_src_buf.eop='0' THEN
              nxt_state <= s_tail;                          -- there will be more output tail data words
            END IF;                                         -- else output tail with one data word is done, remain in state s_head
          ELSE
            -- begin or continue output head
            IF TO_UINT(nof_symbols_reg)>TO_UINT(v_symbol_cnt) THEN
              nxt_src_out_arr(c_head) <= next_src_buf;      -- there will be more output head data words
              nxt_symbol_cnt <= v_symbol_cnt;               -- keep symbol count for state s_head
            ELSIF TO_UINT(nof_symbols_reg)=TO_UINT(v_symbol_cnt) THEN
              -- split at word boundary
              nxt_src_out_arr(c_head)     <= next_src_buf;  -- output last head
              nxt_src_out_arr(c_head).eop <= '1';
              IF next_src_buf.eop='0' THEN
                i_snk_out <= src_in_arr(c_tail);            -- input request for tail output start, now depends on tail src_in_arr ready
                nxt_tail.sop <= '1';                        -- keep the sop via tail, the first tail data itself will arrive with the next_src_buf 
                nxt_state <= s_tail;
              END IF;                                       -- else there will be no output tail, remain in state s_head
            ELSIF TO_UINT(nof_symbols_reg)>=TO_UINT(v_head_symbol_cnt) THEN
              nxt_src_out_arr(c_head)     <= next_src_buf;  -- output last head, there are no symbols left for the tail
              nxt_symbol_cnt <= (OTHERS=>'0');              -- reset symbol count and remain in s_head
            ELSE
              -- split at symbol boundary
              v_cnt        := TO_UVEC(TO_UINT(v_symbol_cnt) - TO_UINT(nof_symbols_reg), c_symbol_cnt_w);  -- use v_cnt to avoid truncation warning
              v_head_empty := v_cnt(c_empty_w-1 DOWNTO 0);
              nxt_src_out_arr(c_head)       <= next_src_buf;-- output last head
              nxt_src_out_arr(c_head).empty <= RESIZE_DP_EMPTY(v_head_empty);
              nxt_src_out_arr(c_head).eop   <= '1';
              nxt_head_empty_reg <= v_head_empty;           -- keep tail start symbol boundary
              nxt_tail           <= next_src_buf;           -- keep tail start data part
              nxt_tail.sop       <= '1';
              IF next_src_buf.eop='0' THEN
                i_snk_out <= src_in_arr(c_tail);            -- input request for tail output start, now depends on tail src_in_arr ready
                nxt_state <= s_tail;                        -- there will be more output tail data words
              ELSE                                          -- need one tail word to output the tail part
                i_snk_out <= c_dp_siso_rst;                 -- no input request for at least one clock cycle, to allow change in nof_symbols_reg and for state s_eop
                nxt_tail.sync <= sync;                      -- force the output tail sync for the tail start
                nxt_input_empty_reg <= v_input_empty;
                nxt_nof_remain_reg  <= TO_UVEC(TO_UINT(v_head_empty) - TO_UINT(v_input_empty), c_empty_w);
                nxt_state <= s_eop;
              END IF;
            END IF;
          END IF;
        END IF;
        
      WHEN s_tail =>
        -- Default ready for input to tail
        i_snk_out <= src_in_arr(c_tail);                    -- input request for tail output
        hold_src_in <= src_in_arr(c_tail);                  -- input request for tail output
        
        -- Act when there is valid input ready for output
        IF next_src_buf.valid='1' THEN
          nxt_tail                <= next_src_buf;          -- keep tail data part, needed in case head had empty symbols
          v_nof_from_this := c_nof_symbols_per_data-TO_UINT(head_empty_reg);  -- default output is next_src_buf in case head had no empty symbols
          nxt_src_out_arr(c_tail) <= func_dp_data_shift(tail, next_src_buf, g_symbol_w, c_nof_symbols_per_data, v_nof_from_this);
          nxt_src_out_arr(c_tail).sop  <= tail.sop;            -- force the output tail sop for the tail start
          nxt_src_out_arr(c_tail).sync <= tail.sop AND sync;   -- force the output tail sync for the tail start
          IF next_src_buf.eop='1' THEN
            i_snk_out <= c_dp_siso_rst;                     -- no input request for at least one clock cycle, to allow change in nof_symbols_reg and/or for state s_eop
            IF TO_UINT(v_input_empty)>=TO_UINT(head_empty_reg) THEN
                                                            -- this is the last tail output, the input eop marks the tail output eop
              nxt_src_out_arr(c_tail).empty <= RESIZE_DP_EMPTY(func_dp_empty_split(v_input_empty, head_empty_reg, c_nof_symbols_per_data));
              nxt_state <= s_head;
            ELSE                                            -- need one more tail word to output the last tail part
              nxt_src_out_arr(c_tail).eop <= '0';           -- the last will be flushed from tail in state s_eop
              nxt_input_empty_reg <= v_input_empty;
              nxt_nof_remain_reg <= TO_UVEC(TO_UINT(head_empty_reg) - TO_UINT(v_input_empty), c_empty_w);
              nxt_state <= s_eop;
            END IF;
          END IF;
        END IF;
        
      WHEN OTHERS =>  -- s_eop                              -- no input request until last tail word has been output
        -- Act when tail output port is ready
        IF src_in_arr(c_tail).ready = '1' THEN              -- finish the last tail part
          nxt_state <= s_head;                              -- no input request yet, allow one more idle cycle and input request again in state s_head
          -- shift tail output because tail will have less empty symbols than the head had
          nxt_src_out_arr(c_tail) <= func_dp_data_shift_last(tail, g_symbol_w, c_nof_symbols_per_data, TO_UINT(nof_remain_reg), TO_UINT(input_empty_reg));
          nxt_src_out_arr(c_tail).empty <= RESIZE_DP_EMPTY(func_dp_empty_split(input_empty_reg, head_empty_reg, c_nof_symbols_per_data));
        END IF;
    END CASE;  -- CASE state
    
    -- Pass on frame level flow control
    i_snk_out.xon   <= func_dp_stream_arr_and(src_in_arr, "XON");
    hold_src_in.xon <= func_dp_stream_arr_and(src_in_arr, "XON");
  END PROCESS;
  
END rtl;
