--------------------------------------------------------------------------------
--
-- Copyright (C) 2014
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE,common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose:
-- . Control a source's output rate by toggling its ready input.
-- Description:
-- . The following diagram shows how src_out depends on the generics (only      
--   two src_out samples shown):                                                
--                                            |<--g_block_period-->|   
--  src_out) |<-----g_init_valid_delay----->[0]                             [1] 
--           ^
--           |
--           init_valid_delay_ref='1'

ENTITY dp_src_out_timer IS
  GENERIC (
    g_init_valid_delay       : NATURAL := 0;
    g_block_period  : NATURAL := 1; -- Block period >1
    g_block_len              : NATURAL := 1 -- Nof words per valid block
  ); 
  PORT (
    rst                  : IN  STD_LOGIC;
    clk                  : IN  STD_LOGIC;

    init_valid_delay_ref : IN STD_LOGIC := '1';

    snk_in               : IN  t_dp_sosi;
    snk_out              : OUT t_dp_siso --rl=1
  );
END dp_src_out_timer;

ARCHITECTURE rtl OF dp_src_out_timer IS

  TYPE t_state IS (s_init, s_wait, s_delay, s_ready, s_not_ready);
  SIGNAL state             : t_state;
  SIGNAL nxt_state         : t_state;

  CONSTANT c_max_count     : NATURAL := largest(largest(1, g_init_valid_delay), g_block_period);

  SIGNAL cycle_cnt         : STD_LOGIC_VECTOR(ceil_log2(c_max_count)-1 DOWNTO 0);
  SIGNAL nxt_cycle_cnt     : STD_LOGIC_VECTOR(ceil_log2(c_max_count)-1 DOWNTO 0);

  SIGNAL prv_snk_out_ready : STD_LOGIC;
  SIGNAL snk_out_ready     : STD_LOGIC;
  SIGNAL nxt_snk_out_ready : STD_LOGIC;

BEGIN

  snk_out.ready <= snk_out_ready;
  snk_out.xon   <= '1';

  -----------------------------------------------------------------------------
  -- FSM that controls ready signal
  -----------------------------------------------------------------------------
  p_state : PROCESS(state, init_valid_delay_ref, cycle_cnt, snk_out_ready, prv_snk_out_ready, snk_in.valid)
  BEGIN

    nxt_state         <= state;
    nxt_cycle_cnt     <= cycle_cnt;
    nxt_snk_out_ready <= snk_out_ready;

    CASE state IS

      WHEN s_wait =>
        -- Wait for our delay reference input to go high
        IF init_valid_delay_ref='1' THEN
          nxt_state <= s_delay;
        END IF;

      WHEN s_delay =>
        -- Delay for X cycles after reference input went high
        nxt_cycle_cnt <= INCR_UVEC(cycle_cnt, 1);
        IF TO_UINT(cycle_cnt) >= g_init_valid_delay THEN
          nxt_state     <= s_ready;
          nxt_cycle_cnt <= (OTHERS=>'0');
          nxt_snk_out_ready <= '1';
        END IF;

      WHEN s_ready =>
        nxt_snk_out_ready <= '1';

        -- Clock out g_block_len words
        IF g_block_period>1 THEN
          nxt_cycle_cnt <= INCR_UVEC(cycle_cnt, 1);
          IF TO_UINT(cycle_cnt)>=g_block_len-1 THEN  
            nxt_state         <= s_not_ready;
            nxt_snk_out_ready <= '0';
          END IF;  
        END IF;
        IF prv_snk_out_ready/=snk_in.valid THEN
          -- We're no longer provided with valid data to time, so we lost timing. Reset.
          nxt_state <= s_init;
        END IF;

      WHEN s_not_ready =>
        nxt_snk_out_ready <= '0';
        -- De-assert ready to get some invalid cycles if required
        nxt_cycle_cnt <= INCR_UVEC(cycle_cnt, 1);
        IF TO_UINT(cycle_cnt)>=g_block_period-1 THEN
          nxt_state         <= s_ready;
          nxt_snk_out_ready <= '1';
          nxt_cycle_cnt     <= (OTHERS=>'0');
        END IF;
        IF prv_snk_out_ready/=snk_in.valid THEN
          -- We're no longer provided with valid data to time, so we lost timing. Reset.
          nxt_state <= s_init;
        END IF;

      WHEN OTHERS =>  -- s_init
        nxt_state         <= s_wait;
        nxt_snk_out_ready <= '0';
        nxt_cycle_cnt     <= (OTHERS=>'0');

    END CASE;

  END PROCESS;

  -----------------------------------------------------------------------------
  -- Registers
  -----------------------------------------------------------------------------
  p_clk: PROCESS(clk, rst)
  BEGIN
    IF rst='1' THEN
      state             <= s_init;
      snk_out_ready     <= '0';
      prv_snk_out_ready <= '0';
      cycle_cnt         <= (OTHERS=>'0');
    ELSIF rising_edge(clk) THEN
      state             <= nxt_state;
      snk_out_ready     <= nxt_snk_out_ready;
      prv_snk_out_ready <= snk_out_ready; 
      cycle_cnt         <= nxt_cycle_cnt;
    END IF;
  END PROCESS;

END rtl;
