-------------------------------------------------------------------------------
--
-- Copyright (C) 2016
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Author:
-- . Daniel van der Schuur
-- Purpose:
-- . Switch one of g_nof_inputs input streams to the output.
-- Description:
-- . Respects frame boundaries (SOP..EOP).
-- . This is basically an MM-controlled dp_mux (with g_mode=4).
-- . There is always one input enabled. If undesired, add your own input and 
--   assign c_dp_sosi_rst to it. Switching to that input will disable the output.
-- Remark:
-- . dp_mux requires a minimum invaled gap of 1 cycle to switch to a new input!
--   . So this does not work for continuous streams!

LIBRARY IEEE, common_lib, mm_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_field_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY dp_switch IS
  GENERIC (
    g_nof_inputs      : NATURAL;          -- Number of inputs
    g_default_enabled : NATURAL := 0;     -- This input number 0..g_nof_inputs-1 will be enabled by default
    g_use_fifo        : BOOLEAN := FALSE; -- This and the generics below are forwarded to dp_mux only
    g_bsn_w           : NATURAL := 16;
    g_data_w          : NATURAL := 16;
    g_empty_w         : NATURAL := 1;
    g_in_channel_w    : NATURAL := 1;
    g_error_w         : NATURAL := 1;
    g_use_bsn         : BOOLEAN := FALSE;
    g_use_empty       : BOOLEAN := FALSE;
    g_use_in_channel  : BOOLEAN := FALSE;
    g_use_error       : BOOLEAN := FALSE;
    g_use_sync        : BOOLEAN := FALSE;
    g_fifo_af_margin  : NATURAL := 4;  -- Nof words below max (full) at which fifo is considered almost full
    g_fifo_size       : NATURAL := 1024;
    g_fifo_fill       : NATURAL := 0
  );
  PORT (
    dp_clk      : IN  STD_LOGIC;
    dp_rst      : IN  STD_LOGIC;

    mm_clk      : IN  STD_LOGIC;
    mm_rst      : IN  STD_LOGIC;

    snk_in_arr  : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
    snk_out_arr : OUT t_dp_siso_arr(g_nof_inputs-1 DOWNTO 0);

    src_out     : OUT t_dp_sosi;
    src_in      : IN  t_dp_siso := c_dp_siso_rdy;

    reg_mosi    : IN  t_mem_mosi;
    reg_miso    : OUT t_mem_miso
  );
END dp_switch;


ARCHITECTURE str OF dp_switch IS

  CONSTANT c_field_arr : t_common_field_arr(0 DOWNTO 0) := (0=> ( field_name_pad("input_select"), "RW", ceil_log2(g_nof_inputs), field_default(g_default_enabled) ));

  SIGNAL mm_fields_out       : STD_LOGIC_VECTOR(field_slv_out_len(c_field_arr)-1 DOWNTO 0);
  SIGNAL dp_mux_sel_ctrl_req : NATURAL;
  SIGNAL dp_mux_sel_ctrl     : NATURAL;

  SIGNAL xonoff_src_out_arr  : t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  SIGNAL xonoff_src_in_arr   : t_dp_siso_arr(g_nof_inputs-1 DOWNTO 0);

  SIGNAL inverted_snk_in_arr : t_dp_sosi_arr(0 TO g_nof_inputs-1);
  SIGNAL inverted_snk_out_arr: t_dp_siso_arr(0 TO g_nof_inputs-1);
  
BEGIN

  ------------------------------------------------------------------------------
  -- A single MM register contains input to select
  ------------------------------------------------------------------------------
  u_mm_fields: ENTITY mm_lib.mm_fields
  GENERIC MAP(
    g_field_arr => c_field_arr
  )
  PORT MAP (
    mm_clk  => mm_clk,
    mm_rst  => mm_rst,

    mm_mosi => reg_mosi,
    mm_miso => reg_miso,
    
    slv_clk => dp_clk,
    slv_rst => dp_rst, 

    slv_out => mm_fields_out
  );
  
  ------------------------------------------------------------------------------
  -- put dp_xonoff block inbetween data path to control data flow. 
  ------------------------------------------------------------------------------
  gen_dp_xonoff_arr : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    u_dp_xonoff: ENTITY work.dp_xonoff
    PORT MAP (
      clk         => dp_clk,
      rst         => dp_rst,
      -- Frame in
      in_sosi     => snk_in_arr(i),
      in_siso     => snk_out_arr(i),
      -- Frame out
      out_siso    => xonoff_src_in_arr(i),     -- flush control done by dp_mux.snk_out_arr
      out_sosi    => xonoff_src_out_arr(i) 
    );
  END GENERATE;

  ------------------------------------------------------------------------------
  -- invert snk_in_arr, dp_mux uses a TO array. 
  ------------------------------------------------------------------------------
  gen_dp_switch_snk_in_arr : FOR i IN 0 TO g_nof_inputs-1 GENERATE
    inverted_snk_in_arr(i) <= xonoff_src_out_arr(i);
    xonoff_src_in_arr(i) <= inverted_snk_out_arr(i);
  END GENERATE;

  ------------------------------------------------------------------------------
  -- Don't allow user to select a non-existent input
  -- . If user requests non-existent input, the default input is forwarded instead.
  ------------------------------------------------------------------------------
  dp_mux_sel_ctrl_req <= TO_UINT(mm_fields_out(field_hi(c_field_arr, "input_select") DOWNTO field_lo(c_field_arr, "input_select")));
  dp_mux_sel_ctrl <= dp_mux_sel_ctrl_req WHEN dp_mux_sel_ctrl_req<g_nof_inputs ELSE g_default_enabled;
    
  ------------------------------------------------------------------------------
  -- DP mux forwards input based on dp_mux_sel_ctrl
  ------------------------------------------------------------------------------
  u_dp_mux : ENTITY work.dp_mux
  GENERIC MAP (
    g_mode            => 4, -- Use sel_ctrl
    g_sel_ctrl_invert => FALSE, -- Dont invert, data is already reverse mapped from DOWNTO >> TO arrays.
    g_nof_input       => g_nof_inputs,
    g_use_fifo        => g_use_fifo,  
    g_bsn_w           => g_bsn_w,  
    g_data_w          => g_data_w,  
    g_empty_w         => g_empty_w,  
    g_in_channel_w    => g_in_channel_w,   
    g_error_w         => g_error_w,  
    g_use_bsn         => g_use_bsn,  
    g_use_empty       => g_use_empty,  
    g_use_in_channel  => g_use_in_channel,  
    g_use_error       => g_use_error,  
    g_use_sync        => g_use_sync,  
    g_fifo_af_margin  => g_fifo_af_margin,  
    g_fifo_size       => array_init(g_fifo_size, g_nof_inputs),
    g_fifo_fill       => array_init(g_fifo_fill, g_nof_inputs)
  )
  PORT MAP (
    clk         => dp_clk,
    rst         => dp_rst,

    sel_ctrl    => dp_mux_sel_ctrl,

    snk_in_arr  => inverted_snk_in_arr,
    snk_out_arr => inverted_snk_out_arr,

    src_out     => src_out,
    src_in      => src_in
  );  

END str;
