-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose : * Insert extra sync pulses that copy the BSN of the incoming sync until a new sync pulse is availabel at the input
--           * The blocksize of the output stream can also be changed. 
-- Description:
--   The BSN is captured on the incoming sync pulse and stored. 
--   The incoming data is repacked into output blocks of size g_nof_data_per_blk. 
--   The output control signals sop and eop are created. Every g_nof_blk_per_sync block 
--   a sync pulse is created at the output. Allong with this sync pulse the stored BSN is 
--   set on the output. The BSN of the non-synced blocks are simply incremented starting 
--   from 1 and resetted every outgoing sync pulse. 
--
-- Remarks:
-- . There is no support for back pressure. 
-- . It does not compensate for missing data or extra data. There is NO reset function. It assumes that the 
--   incoming data is perfectly aligned. Use a dp_sync_checker to assure the incoming data is perfect.      

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL; 

ENTITY dp_sync_insert IS
  GENERIC (
    g_nof_data_per_blk : NATURAL := 128;  
    g_nof_blk_per_sync : NATURAL := 12500
  );             
  PORT (         
    rst        : IN  STD_LOGIC;
    clk        : IN  STD_LOGIC;
    -- Streaming sink
    snk_in     : IN  t_dp_sosi := c_dp_sosi_rst;
    -- Streaming source
    src_out    : OUT t_dp_sosi
  );
END dp_sync_insert;


ARCHITECTURE rtl OF dp_sync_insert IS

  TYPE t_reg IS RECORD  -- local registers
    data_cnt      : NATURAL RANGE 0 TO g_nof_data_per_blk;
    blk_cnt       : NATURAL RANGE 0 TO g_nof_blk_per_sync;
    sync_out_bsn  : STD_LOGIC_VECTOR(c_dp_stream_bsn_w-1 DOWNTO 0);
    snk_in_reg    : t_dp_sosi;
    src_out       : t_dp_sosi;
    sync_detected : STD_LOGIC;
  END RECORD;
  
  CONSTANT c_reg_rst  : t_reg := ( 0, 0, (OTHERS=>'0'), c_dp_sosi_rst, c_dp_sosi_rst, '0');

  -- Define the local registers in t_reg record
  SIGNAL r         : t_reg;
  SIGNAL nxt_r     : t_reg;
  
BEGIN
  
  src_out <= r.src_out;
  
  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      r <= c_reg_rst;
    ELSIF rising_edge(clk) THEN
      r <= nxt_r;
    END IF;
  END PROCESS;  

  
  p_comb : PROCESS(r, snk_in)
    VARIABLE v : t_reg;
  BEGIN
   
    v := r;
    v.src_out := func_dp_stream_reset_control(v.src_out);
    v.snk_in_reg := snk_in;
        
    IF snk_in.sync = '1' THEN
      v.sync_out_bsn := snk_in.bsn;
      IF r.sync_detected = '0' THEN
        v.sync_detected := '1';
      END IF; 
    END IF;

    IF r.snk_in_reg.valid = '1' AND r.sync_detected = '1' THEN
      v.src_out.data  := r.snk_in_reg.data;
      v.src_out.re    := r.snk_in_reg.re;
      v.src_out.im    := r.snk_in_reg.im;
      v.src_out.valid := '1';
      
      v.data_cnt      := r.data_cnt + 1;
      
      IF r.data_cnt = 0 AND r.blk_cnt = 0 THEN 
        v.src_out.sync := '1';
        v.src_out.bsn  := r.sync_out_bsn;
        v.src_out.sop  := '1';        
        v.blk_cnt      := r.blk_cnt + 1;        
      ELSIF r.data_cnt = 0 THEN 
        v.src_out.sop  := '1';
        v.src_out.bsn  := TO_DP_BSN(r.blk_cnt);        
        v.blk_cnt      := r.blk_cnt + 1;        
      ELSIF r.data_cnt = g_nof_data_per_blk-1 THEN 
        v.data_cnt    := 0; 
        v.src_out.eop := '1'; 
      END IF;

      IF r.blk_cnt = g_nof_blk_per_sync THEN
        v.blk_cnt := 0;
      END IF; 
    END IF; 
    
    nxt_r <= v;
     
  END PROCESS;  
 
END rtl;
