--------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
--
-- Purpose: Remove tail from an input block
-- Description:
--   A shift register is used to remove the tail from a stream and output the
--   stripped stream on src_out.
-- Remark:
-- . If g_symbol_w = g_data_w then it is better to use dp_block_resize.vhd,
--   because that provides a simpler solution. The dp_block_resize.vhd can
--   remove 0 or more head data and 0 or more tail data from the input block.

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY dp_tail_remove IS
  GENERIC (
    g_data_w      : NATURAL;
    g_symbol_w    : NATURAL;
    g_nof_symbols : NATURAL -- Nof symbols to be stripped from end of the packet,
                            -- and accounting for the nof empty symbols.
  );
  PORT (
    st_rst  : IN  STD_LOGIC;
    st_clk  : IN  STD_LOGIC;

    snk_out : OUT t_dp_siso;
    snk_in  : IN  t_dp_sosi;

    src_in : IN  t_dp_siso;
    src_out : OUT t_dp_sosi
  );
END dp_tail_remove;


ARCHITECTURE str OF dp_tail_remove IS

  CONSTANT c_output_reg     : BOOLEAN := TRUE;  -- register dp_shiftreg output to ease timing closure
  CONSTANT c_tail_nof_words : NATURAL := ceil_div(g_nof_symbols*g_symbol_w, g_data_w);

  -- The number of words we need to buffer equals c_tail_nof_words, plus one extra because we need to move the EOP to
  -- the last valid data word.
  CONSTANT c_nof_shiftreg_words : NATURAL := c_tail_nof_words+1; 

  SIGNAL rd_sosi_arr        : t_dp_sosi_arr(0 TO c_nof_shiftreg_words-1);
  SIGNAL wr_sosi_arr        : t_dp_sosi_arr(0 TO c_nof_shiftreg_words-1);

BEGIN

  snk_out <= src_in;

  u_src_shift : ENTITY work.dp_shiftreg
  GENERIC MAP (
    g_output_reg     => c_output_reg,
    g_flush_eop      => TRUE,
    g_modify_support => TRUE,
    g_nof_words      => c_nof_shiftreg_words
  )
  PORT MAP (
    rst                 => st_rst,
    clk                 => st_clk,
    -- ST sink
    snk_out             => OPEN,
    snk_in              => snk_in,
    -- Control shift register contents
    cur_shiftreg_inputs => rd_sosi_arr,
    new_shiftreg_inputs => wr_sosi_arr,
    -- ST source
    src_in              => src_in,  -- We correct the stream via new_shiftreg_inputs, so
    src_out             => src_out  -- the shiftreg sources everything but the tail.
  );

  p_shift: PROCESS(rd_sosi_arr)
    VARIABLE v_wr_sosi_arr : t_dp_sosi_arr(0 TO c_nof_shiftreg_words-1);
  BEGIN
    v_wr_sosi_arr := rd_sosi_arr;
    
    IF rd_sosi_arr(0).eop = '1' THEN
      FOR i IN 0 TO c_nof_shiftreg_words-2 LOOP
        v_wr_sosi_arr(i).valid := '0';
      END LOOP;
      v_wr_sosi_arr(0).eop := '0';
      v_wr_sosi_arr(c_nof_shiftreg_words-1).eop := '1';
      v_wr_sosi_arr(c_nof_shiftreg_words-1).empty := rd_sosi_arr(0).empty;
    END IF;

    wr_sosi_arr <= v_wr_sosi_arr;
  END PROCESS;

END str;
