-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;


ENTITY dp_throttle IS
  GENERIC (
    g_dc_period      : NATURAL := 100; -- provides a resolution of 1% (1/100..100/100)
    g_throttle_valid : BOOLEAN := FALSE -- FALSE: Stream passes through, snk_out.ready is AND'ed with pulse
                                        -- TRUE : Throttles src_out.valid instead of snk_out.ready; Onther entity I/O unused.
  );
  PORT (
    rst      : IN  STD_LOGIC;
    clk      : IN  STD_LOGIC;
    -- ST sinks
    snk_out  : OUT t_dp_siso;
    snk_in   : IN  t_dp_sosi := c_dp_sosi_rst;
    -- ST source
    src_in   : IN  t_dp_siso := c_dp_siso_rst;
    src_out  : OUT t_dp_sosi;

    throttle : IN STD_LOGIC_VECTOR(ceil_log2(g_dc_period+1)-1 DOWNTO 0)
  );
END dp_throttle;

ARCHITECTURE str OF dp_throttle IS

  SIGNAL dc_out : STD_LOGIC;

BEGIN

  gen_throttle_ready : IF g_throttle_valid = FALSE GENERATE
    -- Simply AND the ready signal with out duty cycle controlled signal to provide a throttle
    snk_out.ready <= src_in.ready AND dc_out;
    snk_out.xon   <= src_in.xon;
    src_out <= snk_in;
  END GENERATE;

  gen_throttle_valid : IF g_throttle_valid = TRUE GENERATE
    -- Connect ducty cycle controlled signal directly to src_out.valid
    src_out.valid <= dc_out;
  END GENERATE;

  u_common_duty_cycle : ENTITY common_lib.common_duty_cycle
  GENERIC MAP (
    g_rst_lvl => '0', -- Start with '0' on the output so our connected sink is not maxed out after reset
    g_dis_lvl => '0', -- Don't care - dc_out_en is not used.
    g_act_lvl => '1',
    g_per_cnt => g_dc_period,
    g_act_cnt => 0 -- After init, stay in idle state until we write a new DC value
  )
  PORT MAP (
    rst         => rst,
    clk         => clk,

    dc_act_cnt  => throttle,
                 
    dc_out_en   => '1', --We can also disable the output by writing zero to dc_act_cnt.
    dc_out      => dc_out
   );


END str;

