-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;

ENTITY dp_throttle_reg IS
  GENERIC (
    g_dc_period      : NATURAL
  );
  PORT (
    -- Clocks and reset
    mm_rst            : IN  STD_LOGIC;
    mm_clk            : IN  STD_LOGIC;

    st_rst            : IN  STD_LOGIC;
    st_clk            : IN  STD_LOGIC; 
    
    -- Memory Mapped Slave in mm_clk domain
    sla_in            : IN  t_mem_mosi;  -- actual ranges defined by c_mm_reg
    sla_out           : OUT t_mem_miso := c_mem_miso_rst;  -- actual ranges defined by c_mm_reg
    
    throttle          : OUT STD_LOGIC_VECTOR(ceil_log2(g_dc_period+1)-1 DOWNTO 0)
  );
END dp_throttle_reg;


ARCHITECTURE rtl OF dp_throttle_reg IS

  -- Define the actual size of the MM slave register
  CONSTANT c_mm_reg : t_c_mem := (latency  => 1,
                                  adr_w    => 2,
                                  dat_w    => c_word_w,       -- Use MM bus data width = c_word_w = 32 for all MM registers
                                  nof_dat  => 1,
                                  init_sl  => '0');

  SIGNAL mm_throttle : STD_LOGIC_VECTOR(ceil_log2(g_dc_period+1)-1 DOWNTO 0);

BEGIN

  p_mm_reg : PROCESS (mm_rst, mm_clk)
  BEGIN
    IF mm_rst = '1' THEN
      -- Access event, register values
      mm_throttle  <= (OTHERS=>'0');
      ELSIF rising_edge(mm_clk) THEN
      -- Read access defaults
      sla_out.rdval <= '0';
      
      -- Write event defaults
     
      -- Write access: set register value
      IF sla_in.wr = '1' THEN
        CASE TO_UINT(sla_in.address(c_mm_reg.adr_w-1 DOWNTO 0)) IS
          WHEN 0 =>
            mm_throttle <= sla_in.wrdata(ceil_log2(g_dc_period+1)-1 DOWNTO 0);
          
          WHEN OTHERS => NULL;
        END CASE;
        
      END IF;
    END IF;
  END PROCESS;
 
  u_common_reg_cross_domain : ENTITY common_lib.common_reg_cross_domain
  PORT MAP (
    in_rst      => mm_rst,
    in_clk      => mm_clk,
    in_dat      => mm_throttle,
    in_done     => OPEN,
    out_rst     => st_rst,
    out_clk     => st_clk,
    out_dat     => throttle,
    out_new     => OPEN
  );
  
END rtl;
