-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Author : E. Kooistra, 13 Apr 2017
-- Purpose : Throttle snk_in.xon to control the input block rate
-- Description:
--   The XON='1' time is determined by g_nof_block_on*g_block_size, so the
--   number of active snk_in.valid input. The XON='0' cannot be derived from
--   the snk_in because it is off. Therefore the off time is counted in clock
--   cycles. Typically choose g_nof_clk_off = g_block_size * 'nof block off'
--   * (1/'data valid duty_cycle').

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;


ENTITY dp_throttle_xon IS
  GENERIC (
    g_restart_at_sync : BOOLEAN := FALSE;  -- if true restart xon control at each snk_in_sync, else and start at first valid and never restart
    g_block_size      : NATURAL := 100;    -- number of valid data per block marked by sop and eop
    g_nof_block_on    : NATURAL := 10;     -- number of blocks that snk_out.xon is active
    g_nof_clk_off     : NATURAL := 10      -- number of clock cycles that snk_out.xon is kept inactive
  );
  PORT (
    rst      : IN  STD_LOGIC;
    clk      : IN  STD_LOGIC;
    -- ST sinks
    snk_out  : OUT t_dp_siso;
    snk_in   : IN  t_dp_sosi := c_dp_sosi_rst;
    -- ST source
    src_in   : IN  t_dp_siso := c_dp_siso_rdy;
    src_out  : OUT t_dp_sosi
  );
END dp_throttle_xon;

ARCHITECTURE rtl OF dp_throttle_xon IS

  CONSTANT c_nof_block_on : NATURAL := g_nof_block_on-1;  -- adjust by -1 to account for that first block already appears at cnt = 0
  CONSTANT c_nof_valid_on : NATURAL := c_nof_block_on*g_block_size;
  
  SIGNAL cnt     : NATURAL RANGE 0 TO largest(g_block_size*c_nof_block_on, g_nof_clk_off);
  SIGNAL nxt_cnt : NATURAL;
  SIGNAL xon     : STD_LOGIC;
  SIGNAL nxt_xon : STD_LOGIC;

BEGIN

  snk_out.ready <= src_in.ready;           -- pass on ready
  snk_out.xon   <= src_in.xon AND xon;     -- control xon
  
  src_out <= snk_in;                       -- pass on sosi

  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      cnt <= 0;
      xon <= '1';
    ELSIF rising_edge(clk) THEN
      cnt <= nxt_cnt;
      xon <= nxt_xon;
    END IF;
  END PROCESS;
  
  p_comb : PROCESS(snk_in, cnt, xon)
  BEGIN
    -- Default
    nxt_cnt <= cnt;
    nxt_xon <= xon;
    
    -- XON/XOFF control
    IF xon='1' THEN
      -- XON time
      IF snk_in.valid='1' THEN
        IF cnt<c_nof_valid_on-1 THEN
          nxt_cnt <= cnt + 1;
        ELSE
          nxt_cnt <= 0;
          nxt_xon <= '0';
        END IF;
      END IF;
    ELSE
      -- XOFF time
      IF cnt<g_nof_clk_off-1 THEN
        nxt_cnt <= cnt + 1;
      ELSE
        nxt_cnt <= 0;
        nxt_xon <= '1';
      END IF;
    END IF;
    
    -- Overrule if restart at sync
    IF g_restart_at_sync=TRUE AND snk_in.sync='1' THEN
      nxt_cnt <= 1;
      nxt_xon <= '1';
    END IF;

    -- Overrule if g_nof_clk_off=0
    IF g_nof_clk_off=0 THEN
      nxt_cnt <= 1;
      nxt_xon <= '1';
    END IF;
  END PROCESS;
  
END rtl;

