--------------------------------------------------------------------------------
--
-- Copyright (C) 2016
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

-- Author:
-- . Daniel van der Schuur
-- Purpose:
-- . Unfold n input streams into n*2, n*2*2, n*2*2*2, .. output streams
-- Description:
-- . Reversed operation of dp_folder.
-- Remark:
-- . Use dp_deinterleave_one_to_n.vhd when:
--   - for 1 to N, so any N >= 1 (so not only powers of 2)
--   - when all sosi fields need to be preserved (so not only the valid)
--   - if flow control is needed

ENTITY dp_unfolder IS
  GENERIC (
    g_nof_inputs        : NATURAL;          -- Number of inputs
    g_nof_unfolds       : NATURAL := 0;     -- Number of times to unfold
    g_output_block_size : NATURAL := 0;     -- >0: Create SOP/EOP tagged output blocks of this size.
    g_fwd_sync_bsn      : BOOLEAN := FALSE; -- TRUE: forwards (stored) input Sync+BSN (from snk_in_arr(0)) to all output streams
    g_use_channel       : BOOLEAN := FALSE;-- TRUE: Also fold the channel field
    g_output_align      : BOOLEAN := TRUE   -- TRUE: Use pipeline stages to align the outputs
  );
  PORT (
    rst            : IN  STD_LOGIC;
    clk            : IN  STD_LOGIC;

    snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);

    src_out_arr    : OUT t_dp_sosi_arr(g_nof_inputs*pow2(g_nof_unfolds)-1 DOWNTO 0)
  );
END dp_unfolder;

ARCHITECTURE str OF dp_unfolder IS

  COMPONENT dp_unfolder IS
    GENERIC (
      g_nof_inputs   : NATURAL;
      g_nof_unfolds  : NATURAL := 0;
      g_output_block_size : NATURAL := 0;   
      g_fwd_sync_bsn      : BOOLEAN := FALSE;
      g_use_channel       : BOOLEAN := FALSE;
      g_output_align : BOOLEAN := TRUE  
    ); 
    PORT (
      rst            : IN  STD_LOGIC;
      clk            : IN  STD_LOGIC;
  
      snk_in_arr     : IN  t_dp_sosi_arr(g_nof_inputs-1 DOWNTO 0);
  
      src_out_arr    : OUT t_dp_sosi_arr(g_nof_inputs*pow2(g_nof_unfolds)-1 DOWNTO 0)
    );
  END COMPONENT;

  CONSTANT c_nof_demuxes : NATURAL := g_nof_inputs;
  CONSTANT c_nof_outputs : NATURAL := g_nof_inputs*pow2(g_nof_unfolds);

  SIGNAL output_sel_arr           : STD_LOGIC_VECTOR(c_nof_demuxes-1 DOWNTO 0);
  SIGNAL nxt_output_sel_arr       : STD_LOGIC_VECTOR(c_nof_demuxes-1 DOWNTO 0);

  SIGNAL demux_src_out_2arr_2     : t_dp_sosi_2arr_2(c_nof_demuxes-1 DOWNTO 0);
  SIGNAL nxt_demux_src_out_2arr_2 : t_dp_sosi_2arr_2(c_nof_demuxes-1 DOWNTO 0);
  SIGNAL demux_src_out_arr        : t_dp_sosi_arr(2*c_nof_demuxes-1 DOWNTO 0);

  SIGNAL dp_pipeline_snk_in_arr   : t_dp_sosi_arr(c_nof_outputs-1 DOWNTO 0);

  SIGNAL dp_block_gen_snk_in_arr  : t_dp_sosi_arr(c_nof_outputs-1 DOWNTO 0);
  SIGNAL dp_block_gen_src_out_arr : t_dp_sosi_arr(c_nof_outputs-1 DOWNTO 0);

BEGIN

  gen_arch: IF g_nof_unfolds/=0 GENERATE
    -----------------------------------------------------------------------------
    -- Simple 2-output demux logic
    -----------------------------------------------------------------------------
    gen_demux_comb: FOR i IN 0 TO c_nof_demuxes-1 GENERATE
      nxt_output_sel_arr(i) <= NOT output_sel_arr(i) WHEN snk_in_arr(i).valid='1' ELSE output_sel_arr(i);
    END GENERATE;

    p_demux : PROCESS(snk_in_arr, output_sel_arr)
    BEGIN
      FOR i IN 0 TO c_nof_demuxes-1 LOOP
        nxt_demux_src_out_2arr_2(i)(0) <= c_dp_sosi_rst;
        nxt_demux_src_out_2arr_2(i)(1) <= c_dp_sosi_rst;
        IF snk_in_arr(i).valid='1' THEN
          IF output_sel_arr(i)='0' THEN
            nxt_demux_src_out_2arr_2(i)(0).data  <= snk_in_arr(i).data;
            nxt_demux_src_out_2arr_2(i)(0).re    <= snk_in_arr(i).re;
            nxt_demux_src_out_2arr_2(i)(0).im    <= snk_in_arr(i).im;
            nxt_demux_src_out_2arr_2(i)(0).valid <= '1';
            IF g_use_channel=TRUE THEN
              nxt_demux_src_out_2arr_2(i)(0).channel <= snk_in_arr(i).channel;
            END IF;
          ELSE 
            nxt_demux_src_out_2arr_2(i)(1).data  <= snk_in_arr(i).data;
            nxt_demux_src_out_2arr_2(i)(1).re    <= snk_in_arr(i).re;
            nxt_demux_src_out_2arr_2(i)(1).im    <= snk_in_arr(i).im;
            nxt_demux_src_out_2arr_2(i)(1).valid <= '1';
            IF g_use_channel=TRUE THEN
              nxt_demux_src_out_2arr_2(i)(1).channel <= snk_in_arr(i).channel;
            END IF;
          END IF;
        END IF;
      END LOOP;
    END PROCESS;

    -- Registers
    p_clk: PROCESS(clk, rst)
    BEGIN
      IF rst='1' THEN
        demux_src_out_2arr_2 <= (OTHERS=>(OTHERS=>c_dp_sosi_rst));
        output_sel_arr       <= (OTHERS=>'0');
      ELSIF rising_edge(clk) THEN
        demux_src_out_2arr_2 <= nxt_demux_src_out_2arr_2;
        output_sel_arr       <= nxt_output_sel_arr;
      END IF;
    END PROCESS;

    -----------------------------------------------------------------------------
    -- Wire the 2D demux output array to 1D array to match entity I/O type
    -----------------------------------------------------------------------------
    gen_demux_inputs_0: FOR i IN 0 TO c_nof_demuxes-1 GENERATE
       demux_src_out_arr(2*i)   <= demux_src_out_2arr_2(i)(0);
       demux_src_out_arr(2*i+1) <= demux_src_out_2arr_2(i)(1);
    END GENERATE;
  
    -----------------------------------------------------------------------------
    -- g_nof_unfolds>1, so add an unfolder stage.
    -----------------------------------------------------------------------------
    gen_dp_unfolder: IF g_nof_unfolds>1 GENERATE
      u_dp_unfolder : dp_unfolder
      GENERIC MAP (
        g_nof_inputs        => c_nof_demuxes*2, -- Next stage has all our demux outputs as inputs
        g_nof_unfolds       => g_nof_unfolds-1,
        g_output_block_size => g_output_block_size,
        g_fwd_sync_bsn      => g_fwd_sync_bsn,
        g_use_channel       => g_use_channel,
        g_output_align      => g_output_align
      )
      PORT MAP (
        rst         => rst,
        clk         => clk,
                    
        snk_in_arr  => demux_src_out_arr,                   
        src_out_arr => src_out_arr
      );
    END GENERATE;
 
    -----------------------------------------------------------------------------
    -- g_nof_unfolds=1, so this is the last stage. Wire up the outputs.
    -----------------------------------------------------------------------------
    gen_src_out_arr: IF g_nof_unfolds=1 GENERATE

      gen_output_align : IF g_output_align=TRUE GENERATE
        dp_pipeline_snk_in_arr <= demux_src_out_arr;
  
        -----------------------------------------------------------------------------
        -- Pipeline to re-align the unfolder output
        -----------------------------------------------------------------------------
        gen_dp_pipeline : FOR i IN 0 TO c_nof_outputs-1 GENERATE
          u_dp_pipeline : ENTITY dp_lib.dp_pipeline
          GENERIC MAP (
            g_pipeline => 0 + (pow2(g_nof_unfolds) - i REM pow2(g_nof_unfolds)-1)
          )
          PORT MAP (
            rst         => rst,
            clk         => clk,
        
            snk_in      => dp_pipeline_snk_in_arr(i),
            src_out     => dp_block_gen_snk_in_arr(i)
          );
        END GENERATE; 
      END GENERATE;

      no_output_align : IF g_output_align=FALSE GENERATE
        dp_block_gen_snk_in_arr <= demux_src_out_arr;
      END GENERATE;

      -----------------------------------------------------------------------------
      -- Add SOP and EOP to the outputs
      -----------------------------------------------------------------------------
      gen_ctrl : IF g_output_block_size>0 GENERATE
        gen_dp_block_gen : FOR i IN 0 TO c_nof_outputs-1 GENERATE
          u_dp_block_gen : ENTITY work.dp_block_gen
          GENERIC MAP (
            g_use_src_in       => FALSE,
            g_nof_data         => g_output_block_size,
            g_preserve_sync    => TRUE,
            g_preserve_bsn     => TRUE,
            g_preserve_channel => g_use_channel
          )
          PORT MAP(
            rst        => rst,
            clk        => clk,
           
            snk_in     => dp_block_gen_snk_in_arr(i),
            src_out    => dp_block_gen_src_out_arr(i)
          );    
        END GENERATE;
      END GENERATE;
    
      no_ctrl : IF g_output_block_size=0 GENERATE
        dp_block_gen_src_out_arr <= dp_block_gen_snk_in_arr;
      END GENERATE;
    
      -----------------------------------------------------------------------------
      -- Re-add input sync + BSN to all output streams
      -----------------------------------------------------------------------------
      gen_sync_bsn : IF g_fwd_sync_bsn = TRUE GENERATE
        gen_dp_fifo_info: FOR i IN 0 TO c_nof_outputs-1 GENERATE
          u_dp_fifo_info : ENTITY work.dp_fifo_info
          GENERIC MAP (
            g_use_sync => TRUE,
            g_use_bsn  => TRUE
          )
          PORT MAP (
            rst          => rst,
            clk          => clk,
        
            data_snk_in  => dp_block_gen_src_out_arr(i),  -- delayed snk_in data
            info_snk_in  => snk_in_arr(0),                -- original snk_in info
        
            src_in       => c_dp_siso_rdy,
            src_out      => src_out_arr(i)
          );
        END GENERATE;
      END GENERATE;
    
      no_sync_bsn : IF g_fwd_sync_bsn = FALSE GENERATE
        src_out_arr <= dp_block_gen_src_out_arr;
      END GENERATE;
    
        END GENERATE;
    
      END GENERATE;

  -----------------------------------------------------------------------------
  -- Wire output to input if g_nof_unfolds=0
  -----------------------------------------------------------------------------
  gen_wire_out_to_in: IF g_nof_unfolds=0 GENERATE
    dp_block_gen_snk_in_arr <= snk_in_arr;
  END GENERATE;

END str;

