--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------
 
LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.all;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_packetizing_pkg.ALL;

-- Reuse from LOFAR rad_unframe.vhd and rad_unframe(rtl).vhd

-- Purpose:
--   Extract the payload data from a frame:
--     -----------------------
--     | fsn | dat ... | brc |
--     -----------------------
--       sof             eof
--       val   val ...   val
--   and output the frame fsn and brc information as signals:
--     . dat[]
--     . fsn[]
--     . sync
--     . val
--     . sof
--     . eof
--     . sof_sync
--
-- Description:
--   The dp_unframe is the reverse of dp_frame. The output sof and eof mark the
--   start and end of the output dat. The output fsn signal is valid at the
--   output sof and remains valid. The output sync is a pulse and is valid one
--   cycle before output sof. The output err signal reflects the brc value and
--   is valid at the output eof and set to 0 at the next input sof.

ENTITY dp_unframe IS
  GENERIC (
    g_fsn_w          : POSITIVE := c_dp_max_w;
    g_dat_w          : POSITIVE := c_dp_max_w
  );
  PORT (
    rst              : IN  STD_LOGIC;
    clk              : IN  STD_LOGIC;
    
    in_dat           : IN  STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
    in_val           : IN  STD_LOGIC;
    in_sof           : IN  STD_LOGIC;
    in_eof           : IN  STD_LOGIC;
    
    out_fsn          : OUT STD_LOGIC_VECTOR(g_fsn_w-1 DOWNTO 0);
    out_sync         : OUT STD_LOGIC;   -- LOFAR style sync before sof
    sof_sync         : OUT STD_LOGIC;   -- DP style sync at sof
    out_dat          : OUT STD_LOGIC_VECTOR(g_dat_w-1 DOWNTO 0);
    out_val          : OUT STD_LOGIC;
    out_sof          : OUT STD_LOGIC;
    out_eof          : OUT STD_LOGIC;
    out_err          : OUT STD_LOGIC
  );
BEGIN
-- synthesis translate_off
  ASSERT g_fsn_w <= g_dat_w
    REPORT "g_fsn_w must be smaller than or equal to g_dat_w"
      SEVERITY ERROR;
-- synthesis translate_on

END dp_unframe;


ARCHITECTURE rtl OF dp_unframe IS

  -- Use unregistered input to save logic, use registered input to ease achieving timing constrains.
  CONSTANT c_input_reg   : BOOLEAN := FALSE;
  
  CONSTANT c_brc_ok      : STD_LOGIC_VECTOR(in_dat'RANGE) := STD_LOGIC_VECTOR(TO_UNSIGNED(c_dp_brc_ok ,in_dat'LENGTH));
  CONSTANT c_brc_err     : STD_LOGIC_VECTOR(in_dat'RANGE) := STD_LOGIC_VECTOR(TO_UNSIGNED(c_dp_brc_err,in_dat'LENGTH));
  
  -- Registered inputs (see c_input_reg)
  SIGNAL in_dat_reg      : STD_LOGIC_VECTOR(in_dat'RANGE);
  SIGNAL in_val_reg      : STD_LOGIC;
  SIGNAL in_sof_reg      : STD_LOGIC;
  SIGNAL in_eof_reg      : STD_LOGIC;
  
  -- Hold input sof and then adjust it to next input val to put out_sof on first valid out_dat word (in_sof is on fsn)
  SIGNAL sof_hld         : STD_LOGIC;
  SIGNAL nxt_sof_hld     : STD_LOGIC;
  SIGNAL sync_hld        : STD_LOGIC;  -- hold input sync from FSN to adjust to output sof as sof_sync
  SIGNAL nxt_sync_hld    : STD_LOGIC;
  
  SIGNAL i_out_fsn       : STD_LOGIC_VECTOR(out_fsn'RANGE);
  SIGNAL nxt_out_fsn     : STD_LOGIC_VECTOR(out_fsn'RANGE);
  SIGNAL nxt_out_sync    : STD_LOGIC;
  SIGNAL i_out_err       : STD_LOGIC;
  SIGNAL nxt_out_err     : STD_LOGIC;
  
  -- Delay output dat to get out_eof on last valid out_dat word (in_eof is on brc)
  SIGNAL dat_adj         : STD_LOGIC_VECTOR(out_dat'RANGE);
  SIGNAL nxt_dat_adj     : STD_LOGIC_VECTOR(dat_adj'RANGE);
  SIGNAL val_adj         : STD_LOGIC;
  SIGNAL nxt_val_adj     : STD_LOGIC;
  SIGNAL sof_adj         : STD_LOGIC;
  SIGNAL nxt_sof_adj     : STD_LOGIC;
  SIGNAL sync_adj        : STD_LOGIC;
  SIGNAL nxt_sync_adj    : STD_LOGIC;
  
  SIGNAL i_out_dat       : STD_LOGIC_VECTOR(out_dat'RANGE);
  SIGNAL nxt_out_dat     : STD_LOGIC_VECTOR(out_dat'RANGE);
  SIGNAL nxt_out_val     : STD_LOGIC;
  SIGNAL nxt_out_sof     : STD_LOGIC;
  SIGNAL nxt_out_eof     : STD_LOGIC;
  SIGNAL nxt_sof_sync    : STD_LOGIC;
  
BEGIN

  out_fsn <= i_out_fsn;
  out_err <= i_out_err;
  out_dat <= i_out_dat;

  gen_input_reg : IF c_input_reg=TRUE GENERATE
    p_reg : PROCESS (clk, rst)
    BEGIN
      IF rst='1' THEN
        in_dat_reg  <= (OTHERS => '0');
        in_val_reg  <= '0';
        in_sof_reg  <= '0';
        in_eof_reg  <= '0';
      ELSIF rising_edge(clk) THEN
        in_dat_reg  <= in_dat;
        in_val_reg  <= in_val;
        in_sof_reg  <= in_sof;
        in_eof_reg  <= in_eof;
      END IF;
    END PROCESS;
  END GENERATE;
  no_input_reg : IF c_input_reg=FALSE GENERATE
    in_dat_reg  <= in_dat;
    in_val_reg  <= in_val;
    in_sof_reg  <= in_sof;
    in_eof_reg  <= in_eof;
  END GENERATE;
    
  p_clk: PROCESS(clk, rst)
  BEGIN
    IF rst='1' THEN
      sof_hld     <= '0';
      sync_hld    <= '0';
      i_out_fsn   <= (OTHERS => '0');
      out_sync    <= '0';
      i_out_err   <= '0';
      dat_adj     <= (OTHERS => '0');
      val_adj     <= '0';
      sof_adj     <= '0';
      sync_adj    <= '0';
      sof_sync    <= '0';
      i_out_dat   <= (OTHERS => '0');
      out_val     <= '0';
      out_sof     <= '0';
      out_eof     <= '0';
    ELSIF rising_edge(clk) THEN
      sof_hld     <= nxt_sof_hld;
      sync_hld    <= nxt_sync_hld;
      i_out_fsn   <= nxt_out_fsn;
      out_sync    <= nxt_out_sync;
      i_out_err   <= nxt_out_err;
      dat_adj     <= nxt_dat_adj;
      val_adj     <= nxt_val_adj;
      sof_adj     <= nxt_sof_adj;
      sync_adj    <= nxt_sync_adj;
      sof_sync    <= nxt_sof_sync;
      i_out_dat   <= nxt_out_dat;
      out_val     <= nxt_out_val;
      out_sof     <= nxt_out_sof;
      out_eof     <= nxt_out_eof;
    END IF;
  END PROCESS;
  
  p_output: PROCESS(in_dat_reg, in_val_reg, sync_hld, sof_hld, in_sof_reg, in_eof_reg, dat_adj, val_adj, sof_adj, sync_adj, i_out_dat, i_out_fsn, i_out_err)
  BEGIN
    nxt_sof_hld  <= sof_hld;           -- hold input sof to first input dat
    nxt_sync_hld <= sync_hld;          -- hold input sync to first input dat
    nxt_out_fsn  <= i_out_fsn;         -- hold out_fsn during a frame
    nxt_out_sync <= '0';
    nxt_out_err  <= i_out_err;         -- hold out_err until next sof
    
    nxt_dat_adj  <= dat_adj;           -- delay out_dat, val, sof to meet out_eof
    nxt_val_adj  <= val_adj;
    nxt_sof_adj  <= sof_adj;
    nxt_sync_adj <= sync_adj;
    
    nxt_out_dat  <= i_out_dat;
    nxt_out_val  <= '0';
    nxt_out_sof  <= '0';
    nxt_out_eof  <= '0';
    nxt_sof_sync <= '0';
    
    IF in_val_reg='1' THEN
      nxt_sof_hld <= '0';
      nxt_sync_hld <= '0';
      nxt_dat_adj <= in_dat_reg;
      nxt_val_adj <= '0';
      nxt_sof_adj <= '0';
      nxt_sync_adj <= '0';
      nxt_out_dat <= dat_adj;          -- dat output
      nxt_out_val <= val_adj;          -- val output
      nxt_out_sof <= sof_adj;          -- sof output
      nxt_sof_sync <= sync_adj;        -- sync output
      IF in_sof_reg='1' THEN      
        nxt_sof_hld  <= '1';
        nxt_sync_hld <= in_dat_reg(out_fsn'HIGH);   -- sync pulse
        nxt_out_fsn  <= in_dat_reg(out_fsn'RANGE);  -- new fsn
        nxt_out_sync <= in_dat_reg(out_fsn'HIGH);   -- sync pulse
        nxt_out_err  <= '0';                        -- clear previous err
      ELSIF sof_hld='1' THEN
        nxt_val_adj <= '1';            -- val adjust
        nxt_sof_adj <= '1';            -- sof adjust
        nxt_sync_adj <= sync_hld;      -- sync adjust
      ELSIF in_eof_reg='1' THEN        -- eof output
        nxt_out_eof <= '1';
        nxt_out_err <= '0';                         -- set err based on brc, already valid at out_eof
        IF in_dat_reg/=c_brc_ok THEN
          nxt_out_err <= '1';
        END IF;
      ELSE
        nxt_val_adj <= '1';            -- val adjust
      END IF;
    END IF;
  END PROCESS;
  
END rtl;
