-------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose:
--   Assert valid, sop and eop only when they are valid.
-- Description:
--   . Assert valid regarding the ready latency
--   . Assert sop and eop only when valid is active
--   . Pass the SOSI data signals on as wires
--   . If g_boolean_error=TRUE then vector-OR the snk_in.err into a bit 0 of
--     src_out.err, where '0' = OK and '1' = error.
-- Remark:
--   . All assignments are combinatorial, so no clock latency

ENTITY dp_validate IS
  GENERIC (
    g_ready_latency : NATURAL := 0;
    g_boolean_error : BOOLEAN := FALSE
  );
  PORT (
    -- ST sink
    snk_out : OUT t_dp_siso;
    snk_in  : IN  t_dp_sosi;
    -- ST source
    src_in  : IN  t_dp_siso;
    src_out : OUT t_dp_sosi
  );
END dp_validate;


ARCHITECTURE rtl OF dp_validate IS

  SIGNAL valid : STD_LOGIC;
  
BEGIN

  -- SISO ready
  snk_out <= src_in;
  
  -- SOSI
  -- The downstream sink ready latency must match that of the upstream source.
  -- For ready latency 0 the src_out.valid must only be interpreted when the
  -- src_in.ready is asserted. The src_in.ready input needs to be passed on via
  -- the snk_out.ready output of this component.
  -- According do the Avalon specification in case the ready latency > 0 then
  -- the valid will only be asserted by the upstream source after ready latency
  -- cycles when its src_in.ready was asserted. Hence then no (delayed) check
  -- on the src_in.ready is needed.
  valid <= snk_in.valid AND src_in.ready WHEN g_ready_latency=0 ELSE snk_in.valid;
  
  p_sosi : PROCESS(snk_in, valid)
  BEGIN
    src_out <= snk_in;
        
    -- According do the Avalon specification sop and eop are only interpreted
    -- when valid is asserted, so to be save AND them with valid.
    src_out.valid <= valid;
    src_out.sop   <= valid AND snk_in.sop;
    src_out.eop   <= valid AND snk_in.eop;
    
    -- Combine error vector into one bit if necessary
    IF g_boolean_error=TRUE THEN
      src_out.err <= (OTHERS=>'0');
      IF UNSIGNED(snk_in.err)/=0 THEN
        src_out.err(0) <= '1';
      END IF;
    END IF;
  END PROCESS;
  
END rtl;
