-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Scope component to show the concateneated DP SOSI data at the SCLK
--          sample rate
-- Description:
--   The scope supports g_nof_streams streams (signal paths).
--   The DP SOSI data is parallel data that consists of g_wideband_factor 
--   concatenated data samples per field. The data can be in the SOSI data
--   field or in case of complex data in the SOSI re and im fields.
--   The output scope_*_arr can be viewed in analogue format in the Wave
--   Window to observe the signal shape at the SCLK sample rate.
-- Remark:
-- . Only for simulation.
-- . When g_use_sclk=TRUE then the input SCLK is used. Else the SCLK is derived
--   from the DCLK so that it does not have to be applied via an input. This
--   eases the use of this scope within a design.
-- . The concatenated parallel data width is g_wideband_factor * g_dat_w and
--   must fit the sosi.data, re, im fields. Therefor use g_complex to avoid
--   out of range failure on re, im in case data has too large g_dat_w to also
--   fit for re, im, which can happen because c_dp_stream_data_w >>
--   c_dp_stream_dsp_data_w.
-- . In this dp_wideband_sp_arr_scope the input is one or more wideband streams
--   and the input sosi array has size g_nof_streams, so the wideband data is
--   carried by g_wideband_factor concatenated symbols in the data field or in
--   the (re, im) fields.
--   In dp_wideband_wb_arr_scope the input is only one wideband stream
--   and the input sosi array has size g_wideband_factor, so there the wideband
--   data is carried via the sosi array dimension.


LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY dp_wideband_sp_arr_scope IS
  GENERIC (
    g_sim                 : BOOLEAN := FALSE;
    g_use_sclk            : BOOLEAN := TRUE;
    g_complex             : BOOLEAN := FALSE;
    g_nof_streams         : NATURAL := 4;
    g_wideband_factor     : NATURAL := 4;        -- Wideband rate factor = 4 for dp_clk processing frequency is 200 MHz frequency and SCLK sample frequency Fs is 800 MHz
    g_wideband_big_endian : BOOLEAN := FALSE;    -- When true sp_sosi_arr[].data[3:0] = sample[t0,t1,t2,t3], else when false : sp_sosi_arr[].data[3:0] = sample[t3,t2,t1,t0]
    g_dat_w               : NATURAL := 8         -- Actual g_dat_w width of the data field or g_dat_w/2 width in case of the re field, im field
  );
  PORT (
    -- Digital processing clk
    DCLK            : IN STD_LOGIC := '0';
    
    -- Sampling clk, for simulation only
    SCLK            : IN STD_LOGIC := '0';   -- SCLK rate = g_wideband_factor * DCLK rate
    
    -- Streaming input samples for g_nof_streams
    sp_sosi_arr     : IN t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);   -- = [3:0] = Signal Paths [D,C,B,A]
    
    -- Scope output samples for g_nof_streams
    scope_sosi_arr  : OUT t_dp_sosi_integer_arr(g_nof_streams-1 DOWNTO 0)
  );
END dp_wideband_sp_arr_scope;


ARCHITECTURE beh OF dp_wideband_sp_arr_scope IS

  SIGNAL SCLKi           : STD_LOGIC;  -- sampling clk, for simulation only
  SIGNAL scope_cnt_arr   : t_natural_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL st_sosi_arr     : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  
BEGIN

  sim_only : IF g_sim=TRUE GENERATE
    use_sclk : IF g_use_sclk=TRUE GENERATE
      SCLKi <= SCLK;  -- no worry about the delta cycle delay from SCLK to SCLKi
    END GENERATE;
    gen_sclk : IF g_use_sclk=FALSE GENERATE
      proc_common_dclk_generate_sclk(g_wideband_factor, DCLK, SCLKi);
    END GENERATE;
  
    -- View sp_sosi_arr at the sample rate using st_sosi_arr 
    gen_arr : FOR I IN 0 TO g_nof_streams-1 GENERATE
      p_st_sosi_arr : PROCESS(SCLKi)
        VARIABLE vI : NATURAL;
      BEGIN
        IF rising_edge(SCLKi) THEN
          IF g_wideband_big_endian=TRUE THEN
            vI := g_wideband_factor-1-scope_cnt_arr(I);
          ELSE
            vI := scope_cnt_arr(I);
          END IF;
          scope_cnt_arr(I) <= 0;
          IF sp_sosi_arr(I).valid='1' AND scope_cnt_arr(I) < g_wideband_factor-1 THEN
            scope_cnt_arr(I) <= scope_cnt_arr(I) + 1;
          END IF;
          st_sosi_arr(I)         <= sp_sosi_arr(I);
          -- Default set unused sp_sosi_arr(I) fields to 0 to avoid Warning: NUMERIC_STD.TO_INTEGER: metavalue detected within func_dp_stream_slv_to_integer()
          st_sosi_arr(I).data    <= (OTHERS=>'0');
          st_sosi_arr(I).re      <= (OTHERS=>'0');
          st_sosi_arr(I).im      <= (OTHERS=>'0');
          st_sosi_arr(I).empty   <= (OTHERS=>'0');
          st_sosi_arr(I).channel <= (OTHERS=>'0');
          st_sosi_arr(I).err     <= (OTHERS=>'0');
          IF g_complex THEN
            st_sosi_arr(I).re <= RESIZE_DP_DSP_DATA(sp_sosi_arr(I).re((g_wideband_factor-vI)*g_dat_w-1 DOWNTO (g_wideband_factor-vI-1)*g_dat_w));
            st_sosi_arr(I).im <= RESIZE_DP_DSP_DATA(sp_sosi_arr(I).im((g_wideband_factor-vI)*g_dat_w-1 DOWNTO (g_wideband_factor-vI-1)*g_dat_w));            
          ELSE
            st_sosi_arr(I).data <= RESIZE_DP_SDATA(sp_sosi_arr(I).data((g_wideband_factor-vI)*g_dat_w-1 DOWNTO (g_wideband_factor-vI-1)*g_dat_w));
          END IF;
        END IF;
      END PROCESS;
      
      -- Map sosi to SLV of actual g_dat_w to allow observation in Wave Window in analogue format
      scope_sosi_arr(I) <= func_dp_stream_slv_to_integer(st_sosi_arr(I), g_dat_w);
    END GENERATE;
  END GENERATE;
  
END beh;
