-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Provide MM slave register for dp_xonoff
-- Description:
--
--  Read/write register for stream enable/disable
--
--   31             24 23             16 15              8 7               0  wi
--  |-----------------|-----------------|-----------------|-----------------|
--  |                                                                     |X|  0
--  |-----------------------------------------------------------------------|
--
--
--

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;

ENTITY dp_xonoff_reg IS
  GENERIC ( 
    g_default_value : STD_LOGIC := '1'
  );
  PORT (
    -- Clocks and reset
    mm_rst            : IN  STD_LOGIC;
    mm_clk            : IN  STD_LOGIC;
    st_rst            : IN  STD_LOGIC;
    st_clk            : IN  STD_LOGIC;
    
    -- Memory Mapped Slave in mm_clk domain
    sla_in            : IN  t_mem_mosi;
    sla_out           : OUT t_mem_miso;
    
    -- MM registers in st_clk domain
    xonoff_reg        : OUT STD_LOGIC_VECTOR(0 DOWNTO 0)
  );
END dp_xonoff_reg;


ARCHITECTURE str OF dp_xonoff_reg IS

  CONSTANT c_mm_reg : t_c_mem := (latency  => 1,
                                  adr_w    => 1,
                                  dat_w    => c_word_w,
                                  nof_dat  => 1,
                                  init_sl  => g_default_value);

  SIGNAL mm_xonoff_reg : STD_LOGIC_VECTOR(0 DOWNTO 0);

BEGIN

  p_mm_reg : PROCESS (mm_rst, mm_clk)
  BEGIN
    IF mm_rst = '1' THEN
      -- Read access
      sla_out   <= c_mem_miso_rst;
      -- Write access, register values
      mm_xonoff_reg(0) <= g_default_value; 
   
    ELSIF rising_edge(mm_clk) THEN
      -- Read access defaults
      sla_out.rdval <= '0';
      
      -- Write event defaults

      -- Write access: set register value
      IF sla_in.wr = '1' THEN
        CASE TO_UINT(sla_in.address(c_mm_reg.adr_w-1 DOWNTO 0)) IS
          WHEN 0 =>            
             mm_xonoff_reg(0) <= sla_in.wrdata(0);
                  
          WHEN OTHERS => NULL;  -- not used MM addresses
        END CASE;
        
      -- Read access: get register value
      ELSIF sla_in.rd = '1' THEN
        sla_out       <= c_mem_miso_rst;    -- set unused rddata bits to '0' when read
        sla_out.rdval <= '1';               -- c_mm_reg.latency = 1
        CASE TO_UINT(sla_in.address(c_mm_reg.adr_w-1 DOWNTO 0)) IS
          -- Read Block Sync
          WHEN 0 =>            
            sla_out.rddata(0) <= mm_xonoff_reg(0);
   
          WHEN OTHERS => NULL;  
        END CASE;
      END IF;
    END IF;
  END PROCESS;
    
  u_reg_cross_domain : ENTITY common_lib.common_reg_cross_domain
  PORT MAP (
    in_rst     => mm_rst,
    in_clk     => mm_clk,

    in_dat     => mm_xonoff_reg,

    out_rst    => st_rst,
    out_clk    => st_clk,

    out_dat    => xonoff_reg
  );

END str;
