-------------------------------------------------------------------------------
--
-- Copyright (C) 2018
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Author:
-- . Eric Kooistra
-- Purpose: 
-- . Provide MM interface to dp_block_select
-- Description:
--   See dp_block_select.vhd for functional description.
--   All g_nof_streams use the same index_lo and index_hi settings.
--   The default index_lo and index_hi are set by g_index_lo and g_index_hi.
--   The index_lo and index_hi can be read and set via the MM interface.

LIBRARY IEEE, common_lib, mm_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_field_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY mms_dp_block_select IS
  GENERIC (
    g_nof_streams          : NATURAL := 1;
    g_nof_blocks_per_sync  : NATURAL;
    g_index_lo             : NATURAL := 0;
    g_index_hi             : NATURAL
  );
  PORT (
    mm_rst       : IN  STD_LOGIC;
    mm_clk       : IN  STD_LOGIC;
    dp_rst       : IN  STD_LOGIC;
    dp_clk       : IN  STD_LOGIC;
    -- Control
    reg_mosi     : IN  t_mem_mosi;
    reg_miso     : OUT t_mem_miso;
    -- ST sink
    snk_out_arr  : OUT t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
    snk_in_arr   : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
    -- ST source
    src_in_arr   : IN  t_dp_siso_arr(g_nof_streams-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);
    src_out_arr  : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0)
  );
END mms_dp_block_select;


ARCHITECTURE str OF mms_dp_block_select IS

  CONSTANT c_field_arr : t_common_field_arr(1 DOWNTO 0) := ( (field_name_pad("index_hi"), "RW", 32, field_default(g_index_hi) ),
                                                             (field_name_pad("index_lo"), "RW", 32, field_default(g_index_lo) ));

  SIGNAL mm_fields_out : STD_LOGIC_VECTOR(field_slv_out_len(c_field_arr)-1 DOWNTO 0);

  SIGNAL index_lo      : NATURAL;
  SIGNAL index_hi      : NATURAL;
      
BEGIN

  -- Use same control for all streams
  u_mm_fields: ENTITY mm_lib.mm_fields
  GENERIC MAP(
    g_use_slv_in_val  => FALSE,    -- use FALSE to save logic when always slv_in_val='1'
    g_field_arr       => c_field_arr
  )
  PORT MAP (
    mm_clk     => mm_clk,
    mm_rst     => mm_rst,

    mm_mosi    => reg_mosi,
    mm_miso    => reg_miso,
    
    slv_clk    => dp_clk,
    slv_rst    => dp_rst, 

    slv_out    => mm_fields_out
  );
  
  index_lo <= TO_UINT(mm_fields_out(field_hi(c_field_arr, "index_lo") DOWNTO field_lo(c_field_arr, "index_lo")));
  index_hi <= TO_UINT(mm_fields_out(field_hi(c_field_arr, "index_hi") DOWNTO field_lo(c_field_arr, "index_hi")));

  gen_dp_block_select : FOR I IN 0 TO g_nof_streams-1 GENERATE
    u_dp_block_select : ENTITY work.dp_block_select
    GENERIC MAP (
      g_nof_blocks_per_sync => g_nof_blocks_per_sync,
      g_index_lo            => g_index_lo,
      g_index_hi            => g_index_hi
    )
    PORT MAP (
      rst          => dp_rst,
      clk          => dp_clk,
      -- Control
      index_lo     => index_lo,
      index_hi     => index_hi,
      -- ST sink
      snk_out      => snk_out_arr(I),
      snk_in       => snk_in_arr(I),
      -- ST source
      src_in       => src_in_arr(I),
      src_out      => src_out_arr(I)
    );
  END GENERATE;
      
END str;

