-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose : MMS for dp_bsn_aligner
-- Description: See dp_bsn_aligner.vhd

ENTITY mms_dp_bsn_align IS
  GENERIC (
    g_block_size           : NATURAL := 32;  -- > 1, g_block_size=1 is not supported, g_block_size <= g_block_period
    g_block_period         : NATURAL := 256; -- number of clock cycles per block period or block interval (not used in the architecture, used in the comment only)
    g_nof_input            : NATURAL := 2;   -- >= 1
    g_xoff_timeout         : NATURAL := 1536;  -- e.g.  g_bsn_latency * 2  * g_block_period
    g_sop_timeout          : NATURAL := 1024;  -- e.g. (g_bsn_latency + 1) * g_block_period
    g_bsn_latency          : NATURAL := 3;
    g_bsn_request_pipeline : NATURAL := 2;    -- = 1, 2, ..., c_bsn_nof_stages, requested total pipelining of the BSN max and BSN min operation
    g_cross_clock_domain   : BOOLEAN := TRUE  -- use FALSE when mm_clk and st_clk are the same, else use TRUE to cross the clock domain    
  );
  PORT (
    -- Memory-mapped clock domain
    mm_rst      : IN  STD_LOGIC;
    mm_clk      : IN  STD_LOGIC;
    reg_mosi    : IN  t_mem_mosi;
    reg_miso    : OUT t_mem_miso;
    
    -- Streaming clock domain
    dp_rst      : IN  STD_LOGIC;
    dp_clk      : IN  STD_LOGIC;
    -- ST sinks
    snk_out_arr : OUT t_dp_siso_arr(g_nof_input-1 DOWNTO 0);
    snk_in_arr  : IN  t_dp_sosi_arr(g_nof_input-1 DOWNTO 0);
    -- ST source
    src_in_arr  : IN  t_dp_siso_arr(g_nof_input-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);  -- only src_in_arr(0) is used for src_out flow control because all src_out need to stay aligned
    src_out_arr : OUT t_dp_sosi_arr(g_nof_input-1 DOWNTO 0)
  );
END mms_dp_bsn_align;


ARCHITECTURE str OF mms_dp_bsn_align IS
 
  SIGNAL en_evt : STD_LOGIC;
  SIGNAL en_arr : STD_LOGIC_VECTOR(g_nof_input-1 DOWNTO 0);      
  
BEGIN

  u_align : ENTITY work.dp_bsn_align 
  GENERIC MAP(
    g_block_size           => g_block_size,          
    g_block_period         => g_block_period,        
    g_nof_input            => g_nof_input,           
    g_xoff_timeout         => g_xoff_timeout,        
    g_sop_timeout          => g_sop_timeout,         
    g_bsn_latency          => g_bsn_latency,         
    g_bsn_request_pipeline => g_bsn_request_pipeline
  )
  PORT MAP(
    rst         => dp_rst,
    clk         => dp_clk,
    -- ST sinks
    snk_out_arr => snk_out_arr,
    snk_in_arr  => snk_in_arr,
    -- ST source
    src_in_arr  => src_in_arr,  
    src_out_arr => src_out_arr,
    -- MM
    in_en_evt   => en_evt,  
    in_en_arr   => en_arr
  );

  u_reg : ENTITY work.dp_bsn_align_reg
  GENERIC MAP (
    g_nof_input          => g_nof_input,
    g_cross_clock_domain => g_cross_clock_domain
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst   => mm_rst,  
    mm_clk   => mm_clk,  
    st_rst   => dp_rst,  
    st_clk   => dp_clk,  
    
    -- Memory Mapped Slave in mm_clk domain
    sla_in   => reg_mosi,
    sla_out  => reg_miso,
    
    -- MM registers in st_clk domain
    out_en_evt => en_evt,
    out_en_arr => en_arr
  );

END str;

