-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;

-- Purpose : MMS for dp_bsn_monitor
-- Description: See dp_bsn_monitor.vhd

ENTITY mms_dp_bsn_monitor IS
  GENERIC (
    g_nof_streams        : POSITIVE := 1;
    g_cross_clock_domain : BOOLEAN := TRUE;  -- use FALSE when mm_clk and dp_clk are the same, else use TRUE to cross the clock domain
    g_sync_timeout       : NATURAL := 200*10**6;
    g_bsn_w              : NATURAL := c_dp_stream_bsn_w;
    g_error_bi           : NATURAL := 0;
    g_cnt_sop_w          : NATURAL := c_word_w;
    g_cnt_valid_w        : NATURAL := c_word_w;
    g_log_first_bsn      : BOOLEAN := TRUE -- Logs first BSN + cycle count. Useful when using BSN aligner.
  );
  PORT (
    -- Memory-mapped clock domain
    mm_rst      : IN  STD_LOGIC;
    mm_clk      : IN  STD_LOGIC;
    reg_mosi    : IN  t_mem_mosi;
    reg_miso    : OUT t_mem_miso;
    
    -- Streaming clock domain
    dp_rst      : IN  STD_LOGIC;
    dp_clk      : IN  STD_LOGIC;
    sync_in     : IN  STD_LOGIC := '0'; -- Sync input that samples the current bsn number and the number of clockcycles that have expired since the last sop    
    
    in_siso_arr : IN  t_dp_siso_arr(g_nof_streams-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);
    in_sosi_arr : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0)
  );
END mms_dp_bsn_monitor;


ARCHITECTURE str OF mms_dp_bsn_monitor IS

  CONSTANT c_reg_adr_w : NATURAL := ceil_log2(9);

  SIGNAL mon_evt_arr           : STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0);
  SIGNAL mon_sync_timeout_arr  : STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0);
  
  SIGNAL mon_ready_stable_arr  : STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0);
  SIGNAL mon_xon_stable_arr    : STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0);
 
  TYPE t_mon_bsn_arr IS ARRAY(g_nof_streams-1 DOWNTO 0) OF STD_LOGIC_VECTOR(g_bsn_w-1 DOWNTO 0);
  TYPE t_mon_sop_arr IS ARRAY(g_nof_streams-1 DOWNTO 0) OF STD_LOGIC_VECTOR(g_cnt_sop_w-1 DOWNTO 0);
  TYPE t_mon_val_arr IS ARRAY(g_nof_streams-1 DOWNTO 0) OF STD_LOGIC_VECTOR(g_cnt_valid_w-1 DOWNTO 0);
 
  SIGNAL mon_bsn_at_sync_arr         : t_mon_bsn_arr;
  SIGNAL mon_nof_sop_arr             : t_mon_sop_arr;
  SIGNAL mon_nof_err_arr             : t_mon_sop_arr;  -- use g_cnt_sop_w, because at maximium all frames have an error
  SIGNAL mon_nof_valid_arr           : t_mon_val_arr;
  SIGNAL mon_bsn_first_arr           : t_mon_bsn_arr;
  SIGNAL mon_bsn_first_cycle_cnt_arr : t_slv_32_arr(g_nof_streams-1 DOWNTO 0);

  SIGNAL reg_mosi_arr                : t_mem_mosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL reg_miso_arr                : t_mem_miso_arr(g_nof_streams-1 DOWNTO 0); 
      
BEGIN

  u_common_mem_mux : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_streams,
    g_mult_addr_w => c_reg_adr_w
  )
  PORT MAP (
    mosi     => reg_mosi,
    miso     => reg_miso,
    mosi_arr => reg_mosi_arr,
    miso_arr => reg_miso_arr
  );

  gen_stream : FOR i IN 0 TO g_nof_streams-1 GENERATE
      
    u_reg : ENTITY work.dp_bsn_monitor_reg
    GENERIC MAP (
      g_cross_clock_domain => g_cross_clock_domain
    )
    PORT MAP (
      -- Clocks and reset
      mm_rst                  => mm_rst,
      mm_clk                  => mm_clk,
      st_rst                  => dp_rst,
      st_clk                  => dp_clk,
      
      -- Memory Mapped Slave in mm_clk domain
      sla_in                  => reg_mosi_arr(i),
      sla_out                 => reg_miso_arr(i),
      
      -- MM registers in dp_clk domain
      -- . control
      mon_evt                 => mon_evt_arr(i),
      mon_sync_timeout        => mon_sync_timeout_arr(i),
      -- . siso
      mon_ready_stable        => mon_ready_stable_arr(i),
      mon_xon_stable          => mon_xon_stable_arr(i),
      -- . sosi
      mon_bsn_at_sync         => mon_bsn_at_sync_arr(i),
      mon_nof_sop             => mon_nof_sop_arr(i),
      mon_nof_err             => mon_nof_err_arr(i),
      mon_nof_valid           => mon_nof_valid_arr(i),

      mon_bsn_first           => mon_bsn_first_arr(i),
      mon_bsn_first_cycle_cnt => mon_bsn_first_cycle_cnt_arr(i)
    );
    
    u_mon : ENTITY work.dp_bsn_monitor
    GENERIC MAP (
      g_sync_timeout  => g_sync_timeout,
      g_error_bi      => g_error_bi,
      g_log_first_bsn => g_log_first_bsn
    )
    PORT MAP (
      rst                    => dp_rst,
      clk                    => dp_clk,
  
      -- ST interface
      in_siso                => in_siso_arr(i),
      in_sosi                => in_sosi_arr(i),
      sync_in                => sync_in,
      
      -- MM interface
      -- . control
      mon_evt                 => mon_evt_arr(i),  -- pulses when new monitor data is available regarding the previous sync interval
      mon_sync                => OPEN,
      mon_sync_timeout        => mon_sync_timeout_arr(i),
      -- . siso
      mon_ready_stable        => mon_ready_stable_arr(i),
      mon_xon_stable          => mon_xon_stable_arr(i),
      -- . sosi
      mon_bsn_at_sync         => mon_bsn_at_sync_arr(i),
      mon_nof_sop             => mon_nof_sop_arr(i),
      mon_nof_err             => mon_nof_err_arr(i),
      mon_nof_valid           => mon_nof_valid_arr(i),

      mon_bsn_first           => mon_bsn_first_arr(i),
      mon_bsn_first_cycle_cnt => mon_bsn_first_cycle_cnt_arr(i)
    );
  
  END GENERATE;

END str;

