--------------------------------------------------------------------------------
--
-- Copyright (C) 2010
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose:
--   The dp_fifo_fill connected as a Memory Mapped Slave. 
--   
--   
-- Description:
-- Remarks:

LIBRARY IEEE, common_lib, technology_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY mms_dp_fifo_fill IS
  GENERIC (
    g_technology         : NATURAL := c_tech_select_default;
    g_nof_streams        : NATURAL := 3;                    
    g_cross_clock_domain : BOOLEAN := TRUE;  -- use FALSE when mm_clk and st_clk are the same, else use TRUE to cross the clock domain    
    g_data_w             : NATURAL := 16;
    g_bsn_w              : NATURAL := 1;
    g_empty_w            : NATURAL := 1;
    g_channel_w          : NATURAL := 1;
    g_error_w            : NATURAL := 1;
    g_use_bsn            : BOOLEAN := FALSE;
    g_use_empty          : BOOLEAN := FALSE;
    g_use_channel        : BOOLEAN := FALSE;
    g_use_error          : BOOLEAN := FALSE;
    g_use_sync           : BOOLEAN := FALSE;
    g_use_complex        : BOOLEAN := FALSE;  -- TRUE feeds the concatenated complex fields (im & re) through the FIFO instead of the data field.
    g_fifo_fill          : NATURAL := 0;
    g_fifo_size          : NATURAL := 256;    -- (32+2) * 256 = 1 M9K, g_data_w+2 for sop and eop
    g_fifo_af_margin     : NATURAL := 4;      -- Nof words below max (full) at which fifo is considered almost full
    g_fifo_rl            : NATURAL := 1       -- use RL=0 for internal show ahead FIFO, default use RL=1 for internal normal FIFO
  );
  PORT (
    -- Memory-mapped clock domain
    mm_rst      : IN  STD_LOGIC;
    mm_clk      : IN  STD_LOGIC;
    reg_mosi    : IN  t_mem_mosi;
    reg_miso    : OUT t_mem_miso;

    -- Streaming clock domain
    dp_rst      : IN  STD_LOGIC;
    dp_clk      : IN  STD_LOGIC;
    
    -- Monitor FIFO filling
--    wr_ful      : OUT STD_LOGIC;
--    usedw       : OUT STD_LOGIC_VECTOR(ceil_log2(largest(g_fifo_size, g_fifo_fill + g_fifo_af_margin + 2))-1 DOWNTO 0);  -- = ceil_log2(c_fifo_size)-1 DOWNTO 0
--    rd_emp      : OUT STD_LOGIC;
    
    -- ST sink
    snk_out_arr : OUT t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
    snk_in_arr  : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
    -- ST source
    src_in_arr  : IN  t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
    src_out_arr : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0)
  );
END mms_dp_fifo_fill;


ARCHITECTURE str OF mms_dp_fifo_fill IS
  
  CONSTANT c_usedw_w          : NATURAL := ceil_log2(largest(g_fifo_size, g_fifo_fill + g_fifo_af_margin + 2));
  
  SIGNAL fifo_usedw_reg : STD_LOGIC_VECTOR(g_nof_streams*c_word_w-1 DOWNTO 0) := (OTHERS => '0');
  SIGNAL wr_ful_reg     : STD_LOGIC_VECTOR(g_nof_streams-1          DOWNTO 0);
  SIGNAL rd_emp_reg     : STD_LOGIC_VECTOR(g_nof_streams-1          DOWNTO 0);
  
BEGIN

  gen_fifos : FOR I IN 0 TO g_nof_streams-1 GENERATE
    dp_fifo_sc : ENTITY work.dp_fifo_fill
    GENERIC MAP (
      g_technology     => g_technology,                                                                                    
      g_data_w         => g_data_w,                                                                                        
      g_bsn_w          => g_bsn_w,                                                                                         
      g_empty_w        => g_empty_w,                                                                                       
      g_channel_w      => g_channel_w,                                                                                     
      g_error_w        => g_error_w,                                                                                       
      g_use_bsn        => g_use_bsn,                                                                                       
      g_use_empty      => g_use_empty,                                                                                     
      g_use_channel    => g_use_channel,                                                                                   
      g_use_error      => g_use_error,                                                                                     
      g_use_sync       => g_use_sync,                                                                                      
      g_use_complex    => g_use_complex,             
      g_fifo_fill      => g_fifo_fill,
      g_fifo_size      => g_fifo_size,               
      g_fifo_af_margin => g_fifo_af_margin,          
      g_fifo_rl        => g_fifo_rl                  
    )
    PORT MAP (
      rst         => dp_rst,
      clk         => dp_clk,
      -- Monitor FIFO filling
      wr_ful      => wr_ful_reg(I),
      usedw       => fifo_usedw_reg(I*c_word_w + c_usedw_w - 1 DOWNTO I*c_word_w),
      rd_emp      => rd_emp_reg(I),
      -- ST sink
      snk_out     => snk_out_arr(I),
      snk_in      => snk_in_arr(I),
      -- ST source
      src_in      => src_in_arr(I),  
      src_out     => src_out_arr(I)
    );
  END GENERATE;

  u_reg : ENTITY work.dp_fifo_fill_reg 
  GENERIC MAP(
    g_nof_streams        => g_nof_streams,
    g_cross_clock_domain => g_cross_clock_domain
  )
  PORT MAP(
    -- Clocks and reset
    mm_rst  => mm_rst, 
    mm_clk  => mm_clk,
    st_rst  => dp_rst,
    st_clk  => dp_clk, 
    
    -- Memory Mapped Slave in mm_clk domain
    sla_in  => reg_mosi,
    sla_out => reg_miso, 
    
    -- MM registers in st_clk domain
    used_w  => fifo_usedw_reg,
    rd_emp  => rd_emp_reg, 
    wr_ful  => wr_ful_reg
  );

END str;
