-------------------------------------------------------------------------------
--
-- Copyright (C) 2017
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------
--
-- Author : E. Kooistra, 7 apr 2017
-- Purpose : Control force data per parallel stream via MM
-- Description:
--   Address   Data  Access  Description
--     0        [0]     R/W  force enable or default disable for data pass on
--     1     [31:0]     R/W  force sosi data
--     2     [31:0]     R/W  force sosi re
--     3     [31:0]     R/W  force sosi im
--
--   The actual sosi data, re and im width integer is constrained by g_dat_w.
--   Synthesis will optimize away unused bits from the full integer 32b range.
--

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;


ENTITY mms_dp_force_data_parallel IS
  GENERIC (
    g_dat_w                 : NATURAL := 32;    -- must be <= 32 to fit INTEGER range
    g_increment_data        : INTEGER := 0;
    g_increment_re          : INTEGER := 0;
    g_increment_im          : INTEGER := 0;
    g_increment_data_on_sop : BOOLEAN := FALSE;
    g_increment_re_on_sop   : BOOLEAN := FALSE;
    g_increment_im_on_sop   : BOOLEAN := FALSE;
    g_restart_data_on_sync  : BOOLEAN := FALSE;
    g_restart_re_on_sync    : BOOLEAN := FALSE;
    g_restart_im_on_sync    : BOOLEAN := FALSE;
    g_restart_data_on_sop   : BOOLEAN := FALSE;
    g_restart_re_on_sop     : BOOLEAN := FALSE;
    g_restart_im_on_sop     : BOOLEAN := FALSE
  );
  PORT (
    -- Clocks and reset
    mm_rst              : IN  STD_LOGIC;
    mm_clk              : IN  STD_LOGIC;
    dp_rst              : IN  STD_LOGIC;
    dp_clk              : IN  STD_LOGIC;
    -- MM control
    reg_force_data_mosi : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_force_data_miso : OUT t_mem_miso;
    -- ST sink
    snk_out             : OUT t_dp_siso;
    snk_in              : IN  t_dp_sosi;
    -- ST source
    src_in              : IN  t_dp_siso := c_dp_siso_rdy;
    src_out             : OUT t_dp_sosi
  );
END mms_dp_force_data_parallel;

ARCHITECTURE str OF mms_dp_force_data_parallel IS

  -- TYPE t_c_mem IS RECORD
  --   latency   : NATURAL;    -- read latency
  --   adr_w     : NATURAL;
  --   dat_w     : NATURAL;
  --   nof_dat   : NATURAL;    -- optional, nof dat words <= 2**adr_w
  --   init_sl   : STD_LOGIC;  -- optional, init all dat words to std_logic '0', '1' or 'X'
  CONSTANT c_mm_reg         : t_c_mem := (1, 2, g_dat_w, 4, 'X');

  CONSTANT c_mm_reg_init    : STD_LOGIC_VECTOR(c_mem_reg_init_w-1 DOWNTO 0) := (OTHERS=>'0');

  SIGNAL reg_force_data_wr  : STD_LOGIC_VECTOR(c_mm_reg.nof_dat*c_mm_reg.dat_w-1 DOWNTO 0);
  SIGNAL reg_force_data_rd  : STD_LOGIC_VECTOR(c_mm_reg.nof_dat*c_mm_reg.dat_w-1 DOWNTO 0) := (OTHERS=>'0');

  SIGNAL force_en           : STD_LOGIC := '0';
  SIGNAL force_data         : INTEGER := 0;       -- used for sosi.data
  SIGNAL force_re           : INTEGER := 0;       -- used for sosi.re
  SIGNAL force_im           : INTEGER := 0;       -- used for sosi.im

BEGIN

  -- wires
  force_en             <= reg_force_data_wr(0);
  reg_force_data_rd(0) <= reg_force_data_wr(0);  -- other bits are not used and will read '0'
  force_data                                                    <= TO_SINT(reg_force_data_wr(2*c_mm_reg.dat_w-1 DOWNTO   c_mm_reg.dat_w));
  reg_force_data_rd(2*c_mm_reg.dat_w-1 DOWNTO   c_mm_reg.dat_w) <=         reg_force_data_wr(2*c_mm_reg.dat_w-1 DOWNTO   c_mm_reg.dat_w);
  force_re                                                      <= TO_SINT(reg_force_data_wr(3*c_mm_reg.dat_w-1 DOWNTO 2*c_mm_reg.dat_w));
  reg_force_data_rd(3*c_mm_reg.dat_w-1 DOWNTO 2*c_mm_reg.dat_w) <=         reg_force_data_wr(3*c_mm_reg.dat_w-1 DOWNTO 2*c_mm_reg.dat_w);
  force_im                                                      <= TO_SINT(reg_force_data_wr(4*c_mm_reg.dat_w-1 DOWNTO 3*c_mm_reg.dat_w));
  reg_force_data_rd(4*c_mm_reg.dat_w-1 DOWNTO 3*c_mm_reg.dat_w) <=         reg_force_data_wr(4*c_mm_reg.dat_w-1 DOWNTO 3*c_mm_reg.dat_w);

  u_common_reg_r_w_dc : ENTITY common_lib.common_reg_r_w_dc
  GENERIC MAP (
    g_cross_clock_domain => TRUE,
    g_readback           => FALSE,
    g_reg                => c_mm_reg,
    g_init_reg           => c_mm_reg_init
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst         => mm_rst,
    mm_clk         => mm_clk,
    st_rst         => dp_rst,
    st_clk         => dp_clk,

    -- Memory Mapped Slave in mm_clk domain
    sla_in         => reg_force_data_mosi,
    sla_out        => reg_force_data_miso,

    -- MM registers in st_clk domain
    reg_wr_arr     => OPEN,
    reg_rd_arr     => OPEN,
    out_reg        => reg_force_data_wr,   -- readback via ST clock domain
    in_reg         => reg_force_data_rd
  );

  u_dp_force_data_parallel : ENTITY work.dp_force_data_parallel
  GENERIC MAP (
    g_dat_w                 => g_dat_w,
    g_increment_data        => g_increment_data,
    g_increment_re          => g_increment_re,
    g_increment_im          => g_increment_im,
    g_increment_data_on_sop => g_increment_data_on_sop,
    g_increment_re_on_sop   => g_increment_re_on_sop,
    g_increment_im_on_sop   => g_increment_im_on_sop,
    g_restart_data_on_sync  => g_restart_data_on_sync,
    g_restart_re_on_sync    => g_restart_re_on_sync,
    g_restart_im_on_sync    => g_restart_im_on_sync,
    g_restart_data_on_sop   => g_restart_data_on_sop,
    g_restart_re_on_sop     => g_restart_re_on_sop,
    g_restart_im_on_sop     => g_restart_im_on_sop
  )
  PORT MAP (
    rst           => dp_rst,
    clk           => dp_clk,
    -- MM control
    force_en      => force_en,
    force_data    => force_data,
    force_re      => force_re,
    force_im      => force_im,
    -- ST sink
    snk_out       => snk_out,
    snk_in        => snk_in,
    -- ST source
    src_in        => src_in,
    src_out       => src_out
  );
END str;

