-------------------------------------------------------------------------------
--
-- Copyright (C) 2017
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------
--
-- Author : E. Kooistra, 28 jul 2017
-- Purpose : Control force data per serial stream via MM for and array of streams
-- Description:
--   Stream Address   Description
--      0     0:4     See mms_dp_force_data_serial.vhd
--      0     5:7     not used, read 0
--      1     8:12    See mms_dp_force_data_serial.vhd
--      1     13:15   not used, read 0
--      2     16:20   See mms_dp_force_data_serial.vhd
--      2     21:23   not used, read 0
--      etc.
--
--   See description of mms_dp_force_data_serial.vhd.

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;


ENTITY mms_dp_force_data_serial_arr IS
  GENERIC (
    g_nof_streams          : NATURAL := 1;      -- number of parallel streams
    g_dat_w                : NATURAL := 32;     -- number of bits per force data value, must be <= 32 to fit INTEGER range
    g_index_period         : NATURAL := 10;     -- number of indices in time, must be <= 2*31 to fit in NATURAL range
    g_index_sample_block_n : BOOLEAN := FALSE   -- when TRUE sample index in block, else block index in sync interval
  );
  PORT (
    -- Clocks and reset
    mm_rst              : IN  STD_LOGIC;
    mm_clk              : IN  STD_LOGIC;
    dp_rst              : IN  STD_LOGIC;
    dp_clk              : IN  STD_LOGIC;
    -- MM control
    reg_force_data_mosi : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_force_data_miso : OUT t_mem_miso;
    -- ST sink
    snk_out_arr         : OUT t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
    snk_in_arr          : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
    -- ST source
    src_in_arr          : IN  t_dp_siso_arr(g_nof_streams-1 DOWNTO 0) := (OTHERS=>c_dp_siso_rdy);
    src_out_arr         : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0)
  );
END mms_dp_force_data_serial_arr;


ARCHITECTURE str OF mms_dp_force_data_serial_arr IS

  CONSTANT c_mm_reg_adr_w : NATURAL := 3;  -- = ceil_log2(c_mm_reg.adr_w), copied from mms_dp_force_data_serial.vhd

  SIGNAL reg_force_data_mosi_arr : t_mem_mosi_arr(g_nof_streams-1 DOWNTO 0) := (OTHERS=>c_mem_mosi_rst);
  SIGNAL reg_force_data_miso_arr : t_mem_miso_arr(g_nof_streams-1 DOWNTO 0);
  
BEGIN

  u_mem_mux : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_streams,
    g_mult_addr_w => c_mm_reg_adr_w
  )
  PORT MAP (
    mosi     => reg_force_data_mosi,
    miso     => reg_force_data_miso,
    mosi_arr => reg_force_data_mosi_arr,
    miso_arr => reg_force_data_miso_arr
  );

  gen_nof_streams : FOR I IN 0 TO g_nof_streams-1 GENERATE
    u_mms_dp_force_data_serial : ENTITY work.mms_dp_force_data_serial
    GENERIC MAP (
      g_dat_w                => g_dat_w,
      g_index_period         => g_index_period,
      g_index_sample_block_n => g_index_sample_block_n
    )
    PORT MAP (
      -- Clocks and reset
      mm_rst              => mm_rst,
      mm_clk              => mm_clk,
      dp_rst              => dp_rst,
      dp_clk              => dp_clk,
      -- MM control
      reg_force_data_mosi => reg_force_data_mosi_arr(I),
      reg_force_data_miso => reg_force_data_miso_arr(I),
      -- ST sink
      snk_out             => snk_out_arr(I),
      snk_in              => snk_in_arr(I),
      -- ST source
      src_in              => src_in_arr(I),
      src_out             => src_out_arr(I)
    );
  END GENERATE;
END str;

