-------------------------------------------------------------------------------
--
-- Copyright (C) 2017
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------
-- Author : E. Kooistra, 29 mar 2017
-- Purpose : Set gain for one stream via MM
-- Description:
--   Wrapper of mms_dp_gain_arr with g_nof_streams=1. See mms_dp_gain_arr
--   for details.
-- Usage:
-- . See tb_mms_dp_gain_arr which also tests this mms_dp_gain

LIBRARY IEEE, common_lib, common_mult_lib, dp_lib, technology_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;


ENTITY mms_dp_gain IS
  GENERIC (
    g_technology                    : NATURAL := c_tech_select_default;    
    g_complex_data                  : BOOLEAN := TRUE;
    g_complex_gain                  : BOOLEAN := FALSE;
    g_gain_init_re                  : INTEGER := 1;
    g_gain_init_im                  : INTEGER := 0;
    g_gain_w                        : NATURAL := 16;
    g_in_dat_w                      : NATURAL := 9;
    g_out_dat_w                     : NATURAL := 24;  -- 16 + 9 - 1 = 24 (-1 to skip double sign bit in product)
    
    -- pipelining (typically use defaults)
    -- . real multiplier
    g_pipeline_real_mult_input      : NATURAL := 1;      -- 0 or 1
    g_pipeline_real_mult_product    : NATURAL := 1;      -- 0 or 1
    g_pipeline_real_mult_output     : NATURAL := 1;      -- >= 0
    -- . complex multiplier
    g_pipeline_complex_mult_input   : NATURAL := 1;      -- 0 or 1
    g_pipeline_complex_mult_product : NATURAL := 0;      -- 0 or 1
    g_pipeline_complex_mult_adder   : NATURAL := 1;      -- 0 or 1
    g_pipeline_complex_mult_output  : NATURAL := 1       -- >= 0
  );
  PORT (
    -- Clocks and reset
    mm_rst            : IN  STD_LOGIC;
    mm_clk            : IN  STD_LOGIC;
    dp_rst            : IN  STD_LOGIC;
    dp_clk            : IN  STD_LOGIC;

    -- MM access to gain
    reg_gain_re_mosi  : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_gain_re_miso  : OUT t_mem_miso;
    reg_gain_im_mosi  : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_gain_im_miso  : OUT t_mem_miso;

    -- ST
    in_sosi           : IN  t_dp_sosi;
    out_sosi          : OUT t_dp_sosi
  );
END mms_dp_gain;


ARCHITECTURE str OF mms_dp_gain IS

  SIGNAL in_sosi_arr     : t_dp_sosi_arr(0 DOWNTO 0);
  SIGNAL out_sosi_arr    : t_dp_sosi_arr(0 DOWNTO 0);

BEGIN
  -- wires
  in_sosi_arr(0) <= in_sosi;
  out_sosi       <= out_sosi_arr(0);

  u_one : ENTITY work.mms_dp_gain_arr
  GENERIC MAP (
    g_technology                    => g_technology,
    -- functional
    g_nof_streams                   => 1,
    g_complex_data                  => g_complex_data,
    g_complex_gain                  => g_complex_gain,
    g_gain_init_re                  => g_gain_init_re,
    g_gain_init_im                  => g_gain_init_im,
    g_gain_w                        => g_gain_w,
    g_in_dat_w                      => g_in_dat_w,
    g_out_dat_w                     => g_out_dat_w,    
    -- pipelining (typically use defaults)
    -- . real multiplier
    g_pipeline_real_mult_input      => g_pipeline_real_mult_input,
    g_pipeline_real_mult_product    => g_pipeline_real_mult_product,
    g_pipeline_real_mult_output     => g_pipeline_real_mult_output,
    -- . complex multiplier
    g_pipeline_complex_mult_input   => g_pipeline_complex_mult_input,
    g_pipeline_complex_mult_product => g_pipeline_complex_mult_product,
    g_pipeline_complex_mult_adder   => g_pipeline_complex_mult_adder,
    g_pipeline_complex_mult_output  => g_pipeline_complex_mult_output
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => dp_rst,
    dp_clk            => dp_clk,

    -- MM access to gain
    reg_gain_re_mosi  => reg_gain_re_mosi,
    reg_gain_re_miso  => reg_gain_re_miso,
    reg_gain_im_mosi  => reg_gain_im_mosi,
    reg_gain_im_miso  => reg_gain_im_miso,

    -- ST
    in_sosi_arr       => in_sosi_arr,
    out_sosi_arr      => out_sosi_arr
  );

END str;

