-------------------------------------------------------------------------------
--
-- Copyright (C) 2017
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------
-- Author : E. Kooistra, 29 mar 2017
-- Purpose : Set gain per stream via MM
-- Description:
--   There is one MM programmable gain per stream and therefore the same gain
--   applies to all serial data on that stream. There are g_nof_streams in
--   parallel.
--   If the gain is complex then there is a separate MM reg for the imaginary
--   gain. If the data is complex then the in_sosi.re/im are used, else the
--   in_sosi.data is used. The following combinations are possible:
--
--   g_complex_data   g_complex_gain   Description
--     True             True           for amplitude and phase adjustment
--     True             False          for amplitude only adjustment
--     False            False          for real input gain adjustment
--     False            True           for making real input complex
--
-- Remarks:
-- . The in_sosi_arr().valid is propagated to out_sosi_arr().valid.
-- . Typically choose g_out_dat_w = g_gain_w + g_in_dat_w - 1. This skips the
--   double sign bit of the product and assumes that full scale amplitude of
--   the complex gain = full scale re = full scale im to not cause overflow.
-- . Requantization needs to be done outside this component, e.g. using
--   dp_requantize.
-- . There is one gain per parallel stream. If all streams can have the same 
--   gain, then program the same gain for each stream. This avoids potential
--   fanout timing issues if one gain programmable would be used for all
--   streams.
-- . If the gain needs to vary in time per stream, because the stream carries
--   time multiplexed data, then a new component is necessary. The gains per
--   stream may then be stored in a MM RAM, because they are used sequentially.
-- . see also reqant_selection_compensation.m

LIBRARY IEEE, common_lib, common_mult_lib, technology_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;


ENTITY mms_dp_gain_arr IS
  GENERIC (
    g_technology                    : NATURAL := c_tech_select_default;
    g_nof_streams                   : NATURAL := 3;    -- for g_nof_streams=1 there is mms_dp_gain.vhd
    g_complex_data                  : BOOLEAN := TRUE;
    g_complex_gain                  : BOOLEAN := FALSE;
    g_gain_init_re                  : INTEGER := 1;
    g_gain_init_im                  : INTEGER := 0;
    g_gain_w                        : NATURAL := 16;
    g_in_dat_w                      : NATURAL := 9;
    g_out_dat_w                     : NATURAL := 24;  -- 16 + 9 - 1 = 24 (-1 to skip double sign bit in product)

    -- pipelining (typically use defaults)
    -- . real multiplier
    g_pipeline_real_mult_input      : NATURAL := 1;      -- 0 or 1
    g_pipeline_real_mult_product    : NATURAL := 1;      -- 0 or 1
    g_pipeline_real_mult_output     : NATURAL := 1;      -- >= 0
    -- . complex multiplier
    g_pipeline_complex_mult_input   : NATURAL := 1;      -- 0 or 1
    g_pipeline_complex_mult_product : NATURAL := 0;      -- 0 or 1
    g_pipeline_complex_mult_adder   : NATURAL := 1;      -- 0 or 1
    g_pipeline_complex_mult_output  : NATURAL := 1       -- >= 0
  );
  PORT (
    -- Clocks and reset
    mm_rst            : IN  STD_LOGIC;
    mm_clk            : IN  STD_LOGIC;
    dp_rst            : IN  STD_LOGIC;
    dp_clk            : IN  STD_LOGIC;

    -- MM access to gain
    reg_gain_re_mosi  : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_gain_re_miso  : OUT t_mem_miso;
    reg_gain_im_mosi  : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_gain_im_miso  : OUT t_mem_miso;

    -- ST
    in_sosi_arr       : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
    out_sosi_arr      : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0)
  );
END mms_dp_gain_arr;


ARCHITECTURE str OF mms_dp_gain_arr IS

  CONSTANT c_real_multiply  : BOOLEAN := g_complex_data=FALSE AND g_complex_gain=FALSE;
  
  -- TYPE t_c_mem IS RECORD
  --   latency   : NATURAL;    -- read latency
  --   adr_w     : NATURAL;
  --   dat_w     : NATURAL;
  --   nof_dat   : NATURAL;    -- optional, nof dat words <= 2**adr_w
  --   init_sl   : STD_LOGIC;  -- optional, init all dat words to std_logic '0', '1' or 'X'
  CONSTANT c_mm_reg         : t_c_mem := (1, ceil_log2(g_nof_streams), g_gain_w, g_nof_streams, 'X');
  
  CONSTANT c_init_re        : STD_LOGIC_VECTOR(g_nof_streams*g_gain_w-1 DOWNTO 0) := array_sinit(g_gain_init_re, g_nof_streams, g_gain_w);
  CONSTANT c_init_im        : STD_LOGIC_VECTOR(g_nof_streams*g_gain_w-1 DOWNTO 0) := array_sinit(g_gain_init_im, g_nof_streams, g_gain_w);
  
  CONSTANT c_mm_reg_init_re : STD_LOGIC_VECTOR(c_mem_reg_init_w-1 DOWNTO 0) := RESIZE_UVEC(c_init_re, c_mem_reg_init_w);
  CONSTANT c_mm_reg_init_im : STD_LOGIC_VECTOR(c_mem_reg_init_w-1 DOWNTO 0) := RESIZE_UVEC(c_init_im, c_mem_reg_init_w);

  SIGNAL reg_gain_re        : STD_LOGIC_VECTOR(g_nof_streams*g_gain_w-1 DOWNTO 0) := c_init_re;
  SIGNAL reg_gain_im        : STD_LOGIC_VECTOR(g_nof_streams*g_gain_w-1 DOWNTO 0) := c_init_im;

  SIGNAL mult_sosi_arr      : t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
  
BEGIN

  u_common_reg_r_w_dc_re : ENTITY common_lib.common_reg_r_w_dc
  GENERIC MAP (
    g_cross_clock_domain => TRUE,
    g_readback           => FALSE,
    g_reg                => c_mm_reg,
    g_init_reg           => c_mm_reg_init_re
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst         => mm_rst,
    mm_clk         => mm_clk,
    st_rst         => dp_rst,
    st_clk         => dp_clk,

    -- Memory Mapped Slave in mm_clk domain
    sla_in         => reg_gain_re_mosi,
    sla_out        => reg_gain_re_miso,

    -- MM registers in st_clk domain
    reg_wr_arr     => OPEN,
    reg_rd_arr     => OPEN,
    in_reg         => reg_gain_re,
    out_reg        => reg_gain_re   -- readback via ST clock domain
  );

  gen_real_multiply : IF c_real_multiply=TRUE GENERATE
    gen_nof_streams : FOR I IN 0 TO g_nof_streams-1 GENERATE
      u_common_mult : ENTITY common_mult_lib.common_mult
      GENERIC MAP (
        g_technology       => g_technology,
        g_variant          => "IP",
        g_in_a_w           => g_gain_w,
        g_in_b_w           => g_in_dat_w,
        g_out_p_w          => g_out_dat_w,  -- g_out_dat_w = g_gain_w+g_in_dat_w, use smaller g_out_dat_w to truncate MSbits, or larger g_out_dat_w to extend MSbits
        g_nof_mult         => 1,
        g_pipeline_input   => g_pipeline_real_mult_input,
        g_pipeline_product => g_pipeline_real_mult_product,
        g_pipeline_output  => g_pipeline_real_mult_output,
        g_representation   => "SIGNED"  -- or "UNSIGNED"
      )
      PORT MAP (
        rst        => dp_rst,
        clk        => dp_clk,
        in_a       => reg_gain_re((I+1)*g_gain_w-1 DOWNTO I*g_gain_w),
        in_b       => in_sosi_arr(I).data(g_in_dat_w-1 DOWNTO 0),
        in_val     => in_sosi_arr(I).valid,
        out_p      => mult_sosi_arr(I).data(g_out_dat_w-1 DOWNTO 0),
        out_val    => mult_sosi_arr(I).valid
      );
      
      p_out_sosi_arr : PROCESS(mult_sosi_arr)
      BEGIN
        out_sosi_arr(I)       <= c_dp_sosi_rst;  -- to avoid (vsim-8684) No drivers exist on out port out_sosi_arr
        out_sosi_arr(I).valid <= mult_sosi_arr(I).valid;
        out_sosi_arr(I).data  <= RESIZE_DP_SDATA(mult_sosi_arr(I).data(g_out_dat_w-1 DOWNTO 0));  -- sign extend
      END PROCESS;
      
    END GENERATE gen_nof_streams;
  END GENERATE gen_real_multiply;

  gen_complex_multiply : IF c_real_multiply=FALSE GENERATE
    gen_complex_gain : IF g_complex_gain=TRUE GENERATE
      u_common_reg_r_w_dc_im : ENTITY common_lib.common_reg_r_w_dc
      GENERIC MAP (
        g_cross_clock_domain => TRUE,
        g_readback           => FALSE,
        g_reg                => c_mm_reg,
        g_init_reg           => c_mm_reg_init_im
      )
      PORT MAP (
        -- Clocks and reset
        mm_rst         => mm_rst,
        mm_clk         => mm_clk,
        st_rst         => dp_rst,
        st_clk         => dp_clk,

        -- Memory Mapped Slave in mm_clk domain
        sla_in         => reg_gain_im_mosi,
        sla_out        => reg_gain_im_miso,

        -- MM registers in st_clk domain
        reg_wr_arr     => OPEN,
        reg_rd_arr     => OPEN,
        in_reg         => reg_gain_im,
        out_reg        => reg_gain_im   -- readback via ST clock domain
      );
    END GENERATE gen_complex_gain;
    -- ELSE: if g_complex_gain=FALSE then use default reg_gain_im, which is then typically g_gain_init_im=0 for all streams.

    gen_nof_streams : FOR I IN 0 TO g_nof_streams-1 GENERATE
      u_common_complex_mult : ENTITY common_mult_lib.common_complex_mult
      GENERIC MAP (
        g_sim              => FALSE,
        g_sim_level        => 0, -- 0: Simulate variant passed via g_variant for given g_technology
        g_technology       => g_technology,
        g_variant          => "IP",
        g_in_a_w           => g_gain_w,
        g_in_b_w           => g_in_dat_w,
        g_out_p_w          => g_out_dat_w,  -- g_out_dat_w = g_gain_w+g_in_dat_w, use smaller g_out_dat_w to truncate MSbits, or larger g_out_dat_w to extend MSbits
        g_conjugate_b      => FALSE,
        g_pipeline_input   => g_pipeline_complex_mult_input,
        g_pipeline_product => g_pipeline_complex_mult_product,
        g_pipeline_adder   => g_pipeline_complex_mult_adder,
        g_pipeline_output  => g_pipeline_complex_mult_output
      )
      PORT MAP (
        rst        => dp_rst,
        clk        => dp_clk,
        in_ar      => reg_gain_re((I+1)*g_gain_w-1 DOWNTO I*g_gain_w),
        in_ai      => reg_gain_im((I+1)*g_gain_w-1 DOWNTO I*g_gain_w),
        in_br      => in_sosi_arr(I).re(g_in_dat_w-1 DOWNTO 0),
        in_bi      => in_sosi_arr(I).im(g_in_dat_w-1 DOWNTO 0),
        in_val     => in_sosi_arr(I).valid,                          -- only propagate valid, not used internally
        out_pr     => mult_sosi_arr(I).re(g_out_dat_w-1 DOWNTO 0),
        out_pi     => mult_sosi_arr(I).im(g_out_dat_w-1 DOWNTO 0),
        out_val    => mult_sosi_arr(I).valid
      );
      
      p_out_sosi_arr : PROCESS(mult_sosi_arr)
      BEGIN
        out_sosi_arr(I)       <= c_dp_sosi_rst;  -- to avoid (vsim-8684) No drivers exist on out port out_sosi_arr
        out_sosi_arr(I).valid <= mult_sosi_arr(I).valid;
        out_sosi_arr(I).re    <= RESIZE_DP_DSP_DATA(mult_sosi_arr(I).re(g_out_dat_w-1 DOWNTO 0));  -- sign extend
        out_sosi_arr(I).im    <= RESIZE_DP_DSP_DATA(mult_sosi_arr(I).im(g_out_dat_w-1 DOWNTO 0));  -- sign extend
      END PROCESS;
      
    END GENERATE gen_nof_streams;
  END GENERATE gen_complex_multiply;

END str;

