-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Add MM control and multi-stream support to dp_pkt_merge.
-- Description:
-- Remarks:

LIBRARY IEEE, common_lib, mm_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_field_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY mms_dp_packet_merge IS
  GENERIC (
    g_nof_streams : NATURAL;
    g_nof_pkt     : NATURAL
  );
  PORT (
    -- Memory-mapped clock domain
    mm_rst       : IN  STD_LOGIC;
    mm_clk       : IN  STD_LOGIC;

    reg_mosi     : IN  t_mem_mosi;
    reg_miso     : OUT t_mem_miso;
    
    -- Streaming clock domain
    dp_rst       : IN  STD_LOGIC;
    dp_clk       : IN  STD_LOGIC;

    -- ST sinks
    snk_out_arr  : OUT t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
    snk_in_arr   : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
    -- ST source
    src_in_arr   : IN  t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
    src_out_arr  : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0)
  );
END mms_dp_packet_merge;


ARCHITECTURE str OF mms_dp_packet_merge IS

  CONSTANT c_field_arr : t_common_field_arr(0 DOWNTO 0) := (0=> ( field_name_pad("nof_pkt"), "RW", ceil_log2(g_nof_pkt+1), field_default(18) ));

  TYPE t_mm_fields_out_arr IS ARRAY(g_nof_streams-1 DOWNTO 0) OF STD_LOGIC_VECTOR(field_slv_out_len(c_field_arr)-1 DOWNTO 0);
  TYPE t_nof_pkt_arr       IS ARRAY(g_nof_streams-1 DOWNTO 0) OF STD_LOGIC_VECTOR(ceil_log2(g_nof_pkt+1)-1 DOWNTO 0);

  SIGNAL reg_mosi_arr      : t_mem_mosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL reg_miso_arr      : t_mem_miso_arr(g_nof_streams-1 DOWNTO 0); 

  SIGNAL mm_fields_out_arr : t_mm_fields_out_arr;
  SIGNAL nof_pkt           : t_nof_pkt_arr;
      
BEGIN

  u_common_mem_mux : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_streams,
    g_mult_addr_w => ceil_log2(field_nof_words(c_field_arr, c_word_w))
  )
  PORT MAP (
    mosi     => reg_mosi,
    miso     => reg_miso,
    mosi_arr => reg_mosi_arr,
    miso_arr => reg_miso_arr
  );

  gen_stream : FOR i IN 0 TO g_nof_streams-1 GENERATE
      
    u_mm_fields: ENTITY mm_lib.mm_fields
    GENERIC MAP(
      g_field_arr => c_field_arr
    )
    PORT MAP (
      mm_clk  => mm_clk,
      mm_rst  => mm_rst,
  
      mm_mosi => reg_mosi_arr(i),
      mm_miso => reg_miso_arr(i),
      
      slv_clk => dp_clk,
      slv_rst => dp_rst, 
  
      slv_out => mm_fields_out_arr(i)
    );

    nof_pkt(i) <= mm_fields_out_arr(i)(field_hi(c_field_arr, "nof_pkt") DOWNTO field_lo(c_field_arr, "nof_pkt"));
    
    u_dp_merge : ENTITY work.dp_packet_merge
    GENERIC MAP (
      g_nof_pkt   => g_nof_pkt
    )
    PORT MAP (
      rst         => dp_rst,
      clk         => dp_clk,
      nof_pkt     => nof_pkt(i),
      snk_out     => snk_out_arr(i),
      snk_in      => snk_in_arr(i),
      src_in      => src_in_arr(i), 
      src_out     => src_out_arr(i)
    );
  
  END GENERATE;

END str;

