-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY mms_dp_split IS
  GENERIC (
    g_nof_streams     : NATURAL := 1;
    g_data_w          : NATURAL := 16;
    g_symbol_w        : NATURAL := 8;
    g_nof_symbols_max : NATURAL := 8  -- Sets maximum nof_symbols that can be selected
  );
  PORT (
    -- Memory-mapped clock domain
    mm_rst       : IN  STD_LOGIC;
    mm_clk       : IN  STD_LOGIC;

    reg_mosi     : IN  t_mem_mosi;
    reg_miso     : OUT t_mem_miso;
    
    -- Streaming clock domain
    dp_rst       : IN  STD_LOGIC;
    dp_clk       : IN  STD_LOGIC;

    -- ST sinks
    snk_out_arr  : OUT t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
    snk_in_arr   : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
    -- ST source
    src_in_2arr  : IN  t_dp_siso_2arr_2(g_nof_streams-1 DOWNTO 0);
    src_out_2arr : OUT t_dp_sosi_2arr_2(g_nof_streams-1 DOWNTO 0)
  );
END mms_dp_split;


ARCHITECTURE str OF mms_dp_split IS

  CONSTANT c_reg_adr_w : NATURAL := 1;

  SIGNAL reg_mosi_arr  : t_mem_mosi_arr(g_nof_streams-1 DOWNTO 0);
  SIGNAL reg_miso_arr  : t_mem_miso_arr(g_nof_streams-1 DOWNTO 0); 

  TYPE t_nof_symbols_arr IS ARRAY(g_nof_streams-1 DOWNTO 0) OF STD_LOGIC_VECTOR(ceil_log2(g_nof_symbols_max+1)-1 DOWNTO 0);

  SIGNAL nof_symbols   : t_nof_symbols_arr;
      
BEGIN

  u_common_mem_mux : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => g_nof_streams,
    g_mult_addr_w => c_reg_adr_w
  )
  PORT MAP (
    mosi     => reg_mosi,
    miso     => reg_miso,
    mosi_arr => reg_mosi_arr,
    miso_arr => reg_miso_arr
  );

  gen_stream : FOR i IN 0 TO g_nof_streams-1 GENERATE
      
    u_reg : ENTITY work.dp_split_reg
    GENERIC MAP (
      g_nof_symbols => g_nof_symbols_max
    )
    PORT MAP (
      -- Clocks and reset
      mm_rst       => mm_rst,
      mm_clk       => mm_clk,
      st_rst       => dp_rst,
      st_clk       => dp_clk,
      
      -- Memory Mapped Slave in mm_clk domain
      sla_in       => reg_mosi_arr(i),
      sla_out      => reg_miso_arr(i),
      
      -- MM registers in dp_clk domain
      -- . control
      nof_symbols  => nof_symbols(i)
    );
    
    u_dp_split : ENTITY work.dp_split
    GENERIC MAP (
      g_data_w      => g_data_w,
      g_symbol_w    => g_symbol_w,
      g_nof_symbols => g_nof_symbols_max
    )
    PORT MAP (
      rst         => dp_rst,
      clk         => dp_clk,
      nof_symbols => nof_symbols(i),
      snk_out     => snk_out_arr(i),
      snk_in      => snk_in_arr(i),
      src_in_arr  => src_in_2arr(i), 
      src_out_arr => src_out_2arr(i)
    );
  
  END GENERATE;

END str;

