--------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--------------------------------------------------------------------------------

-- Purpose: Checks if the number of SOPs within a sync interval is as expected and 
--          corrects if sync is too early or too late. 

--
-- Description: The dp_sync_checker distinguishes three situations:
--
--              1. SYNC is ok
--              In case the sync interval is as expected the input data is simply
--              passed through to the output.
--
--              2. SYNC is too early
--              When the sync is too early the data is still passed to the output,
--              unitl the specief number of packets (g_nof_blk_per_sync) have been 
--              send. The SYNC signal that was too early is suppressed and will not
--              be visible at the output. Once the g_nof_blk_per_sync packets have 
--              been send to the output the dp_sync_checker will wait until a new 
--              sync has arrived. 
--
--              3. SYNC is too late
--              When an expected SYNC is not found at the input the dp_sync_checker
--              will stop passing data to the output and wait for the next SYNC on 
--              the input. 
--
-- Remarks: Parsing of backpressure is not yet implemented,  since it was not 
--          required for the first purpose of this block (in reorder_transpose). 
--

LIBRARY IEEE, common_lib;                   
USE IEEE.STD_LOGIC_1164.ALL;    
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY mms_dp_sync_checker IS
  GENERIC(     
    g_cross_clock_domain : BOOLEAN := TRUE;
    g_nof_blk_per_sync   : NATURAL := 16
  );                      
  PORT (
    mm_rst                   : IN  STD_LOGIC;  -- reset synchronous with mm_clk
    mm_clk                   : IN  STD_LOGIC;  -- memory-mapped bus clock                       

    dp_clk                   : IN  STD_LOGIC;
    dp_rst                   : IN  STD_LOGIC;
                             
    -- ST sinks              
    snk_out                  : OUT t_dp_siso;
    snk_in                   : IN  t_dp_sosi;
    -- ST source                 
    src_in                   : IN  t_dp_siso := c_dp_siso_rdy;
    src_out                  : OUT t_dp_sosi;
                             
    -- Memory Mapped                                           
    reg_dp_sync_checker_mosi : IN  t_mem_mosi;
    reg_dp_sync_checker_miso : OUT t_mem_miso 
                 
  );
END mms_dp_sync_checker;

ARCHITECTURE str OF mms_dp_sync_checker IS  

  -- Define the actual size of the MM slave register
  CONSTANT c_nof_regs  : POSITIVE := 2; 
  CONSTANT c_mm_reg    : t_c_mem := (latency  => 1,
                                     adr_w    => ceil_log2(c_nof_regs),
                                     dat_w    => c_word_w,     
                                     nof_dat  => c_nof_regs,
                                     init_sl  => '0');

  SIGNAL read_register   : STD_LOGIC_VECTOR(c_nof_regs*c_word_w-1 DOWNTO 0);  
  SIGNAL nof_early_syncs : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0); 
  SIGNAL nof_late_syncs  : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
  SIGNAL reg_rd_arr      : STD_LOGIC_VECTOR(c_nof_regs-1 DOWNTO 0);

BEGIN 

  u_dp_sync_checker : ENTITY work.dp_sync_checker
  GENERIC MAP(
    g_nof_blk_per_sync => g_nof_blk_per_sync
  )                      
  PORT MAP(
    dp_clk                => dp_clk,
    dp_rst                => dp_rst,
    snk_out               => snk_out,         
    snk_in                => snk_in,          
    src_in                => src_in,          
    src_out               => src_out,         
    nof_early_syncs       => nof_early_syncs, 
    nof_late_syncs        => nof_late_syncs, 
    clear_nof_early_syncs => reg_rd_arr(0),
    clear_nof_late_syncs  => reg_rd_arr(1) 
  );

  read_register <= nof_late_syncs & nof_early_syncs;

  u_reg : ENTITY common_lib.common_reg_r_w_dc
  GENERIC MAP (
    g_cross_clock_domain => g_cross_clock_domain,
    g_in_new_latency     => 1,   
    g_readback           => TRUE,
    g_reg                => c_mm_reg
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,
    st_rst      => dp_rst,
    st_clk      => dp_clk,
    
    -- Memory Mapped Slave in mm_clk domain
    sla_in      => reg_dp_sync_checker_mosi,
    sla_out     => reg_dp_sync_checker_miso,
    
    -- MM registers in st_clk domain
    reg_wr_arr  => OPEN,
    reg_rd_arr  => reg_rd_arr,
    in_new      => OPEN,
    in_reg      => read_register,
    out_reg     => OPEN, 
    out_new     => OPEN
  );
              
END str;


