-------------------------------------------------------------------------------
--
-- Copyright (C) 2013
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.dp_stream_pkg.ALL;

ENTITY mms_dp_xonoff IS
  GENERIC (
    g_nof_streams     : NATURAL := 1;
    g_combine_streams : BOOLEAN := FALSE;
    g_bypass          : BOOLEAN := FALSE;
    g_default_value   : STD_LOGIC := '1';
    g_timeout_time    : NATURAL := 0;
    g_sim             : BOOLEAN := FALSE
  );
  PORT (
    -- Memory-mapped clock domain
    mm_rst       : IN  STD_LOGIC;
    mm_clk       : IN  STD_LOGIC;

    reg_mosi     : IN  t_mem_mosi;
    reg_miso     : OUT t_mem_miso;
    
    -- Streaming clock domain
    dp_rst       : IN  STD_LOGIC;
    dp_clk       : IN  STD_LOGIC;

    -- ST sinks
    snk_out_arr  : OUT t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
    snk_in_arr   : IN  t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);
    -- ST source
    src_in_arr   : IN  t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);
    src_out_arr  : OUT t_dp_sosi_arr(g_nof_streams-1 DOWNTO 0);

    -- Optional override (w.r.t MM setting) to force X Off (by assigning '1')
    force_xoff_arr : IN STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0) := (OTHERS=>'0')
  );
END mms_dp_xonoff;


ARCHITECTURE str OF mms_dp_xonoff IS

  CONSTANT c_nof_ctrl_streams : NATURAL := sel_a_b(g_combine_streams, 1, g_nof_streams);
  CONSTANT c_reg_adr_w        : NATURAL := 1;

  SIGNAL reg_mosi_arr  : t_mem_mosi_arr(c_nof_ctrl_streams-1 DOWNTO 0);
  SIGNAL reg_miso_arr  : t_mem_miso_arr(c_nof_ctrl_streams-1 DOWNTO 0); 

  SIGNAL xonoff_reg    : STD_LOGIC_VECTOR(c_nof_ctrl_streams-1 DOWNTO 0);
  SIGNAL xonoff_reg_i  : STD_LOGIC_VECTOR(g_nof_streams-1 DOWNTO 0);
  SIGNAL src_in_arr_i  : t_dp_siso_arr(g_nof_streams-1 DOWNTO 0);  
      
BEGIN

  u_common_mem_mux : ENTITY common_lib.common_mem_mux
  GENERIC MAP (    
    g_nof_mosi    => c_nof_ctrl_streams,
    g_mult_addr_w => c_reg_adr_w
  )
  PORT MAP (
    mosi     => reg_mosi,
    miso     => reg_miso,
    mosi_arr => reg_mosi_arr,
    miso_arr => reg_miso_arr
  );

  gen_reg : FOR i IN 0 TO c_nof_ctrl_streams-1 GENERATE
    gen_no_timeout : IF g_timeout_time = 0 GENERATE 
      u_reg : ENTITY work.dp_xonoff_reg
      GENERIC MAP(
        g_default_value => g_default_value
      )
      PORT MAP (
        -- Clocks and reset
        mm_rst       => mm_rst,
        mm_clk       => mm_clk,
        st_rst       => dp_rst,
        st_clk       => dp_clk,
        
        -- Memory Mapped Slave in mm_clk domain
        sla_in       => reg_mosi_arr(i),
        sla_out      => reg_miso_arr(i),
        
        -- MM registers in dp_clk domain
        -- . control
        xonoff_reg   => xonoff_reg(i DOWNTO i)
      );
    END GENERATE;
    
    gen_with_timeout : IF g_timeout_time > 0 GENERATE
      u_reg : ENTITY work.dp_xonoff_reg_timeout
      GENERIC MAP(
        g_default_value => g_default_value,
        g_mm_timeout    => g_timeout_time,
        g_sim           => g_sim
      )
      PORT MAP (
        -- Clocks and reset
        mm_rst       => mm_rst,
        mm_clk       => mm_clk,
        st_rst       => dp_rst,
        st_clk       => dp_clk,
        
        -- Memory Mapped Slave in mm_clk domain
        sla_in       => reg_mosi_arr(i),
        sla_out      => reg_miso_arr(i),
        
        -- MM registers in dp_clk domain
        -- . control
        xonoff_reg   => xonoff_reg(i DOWNTO i)
      );   
    END GENERATE; 
  END GENERATE;    

  xonoff_reg_i <= xonoff_reg WHEN g_combine_streams = FALSE ELSE (OTHERS => xonoff_reg(0));

  gen_stream : FOR i IN 0 TO g_nof_streams-1 GENERATE    
    -- The xon signal is AND-ed with the register value. 
    src_in_arr_i(i).ready <= src_in_arr(i).ready;
    src_in_arr_i(i).xon   <= src_in_arr(i).xon AND xonoff_reg_i(i);    
  
    u_dp_xonoff : ENTITY work.dp_xonoff
    GENERIC MAP (
      g_bypass => g_bypass
    )
    PORT MAP (
      rst      => dp_rst,
      clk      => dp_clk,

      in_siso  => snk_out_arr(i),
      in_sosi  => snk_in_arr(i),
      out_siso => src_in_arr_i(i), 
      out_sosi => src_out_arr(i),

      force_xoff => force_xoff_arr(i)
    );
  END GENERATE;

END str;


